/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Atom.h"
#include <iostream>

using namespace std;

/* Functions for class Atom. */

/* constructors */
Atom::Atom(StateVar* staVar, const std::string &val, 
	   const std::string &ide)
  : support (NULL) 
{
  if (staVar->inValues(val)) {
    stateVar = staVar;
    value = val;
  }
  else cout << "Atom --> Cannot create the atom : value not in the variable domain\n";
  id = ide;
  ref_trans_pre = new std::vector<Transition*>();
  ref_trans_post = new std::vector<Transition*>();
/*#ifdef GCC2*/
   ref_comps = new hash_map<const char*, Comp*, hash<const char*>, eqstr> ();
/*#else
   ref_comps = new __gnu_cxx::hash_map<const char*, Comp*, __gnu_cxx::hash<const char*>, eqstr>();
#endif*/
  dispid = stateVar->addAtomToLinks(this); 
  label = "unknown";
  supported = new std::vector<Atom*>();
  ref_clauses = new std::vector<Clause*>();
  propagation_number = 0;
  neg = false;
  post_prob = 0;
  fail = false;
  guards_number = 0;
  idle = 0;
  cond = 0;
  dead_end = false;
  dyn_name = 0;
  topo_number = 0;
}

Atom::Atom(StateVar* staVar, const std::string &val, const std::string &ide, 
	   const bool &non)
  : support (NULL)
 {
  if (staVar->inValues(val)) {
    stateVar = staVar;
    value = val;
  }
  else cout << "Atom --> Cannot create the atom : value not in the variable domain\n";
  id = ide;
  ref_trans_pre = new std::vector<Transition*>();
  ref_trans_post = new std::vector<Transition*>();
  ref_comps = new hash_map<const char*, Comp*, hash<const char*>, eqstr> ();
   /*if (this->evalAtom()) stateVar->currentAtom = this;*/
  dispid = stateVar->addAtomToLinks(this); 
  label = "unknown";
  supported = new std::vector<Atom*>();
  ref_clauses = new std::vector<Clause*>();
  propagation_number = 0;
  neg = non;
  post_prob = 0;
  fail = false;
  guards_number = 0;
  idle = 0;
  cond = 0;
  dead_end = false;
  dyn_name = 0;
  topo_number = 0;
}
/* destructor */
Atom::~Atom() {
  delete ref_trans_post;
  ref_trans_post = 0;
  delete ref_trans_pre;
  ref_trans_pre = 0;
  delete ref_clauses;
  ref_clauses = 0;
  delete ref_clauses;
  ref_clauses = 0;
  if (dyn_name != 0)
    delete dyn_name;
}

/**
 * evaluate the atom.
 * @return true if the atom is set to its value, false otherwise.
 * This is used during knowledge compilation only.
 */
bool Atom::evalAtom() {
  if ((*stateVar).getCurrentState()==value) return true;
  else return false;
}

//set the atom true (modify the value of the corresponding state var).
void Atom::set() {
  stateVar->setCurrentState(value);
  /*stateVar->currentAtom = this;*/
  label = "true";
  for (unsigned int i=0; i<stateVar->getAtomLinks()->size(); i++) //set the other labels to 'false'.
    if (this !=(*stateVar->getAtomLinks())[i])
      (*stateVar->getAtomLinks())[i]->setLabel("false");
}
/**
 * set the atom's label.
 * @param label: true, false, unknown.
 */
void Atom::setLabel(const std::string &lab) {
  label = lab;
}
//set the label and update or not the label of the others atoms of the state variable.
void Atom::setLabel(const std::string &lab, const bool &update) {
  label = lab;
  if (update & lab == "true") { 
    for (unsigned int i=0; i<stateVar->getAtomLinks()->size(); i++)
      if (this != (*stateVar->getAtomLinks())[i])
      (*stateVar->getAtomLinks())[i]->setLabel("false");
  }
}
//add a transition to the ref of pre transitions.
void Atom::addRefTransPre(Transition* trans) {
  ref_trans_pre->push_back(trans);
}
//add a transition to the ref of post transitions.
void Atom::addRefTransPost(Transition* trans) {
  ref_trans_post->push_back(trans);
}
//show the Atom. bool ev, to show evaluation, or not.
void Atom::showAtom(bool ev) {
  if (ev) {
    cout << "--- Atom n# ---\n";
    cout << id << " --> [" << evalAtom() << "]\n";
    cout << "label : " << label << endl;
    cout << "propagation number : " << propagation_number << endl;
  }
  else cout << getId ();
}

void Atom::decTopoNumber() {
  topo_number--;
}

void Atom::incTopoNumber() {
  topo_number++;
}

int Atom::getTopoNumber() {
  return topo_number;
}
