/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Clause.h"
#include <iostream>

using namespace std;

/* Functions for class Clause. */

//constructors.
Clause::Clause(Comp* co) {
  cl = new multimap<bool, Atom*, eqbool>();
  ref_comp = co;
  internal = 0;
}
Clause::Clause(Atom* at, bool non, Comp* co) {
  cl = new multimap<bool, Atom*, eqbool>();
  cl->insert(pair<bool, Atom*>(non, at));
  ref_comp = co;
  internal = 0;
}
Clause::Clause(Clause* cla) {
  cl = new multimap<bool, Atom*, eqbool>();
  for (multimap<bool, Atom*, eqbool>::iterator it = cla->cl->begin(); it != cla->cl->end(); it++)
    cl->insert(pair<bool, Atom*>((*it).first, (*it).second));
  ref_comp = cla->ref_comp;
  internal = 0;
}
//destructor.
Clause::~Clause() {
}
//return the sign of an atom of the clause.
bool Clause::getSign(Atom* at) {
  for (multimap<bool, Atom*, eqbool>::iterator it = cl->begin(); it != cl->end(); it++)
    if ((*it).second == at)
      return (*it).first;
  return false; //BEWARE : return false if the atom has not been found...
}
//return the atoms labelled 'unknown' in a unit clause. DEPRECATED.
Atom* Clause::getUnknown() {
  for (multimap<bool, Atom*, eqbool>::iterator it = cl->begin(); it != cl->end(); it++)
    if ((*it).second->getLabel() == "unknown")
      return (*it).second;
  return NULL;
}
//return the atoms that are not labelled 'unknown'. DEPRECATED.
multimap<bool, Atom*, eqbool>* Clause::getNotUnknown() {
  multimap<bool, Atom*, eqbool>* res = new multimap<bool, Atom*, eqbool>();
  for (multimap<bool, Atom*, eqbool>::iterator it = cl->begin(); it != cl->end(); it++)
    if ((*it).second->getLabel() != "unknown")
      res->insert(pair<bool, Atom*>((*it).first, (*it).second));
  return res;
}
//return the atoms that are labelled 'true'.
multimap<bool, Atom*, eqbool>* Clause::getTrueAtoms() {
  multimap<bool, Atom*, eqbool>* res = new multimap<bool, Atom*, eqbool>();
  for (multimap<bool, Atom*, eqbool>::iterator it = cl->begin(); it != cl->end(); it++)
    if ((*it).second->getLabel() == "true")
      res->insert(pair<bool, Atom*>((*it).first, (*it).second));
  return res;
}
//return the atoms that are labelled 'false'.
multimap<bool, Atom*, eqbool>* Clause::getFalseAtoms() {
  multimap<bool, Atom*, eqbool>* res = new multimap<bool, Atom*, eqbool>();
  for (multimap<bool, Atom*, eqbool>::iterator it = cl->begin(); it != cl->end(); it++)
    if ((*it).second->getLabel() == "false")
      res->insert(pair<bool, Atom*>((*it).first, (*it).second));
  return res;
}
//add an atom and check the clause coherence... (also create a StateFormula...).
void Clause::addAtom(Atom* at, bool non) {
  cl->insert(pair<bool, Atom*>(non, at));
}

/* add atom and reference */
void Clause::addAtomRef(Atom* at, bool non) {
  cl->insert(pair<bool, Atom*>(non, at));
  at->ref_clauses->push_back (this);
}
//eval an atom (iterator is the position of the atom in the multimap).
string Clause::evalAtom(multimap<bool, Atom*, eqbool>::iterator it) {
  if ((*it).first & (*it).second->evalAtom()) return "false";
  else if ((*it).first & !(*it).second->evalAtom()) return "true";
  else if ((*it).second->evalAtom()) return "true";
  else if (!(*it).second->evalAtom()) return "false";
  else return 0;
}
//get opposites of an atom (of the clause...). TO BE OPTIMIZED...
//only if the clause is an horn clause...
vector<Atom*>* Clause::getOpposites(Atom* p) {
  bool sign_p;
  vector<Atom*>* res = new vector<Atom*>();
  
  sign_p = getSign(p); /* get the sign of p */
  for (multimap<bool, Atom*, eqbool>::iterator it = cl->begin(); 
       it != cl->end(); it++) { /* get the opposites */
    if ((*it).first != sign_p)
      res->push_back((*it).second);
    else if ((*it).second != p) {
      if (!res->empty()) {
	delete res;
	res = new vector<Atom*>();
      }
      return res;
    }
  }
  return res;
}

/* get the internal flag. */
bool Clause::getInternal() {
  return internal;
}

/* set the internal flag. */
void Clause::setInternal(bool b) {
  internal = b;
}

//set the clause as a support for an atom and update 'supported' for each atom in cl. DEPRECATED.
void Clause::setSupport(Atom* at) {
  at->support = this;
  for (multimap<bool, Atom*, eqbool>::iterator it = cl->begin(); it != cl->end(); it++)
    if ((*it).second != at)
      (*it).second->supported->push_back(at);
}
/**
 * show the clause on the standard output.
 */
void Clause::showClause(bool flag) {
  string sign;
  unsigned int f = 0;
  //if ((*cl->begin()).second == NULL) cout << "null\n";
  if (cl->empty ()) std::cout << "null\n";
  std::cout << "cl size: " << cl->size () << std::endl;
  for (multimap<bool, Atom*, eqbool>::iterator it = cl->begin(); it != cl->end(); it++) {
    if ((*it).first) sign = "non - " ;
    else sign = "";
    cout << sign << (*it).second->getId(); //"[" << (*it).second->getLabel() << "]";
    f++;
    if (f != cl->size()) cout << " \\/ ";
  }
  cout << endl;
}
