/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Clause = vector of propositions (StateFormula with 'and' operator) representing a disjunction of literals.*/

#ifndef CLAUSE_H
#define CLAUSE_H

#include "Atom.h"
#include <string>
#include <vector>
#include <map>

using namespace std;

class Comp;

/**
 * \struct eqbool
 * \brief boolean equality operator for hash_map.
 */

#ifndef EQBOOL
#define EQBOOL
struct eqbool
  {
    bool operator()(bool s1, bool s2)
    {
      return (s1 == s2);
    }
};
#endif

/**
 * \class Clause
 * \brief represent a clause, i.e. a disjunctive set of logical propositions, that can be negated or not.
 */

class Clause {

 public:
  multimap<bool, Atom*, eqbool>* cl;  /**< clause. */
  Comp* ref_comp;  /**< reference to the component it is defined in. */
  bool internal;  /**< internal / external flag. */
  
  Clause(Comp* co);
  Clause(Atom* at, bool non, Comp* co); 
  Clause(Clause* cla);
  ~Clause();
  bool getSign(Atom* at);                         /**< return the sign of an atom of the clause. */
  Atom* getUnknown();                             /**< return the atom labelled 'unknown' in a unit clause. */
  multimap<bool, Atom*, eqbool>* getNotUnknown(); /**< return the atoms that are not labelled 'unknown'. */
  multimap<bool, Atom*, eqbool>* getTrueAtoms();
  multimap<bool, Atom*, eqbool>* getFalseAtoms();
  void addAtom(Atom* at, bool non); /**< add an atom and check the clause coherence... */
  void addAtomRef(Atom* at, bool non);
  string evalAtom(multimap<bool, Atom*, eqbool>::iterator it);
  vector<Atom*>* getOpposites(Atom* p); /**< return the atoms that have the opposite sign of p. */
  void setInternal (bool b);
  bool getInternal ();
  void setSupport(Atom* at);
  void showClause(bool flag);

};

#endif
