/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Comp.h"
#include <iostream>

using namespace std;

/* Functions for class Component */

/* constructors */
Comp::Comp() {
}
Comp::Comp(string na) {
  name = na;
  state_variables = new vector<StateVar*>();
  constraints = new vector<Formula*>();
  component_transitions = new vector<Transition*>();
  atoms = new hash_map<const char*, Atom*, hash<const char*>, eqstr>();
  influences = new vector<influence>();
  variables = new vector<variable>();
  atoms_number = 0;
  var_name = "";
  instanciated = false;
  generated = false;
  share_commanded = false;
  cmd_in = NULL;
  influListcaen = NULL;
  gVarListcaen = NULL;
  initListcaen = NULL;
  globCListcaen = NULL;
  condCListcaen = NULL;
  corres_import_comps = NULL;
  imported_comps = NULL;
  numClauses = 0;
}
//destructor.
Comp::~Comp() {
  purge(*state_variables);
  delete state_variables;
  state_variables = 0;
  purge(*constraints);
  delete constraints;
  constraints = 0;
  purge(*component_transitions);
  delete component_transitions;
  component_transitions = 0;
  delete atoms;
  delete influences;
  delete variables;
  if (corres_import_comps != NULL)
    delete corres_import_comps;
  if (imported_comps != NULL)
    delete imported_comps;
  if (cmd_in != NULL)
    delete cmd_in;
}
//Access to variables
//set the state_variables
void Comp::setStateVariables(vector<StateVar*>* sta) {
  state_variables = sta;
}
//add a state variable to the vector of state variables
void Comp::addStateVar(StateVar* sta_var) {
  sta_var->setComp(this);
  state_variables->push_back(sta_var);
}
/* add a continuous variables. */
void Comp::addVariable(variable va) {
  variables->push_back(va);
}
//return all current atoms.
/*vector<Atom*>* Comp::getAllCurrentAtoms() {
  vector<Atom*>* res = new vector<Atom*>();
  for (unsigned int i=0; i<state_variables->size(); i++)
    res->push_back((*state_variables)[i]->currentAtom);
  return res;
}*/
//compute nominal cover
void Comp::computeNominalCover() {
  int count = 0;
  for (unsigned int i=0; i<component_transitions->size(); i++) {
    if ((*component_transitions)[i]->type == "Nominal") {
      count++;
      nominal_cover = nominal_cover + (*component_transitions)[i]->proba;
    }
  }
  nominal_cover = (double)nominal_cover / (double)count;
}
//set the vector of feasible states
void Comp::setConstraints(vector<Formula*>* fea) {
  constraints = fea;
}
  //return the vector of feasible states
vector<Formula*>* Comp::getConstraints() {
  return constraints;
}
//add a Formula to feasible states
void Comp::addFormulaToCS(Formula* frm) {
  constraints->push_back(frm);
}
//add an influence
void Comp::addInfluence(influence inf) {
  influences->push_back(inf);
}
//set the vector of transitions
void Comp::setCompTransitions(vector<Transition*>* trans) {
  component_transitions = trans;
}
//add a Transition to the component's transitions
void Comp::addTransitionToC(Transition* trans) {
  component_transitions->push_back(trans);
}
//set the name of the component
void Comp::setName(string na) {
  name = na;
}

/**
 * Tools & debug.
 */
void Comp::getTransitionsGuards() {
  for (unsigned int i=0; i<component_transitions->size(); i++) {
    if ((*component_transitions)[i]->proba == 1) {
      (*component_transitions)[i]->pre_at->guards_number++;
      (*component_transitions)[i]->modevar->guards_number++;
    }
    else {
      for (unsigned int j=i+1; j<component_transitions->size(); j++) {
	if ((*component_transitions)[i]->pre->eq((*component_transitions)[j]->pre)) {
	  (*component_transitions)[i]->pre_at->guards_number++;
	  (*component_transitions)[i]->modevar->guards_number++;
	}
      }
    }
  }
}

string Comp::findCorresDef (string name) {
  for (unsigned int i=0; i<corres_import_comps->size (); i++) {
    if (name == (*corres_import_comps)[i].second)
      return (*corres_import_comps)[i].first;
  }
  return "";
}

string Comp::findCorresName (string defname) {
  for (unsigned int i=0; i<corres_import_comps->size (); i++) {
    if (defname == (*corres_import_comps)[i].first)
      return (*corres_import_comps)[i].second;
  }
  return "";
}

/**
 * debug & visu.
 */
//display the component properties
void Comp::showComp() {
  cout << "\n***************************************************" << endl;
  cout << "Comp : " << getName() << endl;
  cout << "***************************************************" << endl;
  /*cout << "nominal cover : " << nominal_cover << endl;*/
  cout << "imported components:\n";
  if (imported_comps != NULL) {
    for (unsigned int i=0; i<imported_comps->size (); i++)
      cout << (*imported_comps)[i]->var_name << endl;
  }
  cout << "States : \n";
  for (unsigned int i=0;i<state_variables->size();i++) {
    cout << (*state_variables)[i]->name << " - " << (*state_variables)[i]->getType() << " -" << endl;
    (*state_variables)[i]->showValues();
    if ((*state_variables)[i]->getType () == "Cond")
      ((CondVar*) (*state_variables)[i])->showConditions ();
    (*state_variables)[i]->showCurrentState();
    /*if ((*state_variables)[i]->type == "Cond") {
      CondVar* cvtmp = (CondVar*)(*state_variables)[i];
      cout << cvtmp->condvar << endl;
      }*/
  }
  /*cout << endl;
    for (unsigned int i=0; i<variables->size(); i++) {
    printVariable ((*variables)[i]);
    }*/
  cout << "\n\ngvars: " << listcaenGVarLn (this->gVarListcaen) << endl;
  listcaenGVar lg = gVarListcaen;
  while (! isNull (lg)) {
    printGVar (stdout, getLGGVar (lg));
    cout << endl;
    lg = getLGNext (lg);
  }
  cout << "\nglobal constraints: " << listcaenExprLn (globCListcaen) << endl;
  listcaenExpr leg = globCListcaen;
  while (! isNull (leg))
    {
      printExpr (getLEExpr (leg));
      cout << endl;
      leg = getLENext (leg);
    }
  cout << "\ninitvars: " << listcaenInitVarLn (initListcaen) << endl;
  cout << "cond exprs: " << listcaenExprLn (condCListcaen) << endl;
  listcaenExpr le = condCListcaen;
  while (! isNull (le)) {
    printExpr (getLEExpr (le));
    cout << endl;
    le = getLENext (le);
  }
  
  cout << "\n---------------------------------------------------";
  cout << "\nConstraints : \n";
  if (constraints != NULL) {
    for (unsigned int i=0; i<constraints->size(); i++)
      (*constraints)[i]->showFormula();
  }
  cout << "---------------------------------------------------";
  cout << "\nInfluences : \n";
  for (unsigned int i=0; i<influences->size(); i++) {
    printInfluence ((*influences)[i]);
  }
  cout << "---------------------------------------------------";
  cout << "\nTransitions : \n";
  for (unsigned int i=0; i<component_transitions->size(); i++) {
    if (component_transitions == NULL) cout << "NULLL\n";
    (*component_transitions)[i]->showTransition();
  }
  cout << "***************************************************\n";
}
