/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "StateFormula.h"
#include <iostream>

using namespace std;

/* Functions for class StateFormula */

/**
 * empty constructor for lightest initialization.
 */
StateFormula::StateFormula() {
  state_formulas = NULL;
  atoms = new multimap<bool, Atom*, eqbool>();
  operat = "";
  neg = false;
  compiled_num = -1;
}
/**
 * Constructor for single atom initialization.
 * @param at is the atom the state formula is initially constructed with.
 */
StateFormula::StateFormula(Atom* at) {
  state_formulas = NULL;
  atoms = new multimap<bool, Atom*, eqbool>();
  atoms->insert(pair<bool, Atom*>(1, at));
  operat = "";
  neg = false;
  compiled_num = -1;
}
//only one atom, with negation. DEPRECATED ?
StateFormula::StateFormula(Atom* at, bool ne) {
  state_formulas = NULL;
  atoms = new multimap<bool, Atom*, eqbool>();
  atoms->insert(pair<bool, Atom*>(ne, at));
  operat = "";
  neg = false;
  compiled_num = -1;
}
//destructor.
StateFormula::~StateFormula() {
  if (state_formulas != NULL)
    purge(*state_formulas);
  delete state_formulas;
  state_formulas = 0;
  delete atoms;
  atoms = 0;
}
//get the operator for the state formula
string StateFormula::getOperat() {
  return operat;
}
//set the operator for the state formula
void StateFormula::setOperat(string op) {
  if ((op=="and") | (op=="or")) 
    operat = op;
  else return;
}
multimap<bool, Atom*, eqbool>* StateFormula::getAtoms() {
  return atoms;
}
//add an atom to the set of atoms
void StateFormula::addAtom(Atom* at) {
  atoms->insert(pair<bool, Atom*>(1, at));
}
void StateFormula::addAtom(Atom* at, bool sign) {
  atoms->insert(pair<bool, Atom*>(sign, at));
}
//return all atoms of the state formula (recursive or not)
vector<Atom*>* StateFormula::getAllAtoms() {
  vector<Atom*>* res = new vector<Atom*>();
  copyAtoms(res);
  return res;
}
//return all atoms, but with their sign.
multimap<bool, Atom*, eqbool>* StateFormula::getAllSignedAtoms() {
  multimap<bool, Atom*, eqbool>* res = new multimap<bool, Atom*, eqbool>();
  copyAtoms(res);
  return res;
}
void StateFormula::copyAtoms(vector<Atom*>* vec) {
  //theoretically, it is not possible to find the same atom many times in the same state formula...
  if (atoms!=NULL) {
    for (multimap<bool, Atom*, eqbool>::iterator i = atoms->begin(); i != atoms->end(); i++) {
      vec->push_back((*i).second);
    }
  }
  if (state_formulas!=NULL) {
    for (unsigned int i=0; i<state_formulas->size(); i++)
      (*state_formulas)[i]->copyAtoms(vec);
  }
}
void StateFormula::copyAtoms(multimap<bool, Atom*, eqbool>* mult) {
  if (atoms != NULL)
    for (multimap<bool, Atom*, eqbool>::iterator i = atoms->begin(); i != atoms->end(); i++)
      mult->insert(pair<bool, Atom*>((*i).first, (*i).second));
  if (state_formulas != NULL)
    for (unsigned j=0; j<state_formulas->size(); j++)
      (*state_formulas)[j]->copyAtoms(mult);
}
//return the vector of StateFormulas
vector<StateFormula*>* StateFormula::getStateFormulas() {
  return state_formulas;
}
//add a StateFormula
void StateFormula::addStateFormula(StateFormula* stafrm) {
  if (state_formulas==NULL) 
    state_formulas = new vector<StateFormula*>();
  state_formulas->push_back(stafrm);
}
/* non recursive equality test. */
bool StateFormula::eq(StateFormula* stf) {
  bool temp;
  multimap<bool, Atom*, eqbool>* all_atoms1 = getAllSignedAtoms();
  multimap<bool, Atom*, eqbool>* all_atoms2 = stf->getAllSignedAtoms();
  for (multimap<bool, Atom*, eqbool>::iterator it1 = all_atoms1->begin(); it1 != all_atoms1->end(); it1++) {
    temp = false;
    for (multimap<bool, Atom*, eqbool>::iterator it2 = all_atoms2->begin(); it2 != all_atoms2->end(); it2++) {
      if ((*it1).first == (*it2).first && (*it1).second == (*it2).second) {
	temp = true;
      }
    }
    if (!temp)
      return false;
  }
  delete all_atoms1;
  delete all_atoms2;
  return true;
}

//show the state formula
void StateFormula::showStateFormula() {
  if (neg) cout << "non - ";
  if ((atoms!=NULL) & (state_formulas!=NULL)) {
    for (unsigned int j=0; j<state_formulas->size(); j++) {
      (*state_formulas)[j]->showStateFormula();
      cout << " " << operat << " ";
    }
    cout << "(";
    for (multimap<bool, Atom*, eqbool>::iterator i = atoms->begin(); i != atoms->end(); i++) {
      if (!(*i).first) cout << "!";
      (*i).second->showAtom(0);
    }
    cout << ")";
  }
  else if (atoms!=NULL) {
    cout << "(";
    for (multimap<bool, Atom*, eqbool>::iterator i = atoms->begin(); i != atoms->end(); i++) {
      if (!(*i).first) cout << "!";
      (*i).second->showAtom(0);
      cout << " " << operat << " ";
    }
    cout << ")";
  }
  else if (state_formulas!=NULL) {
    for (unsigned int j=0; j<state_formulas->size(); j++) {
      (*state_formulas)[j]->showStateFormula();
      cout << " " << operat << " ";
    }
  }
}

