/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* State formula 
ex : ((mode="on" | mode="off") & (cmdin="open"))
     ((StateFormula) & StateFormula = Atom) = StateFormula
*/

#ifndef STATEFORMULA_H
#define STATEFORMULA_H

#include "Atom.h"
#include "../system/purge.h"
#include <string>
#include <vector>
#include <map>

#ifndef EQBOOL
#define EQBOOL
struct eqbool
  {
    bool operator()(bool s1, bool s2)
    {
      return (s1 == s2);
    }
};
#endif

class StateFormula {
 public:
  string operat;
  vector<StateFormula*>* state_formulas;
  multimap<bool, Atom*, eqbool>* atoms;
  bool neg; //DEPRECATED ?
  int compiled_num;  /**< compilation number */
  
 public:
  StateFormula();
  StateFormula(Atom* at); //in case the state formula is only an atom
  StateFormula(Atom* at, bool ne); //only one atom, with negation.
  ~StateFormula();
   string getOperat();
  void setOperat(string op);
  multimap<bool, Atom*, eqbool>* getAtoms();
  void addAtom(Atom* at);
  void addAtom(Atom* at, bool sign);
  vector<Atom*>* getAllAtoms(); //return all atoms of the state formula (recursive or not)
  multimap<bool, Atom*, eqbool>* getAllSignedAtoms();
  void copyAtoms(vector<Atom*>* vec);
  void copyAtoms(multimap<bool, Atom*, eqbool>* mult);
  vector<StateFormula*>* getStateFormulas();
  void addStateFormula(StateFormula* stafrm);
  bool eq(StateFormula* stf);
  void showStateFormula();
};

#endif
