/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "TrajElrt.h"
#include <iostream>

using namespace std;

/* static variables initialization. */
int TrajEl::trajel_created = 0;
int TrajEl::trajel_del = 0;

/* Functions for class TrajEl. */

//constructors.
TrajEl::TrajEl() {
  nextTrans = new map<int, Transition*>(); //not needed for DependentVar...
  nextTransNeg = new map<int, Transition*>();
  supported = new vector<TrajEl*>();
  support = NULL;
  prevTrans = NULL;
  lock = 2;
  label = "";
  propagation_number = 0;
  prev_step = 0;
  prevTrans_time = -1;
  proba = 1;
  trajel_created++;
}
TrajEl::TrajEl(TrajEl* el) {
  at = el->at;
  prev_step = el->prev_step;
  prevTrans = el->prevTrans;
  nextTrans = new map<int, Transition*>();
  nextTransNeg = new map<int, Transition*>();
  for (map<int, Transition*>::iterator i = el->nextTrans->begin(); i != el->nextTrans->end(); ++i) {
    if ((*i).second != NULL)
      (*nextTrans)[(*i).first] = (*i).second;
  }
  for (map<int, Transition*>::iterator i = el->nextTransNeg->begin(); i != el->nextTransNeg->end(); ++i) {
    if ((*i).second != NULL)
      (*nextTransNeg)[(*i).first] = (*i).second;
  }
  label = el->label;
  if (el->support)
    {
      support = el->support;
      
      //debug
      /* std::cout << "new elt: " << getId () << " support clause: ";
	 support->showClause (true); */
      //debug
    }
  else support = NULL;
  propagation_number = el->propagation_number;
  prevTrans_time = el->prevTrans_time;
  supported = new vector<TrajEl*>(); //BEWARE !!! the vector is not recopied here...
  lock = el->lock; //lock is copied...
  proba = el->proba;
  trajel_created++;

  //debug
  /* std::cout << "elt created:\n";
     showTrajEl (true); */
  //debug
}

TrajEl::TrajEl(Atom* a, Transition* tprev, map<int, Transition*>* tnext, 
	       const std::string &lab) {
  at = a;
  prev_step = 0;
  prevTrans = tprev;
  nextTrans = new map<int, Transition*>();
  nextTransNeg = new map<int, Transition*>();
  if (tnext != NULL) {
    for (map<int, Transition*>::iterator i = tnext->begin(); i != tnext->end(); i++) {
      if ((*i).second != NULL)
	(*nextTrans)[(*i).first] = (*i).second;
    }
  }
  label = lab;
  supported = new vector<TrajEl*>();
  support = NULL;
  propagation_number = 0;
  lock = 2;
  prevTrans_time = -1;
  proba = 1;
  trajel_created++;
}
/* destructor. */
TrajEl::~TrajEl() {
  delete nextTrans;
  nextTrans = 0;
  delete nextTransNeg;
  nextTransNeg = 0;
  delete supported;
  supported = 0;
  trajel_del++;
}

/**
 * set the atom.
 * @param a is the atom to be linked with the element.
 */
void TrajEl::setAtom(Atom* a) {
  at = a;
}
bool TrajEl::eq(const TrajEl& el) {
  return at == el.at;
}
bool TrajEl::evalTrajEl() {
  if (label == "true") return true;
  else return false;
}
//get an atom from the state variable.
Atom* TrajEl::getAtomLink(const std::string &id) {
  return at->getAtomLink(id);
}
/* add a transition to next trans and update the set. */
void TrajEl::addToNextTrans(Transition* trans) {
  /* delete transitions associated to the same mode variable as trans. */
  std::vector<int> l_Tmp;
  for (map<int, Transition*>::iterator i = nextTrans->begin(); i != nextTrans->end(); i++) {
    if ((*i).second != NULL) {
      if ((*i).second->modevar == trans->modevar) {
	//nextTrans->erase((*i).first);
	l_Tmp.push_back ((*i).first);
      }
    }
  }
  for (std::vector<int>::iterator iter = l_Tmp.begin ();
       iter != l_Tmp.end (); iter++)
    nextTrans->erase ((*iter));
  /* add new reference. */
  
  //debug
  /* std::cout << "adding next trans: ";
     trans->showTransition (); */
  //debug

  (*nextTrans)[trans->trans_id] = trans;
}
void TrajEl::addToNextTransNeg(Transition* trans) {
  /* delete transitions associated to the same mode variable as trans. */
  std::vector<int> l_Tmp;
  for (map<int, Transition*>::iterator i = nextTransNeg->begin(); i != nextTransNeg->end(); i++)
    if ((*i).second->modevar == trans->modevar)
      l_Tmp.push_back ((*i).first);
      //nextTransNeg->erase((*i).first);
  for (std::vector<int>::iterator iter = l_Tmp.begin ();
       iter != l_Tmp.end (); iter++)
    nextTrans->erase ((*iter));
  /* add new reference. */
  (*nextTransNeg)[trans->trans_id] = trans;
}
/* update the probability of the elt. */
void TrajEl::updateProba(const double &prob) {
  proba = proba * prob;
}
//show the atom. bool ev, to show evaluation or not.
void TrajEl::showAtom(const bool &ev) {
  if (ev) {
    cout << "--- Atom n# ---\n";
    cout << at->getId() << " --> [" << at->evalAtom() << "]\n";
    cout << "label : " << label << endl;
    cout << "propagation number : " << propagation_number << endl;
    cout << "lock : " << lock << endl;
  }
  else cout << at->getId();
}
//show the elements.
void TrajEl::showTrajEl(const bool &ev) {
  showAtom(ev);
  cout << "prevTrans(" << prev_step << ") :\n";
  if (prevTrans != NULL) {
    prevTrans->showTransition();
  }
  for (std::map<int, Transition*>::const_iterator it = nextTrans->begin ();
       it != nextTrans->end (); ++it)
    {
      if ((*it).second)
	(*it).second->showTransition();
    }
  cout << "support: " << support << endl;
  if (support) support->showClause (true);
}
