
/*    
 *  Copyright 1999, 2000, 2001, 2002, 2003 Emmanuel Benazera, Jean-Claude Gouyon
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* caen-algo.c */
#include "caen-algo.h"
#include <strstream>

/* algo-level temps, for macro calls only */
#ifdef DEBUG
static long L1;
#endif
static listcaenLong Ll1;
static listcaenVar Lv1;
static cGraph Cg1;

long done;
unsigned int numClauses = 0;
unsigned int numStf = 0;
unsigned int numTransitions = 0;
unsigned int numGv = 0;
unsigned int clFilesNum = 0;
string nodestr, initstr, gvarstr, condstr, condcstr, globstr, posCond, cross_ref, mfexprstr;
bool in_gen_types = false;

/* 1. caen-load-------------------------------------------------------*/

arDef symCons;
arDef conSymb;
char appliName[80], tmpDir[80];
char *modFileName[MAXMOD];
listcaenParFor modListcaenPar[MAXMOD];
long iModFile=0;
char *nomComp=NULL;

/*  */
void cleanDefs (void) {
  listcaenParFor lpf = tempDefs;

  while (! isNull (lpf)) {
    delDef (&symCons, getParamForNom (getLPFParFor (lpf)));
    lpf = getLPFNext (lpf);
  } /* endif while */
  return;
}

/*
 * merges data lists from components
 */
void mergeLists () {
  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it = reasoner->active_comps->begin(); it != reasoner->active_comps->end(); it++) {
    influListcaen = addListcaenInfl (influListcaen, (*it).second->influListcaen);
    gVarListcaen = addListcaenGVar (gVarListcaen, (*it).second->gVarListcaen);
    globCListcaen = addListcaenExpr (globCListcaen, (*it).second->globCListcaen);
    condCListcaen = addListcaenExpr (condCListcaen, (*it).second->condCListcaen);
    initListcaen = addListcaenInitVar (initListcaen, (*it).second->initListcaen);
  }
}

/*
 * set global constraints link to global variables
 */
void buildGlobCOfGv (listcaenGVar lg) {     /* listcaen of global constraints gv is arg of */
  long i;

  while (! isNull (lg)) {
    listcaenLong la=NULL;
    gVar gv=getLGGVar (lg);
    listcaenExpr le=globCListcaen;
    i = 0;
    while (! isNull (le)) {
      expr e = getLEExpr (le);
      if (hasGvAsArg (e, gv))
	la = catListcaenLong (la, mkListcaenLong (i));
      i++, le = getLENext (le);    } /* end while ! isNull le */
    setGVarLExpr (gv, la);
    lg = getLGNext (lg);
  } /* end while ! isNull lg */
  return;
}

/* TODO: must iterate through imported comps */
void buildCondCOfGv (listcaenGVar lg) {
  long i;

  while (! isNull (lg)) {
    listcaenLong la=NULL;
    gVar gv=getLGGVar (lg);
    listcaenExpr le=condCListcaen;
    i = 0;
    while (! isNull (le)) {
      expr e = getLEExpr (le);
      if (hasGvAsArg (e, gv)) {
	la = catListcaenLong (la, mkListcaenLong (i));
      }
      i++, le = getLENext (le);
    } /* end while ! isNull le */
    setGVarLcExpr (gv, la);
    lg = getLGNext (lg);
  } /* end while ! isNull lg */
  return;
}

StateVar* findDiscVariable (Comp* co, string s) {
  StateVar* va;

  for (unsigned int i=0; i<variables_disc->size(); i++) {
    if (s == (*variables_disc)[i].first)
      return (*variables_disc)[i].second;
  }
  
  if (! isNull (co->imported_comps)) {
    for (unsigned int j=0; j<co->imported_comps->size(); j++) {
      if (!isNull (va = findDiscVariable (co, (*co->imported_comps)[j], s)))
	return va;
    }
  }
  else {
    semError (1, 0, 73, s.c_str ());
    return NULL;
  }
   return NULL;
}

StateVar* findDiscVariable (Comp* co, Comp* co_imp, string s) {
  string co_imp_name = co->findCorresDef (co_imp->getVarName ());
  for (unsigned int i=0; i<variables_disc->size(); i++) {
    if ((co_imp_name + "." + s) == (*variables_disc)[i].first)
      return (*variables_disc)[i].second;
  }
  semError (1, 0, 73, s.c_str ());
  return NULL;
}

Atom* findAtom (string other_name, Comp* co) {
  int dot_pos = other_name.find (".");
  string co_name = other_name.substr (0, dot_pos);
  string co_var_name, stv_name;
  Comp* co_imp;
  StateVar* stv;

  co_imp = findComp (co_name);
  stv_name = other_name.substr (0, other_name.find ("="));
  stv = findDiscVariable (co_imp, stv_name);
  for (unsigned int j=0; j<stv->atoms_links->size (); j++) {
    if ((*stv->atoms_links)[j]->id == other_name)
      return (*stv->atoms_links)[j];
  }
  return NULL;
}

Comp* findComp (string var_name) {
  for (unsigned int i=0; i<comps->size(); i++) {
    if (var_name == *(*comps)[i]->var_name)
      return (*comps)[i]->com;
  }
  return NULL;
}

/* 2. caen-compile----------------------------------------------------*/

void buildCGraph (arBin p) {

  if (! isNull (p)) {
    buildCGraph (getArBinGauche (p));
    buildCNode (getArBinVar (p));    buildCGraph (getArBinDroite (p));
  }
  return;
}


influence cause[MAXCAUSE];
influence effet[MAXEFFET];
long ic, ie;

void buildCNode (variable v) {
  listcaenInfl li=influListcaen;
  influence ifl;
  cGraph cg;
  long i;

  ic = ie = 0;
  while (! isNull (li)) {
    ifl = getLIInfl (li);
      if (getInfluCause (ifl) == v)
	effet [ie++] = ifl;
      if (getInfluEffet (ifl) == v) {
	if (isVarConst (v))
	  semError (0, nligne, 39);
	/* parameter shouldn't be effect in an influence. */
	else if (isVarInput (v))
	  semError (0, nligne, 54);
	/* "Input variable shouldn't be effect in an influence. */
	else
	  cause [ic++] = ifl;
      } /* endif getInfluEffet... */
    li = getLINext (li);
  } /* end while */
  if (ic || ie) {
    cg = mkCGraph (ic, ie);
    for (i=0; i<ic; ++i)
      setCGCauseNth (cg, i, cause[i]);
    for (i=0; i<ie; ++i)
      setCGEffetNth (cg, i, effet[i]);
    setVarNoeud (v, cg);
  } /* endif ic || ie */
  return;
}


long numTree, maxDepth;

bool compCGraphDepths (arBin p) {
  if (! isNull (p)) {
    if (! compCGraphDepths (getArBinGauche (p)))
      return 0;
    if (! compVarDepth (getArBinVar (p)))
      return 0;
    if (! compCGraphDepths (getArBinDroite (p)))
      return 0;
  }
  return 1;
}


bool compVarDepth (variable v) {
  cGraph cg=getVarNoeud(v);

  if (! isNull (cg)) {
    long nc=getCGNCause(cg);
    influence fl;
    long i, isTemporized=1, isStatic=0, isDynamic=0, isIntegral=0, isDynamic2=0, eqTRep=1;
    objet tr = NULL;
    for (i=0; i<nc; ++i) {
      fl=getCGCauseNth (cg, i);
      switch (getInfluType (fl)) {
      case DINFL :
	isDynamic=1;
	if (i == 0)  /* Beware: error in case influences have different types ! */
	  tr = getInfluTRep (fl);
	else {
	  if (! eqObj (tr, getInfluTRep (fl)))
	    eqTRep = 0;
	}
	break;
      case D2INFL :
	isDynamic2=1;
	/* not sure about the following. (beniz 25/05/01). */
	/*if (i == 0)
	  tr = getInfluTRep (fl);
	  else {
	  if (! eqObj (tr, getInfluTRep (fl)))
	  eqTRep = 0;
	  }*/
	break;
      case IINFL :
	isIntegral=1;
	break;
      case SINFL :
      case CINFL :
	isStatic=1;
	if ((REAL == getObjType (getInfluDelay (fl)))
	    && (0.0 == getObjNVal (getInfluDelay (fl))))
	  isTemporized=0;
	break;
      default:
	break;
      } /* end switch */
    } /* end for */
    if (isDynamic) {
      if (isIntegral || isStatic) {
	semError (1, 0, 40);
	/* incompatible types in influences to the same variable. */
	fprintf (stderr, "  ");
	printVarNom (stderr, v);
	fprintf (stderr, "\n");
      }
      else {
	setCGCauseType (cg, DINFL);
	if (! eqTRep)
	  semError (1, 0, 41); /* "unequal response times to the same variable:\
				  please explicitly introduce marginal variables. */
      }
    }
    else if (isDynamic2) {
      setCGCauseType (cg, D2INFL);
    }
    else { /* isDynamic */
      if (isIntegral && isStatic) {
	semError (1, 0, 40);
	/* incompatible types in influences to the same variable. */
	fprintf (stderr, "  ");
	printVarNom (stderr, v);
	fprintf (stderr, "\n");
      }
      else {
	if (isIntegral)
	  setCGCauseType (cg, IINFL);
	if (isStatic)
	  setCGCauseType (cg, SINFL);
      }
    } /* endif isDynamic */
    if (isTemporized) {
      ++numTree;
      if (! compTreeDepths (v, 0))
	return 0;
    }
  } /* endif ! isNull cg */
  return 1;
}

/*  */
bool compTreeDepths (variable v, long d) {
  cGraph cg=getVarNoeud(v);
  long i, ne=getCGNEffet(cg);
  
  if (numTree == getVarTree (v)) {
    semError (1, 0, 53);     /* "closed loop in causal graph around variable: " */
    fprintf (stderr, "  ");
    printVarNom (stderr, v);
    fprintf (stderr, "\n");
    return 0;
  }
  else { /* numTree */
    if (d > getVarDepth (v)) {
      setVarDepth (v, d);
      if (d > maxDepth)
	maxDepth = d;
      for (i=0; i<ne; ++i) {
	influence fl=getCGEffetNth(cg,i);
	if (isUnTemporal (fl)) {          /* type SINFL ou CINFL, et delai nul */
	  setVarTree (v, numTree);
	  if (! compTreeDepths (getInfluEffet (fl), d+1))
	    return 0;
	  ++numTree;
	} /* endif isUnTemporal fl */
      } /* end for */
    } /* endif d */
  } /* endif numTree */
  return 1;
}


void sortCGraph (arBin p) {

  if (! isNull (p)) {
    sortCGraph (getArBinGauche (p));
    sortVariable (getArBinVar (p));
    sortCGraph (getArBinDroite (p));
  }
  return;
}


listcaenVar depthListcaen[MAXDEPTHLISTCAEN];

void sortVariable (variable v) {

  if (! (isVarConst (v) || isVarInput (v))) {
    long d=getVarDepth(v);
    if (d >= 0)
      depthListcaen[d] = catListcaenVar (depthListcaen[d], mkListcaenVar (v));
  }
  return;
}


void checkInits (listcaenVar lv) {

  while (! isNull (lv)) {
    variable v=getLVVar(lv);
    if (isNull (isInInitListcaen (v, initListcaen)) && ! isVarMeasure (v)) {
      semError (3, 0, 46);    /* "variable should be initialized or measured."  */
      fprintf (stderr, "  ");
      printVarNom (stderr, v);
      fprintf (stderr, "\n");
    }
    lv = getLVNext (lv);
  }
  return;
}


/* 3. caen-generate----------------------------------------------------*/

char appliPath[80], runTDir[80], runTFile[80], runTHead[80],
  rTi1File[80], rTi2File[80], defDir[80], defHFile[80], defCPPFile[80],
  clHFile[80], clCPPFile[80];
FILE *runT, *runH, *rTi1, *rTi2, *defCPP, *defH, *clH, *clCPP;
long numVar, numInput, numMeasure, numDisplay, numTemporal,
  numUntemporal, numInflu, numNoeud, numInit, numGVar, numGVarSV, numGlobC, 
  numConditions, numCondC, numVarG, numNoeudG, numInfluG, numMFExpr;
double timeStep;

bool inVector(string s, vector<string>* vec) {
  for (unsigned int i=0; i<vec->size(); i++)
    if (s == (*vec)[i])
      return 1;
  return 0;
}

bool isCommandComp (Comp* co, string ty) {
  if (! isNull (co->cmd_in)) {
    for (unsigned int i=0; i<co->cmd_in->size (); i++) {
      if ((*co->cmd_in)[i]->comp_lk->type == ty)
	return true;
    }
    return false;
  }
  else return false;
}

bool isIncludedInComp (Comp* co, string ty) {
  vector <Comp*>* nc;
  if (! isNull (co->imported_comps)) {
    nc = new vector<Comp*>(*co->imported_comps);
    nc->push_back (co);
    if (isCommandComp (co, ty))
      return true;
    for (unsigned int i=0; i<co->imported_comps->size (); i++) {
      if ((*co->imported_comps)[i]->type == ty) {
	delete nc;
	return true;
      }
      else {    /* process recursively */
	/*nc->push_back ((*co->imported_comps)[i]);*/
	if (isIncludedInComp ((*co->imported_comps)[i], ty, nc)) {
	  delete nc;
	  return true;
	}
      }
    }
    delete nc;
    return false;
  }
  else return false;
}

bool isIncludedInComp (Comp* co, string ty, vector<Comp*>* nc) {
  bool in_nc = false;
  
  if (isCommandComp (co, ty))
    return true;
  
  if (! isNull (co->imported_comps)) {
    for (unsigned int i=0; i<co->imported_comps->size (); i++) {
      in_nc = false;
      for (unsigned int j=0; j<nc->size (); j++)   { /* if not in nc. */
	if ((*nc)[j] == (*co->imported_comps)[i]) {
	  in_nc = true;
	  break;
	}
      }
      if ((*co->imported_comps)[i]->type == ty)
	return true;
      else if (! in_nc) {    /* process recursively */
	nc->push_back ((*co->imported_comps)[i]);
	if (isIncludedInComp ((*co->imported_comps)[i], ty, nc))
	  return true;
      }
    }
    return false;
  }
  else return false;
}

/* update elts numbers when a component's class has already been generated . */
void updateEltsNum (Comp* co) {
  Comp* cot = 0;
  Transition *trt=0, *tr=0;
  expr e, et;
  listcaenExpr le, let;
  gVar gv, gvt;
  listcaenGVar lg, lgt;
  MapFunc *mf, *mft;

  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator comps_it = reasoner->active_comps->begin(); 
       comps_it != reasoner->active_comps->end(); comps_it++) {
    if ((*comps_it).second->type == co->type) {
      cot = (*comps_it).second;
      break;
    }
  }
  
  if (isNull (cot)) {
    cout << "can't find the component from which the class has been generated: " << co->type << endl;
    return;
  }

  for (unsigned int i=0; i<cot->component_transitions->size (); i++) {  /* update transitions number */
    trt = (*cot->component_transitions)[i];
    tr = (*co->component_transitions)[i];
    tr->compiled_num = trt->compiled_num;
    tr->pre->compiled_num = trt->pre->compiled_num;
    tr->post->compiled_num = trt->post->compiled_num;

    if (! isNull (trt->mf)) {
      for (unsigned int j=0; j<trt->mf->size (); j++) {
	mft = (*trt->mf)[j];
	mf = (*tr->mf)[j];
	for (unsigned int k=0; k<mft->mvs->size (); k++) {
	  et = getMFExpr ((*mft->mvs)[k]);
	  e = getMFExpr ((*mf->mvs)[k]);
	  updateExprNum (e, et);
	}
      }
    }
  }

  let = cot->condCListcaen;
  le = co->condCListcaen;
  while (! isNull (let)) {  /* update conditions expr */
    e = getLEExpr (le);
    et = getLEExpr (let);
    updateExprNum (e, et);
    le = getLENext (le);
    let = getLENext (let);
  }

  lgt = cot->gVarListcaen;
  lg = co->gVarListcaen;
  
  if (! isNull (lg)) {     /* all gVars have been generated in lgt already */
    while (! isNull (lgt)) {
      /*printGVar (stdout, getLGGVar (lg));
      cout << endl;*/
      gv = getLGGVar (lg);
      gvt = getLGGVar (lgt);
      
      if (! isNull (gv->co) && (gv->co->type == gvt->co->type)) {  /* have same component's type */
	if (isGVarVar (gv) && isGVarVar (gvt)) {  /* loose check but should be ok */
	  setGVarNum (gv, getGVarNum (gvt));
	  lg = getLGNext (lg);
	}
      }
      else if (isGVarInterm (gv) && isGVarInterm (gvt)) {
	
	/*if (eqObj (getGVarObj (gv), getGVarObj (gvt))) {*/
	  setGVarNum (gv, getGVarNum (gvt));
	  lg = getLGNext (lg);
	/*}*/ 
      }
      
      /*cout << "lg: " << lg << endl;
      cout << "lgt: " << lgt << endl;*/
      /*setGVarNum (gv, getGVarNum (gvt));
      lg = getLGNext (lg);*/
      lgt = getLGNext (lgt);
    }
  }
}

void updateExprNum (expr e, expr et) {
  enum exprType tp=getExprType(e);
  gVar gv=getExprGVar(e);
  gVar gvt=getExprGVar(et);
  expr e1, e1t, e2, e2t;
  
  setExprNum (e, getExprNum (et));

  switch (tp) {
  case CTE:
  case VAR:
    if (! isNull (gvt))
      setGVarNum (gv, getGVarNum (gvt));
    break;
  case PLUS:
  case MOINS:
  case MULT:
  case DIV:
  case EXP:  /* TODO: cross ref here ? */
    e1 = getExprOp1 (e); e1t = getExprOp1 (et);
    updateExprNum (e1, e1t);
    e2 = getExprOp2 (e); e2t = getExprOp2 (et);
    updateExprNum (e2, e2t);
    break;
  case SGN:
    e1 = getExprOp1 (e); e1t = getExprOp1 (et);
    updateExprNum (e1, e1t);
  default:
    break;
  } /* end switch tp */
}

/**
 * generate symbolic name or cross-reference code for imported variables.
 * influ: 0 - no in an influence,
 *        1 - cause in influence,
 *        2 - effect in influence,
 *        3 - in mapping function.
 *        4 - gVar (interm).
 */
string geneContSymbName (variable va, Comp* co, int influ, string nums, string numob, string numj, string numob2) {
  int dot_pos;
  Comp* coi;
  string na = string (getVarNom (va));
  string co_na;

  dot_pos = na.find (".");
  co_na = na.substr (0, dot_pos);
  /*cout << "co_na: " << co_na << endl;
    cout << "var_name: " << co->var_name << endl;
    cout << "influ: " << influ << endl;*/
  if (co_na == co->var_name) {      /* local variable */
    if (! in_gen_types)
      return string (getVarSymbName (va));
  }
  else {                         /* imported variable */
    std::ostrstream os;
    string num;
    coi = findComp (co_na);
    /*if (! (coi = findComp (co_na))->generated && isIncludedInComp (co, co->type)) {*/
    if (! ((coi->generated && isIncludedInComp (co, coi->type))
	   || (isIncludedInComp (co, coi->type) && isIncludedInComp (coi, co->type)))) {
      if (! in_gen_types)
	return co->findCorresDef (co_na) + "->" + string (getVarSymbName (va));
    }
    else {
      os << numInflu;
      num = string (os.str(), os.tellp());
      if (influ == 1) {                 /* in influence cause */
	cross_ref += co->var_name + "->i_i" + num + "->cause = " + coi->var_name + "->" + string (getVarSymbName (va)) + ";\n\t";
      }
      else if (influ == 2) {            /* in influence effet */
	cross_ref += co->var_name + "->i_i" + num + "->effet = " + coi->var_name + "->" + string (getVarSymbName (va)) + ";\n\t";
	
      }
      else if (influ == 0) {            /* in global constraints */
	/* TODO */
      }
      else if (influ == 3) {            /* in mapping function */
	cross_ref += "(*" + co->var_name + "->" + co->type + "_t" + nums 
	  + "->mf)[" + numj + "]->mvs->push_back (mkMapfVar (getVarName ("
	  + coi->var_name + "->" + string (getVarSymbName (va)) 
	  + "), " + numob + "));\n\t";
      }
      else if (influ == 4) { /* in intermediary gVar */
	cross_ref += co->var_name + "->f_" + nums + "->uv.var = " + coi->var_name + "->" 
	  + string (getVarSymbName (va)) + ";\n\t";
      }
      return "NULL";
    }
   
  }
  return "";
}

/* 
 * generate atom symb name (import or not) in conditions and clauses 
 * 
 * sta: the state variable
 * co: the current component
 * cla: if it is to be inserted in a condition (0), a clause (1),
 *      a stateformula (2).
 * num: the number of the element.
 * bol: additive information if necessary.
 * neg: sign flag for stateformulas elements.
 */
string geneDiscSymbName (StateVar* sta, Comp* co, int cla, string num, string bol, string neg, string bol1) {
  int dot_pos;
  string na = sta->name;
  Comp* coi;
  dot_pos = na.find (".");
  string co_na = na.substr (0, dot_pos);
  string codeH;
  
  if (co_na == co->var_name)   {   /* local variable */
    if (cla == 2) {             /* in a stateformula */
      if (bol1 == "")
	return "\t" + co->type + "_stf" + num + "->addAtom (" + co->type + "_" 
	  + sta->getSymbName () + "_" + bol + ", " + neg + ");\n";
      else return "\t" + co->type + "_stf" + num + "->addAtom (" + co->type + "_" 
	     /*+ sta->getSymbName () + "_" + bol + "\" + name + \"_" + bol1 + ", " + neg + ");\n";*/
	     + sta->getSymbName () + "_" + bol + co->name + "_" + bol1 + ", " + neg + ");\n";
    }
    else if (cla == 1) 
      return co->var_name + "->" + co->type + "_" + sta->getSymbName () + "_" + bol;
    else if (cla == 0)         /* in a condition */
      return "\tsetCondAtom (c_" + num + ", " + co->type 
	+ "_" + sta->getSymbName () + "_" + bol + ");\n\t";
  }
  else {                         /* imported variable */
    coi = findComp (co_na);
    string type = coi->type;
    /*bool inc = isIncludedInComp (coi, co->type);*/
    if (cla == 0) {   /* in a condition */
      /*if (coi->generated && inc) {*/
      if (! ((coi->generated && isIncludedInComp (co, coi->type))
	     || (isIncludedInComp (co, coi->type) && isIncludedInComp (coi, co->type)))) {
	if (! in_gen_types)
	  return "\tsetCondAtom (c_" + num + ", "
	    + co->findCorresDef (co_na) + "->" + type + "_" + sta->getSymbName () 
	    + "_" + bol + ");\n\t";
	
      }
      else {
	cross_ref += "\tsetCondAtom (" + co->var_name + "->" + "c_" + num + ", "
	  + co_na + "->" + type + "_" + sta->getSymbName () 
	  + "_" + bol + ");\n\t";
	return "NULL";
      }
    }
    else if (cla == 2) {     /* in a stateformula */
      /*if (coi->generated && inc) {*/
      if (! ((coi->generated && isIncludedInComp (co, coi->type))
	     || (isIncludedInComp (co, coi->type) && isIncludedInComp (coi, co->type)))) {
	/*return  co->type + "_stf" + num + "->addAtom ("
	  + co->findCorresDef (co_na) + "->" + type + "_" + sta->getSymbName ()  
	  + "_" + bol + ");\n";*/
	if (bol1 == "") {
	  if (! in_gen_types) {
	    return co->type + "_stf" + num + "->addAtom ("
	      /*+ "(*" + co->findCorresDef (co_na) + "->atoms)[string (\""
	      + type + "_" + sta->getSymbName () + "_" + bol + "\").c_str ()], " + neg + ");\n";*/
	      + "(*" + co->findCorresDef (co_na) + "->atoms)[\""
	      + type + "_" + sta->getSymbName () + "_" + bol + "\"], " + neg + ");\n";
	  }
	}
	else if (! in_gen_types)
	  return co->type + "_stf" + num + "->addAtom ("
	    + "(*" + co->findCorresDef (co_na) + "->atoms)[string (\""
	    + type + "_" + sta->getSymbName () + "_" + bol + "\" + name + \"_" + bol1 + "\").c_str ()], " + neg + ");\n";
      }
      else {
	if (bol1 == "")
	  cross_ref += "\t" + co->var_name + "->" + co->type + "_stf" + num + "->addAtom (" 
	    + co_na + "->" + type + "_" + sta->getSymbName () + "_" 
	    + bol + ", " + neg + ");\n";
	/*else cross_ref += "\t" + co->var_name + "->" + co->type + "_stf" + num + "->addAtom (" 
	       + co_na + "->" + type + "_" + sta->getSymbName () + "_" 
	       + bol + + "\" + name + \"_" + bol1 + ", " + neg + ");\n";*/
	
	else cross_ref += "\t" + co->var_name + "->" + co->type + "_stf" + num + "->addAtom ("
	       /*+ "(*" + co_na + "->atoms)[string (\""
	       + type + "_" + sta->getSymbName () + "_" 
	       + bol + co->name + "_" + bol1 + "\").c_str ()], " + neg + ");\n";*/
	       + "(*" + co_na + "->atoms)[\""
	       + type + "_" + sta->getSymbName () + "_" 
	       + bol + co->name + "_" + bol1 + "\"], " + neg + ");\n";
	
	/*return "\t" + co->type + "_stf" + num + "->addAtom (" + co->type + "_" 
	  + sta->getSymbName () + "_" + bol + "\" + name + \"_" + bol1 + ", " + neg + ");\n";*/

	return "";
      }
    }
    else if (cla == 1) {                   /* in a clause */
      /*if (! in_gen_types)*/
	return co_na + "->" + type + "_" + sta->getSymbName () + "_" + bol;
    }
  }
  if (! in_gen_types)
    cout << "warning: trouble generating code for a discrete variable: " << sta->getName () << endl;
  return "";
}

string geneCondName (cond cd, Comp* co, long op, string par) {
  int dot_pos;
  Comp* coi;
  string na = string (getCondName (cd));
  string co_na;
  
  dot_pos = na.find (".");
  co_na = na.substr (0, dot_pos);
  coi = findComp (co_na);
  if (co_na == co->var_name) {      /* local variable */
    /*if (! in_gen_types)*/
      return na.substr (dot_pos+1, na.length ());
  }
  /*else if ((coi = findComp (co_na))->generated && isIncludedInComp (coi, co->type)) {*/
  else if (! ((coi->generated && isIncludedInComp (co, coi->type))
	      || (isIncludedInComp (co, coi->type) && isIncludedInComp (coi, co->type)))) {
    /*if (! in_gen_types)*/
      return co->findCorresDef (co_na) + "->" + na.substr (dot_pos+1, na.length ());
  }  
  else {
    if (op == 0)
      cross_ref += coi->var_name + "->" + par + "=" + coi->var_name  /* was: co->var_name + "->" + par + "=" + coi->var_name */ /* hack */
	+ "->" + na.substr (dot_pos+1, na.length ()) + ";\n\t";
    else if (op == 1)
      cross_ref += co->var_name + "->" + par + "->op1=" + coi->var_name
	+ "->" + na.substr (dot_pos+1, na.length ()) + ";\n\t";
    else if (op == 2)
      cross_ref += co->var_name + "->" + par + "->op2=" + coi->var_name
	+ "->" + na.substr (dot_pos+1, na.length ()) + ";\n\t";
    else if (op == 3) /* in mapping function */
      cross_ref += "(*" + co->var_name + "->" + par + "]->cond = " + coi->var_name 
	+ "->" + na.substr (dot_pos+1, na.length ()) + ";\n\t";
    return "NULL";
  }
}

void geneClausesFileHeaders (Comp* co) {
  std::ostrstream os; os << clFilesNum; string cfn = string (os.str (), os.tellp ());
  strcpy (clHFile, runTDir);  /* header file */
  strcat (clHFile, "/cl_");
  strcat (clHFile, co->getVarName ());
  strcat (clHFile, cfn.c_str ());
  strcat (clHFile, ".h");
  clH = fopen (clHFile, "w");
  fprintf (clH, cl_header.c_str ());
  
  fprintf (runH, "extern void init_cl_%s%ld (Scheduler* sMM);\n\t", co->getVarName (), (long) clFilesNum);
  fprintf (runT, "init_cl_%s%ld (sMM);\n\t", co->getVarName (), (long) clFilesNum);
  fprintf (clH, "\n\textern void init_cl_%s%ld (Scheduler* sMM);\n\t", co->getVarName (), (long) clFilesNum);
  
  strcpy (clCPPFile, runTDir);
  strcat (clCPPFile, "/cl_");
  strcat (clCPPFile, co->getVarName ());
  strcat (clCPPFile, cfn.c_str ());
  strcat (clCPPFile, ".cpp");
  clCPP = fopen (clCPPFile, "w");
  fprintf (clCPP, "#include \"cl_%s%ld.h\"\n\nvoid init_cl_%s%ld (Scheduler* sMM) {\n\t", co->getVarName (), (long) clFilesNum, co->getVarName (), (long) clFilesNum);
  
  cl_files->push_back("runtime/cl_" + co->var_name + cfn);
  clFilesNum++;
}

void geneClassHeaders (const char* ty) {
  int i = 0;
  char buf_head[32];
  
  string code;
  code = "/* Generated from " + string(ty) + ".def */\n\n";
  
  while ((buf_head[i] = toupper(ty[i])) != 0)
    i++;
  buf_head[i++] = '_';
  buf_head[i++] = 'H';
  buf_head[i++] = '\0';
  /*code = code + "#ifndef RUNT\n#define RUNT\n#endif\n\n#ifndef " + string(buf_head) + "\n#define " */
  code += "\n#ifndef " + string(buf_head) + "\n#define "
    + string(buf_head) + "\n\n" + "#include \"Comprt.h\"\n";
  /*if (conf_get ("discrete") == 0)*/
  code += "#include \"caen-runt.h\"\n";
  fprintf (defH, code.c_str());
  code = "#include " + string("\"") + string(ty) + ".h\"\n\n";
  fprintf (defCPP, code.c_str());
}

void geneComp (Comp* co) {
  string codeH, codeC;

  in_gen_types = inVector(co->type, generated_types);
  if (in_gen_types)
    updateEltsNum (co);

  /*if (! inVector(co->type, generated_types)) {*/
    geneUse (co);
  /*}*/

  numInflu = 0;
  clFilesNum = 0;

  if (conf_get ("discrete") == 0 && ! in_gen_types) {
    for (unsigned int i=0; i<co->variables->size(); i++)
      geneVariable (defCPP, defH, co, (*co->variables)[i]);
  }
  else if (in_gen_types) numVar += co->variables->size ();  /* update variables number anyway */

  /* global variables */
  /*cout << "co name: " << co->name << endl;
    if (! isNull (co->gVarListcaen))
    cout << "gvar list: " << listcaenGVarLn (co->gVarListcaen) << endl;
    listcaenGVar lgg = co->gVarListcaen;
    while (! isNull (lgg)) {
    printGVar (stdout, getLGGVar (lgg));
    cout << endl;
    lgg = getLGNext (lgg);
    }*/
  if (listcaenGVarLn (co->gVarListcaen))
    geneGlobVs (defCPP, defH, co->gVarListcaen, co);
  
  /* global constraints */
  if (listcaenExprLn (co->globCListcaen))
    geneGlobCs (defCPP, defH, co->globCListcaen, co);
  
  /* conditions as constraints */
  if (listcaenExprLn (co->condCListcaen))
    geneCondCs (defCPP, defH, co->condCListcaen, co);
  
  /* create clauses files */
  if (reasoner->total_clauses > CLAUSES_NUM_BREAK) {
    geneClausesFileHeaders (co);
  }

  vector<variable>* relcdvars = new vector<variable> ();
  for (unsigned int i = 0; i<co->state_variables->size(); i++)
    geneStateVar (co, (*co->state_variables)[i], relcdvars);
  if (! isNull (relcdvars))
    delete relcdvars;

  geneConstraints (co); /* generate code for clauses and influences. */
  numInfluG += numInflu;

  if (conf_get ("continuous") == 0) {
    for (unsigned int j=0; j<co->component_transitions->size(); j++) {
      geneTransition (co, (*co->component_transitions)[j]);
    }
  }
  
  if (conf_get ("discrete") == 0 && ! in_gen_types) {
    if (listcaenInitVarLn (co->initListcaen))
      geneInitV (co->initListcaen, co);
  }
  
  if (! in_gen_types) {
    codeH = "\n};\n#endif\n";
    codeC = "\n}\n";
  }

  if (reasoner->total_clauses > CLAUSES_NUM_BREAK) {
    fprintf (clCPP, "}\n");
    fclose (clCPP);
    fclose (clH);
  }
  
  fprintf (defH, codeH.c_str());
  fprintf (defCPP, codeC.c_str());

  co->generated = true;
}

void geneUse (Comp* co) {
  string codeH = "";
  string usestr1 = "";
  string usestr2 = "";
  string usestr3 = "";
  string codeC = "";
  string forw_decl = "\n";
  Comp* ico;
  string ico_def;
  string imp_co;
  
  if (! isNull (co->imported_comps)) {
    for (unsigned int i=0; i<co->imported_comps->size (); i++) {
      ico = (*co->imported_comps)[i];
      ico_def = co->findCorresDef (ico->getVarName ());
      /*if (! ico->generated && isIncludedInComp (co, ico->type))*/ /* because of C++ cross-references */
      if ((ico->generated && isIncludedInComp (co, ico->type))
	|| (isIncludedInComp (co, ico->type) && isIncludedInComp (ico, co->type))) {
      /*if (isIncludedInComp (co, ico->type) && isIncludedInComp (ico, co->type)) {*/
	forw_decl += "class " + ico->type + ";\n";
	cross_ref += co->var_name + "->imported_comps->push_back (" + ico->var_name + ");\n\t";
      }
      else if (isIncludedInComp (co, ico->type) && ! in_gen_types) {
	codeH += "#include \"" + ico->type + ".h\"\n";
	imp_co = "imported_comps->push_back ( (Comp*) " + ico_def + "i);\n\t";
      }
     
      usestr1 += ", " + ico->type + "* " + ico_def + "i";        /* imported components */
      usestr2 += ico_def + " = " + ico_def + "i;\n\t" + imp_co;       
      usestr3 += "\t" + ico->type + "* " + ico_def + ";\n"; 
    }
  }
  
  if (co->share_commanded && ! in_gen_types)
    codeH += "#include \"" + (*co->cmd_in)[0]->comp_lk->type + ".h\"\n";
  
  codeH += forw_decl + "\n\nclass " + co->type + ": public Comp {\n\npublic:\n\t"
    + co->type + "(string na, vector<CommandVar*>* in, vector<CommandVar*>* out"
    + usestr1 + ");\n\t~" + co->type + "();\n\tvoid init();\n" + usestr3 + "\n";
  
  if (co->share_commanded && ! in_gen_types)
    codeH += "\t" + (*co->cmd_in)[0]->comp_lk->type + "* cmpin;\n";
  
  codeC = co->type + "::" + co->type 
    + "(string na, vector<CommandVar*>* in, vector<CommandVar*>* out"
    + usestr1 + ") : Comp(na) {\n\t\tcmd_in = in;\n\t\t";
  if (co->share_commanded && ! in_gen_types)
    codeC += "cmpin = (" + (*co->cmd_in)[0]->comp_lk->type + "*) (*cmd_in)[0]->comp_lk;\n\t\t";
  
  codeC += "cmd_out = out;\n\t" + usestr2 + "init();\n}\n" + co->type + "::~" + co->type 
    + "() {\n\n}\n" + "void " + co->type 
    + "::init() {\n";
  
  if (! in_gen_types) {
    fprintf (defH, codeH.c_str());
    fprintf (defCPP, codeC.c_str());
  }
}

void geneStateVar (Comp* co, StateVar* sta, vector<variable>* relcdvars) {
  string codeH, codeC, sttype, numc;
  CondVar* cvtmp;
  ModeVar* mvtmp;
  int numrel = 0;
 
  
  if (sta->type == "Mode")
    sttype = "ModeVar";
  else if (sta->type == "Dependent")
    sttype = "DependentVar";
  else if (sta->type == "Observable")
    sttype = "ObsVar";
  else if (sta->type == "Cond")
    sttype = "CondVar";
  else if (sta->type == "Command")
    sttype = "CommandVar";
  
  if (!in_gen_types) {   /* the component model has not been generated already */
    codeH = "\t" + sttype + "* " + co->type + "_" + sta->getSymbName () + ";\n";
    
    codeC = string("\n") + string("\tset<string> ") + co->type + "_" + sta->getSymbName () 
      + "_values" + ";\n";
    for (set<string>::iterator it = sta->values.begin(); it != sta->values.end(); it++) {
      codeC += "\t" + co->type + "_" + sta->getSymbName () + "_values" 
	+ ".insert(\"" + (*it) +"\");\n";
    }
    codeC += "\tstring " + sta->getSymbName () + "name = name + \"." + sta->getSymbName () + "\";\n" 
      + "\t" + co->type + "_" + sta->getSymbName () + " = new " 
      + sttype + "(" + sta->getSymbName () 
      + "name, " + co->type + "_" + sta->getSymbName () + "_values" + ", " + "\"" 
      + sta->default_state + "\"" + ", this);\n\t";

    if (sttype == "CommandVar" && ((CommandVar*) sta)->exo) {
      codeC += co->type + "_" + sta->getSymbName () + "->exo = 1;\n\t";
    }

    codeC += "\taddStateVar(" + co->type
      + "_" + sta->getSymbName () + ");\n\n";
    
    for (unsigned int j=0; j<sta->atoms_links->size(); j++) {
      codeC += "\t" + co->type + "_" + sta->getSymbName () + "_" 
	+ (*sta->atoms_links)[j]->value 
	+ " = new Atom(" + co->type + "_" + sta->getSymbName () + ", \"" 
	+ (*sta->atoms_links)[j]->value + "\", name + \"." + sta->getSymbName () + "=" 
	+ (*sta->atoms_links)[j]->value + "\");\n\t\t(*atoms)[\"" + co->type 
	+ "_" + sta->getSymbName () + "_" + (*sta->atoms_links)[j]->value + "\"]=" 
	+ co->type + "_" + sta->getSymbName () + "_" + (*sta->atoms_links)[j]->value 
	+ ";\n\t";
      
      codeH += "\tAtom* " + co->type + "_" + sta->getSymbName () + "_" 
	+ (*sta->atoms_links)[j]->value + ";\n";
    }

    fprintf (defH, codeH.c_str());
    fprintf (defCPP, codeC.c_str());
  } /* end if generated_types */

  if (conf_get ("discrete") == 0) {
    if (sttype == "CondVar") {
      cvtmp = (CondVar*) sta;
      codeC = "";
	
      std::ostrstream osm;
      string osms;
      osm << cvtmp->conditions->size();
      osms = string (osm.str(), osm.tellp());
      
      /* set conditions reference size in each variable. */
      if (! in_gen_types) {
	for (unsigned int l=0; l<cvtmp->vars->size(); l++) {  
	  if (! inVector ((*cvtmp->vars)[l], relcdvars)) {
	    codeC += string (getVarSymbName ((*cvtmp->vars)[l])) + "->ncd = " + osms
	      + ";\n\t";
	    codeC += string (getVarSymbName ((*cvtmp->vars)[l])) 
	      + "->relCd = (long*) new_malloc (" + osms + "*sizeof (long));\n\t";
	    (*cvtmp->vars)[l]->ncd = (long) cvtmp->conditions->size ();
	    relcdvars->push_back ((*cvtmp->vars)[l]);
	  }
	  else {
	    codeC += string (getVarSymbName ((*cvtmp->vars)[l])) + "->ncd += " + osms
	      + ";\n\t";
	    codeC += string (getVarSymbName ((*cvtmp->vars)[l]))
	      + "->relCd = (long*) realloc (" + string (getVarSymbName ((*cvtmp->vars)[l])) 
	      + "->relCd, " + string (getVarSymbName ((*cvtmp->vars)[l])) + "->ncd*sizeof (long));\n\t";
	    (*cvtmp->vars)[l]->ncd += (long) cvtmp->conditions->size ();
	  }
	}
	numrel = (*cvtmp->vars)[0]->ncd - (long) cvtmp->conditions->size(); /* beware */
      } /* end if generated_types */
      
      for (unsigned int k=0; k<cvtmp->conditions->size(); k++) {
	string numc1;
	string namec = geneCondition(defCPP, defH, (*cvtmp->conditions)[k], co, 0, "");
	
	if (! in_gen_types) {
	  codeC += "\t" + co->type + "_" + sta->getSymbName ()
	    + "->conditions->insert(pair<const char*, int>("
	    + co->type + "_" + sta->getSymbName () + "_"
	    + (*cvtmp->conditions)[k]->at->value
	    + "->getId(), "
	    + posCond
	    + "));\n";
	  codeC += "\t" + co->type + "_" + sta->getSymbName () + "_"
	    + (*cvtmp->conditions)[k]->at->value
	    + "->cond = 1;\n\t";
	  
	  /* add condition ref to each variable. */
	  for (unsigned int l=0; l<cvtmp->vars->size(); l++) {
	    std::ostrstream os1;
	    os1 << numrel; numc1 = string (os1.str(), os1.tellp());
	    codeC += string (getVarSymbName ((*cvtmp->vars)[l])) + "->relCd["
	      + numc1 + "] = " + posCond + ";\n\t";
	    numrel++;
	  }
	}
      }
      fprintf (defCPP, codeC.c_str());
    }  /* if type CondVar */      
    else if (sttype == "ModeVar") {
	mvtmp = (ModeVar*) sta;
	codeC = "";
	std::ostrstream omv;
	string osmv;
	omv << mvtmp->conditions->size();
	osmv = string (omv.str(), omv.tellp());
	string catval;
	
	for (unsigned int k=0; k<mvtmp->conditions->size(); k++) {
	  if (isNull ((*mvtmp->conditions)[k]->at))
	    catval = (*mvtmp->conditions)[k]->op1->at->value;  /* atom negation */
	  else catval = (*mvtmp->conditions)[k]->at->value;
	  string namec = geneCondition(defCPP, defH, (*mvtmp->conditions)[k], co, 0, "");
	  
	  if (! in_gen_types) {
	    codeC += "\t" + co->type + "_" + sta->getSymbName () 
	      + "->conditions->insert(pair<const char*, int>("
	      + co->type + "_" + sta->getSymbName () + "_"
	      + catval
	      + "->getId(), "
	      + posCond
	      + "));\n";
	    codeC += "\t" + co->type + "_" + sta->getSymbName () + "_"
	      + catval
	      + "->cond = 1;\n\t";
	  }
	}
	fprintf (defCPP, codeC.c_str());
    }
  }
  if (sttype == "CommandVar" && ! in_gen_types) {
    codeC = "";
    if (sta->name.find ("cmdout") > 0) { /* for one command out only... should be enough for now */
      codeC = "\tcmd_out = new vector<CommandVar*>();\n\tcmd_out->push_back ("
	+ co->type + "_" + sta->getSymbName () + ");\n\t";
    }
    fprintf (defCPP, codeC.c_str ());
  }
 
  if (conf_get ("continuous") == 0) {  /* the model is discrete or hybrid */
    for (unsigned int k=0; k<sta->internal_clauses->size(); k++)
      geneClause (co, (*sta->internal_clauses)[k], sta);
  }
}

void geneConstraints (Comp* co) {
  string codeC;
  
  /* generate clauses */
  if (conf_get ("continuous") == 0) {
    if (! in_gen_types) {
      codeC = "\texternal_clauses = new vector<Clause*>();\n";
      fprintf (defCPP, codeC.c_str());
    }
    for (unsigned int i=0; i<co->external_clauses->size(); i++)
      geneClause (co, (*co->external_clauses)[i], 0);
  }
  
  /* generate influences */
  if (conf_get ("discrete") == 0) {
    for (unsigned int j=0; j<co->influences->size(); j++) {
      numInflu++;
      geneInfluence (defCPP, defH, (*co->influences)[j], co);
    }
  
    /* generate local graph nodes. */
    numVarG += numVar;
    numVar = numNoeud = 0;
    for (unsigned int k=0; k<co->variables->size(); k++) {
      geneNode (defCPP, defH, co, (*co->variables)[k]);
    }
    numNoeudG += numNoeud;
    numVar = numNoeud = 0;
  }
}

/* clauses are generated in runtime */
void geneClause (Comp* co, Clause* clau, StateVar* sta) {
  string codeC, nums, bol;
  std::ostrstream os;
  numClauses++;
  os << numClauses;
  
  nums = string (os.str(), os.tellp());
  codeC = "\n\tClause* cl" + nums + " = new Clause (" + co->var_name + ");\n";
  for (multimap<bool, Atom*, eqbool>::iterator cl_it = clau->cl->begin();
       cl_it != clau->cl->end(); cl_it++) {
    if ((*cl_it).first)
      bol = "true";
    else bol = "false";
    codeC += "\tcl" + nums + "->addAtomRef("
      + geneDiscSymbName ((*cl_it).second->getStateVar(), co, 1, "", (*cl_it).second->value, "", "")
      + ", " + bol + ");\n";
  }
  if (clau->getInternal())
    codeC += "\tcl" + nums + "->setInternal(1);\n\t" + co->var_name 
	+ "->" + co->type + "_" + sta->getSymbName ()  + "->internal_clauses->push_back(cl" 
	+ nums + ");\n";
    else {
      codeC += "\tcl" + nums + "->setInternal(0);\n\t";
      codeC += co->var_name + "->external_clauses->push_back(cl" + nums + ");\n\t";
    }
  codeC += "sMM->cls->push_back (cl" + nums +");\n\t";
  co->numClauses++;
  if (reasoner->total_clauses <= CLAUSES_NUM_BREAK)
    fprintf (runT, codeC.c_str());
  else {
    if (co->numClauses > CLAUSES_NUM_BREAK) {
      fprintf (clCPP, "}\n");
      fclose (clH);
      fclose (clCPP);
      geneClausesFileHeaders (co);
      co->numClauses = 0;
    }
    fprintf (clCPP, codeC.c_str ());
  }
}

void geneTransition (Comp* co, Transition* tr) {
  string codeC, codeH, nums, nums1, nums2, prob, numid, numob, cost, numj, mcondn, nume;
  unsigned int i1, i2;
  std::ostrstream os;
  std::ostrstream os1;
  std::ostrstream os2;
  std::ostrstream os3;
  std::ostrstream os4;
  std::ostrstream os5;
  
  i1 = numStf;
  i2 = numStf + 1;

  if (! in_gen_types) {
    os << numTransitions;
    tr->compiled_num = numTransitions;
  }
  else os << tr->compiled_num;
  
  nums = string (os.str(), os.tellp());
  
  geneStateformula (co, tr->pre, nums, 1);
  geneStateformula (co, tr->post, nums, 0);

  os1 << i1; nums1 = string (os1.str(), os1.tellp());
  os2 << i2; nums2 = string (os2.str(), os2.tellp());
  os3 << tr->proba; prob = string (os3.str(), os3.tellp());
  os4 << tr->trans_id; numid = string (os4.str(), os4.tellp());
  os5 << tr->cost; cost = string (os5.str (), os5.tellp ());

  if (! in_gen_types) { 
    codeH = "\n\tTransition* " + co->type + "_t" + nums + ";\n";
    codeC = "\n\t" + co->type + "_t" + nums + " = new Transition("
      + co->type + "_stf" + nums1 + ", " + co->type + "_stf" + nums2 + ", "
      + "\"" + tr->type + "\", " + prob + ", " + cost;
    if (tr->idle)
      codeC += ", 1);\n";
    else codeC += ", 0);\n";
    
    codeC += co->type + "_t" + nums + "->trans_id = " + numid + ";\n\t";
  } /* end if generated_types */
  
  if (!isNull (tr->mf)) {
    if (! in_gen_types)
      codeC += co->type + "_t" + nums + "->mf = new vector<MapFunc*> ();\n\t";
    for (unsigned int j=0; j<tr->mf->size (); j++) {
      MapFunc* mf = (*tr->mf)[j];
      if (! in_gen_types)
	codeC += co->type + "_t" + nums + "->mf->push_back (new MapFunc (";
      std::ostrstream os6; os6 << j;
      numj = string (os6.str(), os6.tellp());
      if (! isNull (mf->condi)) {
	if ((mcondn = geneCondition (defCPP, defH, mf->condi, co, 3, co->type + "_t" + nums + "->mf)[" + numj)) != "NULL") {
	  if (! in_gen_types)
	    codeC += mcondn;
	}
      }
      if (! in_gen_types)
	codeC += "));\n\t";
      for (unsigned int i=0; i<mf->mvs->size(); i++) {
	expr e = getMFExpr ((*mf->mvs)[i]);
	geneExpr (defCPP, defH, e, co, 1);
	std::ostrstream os5;
	os5 << getExprNum (e); numob = string (os5.str(), os5.tellp());
	std::ostrstream os6;
	os6 << numMFExpr; nume = string (os6.str(), os6.tellp());
	mfexprstr += "aMF[" + nume + "] = " + co->var_name + "->s_" + numob + ";\n\t";
	string genevar = geneContSymbName (getMFVar ((*mf->mvs)[i]), co, 3, nums, nume, numj, numob);
	if (genevar != "NULL" && ! in_gen_types) {
	  codeC += "(*" + co->type + "_t" + nums + "->mf)[" + numj + "]->mvs->push_back (mkMapfVar (getVarName ("
	    + genevar + "), " + nume + "));\n\t";
	}
	numMFExpr++;
      }
    }
  }
   
  if (! in_gen_types) {
    codeC += "\n\taddTransitionToC(" +  co->type 
      + "_t" + nums + ");\n\t";
    
    fprintf (defH, codeH.c_str ());
    fprintf (defCPP, codeC.c_str());
    numTransitions++;
  }
}

void geneStateformula (Comp* co, StateFormula* stfr, string ntrans, bool pre) {
  string codeC, codeH, nums, gen;
  std::ostrstream os;

  if (! in_gen_types) {
    os << numStf;
    stfr->compiled_num = numStf;
  }
  else os << stfr->compiled_num;
  nums = string (os.str(), os.tellp());
  
  codeH = "\tStateFormula* " +  co->type + "_stf" + nums + ";\n";
  codeC = "\n\t" + co->type + "_stf" + nums + " = new StateFormula();\n";
  for (multimap<bool, Atom*, eqbool>::iterator at_it = stfr->atoms->begin();
       at_it != stfr->atoms->end(); at_it++) {
    /*if ((*at_it).second->dyn_name == "") {*/
    if (isNull ((*at_it).second->dyn_name) || (*at_it).second->dyn_name->count(co->getName ()) == 0) {
      if ((*at_it).first)
	gen = geneDiscSymbName ((*at_it).second->getStateVar (), co, 2, nums, (*at_it).second->value, "1", "");
      else gen = geneDiscSymbName ((*at_it).second->getStateVar (), co, 2, nums, (*at_it).second->value, "0", "");
    }
    else { /* dynamically set the atom name */
      string dyn_name = (*(*at_it).second->dyn_name)[co->getName ()];
      int perc = dyn_name.find ("%");
      string vala = dyn_name.substr (0, perc);
      string valb = dyn_name.substr (perc+1, dyn_name.length ());
      if ((*at_it).first)
	gen = geneDiscSymbName ((*at_it).second->getStateVar (), co, 2, nums, vala, "1", valb);
      else gen = geneDiscSymbName ((*at_it).second->getStateVar (), co, 2, nums, vala, "0", valb);
    }
    if (pre && gen == "") {         /* reference atoms in cross-ref !!! */
      string stan = (*at_it).second->getStateVar ()->name;
      Comp* coi = findComp (stan.substr (0, stan.find (".")));
      cross_ref += "\t" + co->var_name + "->" + co->type + "_t" + ntrans
	+ "->refAtomPre (";
      if ((*at_it).first)
	cross_ref += "1, ";
      else cross_ref += "0, ";
      cross_ref += coi->var_name + "->" + coi->type + "_"
	+ (*at_it).second->getStateVar ()->getSymbName () + "_" + (*at_it).second->value
	+ ");\n";
    }
    else if (gen == "") {
      string stan = (*at_it).second->getStateVar ()->name;
      Comp* coi = findComp (stan.substr (0, stan.find (".")));
      cross_ref += "\t" + co->var_name + "->" + co->type + "_t" + ntrans
	+ "->refAtomPost (" + coi->var_name + "->" + coi->type + "_"
	+ (*at_it).second->getStateVar ()->getSymbName () + "_" + (*at_it).second->value
	+ ");\n";
    }
    codeC += gen + " /* gen */\n";
  }
  codeC +=  "\t" + co->type + "_stf" + nums + "->setOperat(\"" + stfr->getOperat() + "\");\n";
  
  if (! in_gen_types) {
    fprintf (defH, codeH.c_str ());
    fprintf (defCPP, codeC.c_str ());
  }
  numStf++;
}

void geneVariable (FILE* fd, FILE* fdh, Comp* co, variable v) {
  domain d=getVarDom(v);
  long histoLn;
  long numRange = 0;
  const char* obna;
  
  numVar++;

  fprintf (fd, "\n\tobjet ");
  if (isVarConst (v))  {                                /* parameter */
    fprintf (fd, "x");
    obna = "x";
  }
  else        {                                         /* variable */
    fprintf (fd, "d");
    obna = "d";
  }
  fprintf (fd, "_%ld=  (objet) new_malloc(sizeof(s_Objet));\n\t", numVar);
  if (RANGE == getObjType (getDomObj (d))) {
    range r=getDomRVal(d);
    fprintf (fd, "\tsetObjType(%s_%ld,(objType)%d);\n\t", obna, numVar, RANGE);
    fprintf (fd, "\trange r_%ld%ld = {", numVar, numRange);
    geneReal (fd, getRangeInf (r));
    fprintf (fd, ",");
    geneReal (fd, getRangeSup (r));
    fprintf (fd, ",");
    geneReal (fd, getBoundInf (r));
    fprintf (fd, ",");
    geneReal (fd, getBoundSup (r));
    fprintf (fd, "};\n\tsetObjRangeVal(%s_%ld,r_%ld%ld);\n", obna, numVar, numVar, numRange);
  }
  else {
    fprintf (fd, "\tsetObjType(%s_%ld,(objType)%d);\n", obna, numVar, REAL);
    fprintf (fd, "\trange r_%ld%ld = {", numVar, numRange);
    geneReal (fd, getDomNVal (d));
    fprintf (fd, ",%g, 1, 1", 0.0);
    fprintf (fd, "};\n\tsetObjRangeVal(%s_%ld,r_%ld%ld);\n", obna, numVar, numVar, numRange);
  }
  numRange++;
  if (! isVarConst (v)) {                               /* variable */
    fprintf (fd, "\tobjet x_%ld=(objet) new_malloc(sizeof(s_Objet));\n", numVar);
    fprintf (fd, "\tsetObjType(x_%ld,(objType)%d);\n", numVar, UNKNOWN);
    fprintf (fd, "\trange r_%ld%ld = {%g,%g, 1, 1};\n", numVar,
	     numRange, 0.0, 0.0);
    fprintf (fd, "\tsetObjRangeVal(x_%ld,r_%ld%ld);\n\t", numVar, numVar, numRange);
    numRange++;
    
    fprintf (fd, "\tobjet y_%ld=(objet) new_malloc(sizeof(s_Objet));\n", numVar);
    fprintf (fd, "\tsetObjType(y_%ld,(objType)%d);\n", numVar, UNKNOWN);
    fprintf (fd, "\trange r_%ld%ld = {%g,%g, 1, 1};\n", numVar,
	     numRange, 0.0, 0.0);
    fprintf (fd, "\tsetObjRangeVal(y_%ld,r_%ld%ld);\n\t", numVar, numVar, numRange);
    numRange++;
  } /* endif ! isVarConst v */
  if (isVarMeasure (v)) {
    fprintf (fd, "objet z_%ld = (objet) new_malloc(sizeof(s_Objet));\n", numVar);
    fprintf (fd, "\tsetObjType(z_%ld,(objType)%d);\n", numVar, UNKNOWN);
    fprintf (fd, "\trange r_%ld%ld = {%g,%g, 1, 1};\n", numVar,
	     numRange, 0.0, 0.0);
    fprintf (fd, "setObjRangeVal(z_%ld,r_%ld%ld);\n\t", numVar, numVar, numRange);
    numRange++;
  } /* endif  isVarMeasure v */
  if ((histoLn = (long) ceil (getVarMaxDelay (v)/timeStep) + twc)) {  /* history from dynamics and temporal window settings */
    long i;
    fprintf (fd, "objet* m_%ld = (objet*) new_malloc (%ld*sizeof (objet));\n\t", numVar, histoLn);
    for (i=0; i<histoLn; ++i) { /* values */
      fprintf (fd, "m_%ld[%ld] = (objet) new_malloc (sizeof (s_Objet));\n\tsetObjType (m_%ld[%ld], (objType)0);\n\t", numVar, i, numVar, i);
      fprintf (fd, "range r_%ldh%ld = {%g, %g, 1, 1};\n\t", numVar, numRange, 0.0, 0.0);
      fprintf (fd, "setObjRangeVal (m_%ld[%ld], r_%ldh%ld);\n\t", numVar, i, numVar, numRange);
      numRange++;
    } /* end for */
    fprintf (fd, "objet* o_%ld = (objet*) new_malloc (%ld*sizeof (objet));\n\t", numVar, histoLn);
    for (i=0; i<histoLn; ++i) { /* measures */
      fprintf (fd, "o_%ld[%ld] = (objet) new_malloc (sizeof (s_Objet));\n\tsetObjType (o_%ld[%ld], (objType)0);\n\t", numVar, i, numVar, i);
      fprintf (fd, "range r_%ldhm%ld = {%g, %g, 1, 1};\n\t", numVar, numRange, 0.0, 0.0);
      fprintf (fd, "setObjRangeVal (o_%ld[%ld], r_%ldhm%ld);\n\t", numVar, i, numVar, numRange);
      numRange++;
    } /* end for */
    fprintf (fd, "long* u_%ld = (long*) calloc (%ld, sizeof (long));\n\t", numVar, histoLn);
    fprintf (fd, "long* w_%ld = (long*) calloc (%ld, sizeof (long));\n\t", numVar, histoLn);
  } /* endif histoLn */
  fprintf (fd, "\t");
  printVarSymbName (fd, v);
  fprintf (fd, "= (variable) new_malloc (sizeof(s_Variable));\n\t");
  printVarSymbName (fd, v);
  fprintf (fd, "->nom=mkString((name + \".");
  printVarSymbName (fd, v);
  fprintf (fd, "\").c_str());\n\t");
  if (isVarConst (v))    {                              /* parameter */
    printVarSymbName (fd, v);
    fprintf (fd, "->cst = 1;\n\t");
    printVarSymbName (fd, v);
    fprintf (fd, "->dom = NULL;\n\t");
    printVarSymbName (fd, v);
    fprintf (fd, "->deter = 0;\n\t");
    printVarSymbName (fd, v);
    fprintf (fd, "->oldVal = x_%ld;\n\t", numVar);
    printVarSymbName (fd, v);
    fprintf (fd, "->newVal = NULL;\n\t");
  }
  else {      /* variable */
    printVarSymbName (fd, v);
    fprintf (fd, "->cst = 0;\n\t");
    printVarSymbName (fd, v);
    fprintf (fd, "->dom = d_%ld;\n\t", numVar);
    printVarSymbName (fd, v);
    fprintf (fd, "->deter = 0;\n\t");
    printVarSymbName (fd, v);
    fprintf (fd, "->oldVal = x_%ld;\n\t", numVar);
    printVarSymbName (fd, v);
    fprintf (fd, "->newVal = y_%ld;\n\t", numVar);
  }
  if (isVarMeasure (v)) {
    printVarSymbName (fd, v);
    fprintf (fd, "->mesVal = z_%ld;\n\t", numVar);
  }
  else {
    printVarSymbName (fd, v);
    fprintf (fd, "->mesVal = NULL;\n\t");
  }
  fprintf (fd, "resetVarAlarms (");
  printVarSymbName (fd, v);
  fprintf (fd, ");\n\t");
  printVarSymbName (fd, v);
  fprintf (fd, "->histoMisb = NULL;\n\t");
  
  if (isNull (getVarNoeud (v))) {
    printVarSymbName (fd, v);
    fprintf (fd, "->node = NULL;\n\t");
  }
  else {
    fprintf (fd, "%s_n_%ld = (cGraph) new_malloc (sizeof (s_CGraph));\n\t", co->type.c_str(), numVar);
    printVarSymbName (fd, v);
    fprintf (fd, "->node = %s_n_%ld;\n\t", co->type.c_str(), numVar);
  }
  if (histoLn) {
    printVarSymbName (fd, v);
    fprintf (fd, "->histoLn = %ld;\n\t", histoLn);
    printVarSymbName (fd, v);
    fprintf (fd, "->histo = m_%ld;\n\t", numVar);
    printVarSymbName (fd, v);
    fprintf (fd, "->histoMes = o_%ld;\n\t", numVar);
    printVarSymbName (fd, v);
    fprintf (fd, "->histoLowAlrm = u_%ld;\n\t", numVar);
    printVarSymbName (fd, v);
    fprintf (fd, "->histoHighAlrm = w_%ld;\n\t", numVar);
  }
  else {
    fprintf (fd, "resetVarHisto (");
    printVarSymbName (fd, v);
    fprintf (fd, ");\n\t");
  }
  printVarSymbName (fd, v);
  fprintf (fd, "->ncd = 0;\n\t");
  printVarSymbName (fd, v);
  fprintf (fd, "->relCd = NULL;\n\t");
  fprintf (fdh, "\tvariable ");
  printVarSymbName (fdh, v);
  fprintf (fdh, ";\n");
  return;
}


void geneVarV (listcaenVar lv) {
  long first=1;
 
  while (! isNull (lv)) {
    variable v=getLVVar (lv);
    if (first)
      first = 0;
    else
      fprintf (runT, ",");
    fprintf (runT, "\n&v_");
    printVarNom (runT, v);
    lv = getLVNext (lv);
  }
  return;
}


string geneCondition (FILE* fd, FILE* fdh, cond cd, Comp* co, long op, string par) {
  long nc=0;
  string tom, numc, numc1, namec1, namec2;
  std::ostrstream os;
  std::ostrstream os1;
  std::ostrstream os2;
  string res;
  enum condType ct;
  string ncstr;

  nc = getCondNum (cd);
  os << nc; numc = string (os.str(), os.tellp());
  if (par != "")
    res = geneCondName (cd, co, op, par);
  else res = geneCondName (cd, co, op, "c_" + numc);
  
  if (! isCondGened (cd) && ! in_gen_types) {
    int dot_pos = res.find ("->");
    if (res != "NULL" && dot_pos < 0) {
      char buf[700];
      ct=getCondType(cd);
      
      switch (ct) {
      case CDCTE:
	sprintf (buf, "setCondCTE (c_%ld, %d);\n\t", nc, getCondValue (cd));
	break;
      case CDAT:
	os2 << nc; ncstr = string (os2.str(), os2.tellp());
	tom = geneDiscSymbName (getCondAtom (cd)->getStateVar (), co, 0, ncstr, getCondAtom(cd)->getValue(), "", "");
	if (tom != "NULL")
	  sprintf (buf, "setCondCTE (c_%ld, %g);\n\%s;\n\t", nc, 0.0, tom.c_str());
	else sprintf (buf, "c_%ld->at = NULL;\n\t", nc);
	break;
      /*case CDCOMP:
	if (isNull (v = getCondDomVar (cd))) {
	  ob = getCondDomObj (cd);
	  fprintf (fd, "\tobjet e_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", nc);
	  if (RANGE == getObjType (ob)) {
	    range r=getObjRVal (ob);
	    fprintf (fd, "setObjType(e_%ld, (objType)%d);\n\t", nc, RANGE);
	    fprintf (fd, "range rc_%ld = {", nc);
	    geneReal (fd, getRangeInf (r));
	    fprintf (fd, ",");
	    geneReal (fd, getRangeSup (r));
	    fprintf (fd, ",");
	    geneReal (fd, getBoundInf (r));
	    fprintf (fd, ",");
	    geneReal (fd, getBoundSup (r));
	    fprintf (fd, "};\n\tsetObjRangeVal(e_%ld,r_%ld);\n\t", numVar, numVar);
	  }
	  else {
	    fprintf (fd, "setObjType(e_%ld, (objType)%d);\n\t", nc, REAL);
	    geneReal (fd, getObjNVal (ob));
	    fprintf (fd, ",%g", 0.0);
	    fprintf (fd, "};\n\tsetObjRangeVal(e_%ld,r_%ld);\n\t", numVar, numVar);
	  }
	  if (!isNull (getCondAtom (cd))) {
	    tom = co->type + "_" + getCondAtom(cd)->getStateVar()->getSymbName() + "_"
	      + getCondAtom(cd)->getValue();
	  }
	  else tom = "NULL";
	  sprintf (buf, "c_%ld->op1 = NULL;\n\tc_%ld->op2 = NULL;\n\tc_%ld->var1 = %s;\n\tc_%ld->opr = (opRel)%d;\n\tc_%ld->ob = e_%ld;\n\tc_%ld->var2 = NULL;\n\tc_%ld->val = (boolean)0;\n\tc_%ld->at = %s;\n\tc_%ld->num=%s;\n\tc_%ld->nle = 0;\n\tc_%ld->le = NULL;\n\tc_%ld->e = NULL;\n\t", nc, nc, nc, getVarSymbName (getCondVar (cd)), nc, getCondRel (cd), nc, nc, nc, nc, nc, tom.c_str(), nc, numc.c_str(), nc, nc, nc);
	}
	else {
	  if (!isNull (getCondAtom (cd))) {
	    tom = co->type + "_" + getCondAtom(cd)->getStateVar()->getSymbName() + "_"
	      + getCondAtom(cd)->getValue();
	  }
	  else tom = "NULL";
	  sprintf (buf, "c_%ld->op1 = NULL;\n\tc_%ld->op2 = NULL;\n\tc_%ld->var1 = %s;\n\tc_%ld->opr = (opRel)%d;\n\tc_%ld->ob = NULL;\n\tc_%ld->var2 = %s;\n\tc_%ld->val = (boolean)0;\n\tc_%ld->at = %s;\n\tc_%ld->num=%s;\n\tc_%ld->nle = 0;\n\tc_%ld->le = NULL;\n\tc_%ld->e = NULL;\n\t", nc, nc, nc, getVarSymbName (getCondVar (cd)), nc, getCondRel (cd), nc, nc, getVarSymbName (v), nc, nc, tom.c_str(), nc, numc.c_str(), nc, nc, nc);
	}
	break;*/
      case CDOU:
      case CDET:
	namec1 = geneCondition (fd, fdh, getCondOp1 (cd), co, 1, "c_" + numc);
	namec2 = geneCondition (fd, fdh, getCondOp2 (cd), co, 2, "c_" + numc);
	/* beware about the atom... */
	if (!isNull (getCondAtom (cd))) {
	    tom = co->type + "_" + getCondAtom(cd)->getStateVar()->getSymbName() + "_"
	      + getCondAtom(cd)->getValue();
	}
	else tom = "NULL";
	sprintf (buf, "c_%ld->op1 = %s;\n\tc_%ld->op2 = %s;\n\tc_%ld->var1 = NULL;\n\tc_%ld->opr = (opRel)0;\n\tc_%ld->ob = NULL;\n\tc_%ld->var2 = NULL;\n\tc_%ld->val = (boolean)0;\n\tc_%ld->at = %s;\n\tc_%ld->num=%s;\n\tc_%ld->nle = 0;\n\tc_%ld->le = NULL;\n\tc_%ld->e = NULL;\n\t", nc, namec1.c_str(), nc, namec2.c_str(), nc, nc, nc, nc, nc, nc, tom.c_str(), nc, numc.c_str(), nc, nc, nc);
	break;
      case CDNON:
	namec1 = geneCondition (fd, fdh, getCondOp1 (cd), co, 1, "c_" + numc);
	if (! isNull (getCondAtom (cd))) {
	  tom = co->type + "_" + getCondAtom(cd)->getStateVar()->getSymbName() + "_"
	    + getCondAtom(cd)->getValue();
	}
	else tom = "NULL";
	tom = "NULL";
	sprintf (buf, "c_%ld->op1 = %s;\n\tc_%ld->op2 = NULL;\n\tc_%ld->var1 = NULL;\n\tc_%ld->opr = (opRel)0;\n\tc_%ld->ob = NULL;\n\tc_%ld->var2 = NULL;\n\tc_%ld->val = (boolean)0;\n\tc_%ld->at = %s;\n\tc_%ld->num=%s;\n\tc_%ld->nle = 0;\n\tc_%ld->le = NULL;\n\tc_%ld->e = NULL;\n\t", nc, namec1.c_str(), nc, nc, nc, nc, nc, nc, nc, tom.c_str(), nc, numc.c_str(), nc, nc, nc);
	break;
      case CDEXPR:
	long i, j;
	listcaenLong cle;
	fprintf (fd, "long* cle_%ld = (long*) new_malloc (%ld*sizeof (long));\n\t", nc, listcaenLongLn (getCondLe (cd)));
	j = 0;
	cle = getCondLe (cd);
	while (! isNull (cle)) {
	  i = getLLLong (cle);
	  fprintf (fd, "cle_%ld[%ld] = %ld;\n\t", nc, j, i);
	  cle = getLLNext (cle);
	  j++;
	}
	if (! isNull (getCondAtom (cd))) {
	  tom = co->type + "_" + getCondAtom(cd)->getStateVar()->getSymbName() + "_"
	    + getCondAtom(cd)->getValue();
	}
	else tom = "NULL";
	sprintf (buf, "c_%ld->op1 = NULL;\n\tc_%ld->op2 = NULL;\n\tc_%ld->var1 = %s;\n\tc_%ld->opr = (opRel)%d;\n\tc_%ld->ob = NULL;\n\tc_%ld->var2 = NULL;\n\tc_%ld->val = (boolean)0;\n\tc_%ld->at = %s;\n\tc_%ld->num=%s;\n\tc_%ld->nle = %ld;\n\tc_%ld->le = cle_%ld;\n\tc_%ld->e = s_%ld;\n\t", nc, nc, nc, getVarSymbName (getCondVar (cd)), nc, getCondRel (cd), nc, nc, nc, nc, tom.c_str(), nc, numc.c_str(), nc, listcaenLongLn (getCondLe (cd)), nc, nc, nc, getExprNum (getCondExpr (cd)));
	break;
      default :
	implError (6);
	break;
      } /* end switch ct */
      fprintf (fdh, "condition c_%ld;\n\t", nc);
      fprintf (fd, "c_%ld = mkCond ();\n\tsetCondType (c_%ld, %d);\n\t%s", nc, nc, ct, buf);
      setCondGened (cd);
      os1 << numCond; numc1 = string (os1.str(), os1.tellp());
      condstr += "condi[" + numc1 + "] = " + co->var_name + "->c_" + numc + ";\n\t";
      numCond++;
      posCond = numc1;
    } /* endid res != NULL */
  } /* endif isCondGened cd */
  else if (! isCondGened (cd) && in_gen_types) {
    enum condType ct=getCondType(cd);
    switch (ct) {
    case CDAT:
      os2 << nc; ncstr = string (os2.str(), os2.tellp());
      tom = geneDiscSymbName (getCondAtom (cd)->getStateVar (), co, 0, ncstr, getCondAtom(cd)->getValue(), "", "");
      break;
    case CDOU:
    case CDET:
      namec1 = geneCondition (fd, fdh, getCondOp1 (cd), co, 1, "c_" + numc);
      namec2 = geneCondition (fd, fdh, getCondOp2 (cd), co, 2, "c_" + numc);
      break;
    case CDNON:
      namec1 = geneCondition (fd, fdh, getCondOp1 (cd), co, 1, "c_" + numc);
    default:
      break;
    }
    setCondGened (cd);
    os1 << numCond; numc1 = string (os1.str(), os1.tellp());
    condstr += "condi[" + numc1 + "] = " + co->var_name + "->c_" + numc + ";\n\t";
    numCond++;
    posCond = numc1;
  }
  return res;
}

bool inVectorCond (cond cd, vector<cond>* conds) {
  for (unsigned int i=0; i<conds->size(); i++)
    if (cd == (*conds)[i])
      return 1;
  return 0;
}

bool inVector (variable v, vector<variable>* vec) {
  for (unsigned int i=0; i<vec->size(); i++)
    if (v == (*vec)[i])
      return 1;
  return 0;
}

/**
 * generate C-code for an influence (see s_influence in caen_runt.h).
 */
void geneInfluence (FILE* fd, FILE* fdh, influence fl, Comp* co) {
  long delinf, delsup, condHistoLn;
  objet ob=getInfluDelay(fl);
  int i;
  std::ostrstream os;
  string numinf, namec;

  os << numInflu; numinf = string (os.str(), os.tellp());

  if (! in_gen_types)
    fprintf(fd, "\n");
  if (! isNull (getInfluCond (fl)))
    namec = geneCondition (fd, fdh, getInfluCond (fl), co, 0, "i_i" + numinf + "->cd");
  else namec = "";
  
  if (! in_gen_types) {
    condHistoLn = getInfluNActCond(fl) + 1;
    fprintf (fd, "\tenum boolVal t_%ld[%ld]={", numInflu, condHistoLn);
    for (i = 0; i < condHistoLn; ++i) {
      if (i < condHistoLn - 1)
	fprintf (fd, "(boolVal)1,");
      else
	fprintf (fd, "(boolVal)1");
    }
    fprintf (fd, "};");
    /* insert gain. */
    fprintf (fd, "\n\tobjet g_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
    geneObjet (fd, getInfluGain (fl), mkString(("g_"+numinf).c_str()));
    /* delay */
    if (REAL == getObjType(ob))
      delinf = delsup = (long) ceil (getObjNVal (ob)/timeStep);
    else { /* REAL */
      range r=getObjRVal(ob);
      delinf = (long) ceil (getRangeInf (r)/timeStep);
      if (! delinf)
	semError (1, 0, 37);   /* "lower bound of delay shouldn't be null." */
      delsup = (long) ceil (getRangeSup (r)/timeStep);
    }
    /* insert xi. */
    if (! isNull (getInfluXi (fl))) {
      fprintf (fd, "objet xi_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
      geneObjet (fd, getInfluXi (fl), mkString(("xi_"+numinf).c_str()));
    }
    else fprintf (fd, "objet xi_%ld=NULL;\n\t", numInflu);
    /* insert w. */
    if (! isNull (getInfluW (fl))) {
      fprintf (fd,"objet wn_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
      geneObjet (fd, getInfluW (fl), mkString(("wn_"+numinf).c_str()));
    }
    else fprintf (fd, "objet wn_%ld=NULL;\n\t", numInflu);
    /* insert others coefs */
    if (! isNull (getInfluN1 (fl))) {
      fprintf (fd, "objet n1_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
      geneObjet (fd, getInfluN1 (fl), mkString(("n1_"+numinf).c_str()));
      fprintf (fd, "objet n2_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
      geneObjet (fd, getInfluN2 (fl), mkString(("n2_"+numinf).c_str()));
      fprintf (fd, "objet d0_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
      geneObjet (fd, getInfluA0 (fl), mkString(("d0_"+numinf).c_str()));
      fprintf (fd, "objet d1_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
      geneObjet (fd, getInfluA1 (fl), mkString(("d1_"+numinf).c_str()));
      fprintf (fd, "objet d2_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
      geneObjet (fd, getInfluA2 (fl), mkString(("d2_"+numinf).c_str()));
    }
    else {
      fprintf (fd, "objet n1_%ld=NULL;\n\tobjet n2_%ld=NULL;\n\tobjet d0_%ld=NULL;\n\tobjet d1_%ld=NULL;\n\tobjet d2_%ld=NULL;\n\t", numInflu, numInflu, numInflu, numInflu, numInflu);
    }
    
    fprintf (fd, "objet er_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
    geneObjet (fd, getInfluTRep (fl), mkString(("er_"+numinf).c_str()));
    fprintf (fd, "objet p_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
    geneObjet (fd, objZero, mkString(("p_"+numinf).c_str()));
    fprintf (fd, "objet q_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
    geneObjet (fd, objZero, mkString(("q_"+numinf).c_str()));
    fprintf (fd, "objet sa_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
    geneObjet (fd, objZero, mkString(("sa_"+numinf).c_str()));
    fprintf (fd, "objet uu_%ld=(objet) new_malloc(sizeof(s_Objet));\n\t", numInflu);
    geneObjet (fd, objZero, mkString(("uu_"+numinf).c_str()));
    
    fprintf (fd, "i_i%ld=(influence) new_malloc(sizeof(s_Influence));\n\t", numInflu);
    fprintf (fdh, "influence i_i%ld;\n\t", numInflu);
    /* insert name of the influence. */
    fprintf (fd, "i_i%ld->name = mkString(\"%s\");\n\t", numInflu, getInfluNom (fl));
    fprintf (fd, "i_i%ld->it = (influType)%d;\n\t", numInflu, getInfluType (fl));   /* insert type-number, cause_var & effet_var. */
  } /* end if generated_types */
  
  setInfluSymbName (fl, mkString (("i" + numinf).c_str ()));   /* hack. do not remove. */
  string gcont = geneContSymbName (getInfluCause (fl), co, 1, "", "", "", "");
  if (gcont != "NULL" && ! in_gen_types)
    fprintf (fd, "i_i%ld->cause = %s;\n\t", numInflu, gcont.c_str ());
  gcont = geneContSymbName (getInfluEffet (fl), co, 2, "", "", "", "");
  if (gcont != "NULL" && ! in_gen_types)
    fprintf (fd, "i_i%ld->effet = %s;\n\t", numInflu, gcont.c_str ());

  if (! in_gen_types) {
    if (namec != "")  /* insert condition if it exists, NULL otherwise. */
      fprintf (fd, "i_i%ld->cd = %s;\n\t", numInflu, namec.c_str ());
    else
      fprintf (fd, "i_i%ld->cd = NULL;\n\t", numInflu);
    fprintf (fd, "i_i%ld->length_act_cond_hist = %ld;\n\t", numInflu, condHistoLn);  /* insert condhistory-length & influence-number. */
    fprintf (fd, "i_i%ld->act_cond_array = t_%ld;\n\t", numInflu, numInflu);
    /* from gain to variable b. */
    fprintf (fd, "i_i%ld->gain = g_%ld;\n\ti_i%ld->delinf = %ld;\n\ti_i%ld->delsup = %ld;\n\ti_i%ld->tRep = er_%ld;\n\ti_i%ld->a = p_%ld;\n\ti_i%ld->b = q_%ld;\n\ti_i%ld->a2 = sa_%ld;\n\ti_i%ld->b2 = uu_%ld;\n\ti_i%ld->xi = xi_%ld;\n\ti_i%ld->w = wn_%ld;\n\ti_i%ld->n1 = n1_%ld;\n\ti_i%ld->n2 = n2_%ld;\n\ti_i%ld->d0 = d0_%ld;\n\ti_i%ld->d1 = d1_%ld;\n\ti_i%ld->d2 = d2_%ld;\n\t", numInflu, numInflu, numInflu, delinf, numInflu, delsup, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu, numInflu);
    fprintf (fd, "addInfluence (i_i%ld);\n\t", numInflu);
    
#ifdef ENV_OPTI
    if (koa_evlps_opti
	&& (getInfluType (fl) == SINFL || getInfluType (fl) == DINFL 
	    || getInfluType (fl) == IINFL || getInfluType (fl) == D2INFL)) {  /* flag influence for optimization */
      fprintf (fd, "i_i%ld->inf_fct = true;\n\t", numInflu);
    }
    else fprintf (fd, "i_i%ld->inf_fct = false;\n\t", numInflu);
#endif
  } /* end if generated_types */
  return;
}

void geneNode (FILE* fd, FILE* fdh, Comp* co, variable v) {
  string numv, numgv;
  cGraph cg=getVarNoeud(v);
  std::ostrstream os;
  std::ostrstream os1;

  ++numVar; /* ? */
  os << numVar;
  numv = string (os.str(), os.tellp());
  if (! isNull (cg)) {
    long i, nc=getCGNCause(cg), ne=getCGNEffet(cg);
    numNoeud++;
    os1 << numGv; numgv = string (os1.str(), os1.tellp());
    nodestr += "cG[" + numgv + "] = " + co->var_name + "->" + co->type + "_n_" + numv + ";\n\t";
    numGv++;
    if (nc) {
      if (! in_gen_types) {
	fprintf (fdh, "influence* j_%ld;\n\t", numNoeud);
	fprintf (fd, "j_%ld = (influence*) new_malloc (%ld*sizeof (influence));\n\t", numNoeud, nc);
      }
      for (i=0;i<nc; ++i) {
	if (co == getCGCauseNth (cg, i)->co && ! in_gen_types)     /* local influence */
	  fprintf (fd, "j_%ld[%ld] = i_i%ld;\n\t", numNoeud, i, getInfluNum (getCGCauseNth (cg, i)));
	else if (getCGCauseNth (cg, i)->co->generated 
		 && !isIncludedInComp (getCGCauseNth (cg, i)->co, co->type)
		 && ! in_gen_types) {
	  fprintf (fd, "j_%ld[%ld] = %s->i_i%ld;\n\t", numNoeud, i, 
		   co->findCorresDef (getCGCauseNth (cg, i)->co->var_name).c_str (),
		   getInfluNum (getCGCauseNth (cg, i)));
	}
	else {           /* cross_reference */
	  std::ostrstream os2;
	  std::ostrstream os3;
	  std::ostrstream os4;
	  os2 << numNoeud; os3 << i; os4 << getInfluNum (getCGCauseNth (cg, i));
	  string nn = string (os2.str(), os2.tellp());
	  string ni = string (os3.str(), os3.tellp());
	  string nu = string (os4.str(), os4.tellp());
	  cross_ref += co->var_name + "->" + "j_" + nn + "[" + ni
	    + "] = " + getCGCauseNth (cg, i)->co->var_name + "->i_i"
	    + nu + ";\n\t";
	  if (! in_gen_types)
	    fprintf (fd, "j_%ld[%ld] = NULL;\n\t", numNoeud, i);
	}
      } /* end for */
    } /* end nc */
    if (ne) {
      if (! in_gen_types) {
	fprintf (fdh, "influence* k_%ld;\n\t", numNoeud);
	fprintf (fd, "k_%ld = (influence*) new_malloc (%ld*sizeof (influence));\n\t", numNoeud, ne);
      }
      for (i=0;i<ne; ++i) {
	if (co == getCGEffetNth (cg, i)->co && ! in_gen_types)    /* local effect */
	  fprintf (fd, "k_%ld[%ld] = i_i%ld;\n\t", numNoeud, i, getInfluNum (getCGEffetNth (cg, i)));
	else if (getCGEffetNth (cg, i)->co->generated 
		 /*&& co->findCorresDef (getCGEffetNth (cg, i)->co->var_name) != ""*/ /* buggy ? */
		 && !isIncludedInComp (getCGEffetNth (cg, i)->co, co->type)
		 && !in_gen_types) {
	  fprintf (fd, "k_%ld[%ld] = %s->i_i%ld;\n\t", numNoeud, i, 
		   co->findCorresDef (getCGEffetNth (cg, i)->co->var_name).c_str (), getInfluNum (getCGEffetNth (cg, i)));
	}
	else {           /* cross_reference needed or forward use of influences variables. */
	  std::ostrstream os2;
	  std::ostrstream os3;
	  std::ostrstream os4;
	  os2 << numNoeud; os3 << i; os4 << getInfluNum (getCGEffetNth (cg, i));
	  string nn = string (os2.str(), os2.tellp());
	  string ni = string (os3.str(), os3.tellp());
	  string nu = string (os4.str(), os4.tellp());
	  cross_ref += co->var_name + "->" + "k_" + nn + "[" + ni
	    + "] = " + getCGEffetNth (cg, i)->co->var_name + "->i_i"
	    + nu + ";\n\t";
	  if (! in_gen_types)
	    fprintf (fd, "k_%ld[%ld] = NULL;\n\t", numNoeud, i);
	}
      } /* end for */
    } /* end ne */
    if (! in_gen_types) {
      if (nc) {
	fprintf (fd, "%s_n_%ld->ncause = %ld;\n\t", co->type.c_str(), numVar, nc);
	fprintf (fd, "%s_n_%ld->vcause = j_%ld;\n\t", co->type.c_str(), numVar, numNoeud);
	fprintf (fd, "%s_n_%ld->ti = (influType)%d;\n\t", co->type.c_str(), numVar,getCGCauseType (cg));
      }
      else {
	fprintf (fd, "%s_n_%ld->ncause = 0;\n\t", co->type.c_str(), numVar);
	fprintf (fd, "%s_n_%ld->vcause = NULL;\n\t", co->type.c_str(), numVar);
	fprintf (fd, "%s_n_%ld->ti = (influType)%d;\n\t", co->type.c_str(), numVar,getCGCauseType (cg));
      }
      if (ne) {
	fprintf (fd, "%s_n_%ld->neffet = %ld;\n\t", co->type.c_str(), numVar, ne);
	fprintf (fd, "%s_n_%ld->veffet = k_%ld;\n\t", co->type.c_str(), numVar, numNoeud);
      }
      else {
	fprintf (fd, "%s_n_%ld->neffet = 0;\n\t", co->type.c_str(), numVar);
	fprintf (fd, "%s_n_%ld->veffet = NULL;\n\t", co->type.c_str(), numVar);
      }
      fprintf (fdh, "\n\tcGraph %s_n_%ld;\n\t", co->type.c_str(), numVar);
    } /* endif ! isNull cg */
  }  
  return;
}

void geneInitVar (initVar iv) {
  std::ostrstream os;
  string numin;
  os << numInit;
  numin = string (os.str(), os.tellp());
  fprintf (defCPP, "\tobjet a_%ld = (objet) new_malloc (sizeof (s_Objet));\n\t", numInit);
  geneObjet (defCPP, getIVObj (iv), mkString(("a_"+numin).c_str()));
  fprintf (defCPP, "h_%ld = (initVar) new_malloc (sizeof (s_InitVar));\n\t", numInit);
  fprintf (defCPP, "h_%ld->var = ", numInit);
  printVarSymbName (defCPP, getIVVar (iv));
  fprintf (defCPP, ";\n\th_%ld->val = a_%ld;\n\t", numInit, numInit);
  fprintf (defH, "\n\tinitVar h_%ld;", numInit);
  return;
}

void geneInitV (listcaenInitVar li, Comp* co) {
  string numi, numj;
  /*int j = 0;*/

  while (! isNull (li)) {
    std::ostrstream os;
    std::ostrstream os1;
    initVar iv=getLIVInit(li);
    os1 << numInit; numj = string (os1.str(), os1.tellp());
    numInit++;
    os << numInit; numi = string (os.str(), os.tellp());
    geneInitVar (iv);
    initstr += "init[" + numj + "] = " + co->var_name + "->h_" + numi + ";\n\t";
    li = getLIVNext (li);
    /*j++;*/
  }

  return;
}

long geneGVar (FILE* fd, FILE* fdh, gVar gv, Comp* co) {
  long ngv, lll=0, lllc=0, i;
  std::ostrstream os;
  string numgv, gconts;
  variable v;

  if (0 == (ngv = getGVarNum (gv))) {
    v=getGVarVar(gv);
    objet ob=getGVarObj(gv);
    listcaenLong ll = getGVarLExpr (gv);
    listcaenLong llc = getGVarLcExpr (gv);
    if (! in_gen_types)
      fprintf (fd, "\tobjet b_%ld = (objet) new_malloc (sizeof (s_Objet));\n\t", numGVar);
    os << numGVar;
    numgv = string (os.str(), os.tellp());
    if (! in_gen_types) { 
      if (isNull (ob))
	geneObjet (fd, objZero, mkString(("b_"+numgv).c_str()));
      else
	geneObjet (fd, ob, mkString(("b_"+numgv).c_str()));
      if (! isNull (ll)) {    /* global constraints references */
	listcaenLong lb, la=ll;
	lll = listcaenLongLn (ll);
	fprintf (fd, "long* l_%ld = (long*) new_malloc (%ld*sizeof (long));\n\tl_%ld[0] = %ld;\n\t", numGVar, lll+1, numGVar, lll);
	i = 1;
	while (! isNull (lb = getLLNext (la))) {
	  fprintf (fd, "l_%ld[%ld] = %ld;\n\t", numGVar, i, getLLLong (la));
	  la = lb;
	  i++;
	};
	fprintf (fd, "l_%ld[%ld] = %ld;\n\t", numGVar, i, getLLLong (la));
      } /* endif ! isNull ll */
      if (! isNull (llc)) {  /* conditions as constraints, references */
	listcaenLong lb, la=llc;
	lllc = listcaenLongLn (llc);
	fprintf (fd, "long* lc_%ld = (long*) new_malloc (%ld*sizeof (long));\n\tlc_%ld[0] = %ld;\n\t", numGVar, lllc+1, numGVar, lllc);
	i = 1;
	while (! isNull (lb = getLLNext (la))) {
	  fprintf (fd, "lc_%ld[%ld] = %ld;\n\t", numGVar, i, getLLLong (la));
	  la = lb;
	  i++;
	};
	fprintf (fd, "lc_%ld[%ld] = %ld;\n\t", numGVar, i, getLLLong (la));
      }
      fprintf (fdh, "gVar f_%ld;\n\t", numGVar);
      fprintf (fd, "f_%ld = (gVar) new_malloc (sizeof (s_GVar));\n\t", numGVar);
    } /* endif generated_types */
    
    if (isNull (v)) {
      if (! in_gen_types)
	fprintf (fd, "f_%ld->tgv = (gVarType)0;\n\tf_%ld->uv.var = NULL;\n\t", numGVar, numGVar);
    }
    else { /* isNull v */
      std::ostrstream os1; os1 << numGVar;
      string numgv = string (os1.str(), os1.tellp());
      gconts = geneContSymbName (v, co, 4, numgv, "", "", "");
      if (! in_gen_types)
	fprintf (fd, "f_%ld->tgv = (gVarType)1;\n\tf_%ld->uv.var = %s;\n\t", numGVar, numGVar, gconts.c_str ());
    } /* endif isNull v */
    
    if (! in_gen_types) {
      fprintf (fd, "f_%ld->ob = b_%ld;\n\t", numGVar, numGVar);
      if (isNull (ll))
	fprintf (fd, "f_%ld->nex = 0;\n\tf_%ld->ex = NULL;\n\t", numGVar, numGVar);
      else
	fprintf (fd, "f_%ld->nex = %ld;\n\tf_%ld->ex = l_%ld;\n\t", numGVar, lll+1, numGVar, numGVar);
      if (isNull (llc))
	fprintf (fd, "f_%ld->ncex = 0;\n\tf_%ld->cex = NULL;\n\t", numGVar, numGVar);
      else
	fprintf (fd, "f_%ld->ncex = %ld;\n\tf_%ld->cex = lc_%ld;\n\t", numGVar, lllc+1, numGVar, numGVar);
      ngv = numGVar;
      setGVarNum (gv, ngv);
      setGVarGened (gv);
      numGVar++;
    } /* endif generated_types */
    else {
      ngv = getGVarNum (gv);
    }
  } /* endif 0 == ngv */
  else {
    v=getGVarVar(gv);
    if (! isNull (v)) {
      std::ostrstream os1; os1 << getGVarNum (gv); /*numGVar;*/
      string numgv = string (os1.str(), os1.tellp());
      gconts = geneContSymbName (v, co, 4, numgv, "", "", "");
    } /* endif isNull v */
  }
  return ngv;
}


void geneGlobVs (FILE* fd, FILE* fdh, listcaenGVar lg, Comp* co) {
  long ngv=0, ngv2=0;
  string numgv, numgv2;

  while (! isNull (lg)) {
    std::ostrstream os;
    std::ostrstream os2;
    gVar gv=getLGGVar (lg);
    ngv = geneGVar (fd, fdh, gv, co);
    ngv2 = numGVarSV;
    os << ngv;
    os2 << ngv2;
    numgv = string (os.str(), os.tellp());
    numgv2 = string (os2.str(), os2.tellp());
    gvarstr += "gV[" + numgv2 + "] = " + co->var_name + "->f_" + numgv + ";\n\t"; 
    numGVarSV++;
    lg = getLGNext (lg);
  }
  return;
}


long numExpr;

void geneExpr (FILE* fd, FILE* fdh, expr e, Comp* co, bool cond) {
  enum exprType tp=getExprType(e);
  expr e1, e2;
  long ngv=0;
  gVar gv=getExprGVar(e);
  Comp* coi = NULL;
  
  if (0 == getExprNum (e)) {
    setExprNum (e, ++numExpr);
    if (! isNull (gv) && ! isGVarGened (gv)) {
      ngv = geneGVar (fd, fdh, gv, co);
    }
    else if (! isNull (gv))
      ngv = getGVarNum (gv);
    else ngv = -1;
    switch (tp) {
    case CTE:
    case VAR:
      if (ngv != -1)
	coi = getGVarComp (gv);
      if (! in_gen_types)
	fprintf (fd, "s_%ld = (expr) new_malloc (sizeof (s_Expr));\n\ts_%ld->tex = (exprType)%d;\n\t", numExpr, numExpr, getExprType (e));
      if (-1 == ngv) {
	if (! in_gen_types)
	  fprintf (fd, "s_%ld->gv = NULL;\n\ts_%ld->op1 = NULL;\n\ts_%ld->op2 = NULL;\n\t", numExpr, numExpr, numExpr);
      }
      else if (isNull (getGVarComp (gv)) || getGVarComp (gv) == co) {
	if (! in_gen_types)
	  fprintf (fd, "s_%ld->gv = f_%ld;\n\ts_%ld->op1 = NULL;\n\ts_%ld->op2 = NULL;\n\t", numExpr, ngv, numExpr, numExpr);
      }
      else if (! ((coi->generated && isIncludedInComp (co, coi->type))
		  || (isIncludedInComp (co, coi->type) && isIncludedInComp (coi, co->type)))) {
	if (! in_gen_types)
	  fprintf (fd, "s_%ld->gv = %s->f_%ld;\n\ts_%ld->op1 = NULL;\n\ts_%ld->op2 = NULL;\n\t", getExprNum (e), co->findCorresDef (getGVarComp (gv)->var_name).c_str (), ngv, numExpr, numExpr);
      }
      else {
	string nume, numf;
	std::ostrstream os1;
	std::ostrstream os2;
	os1 << getExprNum (e);
	os2 << ngv;
	nume = string (os1.str(), os1.tellp());
	numf = string (os2.str(), os2.tellp());
	cross_ref += co->var_name + "->s_" + nume + "->gv = " + coi->var_name + "->f_" + numf + ";\n\t";
	if (! in_gen_types)
	  fprintf (fd, "s_%ld->op1 = NULL;\n\ts_%ld->op2 = NULL;\n\t", numExpr, numExpr);
      }
      if (! in_gen_types)
	fprintf (fdh, "\n\texpr s_%ld;\n\t", numExpr);
      break;
    case PLUS:
    case MOINS:
    case MULT:
    case DIV:
    case EXP:  /* TODO: cross ref here ? */
      e1 = getExprOp1 (e);
      geneExpr (fd, fdh, e1, co, cond);
      e2 = getExprOp2 (e);
      geneExpr (fd, fdh, e2, co, cond);
      
      if (! in_gen_types) {
	fprintf (fd, "s_%ld = (expr) new_malloc (sizeof (s_Expr));\n\ts_%ld->tex = (exprType)%d;\n\t", getExprNum (e), getExprNum (e), getExprType (e));
	if (-1 == ngv)
	  fprintf (fd, "s_%ld->gv = NULL;\n\ts_%ld->op1 = s_%ld;\n\ts_%ld->op2 = s_%ld;\n\t", getExprNum (e), getExprNum (e), getExprNum (e1), getExprNum (e), getExprNum (e2));
	else
	  fprintf (fd, "s_%ld->gv = f_%ld;\n\ts_%ld->op1 = s_%ld;\n\ts_%ld->op2 = s_%ld;\n\t", getExprNum (e), ngv, getExprNum (e), getExprNum (e1), getExprNum (e), getExprNum (e2));
	fprintf (fdh, "\n\texpr s_%ld;\n\t", getExprNum (e));
      }
      break;
    case SGN:
      e1 = getExprOp1 (e);
      geneExpr (fd, fdh, e1, co, cond);
      if (! in_gen_types) {
	fprintf (fd, "s_%ld = (expr) new_malloc (sizeof (s_Expr));\n\ts_%ld->tex = (exprType)%d;\n\t", getExprNum (e), getExprNum (e), getExprType (e));
	if (-1 == ngv)
	  fprintf (fd, "s_%ld->gv = NULL;\n\ts_%ld->op1 = s_%ld;\n\ts_%ld->op2 = NULL;\n\t", getExprNum (e), getExprNum (e), getExprNum (e1), getExprNum (e));
	else
	  fprintf (fd, "s_%ld->gv = f_%ld;\n\ts_%ld->op1 = s_%ld;\n\ts_%ld->op2 = NULL;\n\t", getExprNum (e), ngv, getExprNum (e), getExprNum (e1), getExprNum (e));
	fprintf (fdh, "\n\texpr s_%ld;\n\t", getExprNum (e));
      }
    default:
      break;
    } /*end switch tp */
  } /* endif 0 == num e */
  return;
}

void geneGlobCs (FILE* fd, FILE* fdh, listcaenExpr le, Comp* co) {
  expr e;
  string nume;
  string numj;
  
  numExpr = 0;
  while (! isNull (le)) {
    std::ostrstream os;
    std::ostrstream os1;
    e = getLEExpr (le);
    if (isExprCte (e))
      implError (9);              /* constant as global constraint */
    geneExpr (fd, fdh, e, co, 0);
    os << getExprNum (e);
    os1 << numGlobC;
    nume = string (os.str(), os.tellp());
    numj = string (os1.str(), os1.tellp());
    globstr += "aG[" + numj + "] = " + co->var_name + "->s_" + nume + ";\n\t";
    le = getLENext (le);
    numGlobC++;
  }
  return;
}

void geneCondCs (FILE* fd, FILE* fdh, listcaenExpr le, Comp* co) {
  expr e;
  string numj, nume;
  
  while (! isNull (le)) {
    std::ostrstream os1;
    std::ostrstream os;
    e = getLEExpr (le);
    geneExpr (fd, fdh, e, co, 1);
    os << getExprNum (e);
    os1 << numCondC;
    nume = string (os.str(), os.tellp());
    numj = string (os1.str(), os1.tellp());
    condcstr += "aC[" + numj + "] = " + co->var_name + "->s_" + nume + ";\n\t";
    numCondC++;
    le = getLENext (le);
  }
  return;
}

/* 4. caen-execute ----------------------------------------------------*/

/* 5. caen-print ------------------------------------------------------*/

void printDic (void) {

  /*fprintf (stdout, "Dictionary:\n");*/
  printArBin (dictionary);
  return;
}


void printRules (void) {
  listcaenInfl lr=influListcaen;

  fprintf (stdout, "Influences (%ld):\n", listcaenInflLn (influListcaen));
  while (! isNull (lr)) {
    printInfluence (getLIInfl (lr));
    lr = getLINext (lr);
  }
  return;
}

void printConditions (void) {
  cout << "Conditions (" << conditions->size () << "):\n";
  for (unsigned int i=0; i<conditions->size (); i++)
    printCondition ((*conditions)[i]);
}

void printComps (void) {
  reasoner->showComps();
}

void printComp (const char* argu) {
  reasoner->showComp(argu);
}

void printAtoms (void) {
  reasoner->showAtoms();
}

void printAllVars () {
  fprintf (stdout, "Discrete variables:\n");
  for (unsigned int i=0; i<variables_disc->size (); i++)
    (*variables_disc)[i].second->showValues ();
  fprintf (stdout, "\nContinuous variables:\n");
  printDic ();
  fprintf (stdout, "\n");
}

void printCGraph (arBin p) {

  if (! isNull (p)) {
    printCGraph (getArBinGauche (p));
    printNode (getArBinVar (p));
    printCGraph (getArBinDroite (p));
  }
  return;
}


void printNode (variable v) {
  cGraph cg=getVarNoeud(v);

  if (! isNull (cg)) {
    long i, nc=getCGNCause(cg), ne=getCGNEffet(cg);
    fprintf (stdout, "  ");
    printVarNom (stdout, v);
    if (nc) {
      fprintf (stdout, ":\n    causes (");
      switch (getCGCauseType (cg)) {
      case DINFL:
	fprintf (stdout, "dynamic");
	break;
      case IINFL:
	fprintf (stdout, "integral");
	break;
      case SINFL:
	fprintf (stdout, "static");
	break;
      default:
	break;
      }
      fprintf (stdout, "): ");
      for (i=0; i<nc; ++i) {
	printVarNom (stdout, getInfluCause (getCGCauseNth (cg, i)));
	fprintf (stdout, "; ");
      }
    }
    else /* nc */
      fprintf (stdout, ":\n    no causes;");
    /* endif nc */
    if (ne) {
      fprintf (stdout, "\n    effects: ");
      for (i=0; i<ne; ++i) {
	printVarNom (stdout, getInfluEffet (getCGEffetNth (cg, i)));
	fprintf (stdout, "; ");
      }
    }
    else /* ne */
      fprintf (stdout, "\n    no effects;");
    /* endif ne */
    fprintf (stdout, "\n    depth: %ld;\n", getVarDepth (v));
  } /* endif ! isNull cg */
  return;
}
 

void printGlobCs (void) {
  listcaenGVar lg=gVarListcaen;
  listcaenExpr le=globCListcaen;
  expr e;
  long i=0;

  fprintf (stdout, "Global constraints variables:\n");
  while (! isNull (lg)) {
    gVar gv=getLGGVar(lg);
    fprintf (stdout, "  ");
    printGVar (stdout, gv);
    fprintf (stdout, ":");
    printGvExprListcaen (stdout, gv);
    fprintf (stdout, ";\n");
    lg = getLGNext (lg);
  }
  fprintf (stdout, "Global constraints:\n");
  while (! isNull (le)) {
    e = getLEExpr (le);
    if (isExprCte (e))
      implError (9);
    fprintf (stdout, "  (%ld) ", ++i);
    printGVar (stdout, getExprGVar (e));
    fprintf (stdout, " = ");
    printExpr (e);
    fprintf (stdout, ";\n");
    le = getLENext (le);
  }
  return;
}

void printCondCs (void) {
  listcaenGVar lg=gVarListcaen;
  listcaenExpr le=condCListcaen;
  expr e;
  long i=0;

  fprintf (stdout, "Conditions constraints variables:\n");
  while (! isNull (lg)) {
    gVar gv=getLGGVar(lg);
    fprintf (stdout, "  ");
    printGVar (stdout, gv);
    fprintf (stdout, ":");
    printGvExprListcaen (stdout, gv);
    fprintf (stdout, ";\n");
    lg = getLGNext (lg);
  }
  fprintf (stdout, "Conditions constraints:\n");
  while (! isNull (le)) {
    e = getLEExpr (le);
    if (isExprCte (e))
      implError (9);
    fprintf (stdout, "  (%ld) ", ++i);
    printGVar (stdout, getExprGVar (e));
    fprintf (stdout, " = ");
    printExpr (e);
    fprintf (stdout, ";\n");
    le = getLENext (le);
  }
  return;
}

#ifdef HAVE_DOT
/* koala-graph --------------------------------------------------------*/
int ceDotPs (const char* fna) {
  int r = 1;
  string com = "dot -Tps " + string ("fig/") + string (fna) + ".dot -o " + "fig/" + string (fna) + ".ps";
  
  if (-1 == system (com.c_str ())) {
    fprintf (stderr, "Fail to convert dot file to ps.\n");
    r = 0;
  }
  return r;
}

void geneDotHeaders (FILE* fd) {
  fprintf (fd, "digraph %s {\n\trankdir=LR;\n\tsize=\"8,5\";\n\tnode [shape = plaintext];\n\t", appliName);
}

void dotNode (variable v, FILE* fd) {
  cGraph cg=getVarNoeud (v);
  influence infc;
  string nv, nv1;
  int pos = 0;

  if (!isNull (cg)) {
    long i, nc=getCGNCause (cg);/*, ne=getCGNEffet (cg);*/
    nv = string (getVarNom (v));
    pos = nv.find (".");
    nv = nv.substr (0, pos) + "_" + nv.substr (pos+1, nv.length ());
    fprintf (fd, "%s;\n", nv.c_str ());
    if (nc) {
      for (i=0; i<nc; ++i) {
	infc = getCGCauseNth (cg, i);
	nv1 = string (getVarNom (getInfluCause (infc)));
	pos = nv1.find (".");
	nv1 = nv1.substr (0, pos) + "_" + nv1.substr (pos+1, nv1.length ());
	fprintf (fd, "%s -> %s [label = \"%s\",", 
		 nv1.c_str (), nv.c_str (), 
		 getInfluNom (infc));
	switch (getCGCauseType (cg)) {
	case D2INFL:
	  fprintf (fd, "color=red ];\n");
	  break;
	case DINFL:
	  fprintf (fd, "color=green4 ];\n");
	  break;
	case IINFL:
	  fprintf (fd, "color=violetred ];\n");
	  break;
	case SINFL:
	  fprintf (fd, "color=orange ];\n");
	  break;
	case CINFL:
	  fprintf (fd, "color=black ];\n");
	  break;
	default:
	  break;
	}
      }
    }
  }
  return;
}

void dotCGraph (arBin p, FILE* fd) {
  if (! isNull (p)) {
    dotCGraph (getArBinGauche (p), fd);
    dotNode (getArBinVar (p), fd);
    dotCGraph (getArBinDroite (p), fd);
  }
  return;
}

void graphComp (const char* argu) {
  reasoner->graphComp(argu);
}

#ifdef HAVE_GV
void launchPs (const char* fna) {
  if (ceDotPs (fna) == 1) {
    string com = "gv " + string("fig/") + string (fna) + ".ps &";
    if (-1 == system (com.c_str ()))
      fprintf (stderr, "Fail to view ps file.\n");
  }
  return;
}
#endif /* HAVE_GV */
#endif /* HAVE_DOT */

/* 12. koala-help -----------------------------------------------------*/

long ceHelp (char *s) {
  printHelp (s);
  return 0;
}

/* 13. koala-quit ----------------------------------------------------------*/

long ceQuit (char *s) {                      /* vuelve a caen con exit (0) */
  done = 1;
  return 0;
}

/* 15. koala-system --------------------------------------------------------*/

long ceSystem (char *s) {
  long r=1;

  if (-1 == system (s)) {
    fprintf (stdout, "Shell command failed.\n");
    r = 0;
  }
  return r;
}


/* 16. caen-comment -------------------------------------------------------*/

long ceComment (char *s) {

  return 1;
}

/* 17. caen-options -------------------------------------------------------*/

long ceOptions (char *s) {

  fprintf (stdout, "Options:\n");
  print_conf ();
  return 1;
}

/***********/
/* delay.c */
/***********/

/*****************************************************************/
/* Algorithm to calculate the length of the activation condition */
/*        historic array for each causal graph's influence       */
/*                   Jose Angel Jimenez Molina                   */
/*                              1999                             */
/*****************************************************************/

/* Calculate, for each causal graph's influence, 
   its activation condition historic array's length */
void actCondArrayLength(void) {

  /*static void traversePath(variable, int, listcaen); */
  void traversePath(variable, int, listcaen);

  listcaen path;
  listcaenVar lv;
  int i;

  createEmpty(&path);
  /* Traverse the causal graph beginning at each measured variable */
  for (i = 0, lv = measureListcaen; i < listcaenVarLn(measureListcaen); i++, lv = getLVNext(lv)) {
    cGraph cg;
    variable var;
      
    /* Get a measured variable from the causal graph */
    var = getLVVar(lv);
    /* Get the variable's causal graph node */
    cg = getVarNoeud(var);
    /* Traverse the causal graph beginning at "var", 
       and calculate each influence's maximum delay path */
    traversePath(var, 0, path);
  }
  return;
}

/* Traverse the causal graph beginning at "var", 
   and calculate each influence's maximum delay path */
/*static */ void traversePath(variable from, int accDelay, listcaen path) {

  /*static long inListcaenVar(variable, listcaenVar);*/ /* ??? */
  long inListcaenVar(variable, listcaenVar);

  listcaen newPath;
  cGraph cg;
  int i;

  /* Get the variable's causal graph node */
  if ((cg = getVarNoeud(from)) == NULL)
    /* Then, the variable is declared as measured, but it is isolated in the causal graph.
       That is, it does not appear in any influence. Maybe it is just there to be used
       in the global constraint equations */
    return;

  if (isPresent(from, path, NULL))
    return;

  /* If the current variable ('from') is NOT already present in the traversed path (no-loop check),
     then update (if necessary) the path's weight and continue to traverse the causal graph */
  
  /* Update traversed path with our current position (variable 'from') */
  assign(&newPath, path);
  addRight(from, &newPath);

  /* For each ascendant influence ... */
  for (i = 0; i < getCGNCause(cg); i++) {
    influence infl = getCGCauseNth(cg, i);
    variable ascendant = getInfluCause(infl);
    long delay, newAccDelay;
        
    /* If the accumulated delay through the current path is greater than the
       activation condition historic array's length, then update its length */
    if (accDelay > getInfluNActCond(infl))
      setInfluNActCond(infl, accDelay);

    /* The search is stopped if the ascendant variable is an input or a measured variable */
    if (!inListcaenVar(ascendant, measureListcaen) && !inListcaenVar(ascendant, inputListcaen)) {
      /* Get the influence's delay */
      switch(getObjType(getInfluDelay(infl))) {
      case REAL:
	delay = (int)getValReal(objVal(getInfluDelay(infl)));
	break;
      case RANGE:
	delay = (int)getRangeSup(getValRange(objVal(getInfluDelay(infl))));
	break;
      default:
	/* ERROR */
	delay = 0;
      }
      /* The accumulated delay takes into account whether the influence we
	 will follow up to the next ascendant variable is static or not */
      /* modif by beniz 14/07/01 CANCELLED. */
      newAccDelay = accDelay + (getInfluType(infl) == SINFL ? delay : delay + 1);
      /*newAccDelay = accDelay + delay;*/
      /* Next trip in our itinerary, destination to "ascendant variable" land */
      traversePath(ascendant, newAccDelay, newPath);
    }
  }
  /* Once back from "ascendant variables'" land, we liberate unuseful allocated memory */
  liberate(&newPath);

  return;
}

/*static*/ long inListcaenVar(variable var, listcaenVar lv) {
  listcaenVar lv_aux;
  int i;

  for (i = 0, lv_aux = lv; i < listcaenVarLn(lv); i++, lv_aux = getLVNext(lv_aux))
    if (var == getLVVar(lv_aux))
      return 1;

  return 0;
}


/******************/
/* End of delay.c */
/******************/

/***********/
/* listcaen.c */
/**********/

/*************************************************************/
/*  Single-Linked Listcaen Abstract Data Type, designed for the  */
/* Fault Isolation Module based on the Hitting-Set Algorithm */
/*                 Jose Angel Jimenez Molina                 */
/*                            1999                           */
/*************************************************************/

void createEmpty(ptListcaen pL) {
  pL->first = NULL;
  pL->last = NULL;
  pL->current = NULL;
  pL->nEl = 0;
  return;
}

void addLeft(ELEMENT e, ptListcaen pL) {
  ptNode newNode;

  if (isEmptyListcaen(*pL)) {
    pL->first = (ptNode)malloc(sizeof(node));
    pL->last = pL->first;
    pL->first->datum = e;
    pL->first->next = NULL;
  }
  else {
    newNode = (ptNode)malloc(sizeof(node));
    newNode->datum = e;
    newNode->next = pL->first;
    pL->first = newNode;
  }
  pL->nEl++;
  return;
}

void addLeftC(ELEMENT e, ptListcaen pL, int datumSize) {
  ptNode newNode;

  if (isEmptyListcaen(*pL)) {
    pL->first = (ptNode)malloc(sizeof(node));
    pL->last = pL->first;
    pL->first->datum = (ELEMENT)malloc(datumSize);
    memcpy(pL->first->datum, e, datumSize);
    pL->first->next = NULL;
  }
  else {
    newNode = (ptNode)malloc(sizeof(node));
    newNode->datum = (ELEMENT)malloc(datumSize);
    memcpy(newNode->datum, e, datumSize);
    newNode->next = pL->first;
    pL->first = newNode;
  }
  pL->nEl++;
  return;
}

void addRight(ELEMENT e, ptListcaen pL) {	
  if (isEmptyListcaen(*pL)) {
    pL->first = (ptNode)malloc(sizeof(node));
    pL->last = pL->first;
    pL->first->datum = e;
    pL->first->next = NULL;
  }
  else {
    pL->last->next = (ptNode)malloc(sizeof(node));
    pL->last = pL->last->next;
    pL->last->datum = e;
    pL->last->next = NULL;
  }
  pL->nEl++;
  return;
}

void addRightC(ELEMENT e, ptListcaen pL, int datumSize) {	
  if (isEmptyListcaen(*pL)) {
    pL->first = (ptNode)malloc(sizeof(node));
    pL->last = pL->first;
    pL->first->datum = (ELEMENT)malloc(datumSize);
    memcpy(pL->first->datum, e, datumSize);
    pL->first->next = NULL;
  }
  else {
    pL->last->next = (ptNode)malloc(sizeof(node));
    pL->last = pL->last->next;
    pL->last->datum = (ELEMENT)malloc(datumSize);
    memcpy(pL->last->datum, e, datumSize);
    pL->last->next = NULL;
  }
  pL->nEl++;
  return;
}

ELEMENT lookLeftC(listcaen l, int datumSize) {
  ELEMENT elm;

  if (!isEmptyListcaen(l)) {
    elm = (ELEMENT)malloc(datumSize);
    memcpy(elm, l.first->datum, datumSize);
    return elm;
  }
  return NULL;
}

ELEMENT lookRightC(listcaen l, int datumSize) {
  ELEMENT elm;

  if (!isEmptyListcaen(l)) {
    elm = (ELEMENT)malloc(datumSize);
    memcpy(elm, l.last->datum, datumSize);
    return elm;
  }
  return NULL;
}

void deleteLeft(ptListcaen pL) {
  ptNode aux;

  if (isEmptyListcaen(*pL))
    return;
  aux = pL->first;
  pL->first = pL->first->next;
  if (pL->first == NULL)
    pL->last = NULL;
  free(aux);
  pL->nEl--;
  return;
}

void deleteLeftC(ptListcaen pL) {
  ptNode aux;

  if (isEmptyListcaen(*pL))
    return;
  aux = pL->first;
  pL->first = pL->first->next;
  if (pL->first == NULL)
    pL->last = NULL;
  free(aux->datum);
  free(aux);
  pL->nEl--;
  return;
}

void deleteRight(ptListcaen pL) {
  ptNode aux;
  
  if (isEmptyListcaen(*pL))
    return;
  if (pL->first == pL->last) {
    free(pL->first);
    pL->first = NULL;
    pL->last = NULL;				
  }
  else{
    aux = pL->first;
    while (aux->next != pL->last)
      aux = aux->next;
    aux->next = NULL;
    free(pL->last);
    pL->last = aux;
  }
  pL->nEl--;
  return;
}

void deleteRightC(ptListcaen pL) {
  ptNode aux;
  
  if (isEmptyListcaen(*pL))
    return;
  if (pL->first == pL->last) {
    free(pL->first->datum);
    free(pL->first);
    pL->first = NULL;
    pL->last = NULL;				
  }
  else{
    aux = pL->first;
    while (aux->next != pL->last)
      aux = aux->next;
    aux->next = NULL;
    free(pL->last->datum);
    free(pL->last);
    pL->last = aux;
  }
  pL->nEl--;
  return;
}

int isPresent(ELEMENT e, listcaen l, int (*IG)(void *, void *)) {
  ptNode aux;
  
  if (isEmptyListcaen(l))
    return NO;

  aux = l.first;
  do {
    if (IG == NULL) {
      if (aux->datum == e)
	return YES;
    }
    else {
      if ((*IG)(aux->datum, e))
	return YES;
    }
    aux = aux->next;
  } while (aux != NULL);
  
  return NO;
}

void assign(ptListcaen pnew, listcaen old) {
  int i;

  createEmpty(pnew);
  for (i = 1; i <= listcaenLength(old); i++) {
    addRight(old.first->datum, pnew);
    old.first = old.first->next;
  }
  return;
}

void assignC(ptListcaen pnew, listcaen old, int datumSize) {
  int i;

  createEmpty(pnew);
  for (i = 1; i <= listcaenLength(old); i++) {
    addRightC(old.first->datum, pnew, datumSize);
    old.first = old.first->next;
  }
  return;
}

void liberate(ptListcaen pL) {
  int i, length;

  length = listcaenLength(*pL);
  for (i = 1; i <= length; i++)
    deleteLeft(pL);
  return;
}

void liberateC(ptListcaen pL) {
  int i, length;

  length = listcaenLength(*pL);
  for (i = 1; i <= length; i++)
    deleteLeftC(pL);
  return;
}

/*****************/
/* End of listcaen.c */
/*****************/
