/*    
 *  Copyright 1999, 2000, 2001, 2002 Jose Angel Jimenez Molina, Emmanuel Benazera
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/* caen-algo.h */

#include <setjmp.h>
#include "caen-objet.h"
#include "koaconf.h"
#include "../itms/itms.h"

/**********/
/* list.h */
/**********/

/*************************************************************/
/*  Single-Linked List Abstract Data Type, designed for the  */
/* Fault Isolation Module based on the Hitting-Set Algorithm */
/*                 Jose Angel Jimenez Molina                 */
/*                            1999                           */
/*************************************************************/

#define NO 0
#define YES 1

typedef void *ELEMENT;

typedef struct NODE{
	ELEMENT datum;
	struct NODE *next;
} node, *ptNode;

typedef struct LISTCAEN {
	ptNode first,last,current;
	int nEl;
} listcaen, *ptListcaen;

#define listcaenLength(l)      ((l).nEl)
#define isEmptyListcaen(l)     (listcaenLength(l) == 0)
#define lookLeft(l) 	   ((!isEmptyListcaen(l)) ? (l).first->datum : NULL)
#define lookRight(l)       ((!isEmptyListcaen(l)) ? (l).last->datum : NULL)

void createEmpty(ptListcaen pL);
void addLeft(ELEMENT e, ptListcaen pL);
void addLeftC(ELEMENT e, ptListcaen pL, int DatumSize);
void addRight(ELEMENT e, ptListcaen pL);
void addRightC(ELEMENT e, ptListcaen pL, int DatumSize);
ELEMENT lookLeftC(listcaen l, int DatumSize);
ELEMENT lookRightC(listcaen l, int DatumSize);
void deleteLeft(ptListcaen pL);
void deleteLeftC(ptListcaen pL);
void deleteRight(ptListcaen pL);
void deleteRightC(ptListcaen pL);
int isPresent(ELEMENT e, listcaen l, int (*IG)(void *, void *));
void assign(ptListcaen, listcaen);
void assignC(ptListcaen, listcaen, int);
void liberate(ptListcaen pL);
void liberateC(ptListcaen pL);

/*****************/
/* End of listcaen.h */
/*****************/

#define MAXMOD 100
#define MDLSUFFIX ".mdl"
#define CESUFFIX ".ce"
#define CPSUFFIX "_cf.cp"
#define CFSUFFIX "_cf.c"
#define HFSUFFIX "_cf.h"
#define CEFSUFFIX "_xt.c"
#define HEFSUFFIX "_xt.h"
#define CRTSUFFIX "_rt.c"
#define HRTSUFFIX "_rt.h"
#define BUILDSUFFIX ".build"
#define CI1SUFFIX "_i1.c"
#define CI2SUFFIX "_i2.c"

#define CLAUSES_NUM_BREAK 1000

extern itms* reasoner;
extern long done;
extern vector<string>* generated_types;
extern arDef symCons;
extern arDef conSymb;
extern char  appliName[], tmpDir[], figDir[];
extern char *nomComp;
extern string cl_header;   /* header for clauses files */
extern vector<string>* cl_files;

extern void cleanDefs ();
extern void mergeLists ();
extern void buildGlobCOfGv (listcaenGVar);
extern void buildCondCOfGv (listcaenGVar);
extern void buildCGraph (arBin);
#define MAXCAUSE 100
#define MAXEFFET 100
extern void buildCNode (variable);
extern long numTree, maxDepth;
extern bool compCGraphDepths (arBin);
extern bool compVarDepth (variable);
extern bool compTreeDepths (variable, long);
#define MAXDEPTHLISTCAEN 100
extern void sortCGraph (arBin);
extern listcaenVar depthListcaen[];
extern void sortVariable (variable);
extern void actCondArrayLength(void);
extern void checkInits (listcaenVar);
extern char appliPath[], runTDir[], runTFile[], runTHead[],
  rTi1File[], rTi2File[], defDir[], defHFile[], defCPPFile[];
extern FILE *runT, *runH, *rTi1, *rTi2, *defCPP, *defH;
extern long numVar, numInput, numMeasure, numDisplay, numTemporal,
  numUntemporal, numInflu, numNoeud, numInit, numGVar, numGVarSV, numGlobC, numConditions,
  numCondC, numVarG, numNoeudG, numInfluG, numMFExpr;
extern int twc;  /* temporal window continuous */

extern vector<cond>* condis;
extern vector<cond>* conditions;
extern vector<pair<string, StateVar*> >* variables_disc;
extern vector<component>* comps;
extern map<int, component>* oco;
extern StateVar* findDiscVariable (Comp* co, string s);
extern StateVar* findDiscVariable (Comp* co, Comp* co_imp, string s);
extern Atom* findAtom (string other_name, Comp* co);
extern Comp* findComp (string var_name);

extern double timeStep;
extern bool inVector(string s, vector<string>* vec);
extern bool isCommandComp (Comp* co, string ty);
extern bool isIncludedInComp (Comp* co, string ty);
extern bool isIncludedInComp (Comp* co, string ty, vector<Comp*>* nc);
void updateEltsNum (Comp* co);
void updateExprNum (expr e, expr et);
extern string geneContSymbName (variable va, Comp* co, int influ, string nums, string numob, string numj, string numob2);
extern string geneDiscSymbName (StateVar* sta, Comp* co, int cla, string num, string bol, string neg, string bol1);
extern string geneCondName (cond cd, Comp* co, long op, string par);
extern void geneClausesFileHeaders (Comp* co);
extern void geneClassHeaders (const char* ty);
extern void geneComp (Comp* co);
extern void geneUse (Comp* co);
/*extern void geneCommands (Comp* co);*/
extern void geneStateVar (Comp* co, StateVar* sta, vector<variable>* rcdv);
extern void geneConstraints (Comp* co);
extern void geneClause (Comp* co, Clause* clau, StateVar* sta);
extern void geneTransition (Comp* co, Transition* tr);
extern void geneStateformula (Comp* co, StateFormula* stfr, string ntrans, bool pre);
extern void geneVarDef (FILE *, listcaenVar);
extern void geneVariable (FILE* fd, FILE* fdh, Comp* co, variable va);
extern void geneVarV (listcaenVar);
extern string geneCondition (FILE* fd, FILE* fdh, cond, Comp* co, long op, string par);
extern bool inVectorCond (cond, vector<cond>*);
extern bool inVector (variable, vector<variable>*);
extern void geneInfluence (FILE* fd, FILE* fdh, influence, Comp* co);
extern void geneNode (FILE* fd, FILE* fdh, Comp* co, variable);
extern void geneInitVar (initVar);
extern void geneInitV (listcaenInitVar, Comp* co);
extern long geneGVar (FILE* fd, FILE* fdh, gVar, Comp*);
extern void geneGlobVs (FILE* fd, FILE* fdh, listcaenGVar, Comp* co);
extern long numExpr;
extern void geneExpr (FILE* fd, FILE* fdh, expr e, Comp* co, bool cond);
extern void geneGlobCs (FILE* fd, FILE* fdh, listcaenExpr, Comp* co);
extern void geneCondCs (FILE* fd, FILE* fdh, listcaenExpr, Comp* co);
extern void printDic (void);
extern void printRules (void);
extern void printConditions (void);
extern void printComps (void);
extern void printComp (const char* argu);
extern void printAtoms (void);
extern void printAllVars (void);
extern void printCGraph (arBin);
extern void printNode (variable);
#ifdef HAVE_DOT
extern int ceDotPs (const char* fna);
extern void geneDotHeaders (FILE* fd);
extern void dotNode (variable v, FILE* fd);
extern void dotCGraph (arBin p, FILE* fd);
extern void graphComp (const char* argu);
extern void launchPs (const char* fna);
#endif /* HAVE_DOT */
extern void printGlobCs (void);
extern void printCondCs (void);
extern void print_conf (void);
extern long ceHelp (char *);
extern long ceQuit (char *);
extern long ceSystem (char *);
extern long ceComment (char *);
