/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/* caen-objet.cpp */
#include "caen-objet.h"
#include <iostream>
#include "Atom.h"
#include "StateFormula.h"

#ifdef READLINE_LIBRARY
#  include "readline.h"
#  include "history.h"
#else
#  include <readline/readline.h>
#  include <readline/history.h>
#endif

/* objets-level temps, for macro calls only */
#ifdef DEBUG
static long L1;
#endif
static valeur Vl1;
static range Rn1;
static arDef Df1;
static objet Obj1;
static domain Dom1;
static coef Cf1;
static arBin Nd1;
static variable Va1;
static listcaenVar Lv1;
static cond Cn1;
static influence In1;
static initVar Iv1;
static mapfVar Mv1;
static expr Ex1;
static listcaenExpr Lex1;
static gVar Gv1;
static listcaenGVar Lg1;
static listcaenInfl Lif1;
static listcaenInitVar Liv1;
/*static listcaenLong Ll1;*/

arBin dictionary;
listcaenParFor tempDefs = NULL;
long loadError;

long numinfs = 0;

/* conversion chaine ("FALSE" ou "TRUE") -> booleen */
enum boolVal a2Bool (char *s) {
  enum boolVal b=CE_FALSE;

  if ((! strcmp (s, "TRUE")) || (! strcmp (s, "true")))
    b = CE_TRUE;
  else if ((strcmp (s, "FALSE")) && (strcmp (s, "false")))
    implError (2);
  return b;
}


void printRange (FILE *fd, range r) {
  
  if (isIncludedInf (r))
    fprintf (fd, "[");
  else fprintf (fd, "]");
  printReal (fd, getRangeInf (r));
  fprintf (fd, ", ");
  printReal (fd, getRangeSup (r));
  if (isIncludedSup (r))
    fprintf (fd, "]");
  else fprintf (fd, "[");
  return;
}


long eqRange (range a, range b) {
  double ainf=getRangeInf(a), asup=getRangeSup(a),
    binf=getRangeInf(b), bsup=getRangeSup(b);
  long r=0;

  if ((eqReal (ainf, binf)) && (eqReal (asup, bsup)))
    r = 1;
  return r;
}


void printReal (FILE *fd, double vd) {

  if (-DBL_MAX == (vd))
    fprintf (fd, "-oo");
  else if (DBL_MAX == vd)
    fprintf (fd, "+oo");
  else
    fprintf (fd, "%g", vd);
  return;
}


range maxRange, posRange, posRangeS, negRange, negRangeS;
valeur valZero, valUn, valMaxDbl, valMaxRange, valFalse, valTrue;
objet objFalse, objTrue, objZero, objUn, objMaxDbl, objPos, objPosS, objNeg, objNegS;
/*expr exprNeg, exprPos;*/

void objInit (void) {

  maxRange = mkRange (-DBL_MAX, DBL_MAX);
  valZero = mkValReal (0.0);
  valUn = mkValReal (1.0);
  valMaxDbl = mkValReal (DBL_MAX);
  valMaxRange = mkValRange (maxRange);
  valFalse = mkValBool (CE_FALSE);
  valTrue = mkValBool (CE_TRUE);
  objFalse = mkObjBool (CE_FALSE);
  objTrue = mkObjBool (CE_TRUE);
  objZero = mkObjReal (0.0);
  objUn = mkObjReal (1.0);
  objMaxDbl = mkObjReal (DBL_MAX);
  posRange = mkRange (0, DBL_MAX);
  posRangeS = mkRange (0, DBL_MAX);
  setNotIncludedInf (posRangeS);
  negRange = mkRange (-DBL_MAX, 0);
  negRangeS = mkRange (-DBL_MAX, 0);
  setNotIncludedSup (negRangeS);
  objPos = mkObjRange (posRange);
  objPosS = mkObjRange (posRangeS);
  objNeg = mkObjRange (negRange);
  objNegS = mkObjRange (negRangeS);
  return;
}

 /**
 * set a as a*b.
 */
void ce_mulObjet (objet a, objet b) {
  enum objType ta=getObjType(a), tb=getObjType(b);

  if (ta == REAL) {
    double da=getObjNVal(a);
    if (da != 0.0) {
      if (tb == REAL) {
	setObjNVal (a, da * getObjNVal (b));
      }
      else { /* tb */
	double dc, dd, de, binf=getRangeInf(getObjRVal(b)),
	  bsup=getRangeSup(getObjRVal(b));
	setObjType (a, RANGE);
	dc = da*binf, dd = da*bsup;
	if (dc > dd)
	  de = dc, dc = dd, dd = de;
	setRangeInf (getObjRVal (a), dc);
	setRangeSup (getObjRVal (a), dd);
	setBounds (getObjRVal (a), getObjRVal (b));
      } /* endif tb */
    }
  }
  else { /* ta */
    double dc, dd, de, ainf=getRangeInf(getObjRVal(a)),
      asup=getRangeSup(getObjRVal(a));
    if (tb == REAL) {
      double db=getObjNVal(b);
      if (db == 0.0) {
	setObjType (a, REAL);               /* ??? range ??? was bug ?? */
	setObjNVal (a, 0.0);
      }
      else { /* db */
	setObjType (a, RANGE);
	dc = db*ainf, dd = db*asup;
	if (dc > dd)
	  de = dc, dc = dd, dd = de;
	setRangeInf (getObjRVal (a), dc);
	setRangeSup (getObjRVal (a), dd);
      }
    }
    else { /* tb */
      double df, dmin, dmax, binf=getRangeInf(getObjRVal(b)),
	bsup=getRangeSup(getObjRVal(b));
      setObjType (a, RANGE);
      dc = ainf*binf, dd = ainf*bsup, de = asup*binf, df = asup*bsup;
      dmin = min (dc, min (dd, min (de, df)));
      dmax = max (dc, max (dd, max (de, df)));
      setRangeInf (getObjRVal (a), dmin);
      setRangeSup (getObjRVal (a), dmax);
    } /* endif tb */
  } /* endif ta */
  return;
}

 /**
 * set a as a/b and return 0 if the division is impossible (1 otherwise).
 */
long ce_divObjet (objet a, objet b) {
  enum objType ta=getObjType(a), tb=getObjType(b);
  long r=1;

  if (ta == REAL) {
    double da=getObjNVal(a);
    if (tb == REAL) {
      double db=getObjNVal(b);
      if (db == 0.0)
	r = 0;                                       /* division impossible */
      else
	setObjNVal (a, da/db);
    }
    else { /* tb */
      double dc, dd, de, binf=getRangeInf(getObjRVal(b)),
	bsup=getRangeSup(getObjRVal(b));
      if (binf*bsup <= 0.0)
	r = 0;                                       /* division impossible */
      else {
	setObjType (a, RANGE);
	dc = da/bsup, dd = da/binf;
	if (dc > dd)
	  de = dc, dc = dd, dd = de;
	setRangeInf (getObjRVal (a), dc);
	setRangeSup (getObjRVal (a), dd);
	setBounds (getObjRVal (a), getObjRVal (b));
      }
    } /* endif tb */
  }
  else { /* ta */
    double dc, dd, de, ainf=getRangeInf(getObjRVal(a)),
      asup=getRangeSup(getObjRVal(a));
    if (tb == REAL) {
      double db=getObjNVal(b);
      if (db == 0.0)
	r = 0;                                       /* division impossible */
      else {
	setObjType (a, RANGE);
	dc = ainf/db, dd = asup/db;
	if (dc > dd)
	  de = dc, dc = dd, dd = de;
	setRangeInf (getObjRVal (a), dc);
	setRangeSup (getObjRVal (a), dd);
      }
    }
    else { /* tb */
      double df, dmin, dmax, binf=getRangeInf(getObjRVal(b)),
	bsup=getRangeSup(getObjRVal(b));
      if (binf*bsup <= 0.0)
	r = 0;                                       /* division impossible */
      else {
	setObjType (a, RANGE);
	dc = ainf/binf, dd = ainf/bsup, de = asup/binf, df = asup/bsup;
	dmin = min (dc, min (dd, min (de, df)));
	dmax = max (dc, max (dd, max (de, df)));
	setRangeInf (getObjRVal (a), dmin);
	setRangeSup (getObjRVal (a), dmax);
      }
    } /* endif tb */
  } /* endif ta */
  return r;
}

/**
 * reverse the sign of an object value.
 */
void ce_minusObjet (objet a) {
  enum objType ta=getObjType(a);

  if (ta == REAL) {
    double da=getObjNVal(a);
    setObjNVal (a, -da);
  }
  else { /* ta */
    double ainf=getRangeInf(getObjRVal(a)),
      asup=getRangeSup(getObjRVal(a));
    setRangeInf (getObjRVal (a), -asup);
    setRangeSup (getObjRVal (a), -ainf);
    /* inclusion */
    if (isIncludedInf (getObjRVal (a)))
      setIncludedSup (getObjRVal (a));
    else setNotIncludedSup (getObjRVal (a));
    if (isIncludedSup (getObjRVal (a)))
      setIncludedInf (getObjRVal (a));
    else setNotIncludedInf (getObjRVal (a));
  } /* endif ta */
  return;
}

/**
 * copy p values into q.
 */
void ce_putObjet (objet p, objet q) {
  enum objType t=getObjType(p);

  setObjType (q, t);
  switch (t) {
  case REAL :
    setObjNVal (q, getObjNVal (p));
    break;
  case RANGE :
    setObjRVal (q, getObjRVal (p));
    break;
  case BOL :
    setObjBVal (q, getObjBVal (p));
    break;
  default:
    break;
  }
  return;
}

/**
 * add an object b to an object a and update valeur.
 */
void ce_addObjet (objet a, objet b) {
  enum objType ta=getObjType(a), tb=getObjType(b);

  if (ta == REAL) {
    double da=getObjNVal(a);
    if (tb == REAL)
      setObjNVal (a, da + getObjNVal (b));
    else { /* tb */
      double binf=getRangeInf(getObjRVal(b)),
	bsup=getRangeSup(getObjRVal(b));
      setObjType (a, RANGE);
      setRangeInf (getObjRVal (a), da + binf);
      setRangeSup (getObjRVal (a), da + bsup);
      setBounds (getObjRVal (a), getObjRVal (b));
    } /* endif tb */
  }
  else { /* ta */
    double ainf=getRangeInf(getObjRVal(a)),
      asup=getRangeSup(getObjRVal(a));
    if (tb == REAL) {
      double db=getObjNVal(b);
      setRangeInf (getObjRVal (a), ainf + db);
      setRangeSup (getObjRVal (a), asup + db);
    }
    else { /* tb */
      double binf=getRangeInf(getObjRVal(b)),
	bsup=getRangeSup(getObjRVal(b));
      setRangeInf (getObjRVal (a), ainf + binf);
      setRangeSup (getObjRVal (a), asup + bsup);
    } /* endif tb */
  } /* endif ta */
  return;
}

/**
 * test if object a value is greater than object b value.
 */
boolVal ce_geObjet (objet a, objet b) {
  boolVal r=CE_FALSE;

  if (REAL == getObjType (a)) {
    double da=getObjNVal(a);
    if (REAL == getObjType (b)) {
      if (da >= getObjNVal (b))
	r = CE_TRUE;
    }
    else { /* REAL b */
      range rb=getObjRVal(b);
      if ((da > getRangeSup (rb))
	  || ((da == getRangeSup (rb))
	      && isIncludedSup (rb)))
	r = CE_TRUE;
    } /* endif REAL b */
  }
  else { /* REAL a */
    range ra=getObjRVal(a);
    if (REAL == getObjType (b)) {
      double db=getObjNVal(b);
      if ((getRangeInf (ra) > db)
	  || ((db == getRangeInf (ra))
	      && isIncludedInf (ra)))
	r = CE_TRUE;
    }
    else { /* REAL b */
      range rb=getObjRVal(b);
      if ((getRangeInf (ra) > getRangeSup (rb))
	  || ((getRangeInf (ra) == getRangeSup (rb))
	      && isIncludedSup (rb)))
	r = CE_TRUE;
    } /* endif REAL b */
  } /* endif REAL a */
  return r;
}

long eqObj (objet a, objet b) {           /* retourne 1 si a==b, 0  si a!=b */
  long r=0;
  
  if (getObjType (a) == getObjType (b)) {
    switch (getObjType(a)) {
    case REAL: {
      r = eqReal (getObjNVal (a), getObjNVal (b));
      break;
    }
    case RANGE: {
      r = eqRange (getObjRVal (a), getObjRVal (b));
      break;
    }
    default:
      break;
    } /* end switch */
  } /* endif */
  return r;
}


void printObjet (FILE *fd, objet ob) {

  switch (getObjType(ob)) {
  case REAL:
    fprintf (fd, " =");
    printReal (fd, getObjNVal (ob));
    break;
  case RANGE:                                              /* in [o1, o2] */
    fprintf (fd, " in ");
    printRange (fd, getObjRVal (ob));
    break;
  default:
    break;
  }
  return;
}

/**
 * generate C-code for a 'real' objet.
 */
void geneReal (FILE *fd, double vd) {

  if (-DBL_MAX == (vd))
    fprintf (fd, "-DBL_MAX");
  else if (DBL_MAX == vd)
    fprintf (fd, "DBL_MAX");
  else
    fprintf (fd, "%g", vd);
  return;
}

/**
 * generace C-code for an objet.
 */
void geneObjet (FILE *fd, objet ob, char* name) {
  range r;

  switch (getObjType(ob)) {
  case RANGE:                                              /* in [o1, o2] */
    r = getObjRVal(ob);
    fprintf (fd, "setObjType(%s, (objType)%d);\n\t", name, RANGE);
    fprintf (fd, "range r_%s = {%g,%g, %i, %i};\n\t", name, getRangeInf (r), getRangeSup (r), getBoundInf (r), getBoundSup (r));
    fprintf (fd, "setObjRangeVal(%s, r_%s);\n\t", name, name);
    break;
  case REAL:
    fprintf (fd, "setObjType(%s, (objType)%d);\n\t", name, REAL);
    /*fprintf (fd, "{(objType)%d,{{", REAL);*/
    fprintf  (fd, "range r_%s = {", name);     
    geneReal (fd, getObjNVal (ob));
    fprintf (fd, ",%g};\n\tsetObjRangeVal(%s, r_%s);\n\t", 0.0, name, name);
    break;
  case BOL:
    implError (5);
    break;
  default:
    break;
  }
  free (name);
  return;
}


long listcaenLongLn (listcaenLong la) {
  long ln=0;
 
  if (! isNull (la))
    ln = 1 + listcaenLongLn (getLLNext (la));
  return ln;
}
 
 
/* concatenation de deux listcaenes d'entierss */
listcaenLong catListcaenLong (listcaenLong la, listcaenLong lb) {
  listcaenLong lc, ld;

  if (isNull (la))
    la = lb;                        /* la vide: lb                           */
  else
    if (! isNull (lb)) {            /* lb vide: la non modifie'e             */
      lc = la;
      while (! isNull (ld = getLLNext (lc)))
	lc = ld;
      setLLNext (lc, lb);           /* concate'nation: lb en queue de la     */
    }
  return la;
}


long storeDef (arDef *q, char *nom, char *txt) {
  arDef p=*q;
  long test, r=1;

  if (isNull (p)) {                             /* noeud (arbre) vide        */
    p = mkArDef (nom, txt);
    *q = p;
  }
  else { /* isNull (p) */
    test = strcmp (nom, getArDefNom(p));
    if (test) {
      if (test <0)                              /* sinon: chercher a gauche  */
	r = storeDef (&getArDefGauche (p), nom, txt);
      else                                      /* ou: chercher a droite     */
	r = storeDef (&getArDefDroite (p), nom, txt);
    }
    else /* test */                                       /* ya fue definido */
      r = 0;
  }
  return r;
}

arDef searchDef (arDef *q, char *nom) {
  arDef r=NULL, p=*q;
  long test;

  if (! isNull (p)) {
    test = strcmp (nom, getArDefNom(p));
    if (test) {
      if (test <0)                              /* sinon: chercher a gauche  */
	r = searchDef (&getArDefGauche (p), nom);
      else                                      /* ou: chercher a droite     */
	r = searchDef (&getArDefDroite (p), nom);
    }
    else /* test */                             /* si test == 0: c'est ici ! */
      r = p;
  }
  return r;
}


void delDef (arDef *q, char *nom) {
  arDef p=*q;
  long test;

  if (! isNull (p)) {
    test = strcmp (nom, getArDefNom(p));
    if (test) {
      if (test <0)                              /* sinon: chercher a gauche  */
	delDef (&getArDefGauche (p), nom);
      else                                      /* ou: chercher a droite     */
	delDef (&getArDefDroite (p), nom);
    }
    else { /* test */                             /* si test == 0: c'est ici ! */
      *q = NULL;
      freeArDef (p);
    } /* endif test */
  } /* endif !isNull p */
  return;
}


void printArDef (arDef p) {

  if (! isNull (p)) {
    printArDef (getArDefGauche (p));
    fprintf (stdout, "  %s: %s\n", getArDefNom (p), getArDefTxt (p));
    printArDef (getArDefDroite (p));
  } /* endif ! isNull p */
  return;
}


long numCstVar;

variable const2Var (double d) {
  variable v;
  char s[10];

  sprintf (s, "cst_%ld", ++numCstVar);
  v = mkVariable (s);
  setVarType (v, REAL), setVarConst (v),
    setVarValue (v, mkObjReal (d));
  return v;
}

 
void printVariable (variable v) {
  domain d=getVarDom (v);

  fprintf (stdout, "  ");
  printVarNom (stdout, v);
  fprintf (stdout, ": ");
  if (isVarInput (v)) {
    fprintf (stdout, "input");                       /* fonction         */
    if (isVarDisplay (v))
      fprintf (stdout, ", display");
    if (isVarMeasure (v))
      fprintf (stdout, ", measure");
    fprintf (stdout, "; ");
  }
  else { /* input v */
    if (isVarDisplay (v)) {
      fprintf (stdout, "display");
      if (isVarMeasure (v))
	fprintf (stdout, ", measure");
      fprintf (stdout, "; ");
    }
    else /* display v */
      if (isVarMeasure (v))
	fprintf (stdout, "measure; ");
  } /* endif input v */
  if (isVarConst (v))
    fprintf (stdout, "parameter ");
  else
    fprintf (stdout, "variable");
  printDomain (stdout, d);
  if (! isVarConst (v))
    fprintf (stdout, "; max delay = %g\n", getVarMaxDelay (v));
  return;
}


void printVarNom (FILE *fd, variable v) {
  fprintf (fd, "%s", getVarNom (v));
  return;
}

void printVarSymbName (FILE* fd, variable v) {
  fprintf (fd, "%s", getVarSymbName (v));
  return;
}

bool isVarInList (listcaenVar la, variable v) {
  
  while (! isNull (la)) {
    variable vi = getLVVar (la);
    if (strcmp(getVarNom (vi), getVarNom (v)) == 0)
      return 1;
    la = getLVNext (la);
  }
  return 0;
}

long listcaenVarLn (listcaenVar la) {
  long ln=0;
 
  if (! isNull (la))
    ln = 1 + listcaenVarLn (getLVNext (la));
  return ln;
}
 
 
/* construction d'une listcaene de variables a` partir d'une listcaene de variables
   et d'une variable */
listcaenVar appVariable (listcaenVar la, variable a) {
  listcaenVar elt, lc, ld;

  if (! isNull (a)) {
    elt = mkListcaenVar (a);                /* cre'e un element de variable-listcaen         */
    if (isNull (la))
      la = elt;                       /* creee la listcaene la                     */
    else {
      lc = la;
      while (! isNull (ld = getLVNext (lc)))
	lc = ld;
      setLVNext (lc, elt);            /*   ... l'ajoute a la listcaene sl          */
    } /* endif isNull la */
  } /* endif ! isNull a */
  return la;
}
 
 
/* concatenation de deux listcaenes de variables */
listcaenVar catListcaenVar (listcaenVar la, listcaenVar lb) {
  listcaenVar lc, ld;

  if (isNull (la))
    la = lb;                        /* la vide: lb                           */
  else
    if (! isNull (lb)) {            /* lb vide: la non modifie'e             */
      lc = la;
      while (! isNull (ld = getLVNext (lc)))
	lc = ld;
      setLVNext (lc, lb);           /* concate'nation: lb en queue de la     */
    }
  return la;
}


/* impression des noms d'une listcaene de variables */
void printListcaenVar (FILE *outs, listcaenVar la) {
  
  if (isNull (la))
    fprintf (outs, ".\n");
  else {
    variable v=getLVVar(la);
    listcaenVar lb=getLVNext (la);
    printVarNom (outs, v);
    if (! isNull (lb))
      fprintf (outs, ", ");
    printListcaenVar (outs, lb);
  }
  return;
}


/* impression des noms et des profondeurs d'une listcaene de variables */
void printUpdateListcaen (FILE *outs, listcaenVar la) {
  
  if (isNull (la))
    fprintf (outs, ";");
  else {
    variable v=getLVVar(la);
    listcaenVar lb=getLVNext (la);
    printVarNom (outs, v);
    fprintf (outs, ":%ld", getVarDepth (v));
    if (! isNull (lb))
      fprintf (outs, ", ");
    printUpdateListcaen (outs, lb);
  }
  return;
}


/* donner une fonction aux variables d'une listcaene */
void setListcaenFct (listcaenVar la, enum interType f) {

  if (! isNull (la)) {
    variable v=getLVVar(la);
    if (isVarConst (v)) {
      switch (f) {
      case INPUT :
      case DISPLAY :
	semError (3, 0, 31);   /* parameter in Display declaration. */
	break;
      case MEASURE :
	semError (1, 0, 30); /* parameter in Input or Measure declaration. */
	break;
      default:
	break;
      } /* end switch */
    }
    else { /* isVarConst v*/
      if (! (getVarFct (v) & f)) {     /* es decir: v ya tiene la funcion f */
	switch (getVarFct (v)) {
	case INPUT :
	  switch (f) {
	  case DISPLAY :
	    setVarFct (v, INDI);
	    break;
	  case MEASURE :
	    semError (0, nligne, 52); /* variable can't be both Input and Measure */
	    break;
	  default:
	    break;
	  } /* end switch f */
	  break;
	case DISPLAY :
	  switch (f) {
	  case INPUT :
	    setVarFct (v, INDI);
	    break;
	  case MEASURE :
	    setVarFct (v, DIME);
	    break;
	  default:
	    break;
	  } /* end switch f */
	  break;
	case INDI :
	  if (f == MEASURE)
	    semError (0, nligne, 52); /* variable can't be both Input and Measure */
	  break;
	case MEASURE :
	  switch (f) {
	  case INPUT :
	    semError (0, nligne, 52); /* variable can't be both Input and Measure */
	    break;
	  case DISPLAY :
	    setVarFct (v, DIME);
	    break;
	  default:
	    break;
	  } /* end switch f */
	case DIME :
	  if (f == INPUT)
	    semError (0, nligne, 52); /* variable can't be both Input and Measure */
	  break;
	default:
	  setVarFct (v, f);
	  break;
	} /* end switch f */
      } /* endif ! (fct v & f) */
      setListcaenFct (getLVNext (la), f);
    } /* endif isVarConst */
  } /* endif ! isNull la */
  return;
}


/* donner un type aux variables d'une listcaene */
void setListcaenType (listcaenVar la, enum objType t) {
  variable v;
  enum objType tv;

  if (! isNull (la)) {
    v = getLVVar (la);
    tv = getTheVarType (v);
    if (tv == UNKNOWN)
      setVarType (v, t);
    else if (tv != t)
      semError (0, nligne, 6);
    setListcaenType (getLVNext (la), t);
  }
  return;
}


/* donner le qualifieur "const" aux variables d'une listcaene */
void setListcaenConst (listcaenVar la) {

  if (! isNull (la)) {
    variable v=getLVVar(la);
    domain d=getVarDom(v);
    if (isNull (d))
      semError (0, nligne, 10);           /* "parameter should be initialized." */
    else
      setVarConst (v);
    setListcaenConst (getLVNext (la));
  } /* endif ! isNull la */
  return;
}


void printDomain (FILE *fd, domain d) {

  if (! isNull (d)) {
    if (RANGE == getObjType (getDomObj (d))) {
      fprintf (fd, " in ");
      printRange (fd, getDomRVal (d));
    }
    else {
      fprintf (fd, " =");
      printReal (fd, getDomNVal (d));
    } /* endif RANGE */
  } /* endif ! isNull d */
  return;
}

variable dom2Var (variable v, enum opRel op, domain d) {
  objet ob=getDomObj(d);
  variable vv=getDomVar(d);
  domain dd;

  if (IN == op) {
    if ((isNull (ob)) || (RANGE != getObjType (ob))) {
      semError (0, nligne, 28);    /* "invalid object beside IN operator." */
      v = NULL;
    }
    else
      setVarDom (v, d);                       /* range */
  }
  else if (EQ == op) {                        /* real ou variable */
    if (isNull (vv)) {
      if (RANGE == getObjType (ob)) {
	semError (0, nligne, 12); /* "invalid relation operator before range." */
	v = NULL;
      }
      else
	setVarDom (v, d);                     /* real */
    }
    else { /* isNull vv */                        /* variable */
      if (! isVarConst (vv)) {
	semError (0, nligne, 9);
	/* "initialization should be done with a parameter." */
	v = NULL;
      }
      else { /* ! isVarConst vv */
	d = getVarDom (vv);
	dd = copyDom (d);
	setVarDom (v, dd);
      } /* endif isVarConst vv */
    } /* endif isNull vv */
  } else { /* EQ == op */
    semError (0, nligne, 21);        /* "invalid operator in variable declaration." */
    v = NULL;
  } /* endif EQ == op */
  return v;
}


/* concatenation de deux listcaenes de parametres formels */
listcaenParFor catListcaenParFor (listcaenParFor la, listcaenParFor lb) {
  listcaenParFor lc, ld;
 
  if (isNull (la))
    la = lb;                                     /* la vide: lb                           */
  else
    if (! isNull (lb)) {                     /* lb vide: la non modifie'e             */
      lc = la;
      while (! isNull (ld = getLPFNext (lc)))
        lc = ld;
      setLPFNext (lc, lb);                /* concate'nation: lb en queue de la     */
    }
  return la;
}


long listcaenParEffLn (listcaenParEff la) {
  long ln=0;
 
  if (! isNull (la))
    ln = 1 + listcaenParEffLn (getLPENext (la));
  return ln;
}
 
 
/* concatenation de deux listcaenes de parametres effectifs*/
listcaenParEff catListcaenParEff (listcaenParEff la, listcaenParEff lb) {
  listcaenParEff lc, ld;
 
  if (isNull (la))
    la = lb;                                     /* la vide: lb                           */
  else
    if (! isNull (lb)) {                     /* lb vide: la non modifie'e             */
      lc = la;
      while (! isNull (ld = getLPENext (lc)))
        lc = ld;
      setLPENext (lc, lb);                /* concate'nation: lb en queue de la     */
    }
  return la;
}

 
/* recherche et insertion dans le dictionnaire organise'
   en arbre binaire; on passe le pointeur du noeud (racine, ou successeur)
   afin de pouvoir modifier la valeur du noeud (inse'rer un symbole);
   la valeur de retour est le noeud qui contient le symbole s.
*/
variable name2Var (arBin *q, char *s) {
  long test;                        /* yytext de'clare' char[] dans lex.yy.c */
  arBin p=*q;
  variable v;

  if (isNull (p)) {                             /* noeud (arbre) vide        */
    v = mkVariable (s);
    p = mkArBin (v);
    *q = p;
  }
  else {
    test = strcmp (s, getArBinNom(p));
    if (test) {
      if (test <0)                              /* sinon: chercher a gauche  */
	v = name2Var (&getArBinGauche (p), s);
      else                                      /* ou: chercher a droite     */
	v = name2Var (&getArBinDroite (p), s);
    }
    else {                                      /* si test == 0: c'est ici ! */
      v = getArBinVar(p);
      setVarOld (v);
    }
  }
  return v;
}


void printArBin (arBin p) {

  if (! isNull (p)) {
    printArBin (getArBinGauche (p));
    printVariable (getArBinVar (p));
    cout << endl;
    printArBin (getArBinDroite (p));
  }
}


coef dom2Coef (enum coefType ct, enum opRel op, domain d) {
  objet ob=getDomObj(d);
  coef cf=NULL;
  variable v=getDomVar(d);

  if (IN == op) {
    if ((isNull (ob)) || (RANGE != getObjType (ob)))
      semError (0, nligne, 28);    /* "invalid object beside IN operator." */
    else
      cf = mkCoefObj (ct, mkObjRange (copyRange (getDomRVal (d))));/* range */
  }
  else if (EQ == op) {
    if (isNull (v)) {
      if (RANGE == getObjType (ob))
	semError (0, nligne, 12); /* "invalid relation operator before range." */
      else
	cf = mkCoefObj (ct, mkObjReal (getDomNVal(d)));
    }
    else { /* isNull v */                         /* variable */
      if (isVarConst (v)) {
	d = getVarDom (v);
	ob = getDomObj (d);
	if (isNull (getDomVar (d))) {
	  if (RANGE == getObjType (ob))
	    cf = mkCoefObj (ct, mkObjRange (copyRange (getDomRVal (d))));
	  else
	    cf = mkCoefObj (ct, mkObjReal (getDomNVal(d)));
	}
	else
	  implError (7);
      }
      else /* isVarConst v */
	cf = mkCoefVar (ct, v);
    } /* endif isNull v */
  } else
    semError (0, nligne, 8);   /* "invalid operator in influence coefficient." */
  return cf;
}


/* concatenation de deux listcaenes de coefficients */
listcaenCoef catListcaenCoef (listcaenCoef la, listcaenCoef lb) {
  listcaenCoef lc, ld;
 
  if (isNull (la))
    la = lb;                                     /* la vide: lb                           */
  else
    if (! isNull (lb)) {                     /* lb vide: la non modifie'e             */
      lc = la;
      while (! isNull (ld = getLCNext (lc)))
        lc = ld;
      setLCNext (lc, lb);                 /* concate'nation: lb en queue de la     */
    }
  return la;
}


long numCond;

/* make a type COMP condition. DEPRECATED ? */
cond mkCondComp (variable v, enum opRel op, domain d) {
  cond c;
  variable vv;

  c = mkCond();
  setCondType (c, CDCOMP);
  setCondVar (c, v);
  setCondRel(c, op);
  if (isNull (vv = getDomVar (d)))
    setCondDomObj (c, getDomObj (d));
  else
    setCondDomVar (c, vv);
  return c;
}

cond mkCondExpr (expr e1, enum opRel op, expr e2) {
  cond c;
  c = mkCond ();
  setCondType (c, CDEXPR);
  setCondVar (c, getGVarVar (getExprGVar (e1)));
  if (getExprType (e2) == CTE)
    setCondDomObj (c, getGVarObj (getExprGVar (e2)));
  setCondRel (c, op);
  setCondExpr (c, e2);
  return c;
}

void printCondition (cond cd) {
  enum condType ct;
  variable v;

  if (! isNull (cd)) {
    ct = getCondType (cd);
    fprintf (stdout, "%s: (", getCondName (cd));
    switch (ct) {
    case CDCTE:
      if (CE_TRUE == getCondValue (cd))
	fprintf (stdout, "true");
      else
	fprintf (stdout, "false");
      break;
    case CDEXPR:
      printGVar (stdout, getExprGVar (getCondExpr (cd)));
      /*cout << endl;*/
      printExpr (getCondExpr (cd));
      if (! isNull (getCondAtom (cd))) {
	/*cout << endl;*/
	getCondAtom (cd)->showAtom (0);
	/*cout << endl;*/
      }
      break;
    case CDCOMP:
      printVarNom (stdout, getCondVar (cd));
      switch (getCondRel (cd)) {
      case EQ:
	fprintf (stdout, " = ");
	break;
      case NE:
	fprintf (stdout, " != ");
	break;
      case LE:
	fprintf (stdout, " <= ");
	break;
      case LT:
	fprintf (stdout, " < ");
	break;
      case GE:
	fprintf (stdout, " >= ");
	break;
      case GT:
	fprintf (stdout, " > ");
	break;
      case IN:
	break;
      default:
	implError (8);
	break;
      }
      if (isNull (v = getCondDomVar (cd))) {
	objet ob=getCondDomObj (cd);
	if (RANGE == getObjType (ob))
	  printRange (stdout, getObjRVal (ob));
	else /* RANGE */
	  printReal (stdout, getObjNVal (ob));
      }
      else /* isNull v */
	printVarNom (stdout, v);
      break;
    case CDAT:
      getCondAtom (cd)->showAtom (0);
      break;
    case CDOU:
      printCondition (getCondOp1 (cd));
      fprintf (stdout, " | ");
      printCondition (getCondOp2 (cd));
      break;
    case CDET:
      printCondition (getCondOp1 (cd));
      fprintf (stdout, " & ");
      printCondition (getCondOp2 (cd));
      break;
    case CDNON:
      fprintf (stdout, "! ");
      printCondition (getCondOp1 (cd));
      break;
    default :
      implError (6);
      break;
    } /* end switch ct */
    fprintf (stdout, ")\n");
  } /* endif ! isNull cd */
  return;
}


influence buildInfluence (Comp* c, enum influType tp, variable cause,
			  variable effet, listcaenCoef lc) {
  influence fl=NULL;
  long isNullTRep, isGain=0, isDelay=0, isXi=0, isPulse=0, isRespTime=0, isN1=0, isN2=0, isA0=0, isA1=0, isA2=0;;
    /*isZeros=0;*/

  if (CINFL == tp) {
    if (! isVarConst (cause))
      semError (0, nligne, 7);
    /*"constant influence should have a parameter as cause." */
  } /* endif CINFL */
  fl = mkInfluence (tp,cause,effet);
  while (! isNull (lc)) {
    coef cf=getLCCoef(lc);
    if (! isNull (cf)) {
      switch (getCoefType (cf)) {
      case GAIN :
	if (isGain)
	  semError (2, nligne, 34); /* double definition of gain (first assumed). */
	else { /* isGain */
	  objet ob=getCoefObj (cf);
	  if (RANGE == getObjType (ob)) {
	    setUncertain (fl, 1);
	    if (hasZero (getObjRVal (ob)))
	      semError (0, nligne, 43); /* "gain range has not a unique sign." */
	  }
	  setInfluGain (fl, ob);
	  isGain = 1;
	} /* endif isGain */
	break;
      case DELAY :
	if (isDelay)
	  semError (2, nligne, 35); /*double definition of delay (first assumed). */
	else { /* isDelay */
	  double d;
	  objet ob=getCoefObj (cf);
	  setInfluDelay (fl, ob);
	  if (REAL == getObjType (ob))
	    d = getObjNVal(ob);
	  else {/* REAL */
	    setUncertain (fl, 1);
	    d = getRangeSup (getObjRVal (ob));
	  }
	  if (d < 0.0)
	    semError (0, nligne, 38);    /* delay shouldn't be negative. */
	  if (d >= 0 && d > getVarMaxDelay (cause))
	    setVarMaxDelay (cause, d);
	  else if (d < 0)
	    setVarMaxDelay (cause, d);
	    /*}*/ /* endif d */
	  isDelay = 1;
	} /* endif isDelay */
	break;
      case XI :
	if (isXi)
	  semError (2, nligne, 60);      /* double definition of xi. */
	else { /* isXi */
	  if (RANGE == getObjType (getCoefObj (cf)))
	    setUncertain (fl, 1);
	  setInfluXi (fl, getCoefObj (cf));
	  isXi = 1;
	}
	break;
      case PULSE :
	if (isPulse)
	  semError (2, nligne, 61);     /* double definition of pulse. */
	else {
	  if (RANGE == getObjType (getCoefObj (cf)))
	    setUncertain (fl, 1);
	  setInfluW (fl, getCoefObj (cf));
	  isPulse = 1;
	}
	break;
      case RESPTIME :
	if (isRespTime)
	  semError (2, nligne, 36);   	/*double definition of response time (first assumed). */
	else {
	  if (RANGE == getObjType (getCoefObj (cf)))
	    setUncertain (fl, 1);
	  setInfluTRep (fl, getCoefObj (cf));
	  isRespTime = 1;
	}
	break;
      case N1 :
	if (isN1)
	  semError (2, nligne, 82);
	else {
	  if (RANGE == getObjType (getCoefObj (cf)))
	    setUncertain (fl, 1);
	  setInfluN1 (fl, getCoefObj (cf));
	  isN1 = 1;
	}
	break;
      case N2 :
	if (isN2)
	  semError (2, nligne, 83);
	else {
	  if (RANGE == getObjType (getCoefObj (cf)))
	    setUncertain (fl, 1);
	  setInfluN2 (fl, getCoefObj (cf));
	  isN2 = 1;
	}
	break;
      case A0 :
	if (isA0)
	  semError (2, nligne, 84);
	else {
	  if (RANGE == getObjType (getCoefObj (cf)))
	    setUncertain (fl, 1);
	  setInfluA0 (fl, getCoefObj (cf));
	  isA0 = 1;
	}
	break;
      case A1 :
	if (isA1)
	  semError (2, nligne, 85);
	else {
	  if (RANGE == getObjType (getCoefObj (cf)))
	    setUncertain (fl, 1);
	  setInfluA1 (fl, getCoefObj (cf));
	  isA1 = 1;
	}
	break;
      case A2 :
	if (isA2)
	  semError (2, nligne, 86);
	else {
	  if (RANGE == getObjType (getCoefObj (cf)))
	    setUncertain (fl, 1);
	  setInfluA2 (fl, getCoefObj (cf));
	  isA2 = 1;
	}
	break;
      default:
	break;
      } /* end switch */
    } /* endif ! isNull cf */
    lc = getLCNext (lc);
  } /* end while */
  /* add delay to effet variable for D2INFL. */
  if (D2INFL == tp) {
    if (getVarMaxDelay (cause) < 2)
      setVarMaxDelay (cause, 2);
    if (getVarMaxDelay (effet) < 2)
      setVarMaxDelay (effet, 2);
    if (! isNull (getInfluN1 (fl))) {
      /* check roots: a0/a2 - (a1/2*a2)^2 >= 0 */
      objet ck = mkObjRange (mkRange (0.5, 0.5));
      ce_mulObjet (ck, getInfluA1 (fl));
      ce_mulObjet (ck, getInfluA1 (fl));
      ce_divObjet (ck, getInfluA2 (fl));
      ce_divObjet (ck, getInfluA2 (fl));
      ce_minusObjet (ck);
      /*objet ck2 = copyObj (getInfluA0 (fl));*/
      objet ck2 = mkObjRange (mkRange (getObjNVal (getInfluA0 (fl)), getObjNVal (getInfluA0 (fl))));
      /*ce_putObjet (getInfluA0 (fl), ck2);*/
      ce_divObjet (ck2, getInfluA2 (fl));
      ce_addObjet (ck, ck2);
      if (! ce_geObjet (ck, objZero)) {
	semError (0, nligne, 87, mkString (strcat (strcat (getVarNom (cause), " --> "), getVarNom (effet))));  /* incorrect second-order influences parameters */
	return NULL;
      }
      free (ck);
      free (ck2);
    }
  }
  else if (CINFL == tp) {
    if (isGain || isDelay || isRespTime)
      semError (0, nligne, 33);/* constant influence should have no coefficients. */
  }
  else { /* CINFL == tp */
    if (! isGain)
      semError (2, nligne, 24);        /* "default value 1 assumed for gain." */
    if (! isDelay)
      semError (2, nligne, 25);        /* "default value 0 assumed for delay." */
    isNullTRep = eqObj (getInfluTRep (fl), objZero);
    if (IINFL == tp) {
      if (isRespTime)
	semError (0, nligne, 32);
      /* "integral influence should have no response time." */
    }
    else if (SINFL == tp) {
      if (isRespTime && (! isNullTRep))
	semError (0, nligne, 22); /* static influence should have no response time. */
    }
    else if (DINFL == tp) {
      if (isNullTRep)
	semError (0, nligne, 23); /* "dynamic influence shouldn't have null response time." */
      if (getVarMaxDelay (cause) == 0)
	setVarMaxDelay (cause, 1);
      if (getVarMaxDelay (effet) == 0)
	setVarMaxDelay (effet, 1);
    } /* endif IINFL == tp */
  } /* endif CINFL == tp */
  fl->co = c;
  numInflu++;
  setInfluNum (fl, numInflu);
  return fl;
}


void printInfluence (influence inf) {
  enum influType tp=getInfluType(inf);
  char *nom=getInfluNom(inf);
  cond cd=getInfluCond(inf);

  fprintf (stdout, "  ");
  fprintf (stdout, "%ld: ", getInfluNum (inf));
  if (! isNull (nom))
    fprintf (stdout, "%s: ", nom);
  if (! isNull (cd)) {
    fprintf (stdout, "condition ");
    printCondition (cd);
    fprintf (stdout, " ");
  }
  printVarNom (stdout, getInfluCause (inf));
  fprintf (stdout, " -");
  switch (tp) {
  case DINFL:
    fprintf (stdout, "D");
    break;
  case D2INFL:
    fprintf (stdout, "D2");
    break;
  case IINFL:
    fprintf (stdout, "I");
    break;
  case SINFL:
    fprintf (stdout, "S");
    break;
  case CINFL:
    fprintf (stdout, "C");
    break;
  default:
    break;
  }
  fprintf (stdout, "-> ");
  printVarNom (stdout, getInfluEffet (inf));
  fprintf (stdout, ": gain ");
  printObjet (stdout, getInfluGain (inf));
  if (! isNull (getInfluXi (inf))) {
    fprintf (stdout, ": xi ");
    printObjet (stdout, getInfluXi (inf));
  }
  if (! isNull (getInfluW (inf))) {
    fprintf (stdout, ": w ");
    printObjet (stdout, getInfluW (inf));
  }
  if (! isNull (getInfluN1 (inf))) {
    fprintf (stdout, ": n1 ");
    printObjet (stdout, getInfluN1 (inf));
    fprintf (stdout, ": n2 ");
    printObjet (stdout, getInfluN2 (inf));
    fprintf (stdout, ": a0 ");
    printObjet (stdout, getInfluA0 (inf));
    fprintf (stdout, ": a1 ");
    printObjet (stdout, getInfluA1 (inf));
    fprintf (stdout, ": a2 ");
    printObjet (stdout, getInfluA2 (inf));
  }
  fprintf (stdout, "; delay ");
  printObjet (stdout, getInfluDelay (inf));
  fprintf (stdout, "; resp. time ");
  printObjet (stdout, getInfluTRep (inf));
  fprintf (stdout, "; condHistoLn = %ld", getInfluNActCond(inf));
  fprintf (stdout, ";\n");
  return;
}


long listcaenInflLn (listcaenInfl li) {
  long ln=0;
 
  if (! isNull (li))
    ln = 1 + listcaenInflLn (getLINext (li));
  return ln;
}
 
listcaenInfl addListcaenInfl (listcaenInfl la, listcaenInfl lb) {
  listcaenInfl lc, ld;
  
  if (isNull (la) && ! isNull (lb)) {
    la = mkListcaenInfl (lb->infl);
    lb = getLINext (lb);
  }
  if (! isNull (lb)) {
    lc = la;
    while (! isNull (ld = getLINext (lc)))
      lc = ld;
    while (! isNull (lb)) {
      setLINext (lc, mkListcaenInfl (lb->infl));
      lb = getLINext (lb);
      lc = getLINext (lc);
    }
    return la;
  }
  else return la;
}
 
/* concatenation de deux listcaenes d'influences */
listcaenInfl catListcaenInfl (listcaenInfl la, listcaenInfl lb) {
  listcaenInfl lc, ld;
 
  if (isNull (la))
    la = lb;                                     /* la vide: lb                           */
  else
    if (! isNull (lb)) {                     /* lb vide: la non modifie'e             */
      lc = la;
      while (! isNull (ld = getLINext (lc)))
        lc = ld;
      setLINext (lc, lb);                 /* concate'nation: lb en queue de la     */
    }
  return la;
}


void setRulesNames (listcaenInfl lf, char *s) {

  while (! isNull (lf)) {
    setInfluNom (getLIInfl (lf), s);
    lf = getLINext (lf);
  }
  return;
}


void andRulesConds (listcaenInfl lf, cond cc, Comp* co) {
  influence fl;
  cond cd, ce;

  while (! isNull (lf)) {
    fl = getLIInfl (lf);
    if (isNull (cd = getInfluCond (fl)))
      ce = cc;
    else {
      ce = mkCondEt (cc, cd);
      setConditionName (ce, co);
      conditions->push_back (ce);
    }
    setInfluCond (fl, ce);
    lf = getLINext (lf);
  }
  return;
}


long isInInfluListcaen (char *s, listcaenInfl lf) {
  long r=0;
 
  if (! isNull (lf)) {
    if (0 == strcmp (s, getInfluNom (getLIInfl (lf))))
      r = 1;
    else
      r = isInInfluListcaen (s, getLINext (lf));
  } /* endif ! isNull lf */
  return r;
}
 
 


long nInterm;

long listcaenGVarLn (listcaenGVar lg) {
  long ln=0;
 
  if (! isNull (lg))
    ln = 1 + listcaenGVarLn (getLGNext (lg));
  return ln;
}
 
 
listcaenGVar isInGVarListcaen (variable v, listcaenGVar lg) {
  listcaenGVar r=NULL;
 
  if (! isNull (lg)) {
    gVar gv=getLGGVar (lg);
    variable w=getGVarVar(gv);
    if (v ==  w)
      r = lg;
    else
      r = isInGVarListcaen (v, getLGNext (lg));
  }
  return r;
}
 
listcaenGVar addListcaenGVar (listcaenGVar la, listcaenGVar lb) {
  listcaenGVar lc, ld;

  if (isNull (la) && ! isNull (lb)) {
    la = mkListcaenGVar (lb->gv);
    lb = getLGNext (lb); 
  }
  if (! isNull (lb)) {
    lc = la;
    while (! isNull (ld = getLGNext (lc)))
      lc = ld;
    while (! isNull (lb)) {
      setLGNext (lc, mkListcaenGVar (lb->gv));
      lb = getLGNext (lb);
      lc = getLGNext (lc);
    }
    return la;
  }
  else return la;
}
 
/* concatenation de deux listcaenes de variables globales*/
listcaenGVar catListcaenGVar (listcaenGVar la, listcaenGVar lb) {
  listcaenGVar lc, ld;

  if (isNull (la))
    la = lb;                        /* la vide: lb                           */
  else
    if (! isNull (lb)) {            /* lb vide: la non modifie'e             */
      lc = la;
      while (! isNull (ld = getLGNext (lc)))
	lc = ld;
      setLGNext (lc, lb);           /* concate'nation: lb en queue de la     */
    }
  return la;
}


/* construction d' une listcaene de variables globales en supprimant une variable globale
   d'une listcaene de variable globales */
listcaenGVar removeGVar (listcaenGVar la, gVar a) {
  listcaenGVar lr;

  if (isNull (la))
    lr = NULL;
  else {
    gVar b=getLGGVar(la);
    listcaenGVar ls=getLGNext (la);
    if (b == a)
      lr = ls;
    else {
      lr = mkListcaenGVar (b);
      ls = removeGVar (ls, a);
      setLGNext (lr, ls);
    }
  }
  return lr;
}


/* impression des noms d'une listcaene de variables globales */
void printListcaenGVar (FILE *outs, listcaenGVar la) {
  
  if (isNull (la))
    fprintf (outs, ".\n");
  else {
    gVar gv=getLGGVar(la);
    listcaenGVar lb=getLGNext (la);
    printGVar (outs, gv);
    if (! isNull (lb))
      fprintf (outs, ", ");
    printListcaenGVar (outs, lb);
  }
  return;
}


void printGVar (FILE *outs, gVar gv) {
  variable v=getGVarVar(gv);
  long ni=getGVarInterm(gv);
  objet ob;

  if (isNull (v)) {
    if (0 == ni) {
      ob = getGVarObj (gv);                   /* isGVarCte gv */
      switch (getObjType(ob)) {
      case REAL:
	printReal (stdout, getObjNVal (ob));
	break;
      case RANGE:
	printRange (stdout, getObjRVal (ob));
	break;
      default:
	break;
      } /* end switch getObjType ob */
    }
    else /* 0 == ni */
      fprintf (outs, "InTeRm%ld", ni);       /* isGVarInterm gv */
  }
  else /* isNull v */
    printVarNom (outs, v);                    /* isGVarVar gv */
  return;
}


void printGvExprListcaen (FILE *outs, gVar gv) {
  listcaenLong ll=getGVarLExpr(gv);

  while (! isNull (ll)) {
    fprintf (outs, " %ld", getLLLong (ll) + 1);
    ll = getLLNext (ll);
  }
  return;
}

/*
 * build expressions as well as 'solution-functions'
 */
extern expr buildExprBinOp (char c, expr e1, expr e2, listcaenExpr &lce, listcaenGVar &lcg, int &numCstrsCond) {
  expr er, expInv1, expInv2, aux1=NULL, aux2=NULL;
  enum exprType op=RIEN, opInv1=RIEN, opInv2=RIEN;
  gVar gv;

  er = mkExpr ();
  switch (c) {
  case '+':
    op=PLUS, opInv1=MOINS , opInv2=MOINS, aux1=er, aux2=e1;
    break;
  case '-':
    op=MOINS, opInv1=PLUS, opInv2=MOINS, aux1=e1, aux2=er;
    break;
  case '*':
    op=MULT, opInv1=DIV, opInv2=DIV, aux1=er, aux2=e1;
    break;
  case '/':
    op=DIV, opInv1=MULT, opInv2=DIV, aux1=e1, aux2=er;
    break;
  case 'e':
    op=EXP, opInv1=EXP;
    break;
  default :
    break;
  }
  setExprType (er, op);
  setExprOp1 (er, e1);
  setExprOp2 (er, e2);
  if (! isExprCte (e1) || ! isExprCte (e2)) {
    gv = mkGVarInterm (++nInterm);
    lcg = catListcaenGVar (lcg, mkListcaenGVar (gv));
    setExprGVar (er, gv);
    lce = catListcaenExpr (lce, mkListcaenExpr (er));
    numCstrsCond++;
    if (! isExprCte (e1)) {
      if (EXP == opInv1) {
	aux1 = mkExprCte (objUn);
	aux2 = mkExpr2 (DIV, aux1, e2);
	expInv1 = mkExpr2 (opInv1, er, aux2);
      }
      else
	expInv1 = mkExpr2 (opInv1, er, e2);
      setExprGVar (expInv1, getExprGVar (e1));
      if (! isNull (lce))
	lce = catListcaenExpr (lce, mkListcaenExpr (expInv1));
      numCstrsCond++;
    } /* endif ! isExprCte e1 */
    if (! isExprCte (e2)) {
      if (RIEN == opInv2)
	semError (0, nligne, 45);       /* "invalid argument in exponentiation." */
      else { /* RIEN */
	expInv2 = mkExpr2 (opInv2, aux1, aux2);
	setExprGVar (expInv2, getExprGVar (e2));
	lce = catListcaenExpr (lce, mkListcaenExpr (expInv2));
	numCstrsCond++;
      } /* RIEN */
    } /* endif ! isExprCte e2 */
  } /* endif ! isExprCte e1 || ! isExprCte e2 */
  return er;
}


expr buildExprSgn (long s, expr e, listcaenExpr &lce, listcaenGVar &lcg, int &numCstrsCond) {
  expr f, fInv;
  gVar gv;

  if (s) {                                      /* signe '-' */
    f = mkExpr1 (SGN, e);
    if (! isExprCte (e)) {
      gv = mkGVarInterm (++nInterm);
      /*gv = mkGVarInterm (0);*/
      lcg = catListcaenGVar (lcg, mkListcaenGVar (gv));
      setExprGVar (f, gv);
      if (! isNull (lce))
	lce = catListcaenExpr (lce, mkListcaenExpr (f));
      fInv = mkExpr1 (SGN, f);
      setExprGVar (fInv, getExprGVar (e));
      lce = catListcaenExpr (lce, mkListcaenExpr (fInv));
      numCstrsCond++;
    } /* endif ! isExprCte e */
  }
  else /* s */
    f = e;
  return f;
}

extern expr buildEgSolFunc (expr e, enum opRel op, listcaenExpr &ccl, listcaenGVar &ccg, int &numCstrsCond) {
  expr ep=0, er;
  
  switch (op) {
  case LT:
    ep = mkExprCte (objNegS);
    break;
  case GE:
    ep = mkExprCte (objPos);
    break;
  case LE:
    ep = mkExprCte (objNeg);
    break;
  case GT:
    ep = mkExprCte (objPosS);
    break;
  default :
    break;
  }
  /*ccl = catListcaenExpr (ccl, mkListcaenExpr (ep));
    numCstrsCond++;*/
  er = buildExprBinOp ('+', e, ep, ccl, ccg, numCstrsCond);
  return er;
}

void printExpr (expr e) {
  enum exprType tp=getExprType(e);
  expr e1, e2;
  objet ob;

  switch (tp) {
  case CTE:
    ob = getGVarObj (getExprGVar (e));
    switch (getObjType(ob)) {
    case REAL:
      printReal (stdout, getObjNVal (ob));
      break;
    case RANGE:
      printRange (stdout, getObjRVal (ob));
      break;
    default:
      break;
    } /* end switch getObjType ob */
    break;
  case VAR:
    printGVar (stdout, getExprGVar (e));
    break;
  case PLUS:
  case MOINS:
  case MULT:
  case DIV:
  case EXP:
    e1 = getExprOp1 (e);
    if (isExprCte (e1) || hasName (e1))
      printGVar (stdout, getExprGVar (e1));
    else {
      fprintf (stdout, "(");
      printExpr (e1);
      fprintf (stdout, ")");
    }
    switch (tp) {
    case PLUS:
      fprintf (stdout, "+");
      break;
    case MOINS:
      fprintf (stdout, "-");
      break;
    case MULT:
      fprintf (stdout, "*");
      break;
    case DIV:
      fprintf (stdout, "/");
      break;
    case EXP:
      fprintf (stdout, "**");
      break;
    default :
      break;
    }
    e2 = getExprOp2 (e);
    if (isExprCte (e2) || hasName (e2))
      printGVar (stdout, getExprGVar (e2));
    else {
      fprintf (stdout, "(");
      printExpr (e2);
      fprintf (stdout, ")");
    }
    break;
  case SGN:
    fprintf (stdout, "-");
    e1 = getExprOp1 (e);
    if (isExprCte (e1) || hasName (e1))
      printGVar (stdout, getExprGVar (e1));
    else {
      fprintf (stdout, "(");
      printExpr (e1);
      fprintf (stdout, ")");
    }
    break;
  default:
    break;
  }
  return;
}


long hasGvAsArg (expr e, gVar gv) {
  long r=0;
  expr e1, e2;

  switch (getExprType (e)) {
  case VAR :
    if (gv == getExprGVar (e))
      r = 1;
    break;
  case PLUS :
  case MOINS :
  case MULT :
  case DIV :
  case EXP :
    e1 = getExprOp1 (e), e2 = getExprOp2 (e);
    if (hasName (e1))
      if (gv == getExprGVar (e1))
	r = 1;
    if (hasName (e2))
      if (gv == getExprGVar (e2))
	r = 1;
    /* hack... ? */
    if (gv == getExprGVar (e))
      r = 1;
    break;
  case SGN :
    e1 = getExprOp1 (e);
    if (hasName (e1))
      if (gv == getExprGVar (e1))
	r = 1;
    break;
  case CTE :
    /*if (gv == getExprGVar (e))
      r = 1;*/
  default:
    break;
  }
  return r;
}


long listcaenExprLn (listcaenExpr le) {
  long ln=0;
 
  if (! isNull (le))
    ln = 1 + listcaenExprLn (getLENext (le));
  return ln;
}

listcaenExpr addListcaenExpr (listcaenExpr la, listcaenExpr lb) {
  listcaenExpr lc, ld;
  
  if (isNull (la) && ! isNull (lb)) {
    la = mkListcaenExpr (lb->ex);
    lb = getLENext (lb);
  }
  if (! isNull (lb)) {
    lc = la;
    while (! isNull (ld = getLENext (lc)))
      lc = ld;
    while (! isNull (lb)) {
      setLENext (lc, mkListcaenExpr (lb->ex));
      lb = getLGNext (lb);
      lc = getLGNext (lc);
    }
    return la;
  }
  else return la;
}
 
/* concatenation de deux listcaenes d'expressions */
listcaenExpr catListcaenExpr (listcaenExpr la, listcaenExpr lb) {
  listcaenExpr lc, ld;
 
  if (isNull (la))
    la = lb;                                     /* la vide: lb                           */
  else
    if (! isNull (lb)) {                     /* lb vide: la non modifie'e             */
      lc = la;
      while (! isNull (ld = getLENext (lc)))
        lc = ld;
      setLENext (lc, lb);                 /* concate'nation: lb en queue de la     */
    }
  return la;
}


/* impression d'un initialiseur de variable */
void printInitVar (FILE *outs, initVar iv) {

  fprintf (outs, "  ");
  printVarNom (outs, getIVVar (iv));
  fprintf (outs, ": ");
  printObjet (outs, getIVObj (iv));
  return;
}

void printMFVar (FILE* fd, mapfVar mv) {
  fprintf (fd, "name: %s\n", getMFVarName (mv));
  cout << "vi: ";
  printObjet (fd, getMFObj (mv));
}
  
objet initDom2Obj (domain d) {
  objet ob=getDomObj(d);
  variable v;

    if (! isNull (v = getDomVar (d))) {         /* real */
      if (! isVarConst (v)) {
	ob = NULL;
      }
      else /* ! isVarConst v */
	ob = initDom2Obj (getVarDom (v));
      /* endif isVarConst v */
    } /* endif isNull v */
  return ob;
}


initVar buildInitVar (variable v, domain d) {
  initVar iv=NULL;
  variable vv;

  if (! isNull (vv = getDomVar (d))) {
    d = getVarDom (vv);
    if (! isVarConst (vv))
      semError (0, nligne, 9);     /* "initializing variable should be const." */
  } /* endif ! isNull vv */
  iv = mkInitVar (v, getDomObj (d));
  return iv;
}

mapfVar buildMapfVar (variable v, domain d) {
  mapfVar mv = NULL;
  variable vv;

  if (! isNull (vv = getDomVar (d))) {
    d = getVarDom (vv);
    if (! isVarConst (vv))
      semError (0, nligne, 9);     /* "initializing variable should be const." */
  }
  mv = mkMapfVar (v, getDomObj (d));
  return mv;
}

mapfVar buildMapfVarExpr (variable v, expr ex) {
  return mkMapfVarExpr (v, ex);
}

long listcaenInitVarLn (listcaenInitVar li) {
  long ln=0;
 
  if (! isNull (li))
    ln = 1 + listcaenInitVarLn (getLIVNext (li));
  return ln;
}
 
listcaenInitVar addListcaenInitVar (listcaenInitVar la, listcaenInitVar lb) {
  listcaenInitVar lc, ld;

  if (isNull (la) && ! isNull (lb)) {
    la = mkListcaenInitVar (lb->init);
    lb = getLIVNext (lb);
  }
  if (! isNull (lb)) {
    lc = la;
    while (! isNull (ld = getLIVNext (lc)))
      lc = ld;
    while (! isNull (lb)) {
      setLIVNext (lc, mkListcaenInitVar (lb->init));
      lb = getLIVNext (lb);
      lc = getLIVNext (lc);
    }
    return la;
  }
  else return la;
}
 
/* concatenation de deux listcaenes d'initialiseurs de variables */
listcaenInitVar catListcaenInitVar (listcaenInitVar la, listcaenInitVar lb) {
  listcaenInitVar lc, ld;

  if (isNull (la))
    la = lb;                        /* la vide: lb                           */
  else
    if (! isNull (lb)) {            /* lb vide: la non modifie'e             */
      lc = la;
      while (! isNull (ld = getLIVNext (lc)))
	lc = ld;
      setLIVNext (lc, lb);           /* concate'nation: lb en queue de la     */
    }
  return la;
}


/* impression d'une listcaene d'initialiseurs de variables */
void printInitListcaen (FILE *outs, listcaenInitVar liv) {
  
  if (! isNull (liv)) {
    listcaenInitVar liw=getLIVNext (liv);
    printInitVar (outs, getLIVInit (liv));
    fprintf (outs, ";\n");
    printInitListcaen (outs, liw);
  }
  return;
}


listcaenInitVar isInInitListcaen (variable v, listcaenInitVar lv) {
  listcaenInitVar r=NULL;
 
  if (! isNull (lv)) {
    initVar iv=getLIVInit(lv);
    variable w=getIVVar(iv);
    if (v ==  w)
      r = lv;
    else
      r = isInInitListcaen (v, getLIVNext (lv));
  }
  return r;
}

#ifdef DEBUG
long memAl, memFr;


void memInit (void) {

  memAl = 0;
  memFr = 0;
  return;
}


void memStats (void) {

  fprintf (stdout, "%ld bytes allocated; %ld bytes freed.\n", memAl, memFr);
  return;
}


void tiMemInit (void) {

  memInit ();
  timeInit ();
  return;
}
#endif /* DEBUG */


listcaenVar inputListcaen, displayListcaen, measureListcaen, temporalListcaen, untemporalListcaen;
listcaenInfl influListcaen;
listcaenGVar gVarListcaen;
listcaenExpr globCListcaen;
listcaenExpr condCListcaen;
listcaenInitVar initListcaen;


maillon *sommetS, *sommetP;

void copyFile (FILE *in, FILE *out) {
  int c;
  
  while (EOF != (c = getc (in)))
    putc (c, out);
  return;
}


int estvideS (void) {

  return (sommetS == NULL);
}


void empileS (char *e) {
  maillon *aux;

  aux = sommetS;
  sommetS = mkMaillon ();
  sommetS->val = e;
  sommetS->adresse = aux;
  return;
}


char *depileS (void) {
  char *aux;
  maillon *bux=sommetS;

  aux = sommetS->val;
  sommetS = sommetS->adresse;
  free (bux);
  return (aux);
}


int estvideP (void) {

  return (sommetP == NULL);
}


void empileP (char *e) {
  maillon* aux;
  aux = sommetP;
  sommetP = mkMaillon ();
  sommetP->val = e;
  sommetP->adresse = aux;
  return;
}


char *depileP (void) {
  char *aux;
  maillon *bux=sommetP;
  aux = sommetP->val;
  sommetP = sommetP->adresse;
  free (bux);
  return (aux);
}

/* pre-processor */
char lessThan[] = "<";
char greatEq[] = ">=";
char lessEq[] = "<=";
char greatThan[] = ">";


void printTable (char **ppc, long n) {
  long i;

  for (i=0; i<n; ++i)
    fprintf (stdout, "   %s\n", *(ppc+i));
}


static char *ligne=(char *)NULL;
extern ceCom cde[];

/* Print out help for ARG, or for all of the commands if ARG is not present. */
void printHelp (char *arg) {
  long i;
  long printed = 0;
     
  for (i=0; cde[i].name!=NULL; i++) {
    if (isNull (arg) || (*arg == '\0')
	|| (strstr (cde[i].name, arg) == cde[i].name)) {
      fprintf (stdout, "%-12s%s.\n", cde[i].name, cde[i].doc);
      printed++;
    }
  }
  if (!printed) {
    fprintf (stdout, "No commands match `%s'.  Possibilities are:\n", arg);
    for (i = 0; cde[i].name!=NULL; i++) {
      if (printed == 6) {                           /* Print in six columns. */
	printed = 0;
	fprintf (stdout, "\n");
      }
      fprintf (stdout, "%-12s", cde[i].name);
      printed++;
    }
    if (printed)
      fprintf (stdout, "\n");
  }
}


void ligneInit (void) {

#ifdef HAVE_LIBREADLINE
  initialize_readline ();                             /* Bind our completer. */
#else /* HAVE_READLINE */
  ligne = (char *) cealloc (81 * sizeof (char));
#endif /* HAVE_READLINE */
  return;
}


#ifdef HAVE_LIBREADLINE
/* Tell the GNU Readline library how to complete.  We want to try to complete
   on command names if this is the first word in the line, or on filenames
   if not. */

char *command_generator ();
char **ce_completion (const char *text, int start, int end);

int initialize_readline (void) {
     
  /* Allow conditional parsing of the ~/.inputrc file. */
  rl_readline_name = "Koala";
  /* Tell the completer that we want a crack first. */
  /*rl_attempted_completion_function = ce_completion;*/ /* broken */
  return 0;
}
     
/* Attempt to complete on the contents of TEXT.  START and END show the
   region of TEXT that contains the word to complete.  We can use the
   entire line in case we want to do some simple parsing.  Return the
   array of matches, or NULL if there aren't any. */
char **ce_completion (const char *text, int start, int end) {
  char **matches=NULL;
     
  /* If this word is at the start of the line, then it is a command
     to complete.  Otherwise it is the name of a file in the current
     directory. */
  /* if (start == 0)
     matches = completion_matches (text, command_generator);*/
  return (matches);
}
     
/* Generator function for command completion.  STATE lets us know whether
   to start from scratch; without any state (i.e. STATE == 0), then we
   start at the top of the listcaen. */
char *command_generator (char *text, long state) {
  static long listcaen_index, len;
  char *name, *s=NULL;
     
  /* If this is a new word to complete, initialize now.  This includes
     saving the length of TEXT for efficiency, and initializing the index
     variable to 0. */
  if (!state) {
    listcaen_index = 0;
    len = strlen (text);
  }
  /* Return the next name which partially matches from the command listcaen. */
  while ((name = cde[listcaen_index].name)) {
    listcaen_index++;
    if (strncmp (name, text, len) == 0)
      s = name;
  }
  /* If no names matched, then return NULL. */
  return s;
}


/* Read a string, and return a pointer to it.  Returns NULL on EOF. */
char *do_gets (char *s) {
  /* If the buffer has already been allocated, return the memory
     to the free pool. */
  if (ligne != (char *)NULL) {
    free (ligne);                                  
    ligne = (char *)NULL;
  }
  /* Get a line from the user. */
  ligne = readline (s);
  /* If the line has any text in it, save it on the history. */
  if (ligne && *ligne)
    add_history (ligne);
  return (ligne);
}

#else /* readline */

char *do_gets (char *s) {
  long again;

  if (! isNull (s))
    fprintf (stdout, "%s", s);
  do {
    again = 0;
    if (isNull (fgets (ligne, 81, stdin)))         /* EOF encontrado         */
      ligne = (char *) NULL;
    else {
      char *p;
      p = strtok (ligne, "\n");                    /* cambia '\n' por '\0'   */
      if (isNull (p)) {
	again = 1;                                 /* ligne empieza por '\n' */
	continue;
      }
    }
  } while (again);
  return ligne;
}
    
#endif /* GNU */

/*CT***************************************************************************
 Tronque n caracteres en debut de la chaine s et retourne cette chaine tronquee
 Detruit la chaine entree
-----------------------------------------------------------------------------*/
char * strTrunk(char *s,int nb) {
  int it,i;

  for (i=0, it=nb ; it < (int) strlen (s); it++,i++) 
    s[i]=s[it];
  s[i]='\0';
  return s;
}


/*****************************************************************************
 Retourne le nombre de mots, separe par un espace, d'une chaine
-----------------------------------------------------------------------------*/
long nbMot(char *s) {
  long i, cptmot;

  i=0;cptmot=0;
  while (i < (long) strlen(s)) {
    if (s[i] == ' ')
      cptmot++;
    i++;
  }
  cptmot++;
  return cptmot;
}
/******************************************************************************
 Extrait le n ieme mot a partir de la fin , separer par des blancs d'une chaine
 num < 0
 retourne un pointeur sur le resultat  
-----------------------------------------------------------------------------*/
char* extrait_mot(long num, char *s) {
  long cptmot=0,i,ii,f;
  char *res;

  i=strlen(s);
  f=i;
  while ( i>0 && cptmot>num ) {
    f=i;
    i--;
    while (i>0 && s[i]!=' ')
      i--;
    if (s[i]==' ' || i==0)
      cptmot--;
  }
  if (cptmot == num) {
    if (s[i] == ' ')
      i++;
    res = (char*)  cealloc(sizeof(char)*( (f-i)+1 ) );
    for (ii=0;i<f;i++,ii++)
      res[ii]=s[i];
    res[ii]='\0';
    return  res ; 
  }
  return s;
}


/* Strip whitespace from the start and end of STRING. (gnu) */
void stripWhite (char *s) {
  long i=0;

  if (! isNull (s)) {
    while (isspace (s[i]))  /* whitespace (s[i])) */
      i++;
    if (i)
      strcpy (s, s+i);
    i = strlen (s) - 1;
    while (i > 0 && isspace (s[i])) /*&& whitespace (s[i]))*/
      i--;
    s[++i] = '\0';
  }
  return;
}

/* Strip numbers in front of a string */
void stripNumbers (char *s) {
  long i=0;

   if (! isNull (s)) {
     while (isdigit (s[i]))
       i++;
    if (i)
      strcpy (s, s+i);
   }
  return;
}

/* Strip quotes from the start and end of STRING */
void stripQuote (char *s) {
  long i=0;

  if (! isNull (s)) {
    while (s[i] == '"')
      i++;
    if (i)
      strcpy (s, s+i);
    i = strlen (s) - 1;
    while (i > 0 && (s[i] == '"'))
      i--;
    s[++i] = '\0';
  }
  return;
}


/* Look up NAME as the name of a command, and return a pointer to that
   command. Return a NULL pointer if NAME isn't a command name. */
ceCom *whichCom (ceCom *com, char *name) {
  long i;
  ceCom *pc=NULL;
  if (isNull (name))
    return pc;
  for (i = 0; com[i].name; i++)
    if (strcmp (name, com[i].name) == 0)
      pc = &com[i];
  return pc;
}


/* Medidas de tiempo */
clock_t tZero, tNow;
struct tms timeBuf;

void timeInit (void) {

  (void) times (&timeBuf);
  tZero = timeBuf.tms_utime;
  return;
}

double spentTime (void) {

  (void) times (&timeBuf);
  tNow = timeBuf.tms_utime;
  return (double) (tNow-tZero) / 60.0;                   /* 60 clock_ticks/s */
}


void implError (long nerr) {
  static const char *msg[]= {
    "too many model definitions.",                                      /* 1 */
    "wrong booleean constant.",                                         /* 2 */
    "too deep causal graph.",                                           /* 3 */
    "file error.",                                                      /* 4 */
    "booleean object in code generation.",                              /* 5 */
    "invalid condition type.",                                          /* 6 */
    "parameter defined from another parameter",                         /* 7 */
    "wrong relation operator in condition",                             /* 8 */
    "constant as global constraint", /* 9 */
    "" /* 10 */
  };
  fprintf (stdout, "Implementation error.\n ...%s\n", msg[nerr-1]);
  return;
}

void semError (long stade, long nligne, long nerr) {
  semError (stade, nligne, nerr, "");
}

void semError (long stade, long nligne, long nerr, const char* smsg) {
  static const char *msg[]= {
    "multiple definition of symbolic constant.",                        /* 1 */
    "model already defined.",                                           /* 2 */
    "undefined model.",                                                 /* 3 */
    "not enough actual parameters in instance.",                        /* 4 */
    "too many actual parameters in instance.",                          /* 5 */
    "contradicting type declarations.",                                 /* 6 */
    "constant influence should have a parameter as cause.",             /* 7 */
    "invalid operator in influence coefficient.",                       /* 8 */
    "initialization should be done with a parameter.",                  /* 9 */
    "parameter should be initialized.",                                /* 10 */
    "range: left bound is greater than right one.",                    /* 11 */
    "invalid relation operator before range.",                         /* 12 */
    "variable already defined.",                                       /* 13 */
    "undefined variable in domain definition.",                        /* 14 */
    "undefined variable in interface listcaen.",                       /* 15 */
    "undefined variable in influence.",                                /* 16 */
    "undefined variable in condition.",                                /* 17 */
    "undefined variable in expression.",                               /* 18 */
    "undefined variable in initialization.",                           /* 19 */
    "unmatching types of parameters.",                                 /* 20 */
    "invalid operator in variable declaration.",                       /* 21 */
    "static influence should have no response time.",                  /* 22 */
    "dynamic influence shouldn't have null response time.",            /* 23 */
    "default value 1 assumed for gain.",                               /* 24 */
    "default value 0 assumed for delay.",                              /* 25 */
    "default value 0 assumed for response time.",                      /* 26 */
    "invalid operator in initialization.",                             /* 27 */
    "invalid object beside IN operator.", 			       /* 28 */
    "no display variable.",                                            /* 29 */
    "parameter in Input or Measure declaration.",        	       /* 30 */
    "parameter in Display declaration.",                               /* 31 */
    "integral influence should have no response time.",                /* 32 */
    "constant influence should have no coefficients.",                 /* 33 */
    "double definition of gain (first assumed).",                      /* 34 */
    "double definition of delay (first assumed).",                     /* 35 */
    "double definition of response time (first assumed).",             /* 36 */
    "lower bound of delay shouldn't be null.",                         /* 37 */
    "delay shouldn't be negative.",                                    /* 38 */
    "parameter shouldn't be effect in an influence.",                  /* 39 */
    "incompatible types in influences to the same variable:",          /* 40 */
    "unequal response times to the same variable: please\n\
          explicitly introduce marginal variables.",                   /* 41 */
    "missing '=' in equation .",                                       /* 42 */
    "gain range has not a unique sign.",                               /* 43 */
    "invalid left hand side in equation.",                             /* 44 */
    "invalid argument in exponentiation.",                             /* 45 */
    "variable should be initialized or measured.",                     /* 46 */
    "variable in condition should better be input or measure.",        /* 47 */
    "component name or rule name already defined.",                    /* 48 */
    "block of rules not allowed inside a model.",                      /* 49 */
    "imbrication of models not allowed.",                              /* 50 */
    "instanciation not allowed inside a model.",                       /* 51 */
    "variable can't be both Input and Measure.",                       /* 52 */
    "closed loop in causal graph around variable:",                    /* 53 */
    "Input variable shouldn't be effect in an influence.",             /* 54 */
    ".",           /* 55 */
    ".",           /* 56 */
    ".",           /* 57 */
    ".",           /* 58 */
    ".",           /* 59 */
    "double definition of xi (first assumed).",                        /* 60 */
    "double definition of pulse (first assumed).",                     /* 61 */
    "double definition of zeros (first assumed).",                     /* 62 */
    "condition defined over different variables.",                     /* 63 */
    "conditions must be declared in a condition variable (CondVar).",  /* 64 */
    "assume identity mapping function.",                               /* 65 */
    "equalities are not recommended in conditions unless you really know what you're doing !",  /* 66 */
    "component not declared.",                                         /* 67 */
    "unknown component type.",                                         /* 68 */
    "wrong number of arguments to component.",                         /* 69 */
    "wrong type for component importation.",                           /* 70 */
    "header file not found.",                                          /* 71 */
    "unknown error during variables parsing.",                         /* 72 */
    "variable not found.",                                             /* 73 */
    "unknown atom.",                                                   /* 74 */
    "unused component variable.",                                      /* 75 */
    "command in undefined.",                                           /* 76 */
    "type not found in use declaration.",                              /* 77 */
    "unknown error during component parsing.",                         /* 78 */
    "undefined or imported variable in expression. \nConditions variables must not be defined on imported variables.",                                                        /* 79 */
    "no component of this type has been instanciated in the model.",    /* 80 */
      "no unknown mode found.",                                         /* 81 */
    "double definition of n1 (first assumed).",                         /* 82 */
    "double definition of n2 (first assumed).",                         /* 83 */
    "double definition of a0 (first assumed).",                         /* 84 */
    "double definition of a1 (first assumed).",                         /* 85 */
    "double definition of a2 (first assumed).",                         /* 86 */
    "incorrect second-order influences parameters: should be broken into two first-order influences."  /* 87 */
  };
  switch (stade) {
  case 0:
    fprintf (stderr, "Line %ld: semantics error.\n  ...%s: %s\n",
	     nligne, msg[nerr-1], smsg);
    loadError = 1;
    exit (1);
    break;
  case 1:
    fprintf (stderr, "Semantics error.\n  ...%s: %s\n", msg[nerr-1], smsg);
    loadError = 1;
    exit (1);
    break;
  case 2:
    fprintf (stderr, "Line %ld: warning.\n  ...%s: %s\n", nligne, msg[nerr-1], smsg);
    break;
  case 3:
    fprintf (stderr, "Warning.\n  ...%s: %s\n", msg[nerr-1], smsg);
  }
  return;
}

