/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/* caen-objet.h */
#ifndef CAEN_OBJET
#define CAEN_OBJET

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string>
#include <limits>
#include <float.h>
//#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_LIBREADLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/times.h>
#include <vector>
#include "../system/config.h"
#include "../config.h"

class Atom;
class StateFormula;
class Comp;

using namespace std;

#define isNull(p)           ((p) == NULL)

extern long loadError;

/* macros d'allocation
   C'est la couche de gestion de la me'moire de Ca-En. Allocation par malloc;
*/
#ifdef DEBUG
#define cealloc(n)     (L1 = (long) (n), memAl += L1, (char *)malloc (L1))
#define cefree(p)      (memFr += (long) sizeof (*p), free (p))
#else /* DEBUG */
#define cealloc(n)     ((char *)malloc (n))
#define cefree(x)      (free (x))
#endif /* DEBUG */

#ifndef CESTRING
#define CESTRING
/* les macros pour les chai^nes de caracte`res */
#define allocString(s)      (cealloc(strlen(s)+1))   /* room for the \0, too */
#define mkString(s)         (strcpy(allocString(s),s))
#define isEmpty(s)          (*(s) == '\0')
/* gets memory for s + t, then copies s, then appends t */
#define catString(s,t)      (isNull(t) ? s \
			     : (S1 = strcat (strcpy (cealloc (strlen(s) \
				  + strlen(t)  + 1), s), t), free (s), S1))
#endif

/* les valeurs booleennes */
enum boolVal {CE_FALSE=0, CE_TRUE=1};

extern enum boolVal a2Bool (char *);

/**
 * \struct s_range
 * \brief le type "range" avec ses ope'rateurs.
 */
typedef struct s_range {
  double inf;
  double sup;
  bool incinf;              /**< inf is included or not */
  bool incsup;              /**< sup is included or not */
} S_range, *range;

#define allocRange()            ((range) cealloc(sizeof(S_range)))
#define getRangeInf(r)          ((r)->inf)
#define getRangeSup(r)          ((r)->sup)
#define isRangeThin(r)          ((r)->inf == (r)->sup)
#define setRangeInf(r,i)        ((r)->inf = i)
#define setRangeSup(r,s)        ((r)->sup = s)
#define isIncludedInf(r)        (r->incinf == 1)
#define isIncludedSup(r)        (r->incsup == 1)
#define setNotIncludedInf(r)    (r->incinf = 0)
#define setNotIncludedSup(r)    (r->incsup = 0)
#define setIncludedInf(r)       (r->incinf = 1)
#define setIncludedSup(r)       (r->incsup = 1)
#define getBoundInf(r)          (r->incinf)
#define getBoundSup(r)          (r->incsup)
#define setBoundInf(r,b)        (r->incinf = b)
#define setBoundSup(r,b)        (r->incsup = b)
#define setBounds(r1, r2)       (r1->incinf = r2->incinf, \
                                 r1->incsup = r2->incsup)
#define mkRange(i,s)            (Rn1 = allocRange (), setRangeInf (Rn1, i), \
				 setRangeSup (Rn1, s), Rn1->incinf = 1, Rn1->incsup = 1, Rn1)
#define copyRange(r)            (Rn1 = allocRange (), \
				 setRangeInf (Rn1, getRangeInf (r)), \
				 setRangeSup (Rn1, getRangeSup (r)), \
                                 Rn1->incinf = r->incinf, Rn1->incsup = r->incsup, Rn1)
#define hasZero(r)              (getRangeInf (r) * getRangeSup (r) < 0)

extern void printRange (FILE *, range);
extern long eqRange (range a, range b);

extern void printReal (FILE *, double);
extern void geneReal (FILE *, double);
#define eqReal(a,b)             (a == b) 

/**
 * \enum objType
 * \brief les types d'objets Ca-En.
 */
enum objType {UNKNOWN, REAL, RANGE, BOL};

/**
 * \union u_val
 * \brief union de valeurs de types #.
 */
typedef union u_val {
  double vdouble;
  range vrange;
  enum boolVal vbool;
} U_val, *valeur;

#define allocValeur()        ((valeur) cealloc (sizeof (U_val)))
#define getValReal(v)        ((v)->vdouble)
#define getValRange(v)       ((v)->vrange)
#define getValBool(v)        ((v)->vbool)
#define setValReal(v,d)      ((v)->vdouble=d)
#define setValRange(v,r)     ((v)->vrange=r)
#define setValBool(v,b)      ((v)->vbool=b)
#define mkValReal(d)         (Vl1 = allocValeur (), setValReal (Vl1, d), Vl1)
#define mkValRange(r)        (Vl1 = allocValeur (), setValRange (Vl1, r), Vl1)
#define mkValBool(b)         (Vl1 = allocValeur (), setValBool (Vl1, b), Vl1)


/**
 * \struct s_objet
 * \brief le type "objet" avec ses ope'rateurs.
 */
typedef struct s_objet {
  enum objType type;
  valeur val;
} S_objet, *objet;

#define allocObj()            ((objet) cealloc(sizeof(S_objet)))
#define allocVObj(n)          ((objet *) cealloc((n)*sizeof(objet)))
#define getObjType(b)         ((b)->type)
#define objVal(b)             ((b)->val)
#define getObjNVal(b)         (getValReal (objVal (b)))
#define getObjRVal(b)         (getValRange (objVal (b)))
#define getObjBVal(b)         (getValBool (objVal (b)))
#define setObjType(b,t)       ((b)->type = (objType)t)
#define setObjVal(b,v)        (objVal (b) = v)
#define setObjNVal(b,d)       (setValReal (objVal (b), d))
#define setObjRVal(b,r)       (setValRange (objVal (b), r))
#define setObjBVal(obj,b)     (setValBool (objVal (obj), b))
#define mkObjet(v)           (Obj1 = allocObj (), setObjType (Obj1, UNKNOWN), \
			         setObjVal (Obj1, v), Obj1)
#define mkObjReal(d)          (Obj1 = allocObj (), setObjType (Obj1, REAL), \
			         setObjVal (Obj1, mkValReal (d)), Obj1)
#define mkObjRange(r)         (Obj1 = allocObj (), setObjType (Obj1, RANGE), \
			         setObjVal (Obj1, mkValRange (r)), Obj1)
#define mkObjBool(b)          (Obj1 = allocObj (), setObjType (Obj1, BOL), \
			         setObjVal (Obj1, mkValBool (b)), Obj1)
#define copyObj(ob)           (Obj1 = allocObj (), \
			       setObjType (Obj1, getObjType (ob)), \
			       setObjVal (Obj1, objVal (ob)), Obj1)

extern range maxRange;
extern valeur valZero, valUn, valMaxDbl, valMaxRange, valFalse, valTrue;
extern objet objFalse, objTrue, objZero, objUn, objMaxDbl;
extern void objInit (void);
extern long eqObj (objet, objet);
extern void mulObjet (objet, objet);
extern long ce_divObjet (objet, objet);
extern void ce_minusObjet (objet);
extern void ce_putObjet (objet, objet);
extern void ce_addObjet (objet, objet);
extern boolVal ce_geObjet (objet, objet);
extern void printObjet (FILE *, objet);

/**
 * generace C-code for an objet.
 */
extern void geneObjet (FILE *, objet, char*);


/**
 * \struct s_listcaenlong
 * \brief le type "listcaene d'entiers", avec ses ope'rateurs.
 */
typedef struct s_listcaenLong {
  long lg;
  struct s_listcaenLong *next;
} S_listcaenLong, *listcaenLong;

#define allocLLong()          ((listcaenLong) cealloc (sizeof(S_listcaenLong)))
#define getLLLong(ll)         (ll->lg)
#define getLLNext(ll)         (ll->next)
#define setLLLong(ll,a)       (ll->lg = a)
#define setLLNext(ll,n)       (ll->next = n)
#define mkListcaenLong(a)         (Ll1 = allocLLong (), setLLLong(Ll1, a), \
			       setLLNext (Ll1, NULL), Ll1)

extern long listcaenLongLn (listcaenLong);
extern listcaenLong catListcaenLong (listcaenLong, listcaenLong);


/**
 * \struct s_arDef
 * \brief le type "arDef: arbre binaire des definitions #define"
 * avec ses ope'rateurs
 */
typedef struct s_arDef {
  struct s_arDef *gauche;
  struct s_arDef *droite;
  char *nom;
  char *txt;
} S_arDef, *arDef;
#define allocArDef()        ((arDef) cealloc(sizeof(S_arDef)))
#define getArDefGauche(p)   (p->gauche)
#define getArDefDroite(p)   (p->droite)
#define getArDefNom(p)      (p->nom)
#define getArDefTxt(p)      (p->txt)
#define setArDefGauche(p,q) (p->gauche = q)
#define setArDefDroite(p,q) (p->droite = q)
#define setArDefNom(p,n)    (p->nom = n)
#define setArDefTxt(p,t)    (p->txt = t)
#define mkArDef(n,t)        (Df1 = allocArDef (), \
                             setArDefGauche (Df1,NULL), \
			     setArDefDroite (Df1,NULL), \
			     setArDefNom (Df1,n), \
			     setArDefTxt (Df1,t), Df1)
#define freeArDef(a)        (cefree (a))
     
extern long storeDef (arDef *, char *, char *);
extern arDef searchDef (arDef *, char *);
extern void delDef (arDef *, char *);
extern void printArDef (arDef);


/**
 * \enum interType
 * \brief les fonctions des variables (les types d'interfaces).
 */
enum interType {NOINTER, INPUT, DISPLAY, INDI, MEASURE, DIME=6};

/**
 * \struct s_variable
 * \brief le type "variable" avec ses ope'rateurs.
 */
typedef struct s_variable {
  char *nom;
  char *symb_name;              /* symbolic name to be used in generated C++ code. */
  long lnew;
  enum objType type;                  /* one of {UNKNOWN, REAL, RANGE}  */
  long cst;
  enum interType fct;           /* one of {NOINTER, INPUT, DISPLAY, MEASURE}  */
  struct s_domain *dom;
  objet value;
  struct s_cGraph *noeud;
  long depth;
  long tree;                    /* ???? */
  double maxDelay;
  long ncd;
} S_variable, *variable;

#define allocVariable()       ((variable) cealloc(sizeof(S_variable)))
#define allocVVar(n)          ((variable *) cealloc((n)*sizeof(variable)))

#define eqVar(v1, v2)         ((strcmp(v1->nom, v2->nom) == 0))

#define getVarNom(v)          (v->nom)
#define getVarSymbName(v)        (v->symb_name)
#define isVarNew(v)           (v->lnew)
#define getTheVarType(v)         (v->type)
#define isVarConst(v)         (v->cst)
#define getVarFct(v)          (v->fct)
#define isVarInput(v)         (v->fct & INPUT)
#define isVarDisplay(v)       (v->fct & DISPLAY)
#define isVarMeasure(v)       (v->fct & MEASURE)
#define getVarDom(v)          (v->dom)
#define getVarValue(v)        (v->value)
#define getVarNoeud(v)        (v->noeud)
#define getVarDepth(v)        (v->depth)
#define getVarTree(v)         (v->tree)
#define getVarMaxDelay(v)     (v->maxDelay)

#define setVarNom(v,s)        (v->nom = s)
#define setVarSymbName(v,s)      (v->symb_name = s)
#define setVarNew(v)          (v->lnew = 1)
#define setVarOld(v)          (v->lnew = 0)
#define setVarType(v,t)       (v->type = t)
#define setVarConst(v)        (v->cst = 1)
#define unSetVarConst(v)      (v->cst = 0)
#define setVarFct(v,f)        (v->fct = f)
#define unSetVarFct(v)        (v->fct = NOINTER)
#define setVarDom(v,d)        (v->dom = d)
#define setVarValue(v,w)      (v->value = w)
#define setVarNoeud(v,n)      (v->noeud = n)
#define setVarDepth(v,d)      (v->depth = d)
#define setVarTree(v,t)       (v->tree = t)
#define setVarMaxDelay(v,m)   (v->maxDelay = m)

#define mkVariable(s)         (Va1 = allocVariable(), \
			      setVarNom (Va1, mkString (s)), \
                              setVarNew (Va1), \
			      setVarType (Va1, UNKNOWN), \
                              unSetVarConst (Va1), \
			      unSetVarFct (Va1), \
			      setVarDom (Va1, NULL), \
			      setVarValue (Va1, NULL), \
			      setVarNoeud (Va1, NULL), \
			      setVarDepth (Va1, -1), \
			      setVarTree (Va1, 0), \
			      setVarMaxDelay (Va1, 0.0), \
                              Va1->ncd = 0, Va1)

#define rVal(a)               (getObjRVal (getVarValue (a)))
#define bVal(a)               (getObjBVal (getVarValue (a)))

#define MAXESPDEPTH 100
extern long numCstVar;
extern variable const2Var (double);
extern void printVariable (variable);
extern void printVarNom (FILE *, variable);
extern void printVarSymbName (FILE *, variable);


/**
 * \struct s_listcaenVar
 * _brief le type "listcaene de variables", avec ses ope'rateurs.
 */
typedef struct s_listcaenVar {
  variable var;
  struct s_listcaenVar *next;
} S_listcaenVar, *listcaenVar;

#define allocLVar()          ((listcaenVar) cealloc(sizeof(S_listcaenVar)))
#define getLVVar(lv)         (lv->var)
#define getLVNext(lv)        (lv->next)
#define setLVVar(lv,a)       (lv->var = a)
#define setLVNext(lv,n)      (lv->next = n)
#define mkListcaenVar(a)         (Lv1 = allocLVar (), setLVVar(Lv1, a), \
			      setLVNext (Lv1, NULL), Lv1)

extern bool isVarInList (listcaenVar, variable);
extern long listcaenVarLn (listcaenVar);
extern listcaenVar appVariable (listcaenVar, variable);
extern listcaenVar catListcaenVar (listcaenVar, listcaenVar);
extern void printListcaenVar (FILE *, listcaenVar);
extern void printUpdateListcaen (FILE *, listcaenVar);
extern void setListcaenFct (listcaenVar, enum interType);
extern void setListcaenType (listcaenVar, enum objType);
extern void setListcaenConst (listcaenVar);


/**
 * \enum opRel
 * \brief les operateurs de relation.
 */
enum opRel {NONE, EQ, NE, LE, LT, GE, GT, IN};

/**
 * \struct s_domain
 * \brief le type "domain" (d'une variable) avec ses ope'rateurs.
 */
typedef struct s_domain {
  objet ob;
  variable var;
} S_domain, *domain;

#define allocDom()            ((domain) cealloc(sizeof(S_domain)))
#define getDomObj(d)          ((d)->ob)
#define getDomVar(d)          ((d)->var)
#define getDomNVal(d)         (getObjNVal (getDomObj (d)))
#define getDomRVal(d)         (getObjRVal (getDomObj (d)))
#define setDomObj(d,o)        ((d)->ob = o)
#define setDomVar(d,v)        ((d)->var = v)
#define mkDomReal(d)          (Dom1 = allocDom (), \
			       setDomObj (Dom1, mkObjReal (d)), \
			       setDomVar (Dom1, NULL), Dom1)
#define mkDomRange(r)         (Dom1 = allocDom (), \
			       setDomObj (Dom1, mkObjRange (r)), \
			       setDomVar (Dom1, NULL), Dom1)
#define mkDomVar(v)           (Dom1 = allocDom (), \
			       setDomObj (Dom1, NULL), \
			       setDomVar (Dom1, v), Dom1)
#define copyDom(d)            (Dom1 = allocDom (), \
			       setDomObj (Dom1, getDomObj (d)), \
			       setDomVar (Dom1, getDomVar (d)), Dom1)

extern void printDomain (FILE *, domain);
extern variable dom2Var (variable, enum opRel, domain);


/**
 * \struct s_paramFor
 * \brief le type "paramFor" (parametre formel de composant) avec ses operateurs.
 */
typedef struct s_paramFor {
  char *nom;
  enum objType type;
  domain dom;
} S_paramFor, *paramFor;

#define allocParamFor()        ((paramFor) cealloc(sizeof(S_paramFor)))
#define getParamForNom(p)      (p->nom)
#define getParamForType(p)     (p->type)
#define getParamForDomain(p)   (p->dom)
#define setParamForNom(p,s)    (p->nom = s)
#define setParamForType(p,t)   (p->type = (objType)t)
#define setParamForDomain(p,d) (p->dom = d)
#define mkNumParamFor(s,d)     (Pf1 = allocParamFor(), \
			        setParamForNom (Pf1, s), \
			        setParamForType (Pf1, REAL), \
			        setParamForDomain (Pf1, d), Pf1)
#define mkBoolParamFor(s)      (Pf1 = allocParamFor(), \
			        setParamForNom (Pf1, s), \
			        setParamForType (Pf1, BOL), \
			        setParamForDomain (Pf1, NULL), Pf1)


/**
 * \struct s_listcaenParFor
 * \brief le type "listcaene de parametres formels", avec ses operateurs.
 */
typedef struct s_listcaenParFor {
  paramFor par;
  struct s_listcaenParFor *next;
} S_listcaenParFor, *listcaenParFor;
 
#define allocLParFor()        ((listcaenParFor) cealloc(sizeof(S_listcaenParFor)))
#define getLPFParFor(lp)      (lp->par)
#define getLPFNext(lp)        (lp->next)
#define setLPFParFor(lp,v)    (lp->par = v)
#define setLPFNext(lp,n)      (lp->next = n)
#define mkListcaenParFor(v)       (Lf1 = allocLParFor (), setLPFParFor (Lf1, v), \
                               setLPFNext(Lf1, NULL), Lf1)

extern listcaenParFor tempDefs;
extern listcaenParFor catListcaenParFor (listcaenParFor, listcaenParFor);


/**
 * \struct s_paramEff
 * \brief le type "paramEff" (parametre effectif de composant) avec ses operateurs.
 */
typedef struct s_paramEff {
  char *nom;
  enum objType type;
} S_paramEff, *paramEff;

#define allocParamEff()        ((paramEff) cealloc(sizeof(S_paramEff)))
#define getParamEffNom(p)      (p->nom)
#define getParamEffType(p)     (p->type)
#define setParamEffNom(p,s)    (p->nom = s)
#define setParamEffType(p,t)   (p->type = (objType)t)
#define mkRealParamEff(s)      (Pe1 = allocParamEff(), \
			        setParamEffNom (Pe1, s), \
			        setParamEffType (Pe1, REAL), Pe1)
#define mkBoolParamEff(s)      (Pe1 = allocParamEff(), \
			        setParamEffNom (Pe1, s), \
			        setParamEffType (Pe1, BOL), Pe1)


/**
 * \struct s_listcaenParEff
 * \brief le type "listcaene de parametres effectifs", avec ses operateurs.
 */
typedef struct s_listcaenParEff {
  paramEff par;
  struct s_listcaenParEff *next;
} S_listcaenParEff, *listcaenParEff;
 
#define allocLParEff()          ((listcaenParEff) cealloc(sizeof(S_listcaenParEff)))
#define getLPEParEff(lp)        (lp->par)
#define getLPENext(lp)          (lp->next)
#define setLPEParEff(lp,v)      (lp->par = v)
#define setLPENext(lp,n)        (lp->next = n)
#define mkListcaenParEff(v)         (Le1 = allocLParEff (), setLPEParEff(Le1, v), \
                                 setLPENext(Le1, NULL), Le1)

extern long listcaenParEffLn (listcaenParEff);
extern listcaenParEff catListcaenParEff (listcaenParEff, listcaenParEff);


/**
 * \struct s_arBin
 * \brief le type "arBin: l'arbre binaire du dictionnaire", avec ses ope'rateurs.
 */
typedef struct s_arBin {
  struct s_arBin *gauche;
  struct s_arBin *droite;
  variable var;
} S_arBin, *arBin;
#define allocArBin()        ((arBin) cealloc(sizeof(S_arBin)))
#define getArBinGauche(p)   (p->gauche)
#define getArBinDroite(p)   (p->droite)
#define getArBinVar(p)      (p->var)
#define getArBinNom(p)      (getVarNom (getArBinVar (p)))
#define setArBinGauche(p,q) (p->gauche = q)
#define setArBinDroite(p,q) (p->droite = q)
#define setArBinVar(p,a)    (p->var = a)
#define mkArBin(a)          (Nd1 = allocArBin(), \
                             setArBinGauche(Nd1,NULL), \
                             setArBinDroite(Nd1,NULL), \
                             setArBinVar(Nd1,a), Nd1)

extern arBin dictionary;
extern variable name2Var (arBin *, char *);
extern void printArBin (arBin);


/**
 * \enum coefType
 * \brief les types de coefficients d'une influence.
 */
/*enum coefType {GAIN, DELAY, RESPTIME, XI, PULSE, ZEROS};*/
enum coefType {GAIN, DELAY, RESPTIME, XI, PULSE, N1, N2, A0, A1, A2};

/**
 * \struct s_coef
 * \brief le type "coef": coefficient d'une influence, avec ses operateurs.
 */
typedef struct s_coef {
  enum coefType ct;
  objet ob;
  variable var;
} S_coef, *coef;
#define allocCoef()        ((coef) cealloc(sizeof(S_coef)))
#define getCoefType(c)     (c->ct)
#define getCoefObj(c)      (c->ob)
#define getCoefVar(c)      (c->var)
#define setCoefType(c,t)   (c->ct = t)
#define setCoefObj(c,v)    (c->ob = v)
#define setCoefVar(c,v)    (c->var = v)
#define mkCoefObj(t,o)     (Cf1 = allocCoef(), \
                            setCoefType (Cf1, t), \
                            setCoefObj (Cf1, o), \
			    setCoefVar (Cf1, NULL), Cf1)
#define mkCoefVar(t,v)     (Cf1 = allocCoef(), \
                            setCoefType (Cf1, t), \
                            setCoefObj (Cf1, NULL), \
			    setCoefVar (Cf1, v), Cf1)

extern coef dom2Coef (enum coefType, enum opRel, domain);


/**
 * \struct s_listcaenCoef
 * \brief le type "listcaene de coefficients", avec ses operateurs.
 */
typedef struct s_listcaenCoef {
  coef cf;
  struct s_listcaenCoef *next;
} S_listcaenCoef, *listcaenCoef;
 
#define allocLCoef()         ((listcaenCoef) cealloc(sizeof(S_listcaenCoef)))
#define getLCCoef(lp)        (lp->cf)
#define getLCNext(lp)        (lp->next)
#define setLCCoef(lp,c)      (lp->cf = c)
#define setLCNext(lp,n)      (lp->next = n)
#define mkListcaenCoef(c)        (Lc1 = allocLCoef (), setLCCoef(Lc1, c), \
                              setLCNext(Lc1, NULL), Lc1)

extern listcaenCoef catListcaenCoef (listcaenCoef, listcaenCoef);

/**
 * \struct s_gVar
 * \brief le type "gVar": variable pour le calcul des contraintes globales, avec ses operateurs.
 */
typedef struct s_gVar {
  long num;
  long gened;
  variable var;
  long interm;
  objet ob;
  Comp* co;
  listcaenLong le;            /**< set of related global constraints (positions in globCListcaen) */
  listcaenLong lce;           /**< set of related conditions cstr (positions in condCListcaen) */
} S_gVar, *gVar;

#define allocGVar()           ((gVar) cealloc(sizeof(S_gVar)))
#define getGVarNum(g)         (g->num)
#define getGVarVar(g)         (g->var)
#define getGVarInterm(g)      (g->interm)
#define getGVarObj(g)         (g->ob)
#define getGVarComp(g)        (g->co)
#define getGVarLExpr(g)       (g->le)
#define getGVarLcExpr(g)      (g->lce)
#define getGVarGened(g)       (g->gened)
#define setGVarUnGened(g)     (g->gened = 0)
#define setGVarGened(g)       (g->gened = 1)
#define setGVarNum(g,n)       (g->num = n)
#define setGVarVar(g,v)       (g->var = v)
#define setGVarInterm(g,i)    (g->interm = i)
#define setGVarObj(g,o)       (g->ob = o)
#define setGVarComp(g,c)      (g->co = c)
#define setGVarLExpr(g,l)     (g->le = l)
#define setGVarLcExpr(g,l)    (g->lce = l)
#define isGVarCte(g)          (isNull (getGVarVar (g)) \
                               && (0 == getGVarInterm (g)))
#define isGVarVar(g)          (! isNull (getGVarVar (g)))
#define isGVarInterm(g)       (0 != getGVarInterm (g))
#define isGVarGened(g)        (0 != getGVarGened (g))

#define mkGVarCte(o)          (Gv1 = allocGVar(), \
                               setGVarNum (Gv1, 0), \
                               setGVarUnGened (Gv1), \
			       setGVarVar (Gv1, NULL), \
			       setGVarInterm (Gv1, 0), \
                               setGVarObj (Gv1, o), \
                               setGVarComp (Gv1, NULL), \
                               setGVarLExpr (Gv1, NULL), \
                               setGVarLcExpr (Gv1, NULL), Gv1)

#define mkGVarVar(v)          (Gv1 = allocGVar(), \
                               setGVarNum (Gv1, 0), \
                               setGVarUnGened (Gv1), \
			       setGVarVar (Gv1, v), \
			       setGVarInterm (Gv1, 0), \
                               setGVarObj (Gv1, NULL), \
                               setGVarComp (Gv1, NULL), \
                               setGVarLExpr (Gv1, NULL), \
                               setGVarLcExpr (Gv1, NULL), Gv1)

#define mkGVarInterm(i)       (Gv1 = allocGVar(), \
                               setGVarNum (Gv1, 0), \
                               setGVarUnGened (Gv1), \
			       setGVarVar (Gv1, NULL), \
			       setGVarInterm (Gv1, i), \
                               setGVarObj (Gv1, NULL), \
                               setGVarComp (Gv1, NULL), \
                               setGVarLExpr (Gv1, NULL), \
                               setGVarLcExpr (Gv1, NULL), Gv1)
extern long nInterm;


/**
 * \struct s_listcaenGVar
 * \brief le type "listcaene de variables globales", avec ses ope'rateurs.
 */
typedef struct s_listcaenGVar {
  gVar gv;
  struct s_listcaenGVar *next;
} S_listcaenGVar, *listcaenGVar;

#define allocLGVar()          ((listcaenGVar) cealloc(sizeof(S_listcaenGVar)))
#define getLGGVar(lg)         (lg->gv)
#define getLGNext(lg)         (lg->next)
#define setLGGVar(lg,a)       (lg->gv = a)
#define setLGNext(lg,n)       (lg->next = n)
#define mkListcaenGVar(a)         (Lg1 = allocLGVar (), setLGGVar(Lg1, a), \
			      setLGNext (Lg1, NULL), Lg1)

extern long listcaenGVarLn (listcaenGVar);
extern listcaenGVar isInGVarListcaen (variable, listcaenGVar);
extern listcaenGVar addListcaenGVar (listcaenGVar, listcaenGVar);
extern listcaenGVar catListcaenGVar (listcaenGVar, listcaenGVar);
extern listcaenGVar removeGVar (listcaenGVar, gVar);
extern void printListcaenGVar (FILE *, listcaenGVar);

/**
 * \enum exprType
 * \brief les types d' expressions.
 */
enum exprType {CTE, VAR, PLUS, MOINS, MULT, DIV, EXP, SGN, RIEN};

/**
 * \struct s_expr
 * \brief le type "expr" avec ses ope'rateurs.
 */
typedef struct s_expr {
  long num;
  long gened;
  gVar gv;
  enum exprType type;    /* {RIEN, CTE, VAR, PLUS, MOINS, MULT, DIV, EXP, SGN}; */
  struct s_expr *op1;
  struct s_expr *op2;
} S_expr, *expr;

#define allocExpr()         ((expr) cealloc(sizeof(S_expr)))

#define getExprNum(e)       (e->num)
#define isExprGened(e)      (1 == e->gened)
#define getExprGVar(e)      (e->gv)
#define getExprType(e)      (e->type)
#define isExprCte(e)        (CTE == getExprType (e))
#define hasName(e)          (Gv1 = getExprGVar (e), \
                              ! (isNull (Gv1) || isGVarCte (Gv1)))
#define getExprOp1(e)       (e->op1)
#define getExprOp2(e)       (e->op2)

#define setExprNum(e,n)     (e->num = n)
#define setExprUnGened(e)   (e->gened = 0)
#define setExprGened(e)     (e->gened = 1)
#define setExprGVar(e,g)    (e->gv = g)
#define setExprType(e,t)    (e->type = t)
#define setExprOp1(e,f)     (e->op1 = f)
#define setExprOp2(e,f)     (e->op2 = f)

#define mkExpr()            (Ex1 = allocExpr(), \
                             setExprNum (Ex1, 0), \
                             setExprUnGened (Ex1), \
			     setExprGVar (Ex1, NULL), \
			     setExprType (Ex1, RIEN), \
			     setExprOp1 (Ex1, NULL), \
			     setExprOp2 (Ex1, NULL), Ex1)

#define mkExprCte(c)        (Ex1 = mkExpr(), \
                             setExprType (Ex1, CTE), \
                             setExprGVar (Ex1, mkGVarCte (c)), Ex1)
 
#define mkExprGVar(g)       (Ex1 = mkExpr(), \
                             setExprType (Ex1, VAR), \
                             setExprGVar (Ex1, g), Ex1)
 
#define mkExpr1(t,e)        (Ex1 = mkExpr(), \
			     setExprType (Ex1, t), \
			     setExprOp1 (Ex1, e), Ex1)

#define mkExpr2(t,e1,e2)    (Ex1 = mkExpr(), \
			     setExprType (Ex1, t), \
			     setExprOp1 (Ex1, e1), \
			     setExprOp2 (Ex1, e2), Ex1)

extern void printGVar (FILE *, gVar);
extern void printGvExprListcaen (FILE *, gVar);
extern void printExpr (expr);
extern long hasGvAsArg (expr, gVar);

/**
 * \struct s_listcaenExpr
 * \brief le type "listcaene d'expressions", avec ses operateurs.
 */
typedef struct s_listcaenExpr {
  expr ex;
  struct s_listcaenExpr *next;
} S_listcaenExpr, *listcaenExpr;
 
#define allocLExpr()         ((listcaenExpr) cealloc(sizeof(S_listcaenExpr)))
#define getLEExpr(le)        (le->ex)
#define getLENext(le)        (le->next)
#define setLEExpr(le,e)      (le->ex = e)
#define setLENext(le,n)      (le->next = n)
#define mkListcaenExpr(v)        (Lex1 = allocLExpr (), setLEExpr(Lex1, v), \
                              setLENext(Lex1, NULL), Lex1)

extern long listcaenExprLn (listcaenExpr);
extern listcaenExpr addListcaenExpr (listcaenExpr, listcaenExpr);
extern listcaenExpr catListcaenExpr (listcaenExpr, listcaenExpr);

extern expr buildExprBinOp (char, expr, expr, listcaenExpr&, listcaenGVar&, int&);
extern expr buildExprSgn (long, expr, listcaenExpr&, listcaenGVar&, int&);
extern expr buildEgSolFunc (expr, enum opRel, listcaenExpr&, listcaenGVar&, int&);

/**
 * \enum condType
 * \brief les types de conditions.
 */
enum condType {CDRIEN, CDCTE, CDCOMP, CDOU, CDET, CDNON, CDAT, CDEXPR};

/**
 * \struct s_cond
 * \brief le type "cond" avec ses ope'rateurs.
 */
typedef struct s_cond {
  char* name;
  long num;
  long gened;
  enum condType type;    /* {CDRIEN, CDCTE, CDCOMP, CDOU, CDET, CDNON, CDAT, CDEXPR} */
  struct s_cond *pere;
  struct s_cond *op1;
  struct s_cond *op2;
  variable var;          /**< variable tested in the condition */
  enum opRel pr;
  objet domob;           /**< domain */
  variable domvar;
  enum boolVal val;
  Atom* at;              /**< atom at discrete level */
  expr e;
  listcaenLong le;       /**< related expr (sol-functions) */
} S_cond, *cond;

#define allocCond()         ((cond) cealloc(sizeof(S_cond)))
#define getCondName(c)      (c->name)
#define getCondNum(c)       (c->num)
#define isCondGened(c)      (c->gened)
#define getCondType(c)      (c->type)
#define getCondPere(c)      (c->pere)
#define getCondOp1(c)       (c->op1)
#define getCondOp2(c)       (c->op2)
#define getCondVar(c)       (c->var)
#define getCondRel(c)       (c->pr)
#define getCondDomObj(c)    (c->domob)
#define getCondDomVar(c)    (c->domvar)
#define getCondValue(c)     (c->val)
#define getCondAtom(c)      (c->at)
#define getCondExpr(c)      (c->e)
#define getCondLe(c)        (c->le)

#define setCondName(c,s)    (c->name = s)
#define setCondNum(c,n)     (c->num = n)
#define setCondUnGened(c)   (c->gened = 0)
#define setCondGened(c)     (c->gened = 1)
#define setCondType(c,t)    (c->type = t)
#define setCondPere(c,p)    (c->pere = p)
#define setCondOp1(c,o)     (c->op1 = o)
#define setCondOp2(c,o)     (c->op2 = o)
#define setCondVar(c,v)     (c->var = v)
#define setCondRel(c,o)     (c->pr = o)
#define setCondDomObj(c,do) (c->domob = do)
#define setCondDomVar(c,dv) (c->domvar = dv)
#define setCondValue(c,b)   (c->val = b)
#define setCondAtom(c,a)   (c->at = a)
#define setCondExpr(c,a)   (c->e = a)
#define setCondLe(c,a)     (c->le = a)

extern long numCond;
#define mkCond()            (Cn1 = allocCond(), \
			     setCondNum (Cn1, ++numCond), \
			     setCondUnGened (Cn1), \
			     setCondType (Cn1, CDRIEN), \
			     setCondPere (Cn1, NULL), \
			     setCondOp1 (Cn1, NULL), \
			     setCondOp2 (Cn1, NULL), \
			     setCondVar (Cn1, NULL), \
			     setCondRel(Cn1, NONE), \
			     setCondDomObj (Cn1, NULL), \
			     setCondDomVar (Cn1, NULL), \
			     setCondValue (Cn1, CE_FALSE), \
                             setCondAtom (Cn1, NULL), \
                             setCondLe (Cn1, NULL), Cn1)

#define mkCondCte(b)        (Cn1 = mkCond(), \
			     setCondType (Cn1, CDCTE), \
                             setCondValue (Cn1, b), Cn1)

#define mkCondNon(e)        (Cn1 = mkCond(), \
			     setCondType (Cn1, CDNON), \
                             setCondOp1 (Cn1, e), \
			     setCondPere (e, Cn1), Cn1)

extern cond mkCondComp (variable, enum opRel, domain);
extern cond mkCondExpr (expr, enum opRel, expr);
extern void setConditionName (cond cd, Comp* co);

#define mkCondOu(e,f)       (Cn1 = mkCond(), \
			     setCondType (Cn1, CDOU), \
                             setCondOp1 (Cn1, e), \
                             setCondOp2 (Cn1, f), \
			     setCondPere (e, Cn1), \
			     setCondPere (f, Cn1), Cn1)

#define mkCondEt(e,f)       (Cn1 = mkCond(), \
			     setCondType (Cn1, CDET), \
                             setCondOp1 (Cn1, e), \
                             setCondOp2 (Cn1, f), \
			     setCondPere (e, Cn1), \
			     setCondPere (f, Cn1), Cn1)

extern void printCondition (cond);

/**
 * \struct s_listcaenCond
 * \brief le type "listcaen de conditions", avec ses operateurs.
 */
/*typedef struct s_listcaenCond {
  cond cd;
  strcut s_listcaenCond *next;
  } S_listcaenCond, *listcaenCond;*/

/**
 * \enum influType
 * \brief les types d' influences.
 */
enum influType {NOINFL, DINFL, IINFL, SINFL, CINFL, D2INFL};

/**
 * \struct s_influence
 * \brief type "influence", with its operators.
 */
typedef struct s_influence {
  long num;
  char *name;
  char* symb_name;
  enum influType it;
  variable cause;
  variable effet;
  cond cd;
  long length_act_cond_hist;
  objet gain;
  objet delai;
  objet tRep;
  /* CE_D2INFL only. */
  objet xi;
  objet w;
  objet d0;
  objet d1;
  objet d2;
  objet n1;
  objet n2;
  Comp* co;              /**< component link */
  bool uncertain;        /**< the influence has uncertain parameters or not */
} S_influence, *influence;

#define allocInfluence()    ((influence) cealloc(sizeof(S_influence)))
#define allocVInfluence(n)  ((influence *) cealloc((n)*sizeof(influence)))
#define getInfluNum(i)      (i->num)
#define getInfluNom(i)      (i->name)
#define getInfluSymbName(i)  (i->symb_name)
#define getInfluType(i)     (i->it)
#define getInfluCause(i)    (i->cause)
#define getInfluEffet(i)    (i->effet)
#define getInfluCond(i)     (i->cd)
#define getInfluNActCond(i) (i->length_act_cond_hist)
#define getInfluGain(i)     (i->gain)
#define getInfluDelay(i)    (i->delai)
#define getInfluTRep(i)     (i->tRep)
#define getInfluW(i)        (i->w)
#define getInfluXi(i)       (i->xi)
#define getInfluN1(i)       (i->n1)
#define getInfluN2(i)       (i->n2)
#define getInfluA0(i)       (i->d0)
#define getInfluA1(i)       (i->d1)
#define getInfluA2(i)       (i->d2)
#define isUncertain(i)      (i->uncertain)

#define isUnTemporal(i)      (((SINFL == getInfluType (i)) \
			     || (CINFL == getInfluType (i))) \
			      && (REAL == getObjType (getInfluDelay (i))) \
			     && (0.0 == getObjNVal (getInfluDelay (i))))

#define setInfluNum(i,n)    (i->num = n)
#define setInfluNom(i,s)    (i->name = s)
#define setInfluSymbName(i,s) (i->symb_name = s)
#define setInfluType(i,t)   (i->it = t)
#define setInfluCause(i,c)  (i->cause = c)
#define setInfluEffet(i,e)  (i->effet = e)
#define setInfluCond(i,c)   (i->cd = c)
#define setInfluNActCond(i,n) (i->length_act_cond_hist = n)
#define setInfluGain(i,g)   (i->gain = g)
#define setInfluDelay(i,d)  (i->delai = d)
#define setInfluTRep(i,r)   (i->tRep = r)
#define setInfluW(i,r)      (i->w = r)
#define setInfluXi(i,r)     (i->xi = r)
#define setInfluN1(i,n)     (i->n1=n)
#define setInfluN2(i,n)     (i->n2=n)
#define setInfluA0(i,n)     (i->d0=n)
#define setInfluA1(i,n)     (i->d1=n)
#define setInfluA2(i,n)     (i->d2=n)
#define setUncertain(i,b)   (i->uncertain=b)
#define mkInfluence(t,c,e)  (In1 = allocInfluence(), \
                             setInfluNum (In1, 0), \
                             setInfluNom (In1, NULL), \
                             setInfluType (In1, t), \
                             setInfluCause (In1, c), \
                             setInfluEffet (In1, e), \
                             setInfluCond (In1, NULL), \
                             setInfluNActCond (In1, 0), \
                             setInfluGain (In1, objUn), \
                             setInfluDelay (In1, objZero), \
                             setInfluTRep (In1, objZero),  \
                             setInfluW (In1, NULL), \
                             setInfluXi (In1, NULL), \
                             setInfluN1 (In1, NULL), \
                             setInfluN2 (In1, NULL), \
                             setInfluA0 (In1, NULL), \
                             setInfluA1 (In1, NULL), \
                             setUncertain (In1, 0), \
                             setInfluA2 (In1, NULL), In1)

extern influence buildInfluence (Comp*, enum influType, variable, variable, listcaenCoef);
extern void printInfluence (influence);
extern long numInflu;

/**
 * \struct s_listcaenInfl
 * \brief le type "listcaen d'influences", avec ses operateurs.
 */
typedef struct s_listcaenInfl {
  influence infl;
  struct s_listcaenInfl *next;
} S_listcaenInfl, *listcaenInfl;
 
#define allocLInfl()         ((listcaenInfl) cealloc(sizeof(S_listcaenInfl)))
#define getLIInfl(lp)        (lp->infl)
#define getLINext(lp)        (lp->next)
#define setLIInfl(lp,v)      (lp->infl = v)
#define setLINext(lp,n)      (lp->next = n)
#define mkListcaenInfl(v)        (Lif1 = allocLInfl (), setLIInfl(Lif1, v), \
                              setLINext(Lif1, NULL), Lif1)

extern long listcaenInflLn (listcaenInfl);
extern listcaenInfl addListcaenInfl (listcaenInfl, listcaenInfl);
extern listcaenInfl catListcaenInfl (listcaenInfl, listcaenInfl);
extern void setRulesNames (listcaenInfl, char *);
extern void andRulesConds (listcaenInfl, cond, Comp* co);
extern long isInInfluListcaen (char *, listcaenInfl);

/**
 * \struct s_cGraph
 * \brief le type "graphe causal", avec ses operateurs.
 */
typedef struct s_cGraph {
  long ncause;
  influence *vcause;
  enum influType ti;
  long neffet;
  influence *veffet;
} S_cGraph, *cGraph;

#define allocCGraph()          ((cGraph) cealloc(sizeof(S_cGraph)))
#define getCGNCause(cg)        (cg->ncause)
#define getCGCauseNth(cg,n)    (*(cg->vcause+(n))) /* n desde 0 hasta ncause-1 */
#define getCGCauseType(cg)     (cg->ti)
#define getCGNEffet(cg)        (cg->neffet)
#define getCGEffetNth(cg,n)    (*(cg->veffet+(n))) /* n desde 0 hasta neffet-1 */
#define setCGNCause(cg,nc)     (cg->ncause = nc)
#define setCGVCause(cg,vc)     (cg->vcause = vc)
#define setCGCauseNth(cg,n,c)  (*(cg->vcause+(n)) = c)     /* n : 0 .. ncause-1 */
#define setCGCauseType(cg,t)   (cg->ti = t)
#define setCGNEffet(cg,e)      (cg->neffet = e)
#define setCGVEffet(cg,ve)     (cg->veffet = ve)
#define setCGEffetNth(cg,n,e)  (*(cg->veffet+(n)) = e)      /* n : 0 .. neffet-1 */
#define mkCGraph(nc,ne)        (Cg1 = allocCGraph(), \
                                setCGNCause (Cg1, nc), \
                                nc ? setCGVCause (Cg1, allocVInfluence(nc)) \
				: setCGVCause (Cg1, NULL), \
                                setCGCauseType (Cg1, NOINFL), \
                                setCGNEffet (Cg1, ne), \
                                ne ? setCGVEffet (Cg1, allocVInfluence(ne)) \
				: setCGVEffet (Cg1, NULL), Cg1)

#ifdef DEBUG
extern void memInit (void);
extern void memStats (void);
extern void tiMemInit (void);
#endif /* DEBUG */


extern void copyFile (FILE *, FILE *);

extern int estvideS (void);
extern char *depileS (void);
extern void empileS (char *);
extern int estvideP (void);
extern char *depileP (void);
extern void empileP (char *);

typedef struct elmlst {
  struct elmlst *adresse;
  char  *val;
} maillon;

#define mkMaillon()         (maillon*) cealloc (sizeof (maillon));

/* extern long tableSym (char ***, long, char *); */
extern void printTable (char **, long);

extern char* strTrunk (char *, int);
extern char* extrait_mot (long, char *);
extern long nbMot(char *);

#define allocVLong(n)       ((long *) cealloc(((n)+1)*sizeof(long)))
#define allocTLong(n)         ((long **) cealloc((n)*sizeof(long *)))


/**
 * \struct s_initVar
 * \brief le type "initialiseur de variable", avec ses ope'rateurs.
 */
typedef struct s_initVar {
  variable va;
  objet vi;
} S_initVar, *initVar;

#define allocInitVar()        ((initVar) cealloc (sizeof (S_initVar)))
#define getIVVar(iv)          (iv->va)
#define getIVObj(iv)          (iv->vi)
#define setIVVar(iv,v)        (iv->va = v)
#define setIVObj(iv,ob)       (iv->vi = ob)
#define mkInitVar(v,ob)       (Iv1 = allocInitVar (), setIVVar (Iv1, v), \
                               setIVObj (Iv1, ob), Iv1)

extern void printInitVar (FILE *, initVar);
extern objet initDom2Obj (domain);
extern initVar buildInitVar (variable, domain);


/**
 * \struct s_listcaenInitVar
 * \brief le type "listcaene d'initialiseurs de variables".
 */
typedef struct s_listcaenInitVar {
  initVar init;
  struct s_listcaenInitVar *next;
} S_listcaenInitVar, *listcaenInitVar;
 
#define allocLInitVar()      ((listcaenInitVar) cealloc(sizeof(S_listcaenInitVar)))
#define getLIVInit(li)       (li->init)
#define getLIVNext(li)       (li->next)
#define setLIVInit(li,v)     (li->init = v)
#define setLIVNext(li,n)     (li->next = n)
#define mkListcaenInitVar(v)     (Liv1 = allocLInitVar (), setLIVInit(Liv1, v), \
                              setLIVNext(Liv1, NULL), Liv1)

extern long listcaenInitVarLn (listcaenInitVar);
extern listcaenInitVar addListcaenInitVar (listcaenInitVar, listcaenInitVar);
extern listcaenInitVar catListcaenInitVar (listcaenInitVar, listcaenInitVar);
extern void printInitListcaen (FILE *, listcaenInitVar);
extern listcaenInitVar isInInitListcaen (variable, listcaenInitVar);

/**
 * \struct s_mapfVar
 * \brief type for initialization variables in mapping functions.
 * TODO : must add expression.
 */
typedef struct s_mapfVar {
  variable v;
  objet vi;
  expr e;
} S_mapfVar, *mapfVar;

#define allocMapfVar()         ((mapfVar) cealloc (sizeof (S_mapfVar)))
#define getMFVar(mv)           (mv->v)
#define getMFVarName(mv)       (getVarNom (mv->v))
#define getMFObj(mv)           (mv->vi)
#define getMFExpr(mv)          (mv->e)
#define setMFVar(mv,n)         (mv->v = n)
#define setMFObj(mv,ob)        (mv->vi = ob)
#define setMFExpr(mv,ex)       (mv->e = ex)
#define mkMapfVar(va,ob)       (Mv1 = allocMapfVar (), setMFVar (Mv1, va), \
                               setMFExpr (Mv1, NULL), \
                               setMFObj (Mv1, ob), Mv1)
#define mkMapfVarExpr(va,ex)       (Mv1 = allocMapfVar (), setMFVar (Mv1, va), \
                               setMFObj (Mv1, NULL), \
                               setMFExpr (Mv1, ex), Mv1)

extern void printMFVar (FILE* fd, mapfVar);
extern mapfVar buildMapfVar (variable, domain);
extern mapfVar buildMapfVarExpr (variable, expr);

#ifndef S_COMPONENT
#define S_COMPONENT

typedef struct s_component {
  string *type;
  string *var_name;
  string *name;
  Comp* com;
  vector<string>* comp_args;
} S_component, *component;

#endif /* S_COMPONENT */

#define allocComponent()            ((component) cealloc(sizeof(S_component)))
#define freeComponent(c)            (delete c->type, \
                                     delete c->var_name, \
                                     delete c->name, \
                                     delete c->com, \
                                     delete c->comp_args, \
                                     cefree (c))
#define getComponentType(c)         (*c->type)
#define getComponentVarName(c)     (*c->var_name)
inline bool isInComponentArgs(component c, string s) {
  for (unsigned int i=0; i<c->comp_args->size (); i++) {
    int dot_pos;
    if ((*c->comp_args)[i] == s)
      return true;
    else if ((dot_pos = (*c->comp_args)[i].find (".")) > 0)
      if ((*c->comp_args)[i].substr (0, dot_pos) == s)
	return true;
  }
  return false;
}
inline bool isACommandArg(component c, string s) {
  for (unsigned int i=0; i<2; i++) {
    int dot_pos;
    if ((*c->comp_args)[i] == s)
      return true;
    else if ((dot_pos = (*c->comp_args)[i].find (".")) > 0)
      if ((*c->comp_args)[i].substr (0, dot_pos) == s)
	return true;
  }
  return false;
}

#ifndef DEBUG
#define NCARPERLINE 50
#endif /* DEBUG */
/* El tipo LFunction es definido como Function en <readline/keymaps.h>: 
              es mas legible `LFunction *func;' que `long (*)()func;' (gnu)  */
typedef long LFunction (char *);
typedef struct {                /* Structura para caber los mandos de Ca-En */
  char *name;                   /* User printable name of the function. */
  LFunction *func;                       /* Function to call to do the job.      */
  const char *doc;                    /* Documentation for this function.     */
} ceCom;

extern char *do_gets (char *);
#ifdef GNU
#ifdef __cplusplus
extern "C" {
#endif
  extern int initialize_readline (void);
  extern char* command_generator ();
  extern char** ce_completion (const char *tex, int start, int end);
#ifdef __cplusplus
}
#endif
#else /* GNU */
#define whitespace(c) (((c) == ' ') || ((c) == '\t'))
#endif /* GNU */
extern void ligneInit (void);
extern void stripWhite (char *);
extern void stripNumbers (char *);
extern void stripQuote (char *);
extern ceCom *whichCom (ceCom *, char *);
extern void printHelp (char *);


/* Medidas de tiempo */
extern void timeInit (void);
extern double spentTime (void);

extern long memAl, memFr;
extern maillon *sommetS, *sommetP;
extern long nligne;                         /* no. de ligne de l'instruction */
extern long numTemp, isQuant;
extern long etape, espace;
extern listcaenVar inputListcaen, displayListcaen, measureListcaen, temporalListcaen,
  untemporalListcaen;
extern listcaenInfl influListcaen;
extern listcaenGVar gVarListcaen;
extern listcaenExpr globCListcaen;
extern listcaenExpr condCListcaen;
extern listcaenInitVar initListcaen;
extern initVar *formerAtV;
extern long nFormerAt, period, nbCycleInit;
#ifdef DEBUG
extern long formCount;
extern long nElt;
#endif /* DEBUG */
extern long numfct;
extern long nNoeudOu, nNoeudEt, nFeuille, nArbre, nBrancheOu, nBrancheEt,
            nActMul, nDCode, nTCode, nResMul, nResion, nTest, depth,
            nElag, nWarn;
extern double compTime;
extern vector<cond>* conditions;

extern void implError (long);
extern void semError (long, long, long);
extern void semError (long, long, long, const char*);

#endif
