/*
 * Parser for koala models.
 * Benazera Emmanuel - LAAS-CNRS / DISCO - October 2001.
 */

%{

#include "../caen/caen-algo.h"
#include "../behav/Comp.h"
#include "../itms/itms.h"
#include <stdio.h>
#include <string>
#include <unistd.h>
#include <vector>
#include <fstream>

  /*
   * \struct parser_control.
   * \brief passing data to headers internal parser.
   */
#ifndef PARSER_CONTROL
#define PARSER_CONTROL
  struct parser_control {
    vector<pair<string, StateVar*> >* variables_disc;
    const char* def_filename;
  };
#endif 

  extern void parseHybridDefFile(const char* deffilename, vector<string>* ar, Comp* co);
  extern int parseContFile(const char* deffilename);
  extern FILE* DefHybridParserin;
  extern itms* reasoner;
  extern char appliPath[], appliName[];
  extern char* mdlfile;
  extern string cl_header; /* header for clauses files */
  extern vector<cond>* conditions;
  extern vector<StateFormula*>* stfs;
  extern int ModelHybridParserlex();
  extern int ModelHybridParserln();
  int ModelHybridParsererror(char* s);
  bool findDefFile(const char* deffilename);
  bool knownType(const char* ty);
  bool crossRefs (component co1, component co2);
  void rankComponents (component co);
  void codeComponent(component co);
  void codeAddComp();
  component searchCompVN (string vn);
  void compToDot(component co);
  void toDot();
  int parseModelFile(const char* fileIn);
  void verifyInstances ();
  void showModel();
  FILE* def_file; /* connection with the parser for .def files. */
  extern long  numConditions;
  extern Comp* parseDef(struct parser_control spc, const char* filename);

  char buf_cpp[32];
  char buf_h[32];
  char* out_cpp;
  char* out_h;
  string outh;
  string defname;
  string nameshort;
  string comp_name_mdl;
  string comp_var_name;
  long nligne;                                /* no. de ligne courante          */
  long braceLevel;
  char fileIn[40];
  char* fileInshort;;
  string code;
  map<int, component>* oco;
  char out[32];
  char class_buf_cpp[1000000];
  char* class_char_cpp;
  char class_buf_h[1000000];
  char* class_char_h;
  /* buffer for dot file generation. */
  string dot_buf;
  vector<pair<string, StateVar*> >* variables_disc;
  char* full_path_filename;
  /*char* current_filename;*/
  vector<string>* declarations;
  extern vector<component>* comps;
  component current_comp;
  listcaenExpr var_expr;

  %}

%union {
  FILE* input;
  vector<string>* declarations; /**< store model files (components) headers. */
  string* var;
  const char* type;
  vector<string>* args;
}

%token <declarations> USE
%token <args> ARGUMENTS
%token <var> VARIABLE
%token <type> TYPE
%token LB RB

%start model

%%

model: use_decl declarations {
  if (conf_get ("discrete") == 0) {
    /*map<int, component>:: iterator it = oco->end ();
    it--;
    while (it != oco->begin ()) {
      cout << getComponentVarName ((*it).second) << endl;
      codeComponent ((*it).second);
      it--;
    }
    cout << getComponentVarName ((*oco->begin ()).second) << endl;
    codeComponent ((*oco->begin ()).second);
    codeAddComp();*/
    for (map<int, component>::iterator it = oco->begin ();
       it != oco->end (); it++) {
      /*cout << getComponentVarName ((*it).second) << endl;*/
      codeComponent ((*it).second);
    }
    codeAddComp ();
  }
  else {  /* discrete only model */
    for (map<int, component>::iterator it = oco->begin ();
       it != oco->end (); it++) {
      codeComponent ((*it).second);
    }
    codeAddComp ();
  }
  numConditions = (long) conditions->size ();
  delete stfs;
  stfs = 0;
  /*delete oco;*/ /* used in prime implicants */
  cefree (var_expr);
}
       |
;

use_decl: USE {
  variables_disc = new vector<pair<string, StateVar*> >();
  declarations = $1;
  comps = new vector<component> ();
  stfs = new vector<StateFormula*> ();

  /* verify that the linked files exists. */
  for (unsigned int i=0; i<$1->size(); i++) {
    defname = (*$1)[i] + ".def";
    if (!findDefFile(defname.c_str())) {
      exit(0);
    }
    else {
      code = "#include \"../C/" + (*$1)[i] + ".h\"\n";
      cl_header += code;
      strcat(class_char_h, code.c_str());
    }
  }
  /* include others files into code. */
  code = "#include \"Scheduler.h\"\n#include \"caen-runt.h\"\n#include \"purge.h\"\n#include <string>\n#include <vector>\n\n";
  strcat(class_char_h, code.c_str());
  cl_header += "#include \"Scheduler.h\"\n#include \"Clause.h\"\n\n";
  code = "#include \"" + outh + "\"\n\nusing namespace std;\n\nvoid startinit (Scheduler* sMM) {\n\n";
  strcat(class_char_cpp, code.c_str());
  code = "";
  oco = new map<int, component> ();
  var_expr = NULL;
}
| {
  fprintf (stdout, "Error: no 'use' declaration found.\n");
  exit (0);
}
;

declarations: declaration ';' declarations {
}
| declaration ';'
|
;

declaration: TYPE VARIABLE {
  component co;
  if (knownType($1)) {
    co = allocComponent ();
    co->type = new string ($1);
    co->var_name = new string (*$2);
    comp_var_name = *co->var_name;
    defname = string ($1) + ".def";
    struct parser_control spc;
    spc.variables_disc = variables_disc;
    spc.def_filename = defname.c_str();
    if ((co->com = parseDef(spc, defname.c_str())) == NULL)
      ModelHybridParsererror ("Variables parsing error\n");
    comps->push_back(co);
  }
  else semError(0, ModelHybridParserln(), 68);
}
| VARIABLE ARGUMENTS {
  component co;
  if ((co = searchCompVN (*$1)) != NULL) {
    co->comp_args = new vector<string>();
    for (unsigned int i=1; i<$2->size(); i++) {
      co->comp_args->push_back((*$2)[i]);
    }
    comp_name_mdl = (*$2)[0];  /* set name */
    co->name = new string (comp_name_mdl);
    co->com->setName (comp_name_mdl);
    comp_var_name = *co->var_name;
    defname = getComponentType (co) + ".def";
    parseHybridDefFile(defname.c_str(), co->comp_args, co->com);
    rankComponents (co);
    cout << defname << " --- " << comp_var_name << " \t\t\t-----> [OK]\n";
  }
  else semError (0, ModelHybridParserln (), 67);
}
|
;           

%%

int ModelHybridParsererror(char *s) {
  fprintf(stdout, "%s in main .mdl parser. Line %i\n", s, ModelHybridParserln ());
  exit(0);
}

bool findDefFile(const char* deffilename) {
  FILE* test_def;
  if ((test_def = fopen(deffilename, "r")) != NULL) {
    fclose(test_def);
    return true;
  }
  cout << "Error => file not found : " << deffilename << endl;
  return false;
}

bool knownType(const char* ty) {
  string tys = string(ty);
  for (unsigned int i=0; i<declarations->size(); i++)
    if ((*declarations)[i] == tys)
      return 1;
  return 0;
}

int parseModelFile(const char* fileIn) {
  int pos = 0;
  out_cpp = buf_cpp;
  out_h = buf_h;
  
  strcpy(out_cpp, appliPath);
  strcat(out_cpp, "runtime/");
  strcat (buf_cpp, appliName);
  strcpy(out, buf_cpp);
  strcpy(out_h, buf_cpp);
  strcat(out_cpp, "_rt.cpp\0");
  strcat(out_h, "_rt.h\0");
  
  outh = string(buf_h);
  pos = outh.find("/");
  outh = outh.substr(pos+1, outh.size());

  class_char_cpp = class_buf_cpp;
  class_char_h = class_buf_h;
  
  int res=0;
  res = ModelHybridParserparse();
  verifyInstances ();
  FILE* cpf;
  if ((cpf = fopen (out_cpp, "w")) != NULL) { /* otherwise don't write when not desired */
    fprintf (cpf, class_char_cpp);
    fclose (cpf);
    FILE* hf = fopen (out_h, "w");
    fprintf (hf, class_char_h);
    fclose (hf);
  }
  
  class_char_cpp = 0;
  class_char_h = 0;
  
  return res;
}

void verifyInstances () {
  for (unsigned int i=0; i<comps->size(); i++) {
    if (! (*comps)[i]->com->instanciated)
      semError (3, ModelHybridParserln (), 75, (*(*comps)[i]->name).c_str ());
  }
}

bool crossRefs (component co1, component co2) {
  if (isIncludedInComp (co1->com, co1->com->type)
      && isIncludedInComp (co2->com, co1->com->type)) {
    return true;
  }
  return false;
}

/* 
 * rank components creation in C++ generated code
 * according to cross_references and inclusions
 */
void rankComponents (component co) {
  component ci, c;
  int ci_rank, c_rank;
  int co_rank = (int) oco->size () + 1;
  bool inserted = false;
  bool cref;

  if (oco->empty ()) {
    oco->insert (pair<int, component> (co_rank, co));
    return;
  }
  
  for (map<int, component>::iterator it = oco->begin ();
       it != oco->end (); it++) {
    ci = (*it).second;
    ci_rank = (*it).first;
    
    if ((co != ci) & (cref = crossRefs (co, ci))
	/*&& ! isACommandArg (co, getComponentVarName (ci))*/
	&& (co_rank >= ci_rank)) {          /* components cross reference: moves co up ci. */
      co_rank = ci_rank;
      map<int, component>::iterator it2 = oco->end ();
      map<int, component>::iterator itci = oco->find (ci_rank);
      while (true) {
	it2--;
	if ((*it2).second == co)    /* co already inserted */
	  oco->erase (it2);
	else {
	  c_rank = (*it2).first;
	  c = (*it2).second;
	  oco->erase (it2);
	  oco->insert (pair <int, component> (c_rank+1, c));
	}
	if (it2 == itci) {
	  oco->insert (pair<int, component> (co_rank, co));
	  break;
	}
      }
      it = oco->begin ();
    }
    else if ((co != ci) && ((! cref && isInComponentArgs (co, getComponentVarName (ci)))
	      || isACommandArg (co, getComponentVarName (ci)))
	     && (ci_rank >= co_rank)) {                        /* co needs ci, so ci moves up co */
      /*cout << getComponentVarName (co) << " needs " << getComponentVarName (ci) << endl;*/
      ci_rank = co_rank;
      map<int, component>::iterator itco = oco->find (ci_rank);
      map<int, component>::iterator it3 = oco->end ();
      while (true) {
	it3--;
	if ((*it3).second == ci) {
	  oco->erase (it3);
	}
	else {
	  c_rank = (*it3).first;
	  c = (*it3).second;
	  oco->erase (it3);
	  oco->insert (pair<int, component> (c_rank+1, c));
	}
	if (it3 == itco)
	  break;
      }
      oco->insert (pair<int, component> (ci_rank, ci));
      it = oco->begin ();
    }
    else if (! inserted) {
      oco->insert (pair<int, component> (co_rank, co));
      inserted = true;
    }
  }
}

void codeComponent(component co) {
  int dot_pos;
  string codeH;
  
  code = string ("\t") + getComponentVarName (co) + " = new " + *co->type + "(" + "\"" + *co->name + "\", ";
  /*code = string ("\t") + getComponentVarName (co) + " = new " + *co->type + "(" + "\"" + *co->var_name + "\", ";*/
    codeH = string("\n\t") + *co->type + "* " + getComponentVarName (co) + ";";
    cl_header += "\n\textern " + *co->type + "* " + getComponentVarName (co) + ";";
    strcat(class_char_h, codeH.c_str());
    for (unsigned int i=0; i<co->comp_args->size(); i++) {
      if (i != co->comp_args->size() - 1) {
	if ((dot_pos = (*co->comp_args)[i].find(".")) < 0)
	  code.append((*co->comp_args)[i] + ", ");
	else {
	  /* hack for commands in & out. */
	  if ((*co->comp_args)[i].substr(dot_pos+1, (*co->comp_args)[i].length()-1) == "cmdout")
	    code.append((*co->comp_args)[i].substr(0, dot_pos) + "->cmd_out, ");
	}
      }
      else {
	code.append((*co->comp_args)[i] + ");\n");
      }
    }
    code += string ("\t") + getComponentVarName (co) + "->type = \"" + *co->type + "\";\n";
    code += string ("\t") + getComponentVarName (co) + "->var_name = \"" + getComponentVarName (co) + "\";\n";
    strcat(class_char_cpp, code.c_str());
    code = "";
}

void codeAddComp() {
  for (unsigned int i=0; i<comps->size(); i++) {
    code.append("\tsMM->addComp(" + *(*comps)[i]->var_name + ");\n");
  }
  strcat(class_char_cpp, code.c_str());
}

/* search a comp given its var_name */
component searchCompVN (string vn) {
  for (unsigned int i=0; i<comps->size(); i++) {
    if (*(*comps)[i]->var_name == vn)
      return (*comps)[i];
  }
  semError (1, ModelHybridParserln(), 67);  /* error component not declared ! */
  return NULL;
}

void compToDot(component co) {
  int dot_pos;
  dot_buf.append(*co->name + ";\n");
  for (unsigned int i=0; i<co->comp_args->size(); i++) {
    if ((*co->comp_args)[i] != "0") {
      if ((dot_pos = (*co->comp_args)[i].find(".")) < 0)
	dot_buf.append(*searchCompVN((*co->comp_args)[i])->name + " -> " + *co->name + ";\n");
      else {
	dot_buf.append(*searchCompVN((*co->comp_args)[i].substr(0, dot_pos))->name + " -> " + *co->name + "[ label=\"" + (*co->comp_args)[i].substr(dot_pos+1, (*co->comp_args)[i].length()-1) + "\"];\n");
	
      }
    }
  }
}

void toDot() {
  int i=0;
  char dotf[32];
  while ((dotf[i] = fileIn[i]) != '.')
    i++;
  dotf[i++] = '\0';
  strcat(dotf, ".dot");
  ofstream f_h(dotf, ios::trunc);
  dot_buf.append("}");
  f_h << dot_buf;
}

void showModel() {
  cout << endl << class_buf_cpp << endl;
}
