
/*    
 *  Copyright 2000, 2001, 2002, 2003 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* koala-exec.h */

#ifndef RUNT
#define RUNT
#endif

#define DEBUG

#include <ctype.h>
#include <stdlib.h>
#include "koala-runt.h"
#include "../config.h"
#ifdef HAVE_LIBREADLINE
#include <readline/readline.h>
#include <readline/history.h>
#ifdef __cplusplus
extern "C" {
#endif
  extern void initialize_readline (void);
  extern char* command_generator ();
  extern char** ce_completion (const char*, int, int);
#ifdef __cplusplus
}
#endif
#endif /* HAVE_LIBREADLINE */
extern char dataInFile [], inLogFile [], dataOutFile [];
extern FILE *dataIn, *logS, *dataOut;
extern multimap<long, runtime, less<long> >* nrt;
#ifdef DEBUG
extern long totalAllo, totalDeallo;
#endif

/* El tipo LFunction es definido como Function en <readline/keymaps.h>: 
              es mas legible `LFunction *func;' que `long (*)()func;' (gnu)  */
typedef long LFunction (char *);
typedef struct {                /* Structura para caber los mandos de Ca-En */
  char *name;                   /* User printable name of the function. */
  LFunction *func;                       /* Function to call to do the job.      */
  const char *doc;                    /* Documentation for this function.     */
} ceCom;

extern char *do_gets (char *);
#ifndef GNU
#define whitespace(c) (((c) == ' ') || ((c) == '\t'))
#endif /* GNU */

/* flags and counters */
extern long guards_int;
extern long h_pi;               /* have compiled prime implicants flag */
extern long disp_num;           /* counter for display ordering */
extern string model_name;
extern bool discrete;

extern void ligneInit (void);
extern void stripWhite (char *);
extern ceCom *whichCom (ceCom *, char *);

extern runtime koala_rt;
extern void startinit (Scheduler* sMM);
extern void localInit (runtime rt, Scheduler* sMM);

extern long ceQuit (char *);
extern long ceHelp (char *s);
extern void printHelp (char *);
extern long ceComment (char *);
extern long ceInit (char *);
extern long ceDiagnosis (char *);
extern long prepareEI (void);
extern long prepareSc (void);
extern long prepareMS (void);
extern long ceDetect (char *);
extern long ceSimul (char *);
extern long ceDisc (char *);
extern long cePrint (char *);
extern long prepareMO (void);
extern long prepareDO (void);
extern long ceSystem (char *);
extern long ceGnup (char *);
extern long ceDebug (char *);
extern long deExit (char *);
extern long deHelp (char *);
extern long dePrint (char *);


extern void initDispAll (runtime, double);
extern void writeDispAll (FILE*, multimap<double, hybTrajectory*, greater<double> >*);
void graphTrajs (multimap<double, hybTrajectory*, greater<double> >*);

/* required functions in runtime: */
extern long koala_getInputs (const char*, long, double*);
extern long koala_getMeasures (const char*, long, double*);
extern void koala_setRTDisplays (long, runtime, double);
extern long koala_setDisplays (const char*, long, double*, double*, long);
