
/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* caen-runt.h */

#ifndef _CE_RUNT_H
#define _CE_RUNT_H
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <map>
#include <queue>
#include "../behav/Atom.h"

using namespace std;

/*#define DEBUG*/

extern "C" {void opt_nelder
(
double (*Nelder_cout)   /* ptr de la fonction de calcul du cout */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)    */
int    *iuser,  /* ptr d'entiers (variables utilisateur)     */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      + = nb d'evaluations du cout   */
		/*	     -1 = max iterations atteint     */
		/*	    -99 = erreur dans Nelder_cout    */
);}

#define ENV_OPTI   /* envelopes optimization */

#ifndef EQBOOL
#define EQBOOL
struct eqbool
  {
    bool operator()(bool s1, bool s2)
    {
      return (s1 == s2);
    }
};
#endif

#define isNull(p)               ((p) == NULL)
#define max(a,b)                ((a) < (b) ? (b) : (a))
#define min(a,b)                ((a) < (b) ? (a) : (b))

#ifdef DEBUG
#define cealloc(n)     (L1 = (long) (n), memAl += L1, (char *)malloc (L1))
#define cefree(p)      (memFr += (long) sizeof (*p), free (p))
#else /* DEBUG */
#define cealloc(n)     ((char *)malloc (n))
#define cefree(x)      (free (x))
#endif /* DEBUG */

/* les macros pour les chai^nes de caracte`res */
#define allocString(s)      (cealloc(strlen(s)+1))   /* room for the \0, too */
#define mkString(s)         (strcpy(allocString(s),s))
#define isEmpty(s)          (*(s) == '\0')
/* gets memory for s + t, then copies s, then appends t */
#define catString(s,t)      (isNull(t) ? s \
			     : (S1 = strcat (strcpy (cealloc (strlen(s) \
				  + strlen(t)  + 1), s), t), free (s), S1))

#ifdef DEBUG
static long L1;
static long memAl;
static long memFr;
#endif

/* copy of arrays */
inline long* copyLong (long* a, long s) {
  if (isNull (a) || s == 0)
    return NULL;
  else {
    long* lnc = (long*) cealloc (s*sizeof(long));
    for (int i=0; i<s; i++)
      lnc[i] = a[i];
    return lnc;	       
  }	
}

inline double* copyDouble (double* a, long s) {
  if (isNull (a))
    return NULL;
  else {
    double* dnc = (double*) cealloc (s*sizeof(double));
    for (int i=0; i<s; i++)
      dnc[i] = a[i];
    return dnc;
  }
}

/*
 * boolean values
 */
#ifndef TY_BOOLEAN
#define TY_BOOLEAN
typedef enum boolVal {CE_FALSE, CE_TRUE} boolean;
#endif

inline boolean* copyBoolean (boolean* a, long s) {
  if (isNull (a))
    return NULL;
  else {
    boolean* bnc = (boolean*) cealloc (s*sizeof(boolean));
    for (int i=0; i<s; i++)
      bnc[i] = a[i];
    return bnc;
  } 
}

extern void ce_printReal (FILE *, double);

/**
 * \struct range
 * \brief the "range" type, along with its operators.
 */
typedef struct {
  double inf;
  double sup;
  bool incinf;              /**< inf is included or not */
  bool incsup;              /**< sup is included or not */
} range;

/**
 * get range inf.
 */
#define getRangeInf(r)          ((r).inf)

/**
 * get range sup.
 */
#define getRangeSup(r)          ((r).sup)
#define setRangeInf(r,i)        ((r).inf = i)
#define setRangeSup(r,s)        ((r).sup = s)

/**
 * set range values (inf, sup).
 */
#define isIncludedInf(r)        ((r).incinf == 1)
#define isIncludedSup(r)        ((r).incsup == 1)
#define setIncludedInf(r)       ((r).incinf = 1)
#define setIncludedSup(r)       ((r).incsup = 1)
#define setNotIncludedInf(r)    ((r).incinf = 0)
#define setNotIncludedSup(r)    ((r).incsup = 0)
#define getBoundInf(r)          ((r).incinf)
#define getBoundSup(r)          ((r).incsup)
#define setBounds(r1, r2)       ((r1).incinf = (r2).incinf, \
                                 (r1).incsup = (r2).incsup)
#define setRange(r,i,s)         (setRangeInf (r, i), setRangeSup (r,s))
#define isRangeThin(r)          ((r).inf == (r).sup)
#define isInRange(d,r)          (((d > getRangeInf (r)) \
				 && (d < getRangeSup (r))) \
                                 || ((d == getRangeInf (r)) \
                                 && isIncludedInf (r)) \
                                 || ((d == getRangeSup (r)) \
                                 && isIncludedSup (r)))

extern void ce_printRange (FILE *, range);

/**
 * \enum objType
 * \brief types of objects.
 */
enum objType {CE_UNKNOWN, CE_REAL, CE_RANGE, CE_BOOL};

/**
 * \union u_Val
 * \brief the "valeur" type.
 */
typedef union u_Val {
  range vrange;
  double vdouble;
  boolean vbool;
} valeur;

/**
 * \struct s_Objet
 * \brief the "objet" type, along with its operators.
 */
typedef struct s_Objet {
  enum objType type;
  valeur val;
} s_objet, *objet;

#define allocObj()            ((objet) cealloc(sizeof(s_Objet)))
#define allocVObj(n)          ((objet *) cealloc((n)*sizeof(objet)))
#define getObjType(b)         ((b)->type)

/**
 * return real value of the valeur of the object.
 */
#define getObjRealVal(b)      ((b)->val.vdouble)

/**
 * return range value (struct) of the valeur of the object.
 */
#define getObjRangeVal(b)     ((b)->val.vrange)

/**
 * boolean value of the valeur of the object.
*/
#define getObjBoolVal(b)      ((b)->val.vbool)

/**
 * set the object's type.
*/
#define setObjType(b,t)       ((b)->type = t)

/**
 * set the object's valeur real value (vdouble).
 */
#define setObjRealVal(b,d)    (b->val.vdouble = d)

/**
 * set the object's valeur range.
 */
#define setObjRangeVal(b,r)   ((b)->val.vrange = r)

/**
 * set the object's valeur bool value.
 */
#define setObjBoolVal(obj,b)  (obj->val.vbool = b)

/**
 * reset the object (as a REAL obj).
*/
#define razObjet(q)           (setObjType (q, CE_REAL), \
			       setObjRealVal (q, 0.0))

inline objet mkObjReal (double d) {
  objet Obj = allocObj ();
  setObjType (Obj, CE_REAL);
  setObjRealVal (Obj, d);
  return Obj;
}

inline objet copyObjet (objet b) {
  if (isNull (b))
    return NULL;
  else {
    objet Objc = allocObj ();
    setObjType (Objc, getObjType (b));
    if (getObjType (b) == CE_REAL)
      setObjRealVal (Objc, getObjRealVal (b));
    else if (getObjType (b) == CE_RANGE |
	     getObjType (b) == CE_UNKNOWN)
      setObjRangeVal (Objc, getObjRangeVal (b));
    else if (getObjType (b) == CE_BOOL)
      setObjBoolVal (Objc, getObjBoolVal (b));
    return Objc;
  }
}

inline objet* copyArrayObj (objet* a, long s) {
  if (isNull (a))
    return NULL;
  else {
    objet* Objnc = allocVObj (s);
    for (int i=0; i<s; i++)
      Objnc[i] = copyObjet (a[i]);
    return Objnc;
  }
}

#define freeObjet(o)    (cefree (o), o=0)

inline void freeArrayObjets (objet* a, long s) {
  if (!isNull (a)) {
    for (int i=0; i<s; i++)
      freeObjet (a[i]);
    cefree (a);
    a = 0;
  }
}

/**
 * eval and print object value to fd.
 */
extern void ce_printObjet (FILE *, objet);

/**
 * copy p values into q.
 */
extern void ce_putObjet (objet, objet);

/**
 * add an object b to an object a and update valeur.
 */
extern void ce_addObjet (objet, objet);

/**
 * substract an object b to an object a and update valeur.
 */
extern void ce_subObjet (objet, objet);

/**
 * reverse the sign of an object value.
 */
extern void ce_minusObjet (objet);

/**
 * set an object value to its exponential value.
 */
extern void ce_expObjet (objet);

/**
 * set an object value to its log value.
 */
extern void ce_logObjet (objet);

/**
 * take the sqrt of an object value.
 */
extern void ce_sqrtObjet (objet);

/* sinus optimization function */
double sin_cost (int *n, double *X, double *Xmin, double *Xmax, double *duser, int *iuser, int *iret, int *ier);

/**
 * take the sin of an objet value.
 */
extern void ce_sinObjet (objet);

/* cosinus optimization function */
double cos_cost (int *n, double *X, double *Xmin, double *Xmax, double *duser, int *iuser, int *iret, int *ier);

/**
 * take cos of an objet value.
 */
extern void ce_cosObjet (objet);

/**
 * set a as a*b.
 */
extern void ce_mulObjet (objet, objet);

/**
 * set object a as a/b and return 0 if the division is impossible (1 otherwise).
 */
extern long ce_divObjet (objet, objet);

/**
 * intersect two objects values and returns CE_FALSE.
 * result is in first argument.
 */
extern boolean ce_interObjet (objet, objet);

/**
 * set first object argument as the union of both objects.
 */
extern void ce_uniObjet (objet, objet);

/**
 * return the sign of the object value (-1 neg, 1 pos, 0 unknown).
 */
extern long ce_sgnObjet (objet);

/**
 * test if two objects have the same values.
 */
extern boolean ce_eqObjet (objet, objet);

/**
 * test if two objects have different types or values.
 */
extern boolean ce_diffObjet (objet, objet);

/**
 * test if object a value is lower than object b value.
 */
extern boolean ce_ltObjet (objet, objet);

/**
 * test if object a value is greater than object b value.
 */
extern boolean ce_geObjet (objet, objet);

/**
 * test if object a value is in object b value.
 */
extern boolean ce_inObjet (objet, objet);

extern objet ce_mkSpObjet (double, double);

extern objet* ce_inSpObjet (objet, objet);
/**
 * \struct s_Variable
 * \brief the "variable" type, along with its operators.
 */
typedef struct s_Variable {
  const char *nom;                                                  /**< name */
  long cst;
  objet dom;
  long deter;                          /**< if the variable value is determined or not. */
  objet oldVal;                        /**< old value of the variable. */
  objet newVal;
  objet mesVal;
  long lowalrm;
  long highalrm;
  long lowmisbehaving;
  long highmisbehaving;
  long misbehaving;
  long *histoMisb;
  struct s_CGraph *node;               /**< corresponding node in the causal graph. */
  long histoLn;                        /**< size of histo. */
  objet *histo;
  objet *histoMes;
  long *histoLowAlrm;
  long *histoHighAlrm;
  long ncd;
  long *relCd;       /**< conditions referenced by their position in runtime (condi) . */
} s_var, *variable;

#define allocVariable()       ((variable) cealloc(sizeof(s_Variable)))
#define allocVVar(n)          ((variable *) cealloc((n)*sizeof(variable)))
#define resetVarAlarms(v)        (v->lowalrm = 0, \
                                  v->highalrm = 0, \
                                  v->lowmisbehaving = 0, \
                                  v->highmisbehaving = 0, \
                                  v->misbehaving = 0)
#define resetVarHisto(v)         (v->histoLn = 0, \
                                  v->histo = NULL, \
                                  v->histoMes = NULL, \
                                  v->histoLowAlrm = NULL, \
                                  v->histoHighAlrm = NULL)
/**
 * \def getVarName(v)
 * return the name of variable \a v
 */
#define getVarName(v)            (v->nom)
#define isParameter(v)           (v->cst)
#define getVarDom(v)             (v->dom)
#define setVarDom(v,d)           (v->dom = d)
#define isDetermined(v)          (v->deter)
#define varOldValue(v)           (v->oldVal)
#define varNewValue(v)           (v->newVal)
#define varMesValue(v)           (v->mesVal)
#define isVarLowAlrm(v)          (v->lowalrm)
#define isVarHighAlrm(v)         (v->highalrm)
#define isVarLowMisbehaving(v)   (v->lowmisbehaving)
#define isVarHighMisbehaving(v)  (v->highmisbehaving)
#define isVarMisbehaving(v)      (v->misbehaving)
#define getVarHistoMisb(v)       (v->histoMisb)
#define getVarHistoMisbNth(v,n)  (*(v->histoMisb+(n)))
#define getVarNode(v)            (v->node)
#define getVarHistoLn(v)         (v->histoLn)
#define varHistoNth(v,i)         (*(v->histo+(i)))
#define varHistoMesNth(v,i)      (*(v->histoMes+(i)))
#define isVarHistoLowAlrm(v,i)   (*(v->histoLowAlrm+(i)))
#define isVarHistoHighAlrm(v,i)  (*(v->histoHighAlrm+(i)))
#define getVarNcd(v)             (v->ncd)
#define getVarRelCdNth(v,i)      (*(v->relCd+(i)))
#define setDetermined(v)         (v->deter = 1)
#define setUndetermined(v)       (v->deter = 0)
#define affectRealOld(v,d)       (setObjRealVal (varOldValue (v), d), \
			  	  setObjType (varOldValue (v), CE_REAL))
#define affectRealNew(v,d)       (setObjRealVal (varNewValue (v), d), \
			  	  setObjType (varNewValue (v), CE_REAL))
#define affectRealMes(v,d)       (setObjRealVal (varMesValue (v), d), \
			  	  setObjType (varMesValue (v), CE_REAL))
#define affectRange(v,r)         (setObjRangeVal (varOldValue (v), r), \
				  setObjType (varOldValue (v), CE_RANGE))
/**
 * reset low level alarm.
 */
#define razVarLowAlrm(v)         (v->lowalrm = 0)

/**
 * reset high level alarm.
 */
#define razVarHighAlrm(v)        (v->highalrm = 0)
#define setVarLowAlrm(v)         (v->lowalrm = 1)
#define setVarHighAlrm(v)        (v->highalrm = 1)
#define incVarLowAlrm(v,s)       ((v->lowalrm = (L1 = (v->lowalrm >= s-1)) \
                                  ? s : v->lowalrm + 1), L1)
#define incVarHighAlrm(v,s)      ((v->highalrm = (L1 = (v->highalrm >= s-1)) \
                                  ? s : v->highalrm + 1), L1)
#define razVarHistoLowAlrm(v,i)  (*(v->histoLowAlrm + (i)) = 0)
#define razVarHistoHighAlrm(v,i) (*(v->histoHighAlrm + (i)) = 0)
#define setVarHistoLowAlrm(v,i)  (*(v->histoLowAlrm + (i)) = 1)
#define setVarHistoHighAlrm(v,i) (*(v->histoHighAlrm + (i)) = 1)
#define setVarLowMisbehaving(v)  (v->lowmisbehaving = 1)
#define setVarHighMisbehaving(v) (v->highmisbehaving = 1)
#define setVarMisbehaving(v)     (v->misbehaving = 1)

/**
 * reset the variable low misbehaving flag.
 */
#define setVarLowOk(v)           (v->lowmisbehaving = 0)

/**
 * reset the variable high misbehaving flag.
 */
#define setVarHighOk(v)          (v->highmisbehaving = 0)

/**
 * reset the variable global misbehaving flag.
 */
#define setVarOk(v)              (v->misbehaving = 0)
#define setVarHistoMisb(v,h)     (v->histoMisb = h)
#define setVarHistoMisbNth(v,n,m) (*(v->histoMisb+(n)) = m)

#define freeVariable(v)           (!isNull(v) ? \
                                  (freeObjet (v->oldVal), \
                                   freeObjet (v->newVal), \
                                   freeObjet (v->mesVal), \
                                   cefree (v->histoMisb), \
                                   freeArrayObjets (v->histo, v->histoLn), \
                                   freeArrayObjets (v->histoMes, v->histoLn), \
                                   cefree (v->histoLowAlrm), \
                                   cefree (v->histoHighAlrm), \
                                   cefree (v->relCd), \
                                   cefree (v), \
                                   v=0) : 0)

inline void freeArrayVariables (variable* a, long s) {
  if (!isNull (a)) {
    for (int i=0; i<s; i++)
      freeVariable (a[i]);
    cefree (a);
    a=0;
  }
}


/**
 * print the var name to the file descriptor.
 */
extern void ce_printVarNom (FILE *, variable);

extern void ce_printVarValueNV (FILE *, variable);

/**
 * print the var value to the file descriptor.
 */
extern void ce_printVarValue (FILE *, variable);

/**
 * \enum gVarType
 * \brief global variables types.
 */
enum gVarType {CE_CONST, CE_VAR, CE_INTERM};

/**
 * \struct s_GVar
 * \brief le type "gVar": variable pour le calcul des contraintes globales,
 * avec ses operateurs.
 */
typedef struct s_GVar {
  enum gVarType tgv;
  union {
    variable var;
    long interm;
  } uv;
  objet ob;
  long nex;
  long *ex;
  long ncex;
  long *cex;
} s_gVar, *gVar;

#define allocGVar()           ((gVar) cealloc (sizeof(s_GVar)))
#define allocVGVar(n)          ((gVar*) cealloc (n*sizeof(gVar)))
#define getGVType(gv)           (gv->tgv)
#define getGVVar(gv)            (gv->uv.var)
#define getGVInterm(gv)         (gv->uv.interm)
#define gVVal(gv)               (gv->ob)
#define getGVNGlobC(gv)         (gv->nex)
#define getGVGlobCNth(gv,n)     (*(gv->ex+(n)))    /* n desde 1 hasta nGlobC */
#define getGVNCondC(gv)         (gv->ncex)
#define getGVCondCNth(gv,n)     (*(gv->cex+(n)))
#define freeGVar(gv)            (!isNull (gv) ? (freeObjet (gv->ob), \
                                 cefree (gv->ex), \
                                 cefree (gv->cex), \
                                 cefree (gv), \
                                 gv=0) : 0)

inline void freeArrayGVar (gVar* a, long s) {
  if (!isNull (a)) {
    for (int i=0; i<s; i++)
      freeGVar (a[i]);
    cefree (a);
    a=0;
  }
}

/**
 * \enum exprType
 * \brief les types d' expressions.
 */
enum exprType {CE_CSTE, CE_VBL, CE_PLUS, CE_MOINS, CE_MULT,
	       CE_DIV, CE_EXP, CE_SGN};

/**
 * \struct s_Expr
 * \brief le type "expr": representation des contraintes globales.
 */
typedef struct s_Expr {
  enum exprType tex;
  gVar gv;
  struct s_Expr *op1;
  struct s_Expr *op2;
} s_expr, *expr;

#define allocExpr()         ((expr) cealloc(sizeof(s_Expr)))
#define allocVExpr(n)       ((expr*) cealloc((n)*sizeof(expr)))
#define getExprType(e)           (e->tex)
#define getExprGVar(e)           (e->gv)
#define getExprOp1(e)            (e->op1)
#define getExprOp2(e)            (e->op2)
#define isExprCte(e)             (CE_CSTE == getExprType (e))
#define hasName(e)               (! isNull (getExprGVar (e)))
#define freeExpr(e)         (!isNull (e) ? \
                             (cefree (e), \
                             e=0) : 0)

inline void freeArrayExpr (expr* a, long s) {
  if (!isNull (a)) {
    for (int i=0; i<s; i++)
      freeExpr (a[i]);
    cefree (a);
    a=0;
  }
}

/* types of conditions */
enum condType {CE_RIEN, CE_CTE, CE_COMP, CE_OU, CE_ET, CE_NON, CE_AT, CE_EXPR};

/* relation operators */
enum opRel {CE_NONE, CE_EQ, CE_NE, CE_LE, CE_LT, CE_GE, CE_GT, CE_IN};

/**
 * \struct s_Condition
 * \brief the "condition" type, along with its operators.
 */
typedef struct s_Condition {
  enum condType type;        /**< type of the condition {CE_RIEN, CE_CTE, CE_COMP, CE_AT, CE_EXPR} */
  struct s_Condition *op1;          
  struct s_Condition *op2; 
  variable var1;
  enum opRel opr;           /**< operator of the condition {CE_OU, CE_ET, CE_NON} */
  objet ob;
  variable var2;
  boolean val;
  Atom* at;
  int num;                  /**< ref in condi (runtime) */
  expr e;
  long nle;
  long* le;                 /**< related expr (sol-functions) */
} s_condition, *condition;

#define allocCondition()         ((condition) cealloc(sizeof(s_Condition)))
#define allocVCondition(n)       ((condition *) cealloc((n)*sizeof(condition)))
#define setCondCTE(f,v)     (f->op1 = NULL, \
                             f->op2 = NULL, \
                             f->var1 = NULL, \
                             f->opr = (opRel)0, \
                             f->ob = NULL, \
                             f->var2 = NULL, \
                             f->val = (boolean)v, \
                             f->at = NULL)
#define getCondNum(f)       (f->num)
#define getCondAtom(f)      (f->at)
#define getCondType(f)      (f->type)
#define getCondOp1(f)       (f->op1)
#define getCondOp2(f)       (f->op2)
#define getCondVar1(f)      (f->var1)
#define getCondOpRel(f)     (f->opr)
#define getCondObj(f)       (f->ob)
#define getCondVar2(f)      (f->var2)
#define getCondValue(f)     (f->val)
#define getCondNle(f)       (f->nle)
#define getCondLe(f)        (f->le)
#define getCondLeNth(f,i)   (*(f->le+(i)))
#define getCondExpr(f)      (f->e)
#define setCondValue(f,b)   (f->val = b)
#define setCondType(f,b)    (f->type = (condType)b)
#define setCondAtom(f,a)    (f->at = a)
#define setCondNum(f,i)    (f->num = i)
#define freeCond(f)         (!isNull (f) ? \
                             (cefree (f->le), \
			      cefree (f), \
                             f=0) : 0)

inline condition mkCond() {
  condition f = allocCondition ();
  setCondType (f, 0);
  f->op1 = NULL;
  f->op2 = NULL;
  f->var1 = NULL;
  f->ob = NULL;
  f->var2 = NULL;
  f->at = NULL;
  f->e = NULL;
  f->le = 0;
  return f;
}

inline void freeArrayCond (condition* a, long s) {
  if (!isNull (a)) {
    for (int i=0; i<s; i++)
      freeCond (a[i]);
    cefree (a);
    a=0;
  }
}

/**
 * print condition to the file descriptor.
 */
extern void ce_printCondition (FILE *, condition);

/**
 * print condition variables values in the file descriptor.
 */
extern void ce_printCondValues (FILE *, condition);

/**
 * \enum influType
 * \brief types of influences.
 * - nothing : CE_NOINFL,
 * - dynamic 1st order : CE_DINFL,
 * - integral 1st order : CE_IINFL,
 * - static : CE_SINFL,
 * - constant : CE_CINFL,
 * - dynamic 2nd order : CE_D2INFL.
 */
enum influType {CE_NOINFL, CE_DINFL, CE_IINFL, CE_SINFL, CE_CINFL, CE_D2INFL};

/**
 * \struct s_Influence
 * \brief the "influence" type, along with its operators.
 */
typedef struct s_Influence {
  const char *name;
  enum influType it;          /**< type of the influence. */
  variable cause;             /**< source variable. */
  variable effet;             /**< dest variable. */
  condition cd;               /**< condition of the influence. */
  long length_act_cond_hist;
  boolean *act_cond_array;    
  objet gain;                 /**< gain of the influence. */
  long delinf;                /* delay sup & inf ?? */
  long delsup;
  objet tRep;                 /**< response time. */
  objet a;
  objet b;
  objet a2;
  objet b2;
  objet xi;
  objet w;
  objet n1;
  objet n2;
  objet d0;
  objet d1;
  objet d2;
#ifdef ENV_OPTI
   /*double (*inf_fct) (int*, double*, double*, double*, double*, int*, int*, int*);*/  /* fct pointer for optimization code */
  bool inf_fct;
#endif
} s_influence, *influence;

#define allocInfluence()    ((influence) cealloc(sizeof(s_Influence)))
#define allocVInfluence(n)  ((influence *) cealloc((n)*sizeof(influence)))     

/**
 * eval a dynamic influence following :
 * Y(t+1) = a*Y(t) + (1-a)*Sigma(Ki*Xi(t-di)).
 * Also compute Tr = -3*Ts/ln(a).
 */
extern void ce_evalInitDynInfl (influence);

/* ? */
extern void ce_pushExternCond (influence, boolean);

#define getInfluNom(i)          (i->name)
#define getInfluType(i)         (i->it)
#define getInfluCause(i)        (i->cause)
#define getInfluEffet(i)        (i->effet)
#define getInfluCond(i)         (i->cd)
#define getInfluNActCond(i)     (i->length_act_cond_hist)
#define getInfluActCondNth(i,n) (*(i->act_cond_array + (n)))
#define getInfluGain(i)         (i->gain)
#define getInfluDelInf(i)       (i->delinf)
#define getInfluDelSup(i)       (i->delsup)
#define getInfluTRep(i)         (i->tRep)
#define getInfluA(i)            (i->a)
#define getInfluB(i)            (i->b)
#define setInfluA(i,v)          (i->a = v)
#define setInfluB(i,v)          (i->b = v)
#define setInfluNActCond(i,n)   (i->length_act_cond_hist = n)
#define setInfluActCondNth(i,n,c) (*(i->act_cond_array + (n)) = c)
#define getInfluXi(i)           (i->xi)
#define getInfluW(i)            (i->w)
#define getInfluN1(i)           (i->n1)
#define getInfluN2(i)           (i->n2)
#define getInfluD0(i)           (i->d0)
#define getInfluD1(i)           (i->d1)
#define getInfluD2(i)           (i->d2)
#define getInfluA2(i)           (i->a2)
#define getInfluB2(i)           (i->b2)
#define getInfOptFct(i)         (i->inf_fct)
#define freeInfluence(i)        (!isNull (i) ? (cefree (i->act_cond_array), \
                                 freeObjet (i->gain), \
                                 freeObjet (i->tRep), \
                                 freeObjet (i->a), \
                                 freeObjet (i->b), \
                                 freeObjet (i->a2), \
                                 freeObjet (i->b2), \
                                 freeObjet (i->xi), \
                                 freeObjet (i->w), \
                                 freeObjet (i->n1), \
                                 freeObjet (i->n2), \
                                 freeObjet (i->d0), \
                                 freeObjet (i->d1), \
                                 freeObjet (i->d2), \
                                 cefree (i), \
                                 i=0) : 0)

inline void freeArrayInfluences (influence* a, long s) {
  if (!isNull (a)) {
    for (int i=0; i<s; i++)
      freeInfluence (a[i]);
    cefree (a);
    a=0;
  }
}

extern void printInfluence (influence);

/**
 * \struct s_CGraph
 * \brief the "causal graph" type , along with its operators. Node of a causal graph.
 */
typedef struct s_CGraph {
  long ncause;            /**< nb of effect influences to that node. */
  influence *vcause;      /**< effect influences to that node. */
  enum influType ti;      /**< influence type of the node. */
  long neffet;            /**< nb of causal influences from that node. */
  influence *veffet;      /**< causal influences from that node. */
} s_cGraph, *cGraph;

#define allocCGraph()          ((cGraph) cealloc(sizeof(s_CGraph)))
#define allocVCGraph(n)  ((cGraph *) cealloc((n)*sizeof(cGraph)))
#define getCGNCause(cg)        (cg->ncause)
#define getCGCauseNth(cg,n)    (*(cg->vcause+(n))) /* n desde 0 hasta ncause-1 */
#define getCGCauseType(cg)     (cg->ti)
#define getCGNEffet(cg)        (cg->neffet)
#define getCGEffetNth(cg,n)    (*(cg->veffet+(n))) /* n desde 0 hasta neffet-1 */
#define freeCGraph(cg)         (cefree (cg), \
                                cg = 0)

inline void freeArrayCGraph(cGraph* a, long s) {
  for (int i=0; i<s; i++) {
    freeCGraph (a[i]);
    a[i]=0;
  }
  cefree (a);
  a=0;
}

/**
 * \struct s_InitVar
 * \brief the "variable initializer" type , along with its operators.
 */
typedef struct s_InitVar {
  variable var;
  objet val;
} s_initVar, *initVar;

#define allocInitVar()       ((initVar) cealloc (sizeof (s_InitVar)))
#define allocVInitV(n)       ((initVar*) cealloc (n*sizeof (initVar)))
#define getIVVar(iv)           (iv->var)
#define getIVVal(iv)           (iv->val)
#define freeInitVar(iv)        (!isNull (iv) ? \
                                (freeObjet (iv->val), \
                                cefree (iv), \
                                iv = 0) : 0)

inline void freeArrayIV (initVar* a, long s) {
  if (!isNull(a)) {
    for (int i=0; i<s; i++)
      freeInitVar (a[i]); 
    cefree (a);
    a=0;
  }
}

/**
 * print the InitVar value on the file descriptor.
 */
extern void ce_printInitVar (FILE *, initVar);

/**
 * \struct s_mapfVar
 * \brief type for initialization variables in mapping functions.
 * TODO : must add expression.
 */
typedef struct s_mapfVar {
  const char* var_name;
  /*objet vi;*/
  long e;  /* expression location */
} S_mapfVar, *mapfVar;

#define allocMapfVar()         ((mapfVar) cealloc (sizeof (S_mapfVar)))
#define getMFVarName(mv)       (mv->var_name)
#define getMFExpr(mv)          (mv->e)
#define setMFVarName(mv,n)     (mv->var_name = n)
#define setMFExpr(mv,ex)       (mv->e = ex)

inline mapfVar mkMapfVar (const char* n, long ex) {
  mapfVar Mv = allocMapfVar ();
  setMFVarName (Mv, n);
  setMFExpr (Mv, ex);
  return Mv;
}

#ifdef DEBUG
extern long totalAllo, totalDeallo;
#endif

void *new_malloc(size_t);
void *new_calloc(size_t, size_t);
void new_free(void *, size_t);

typedef enum answer {NO, YES} ANSWER;

typedef enum status {MEM_ERROR, OK} STATUS;

typedef void *ELEMENT;

typedef struct NODE{
	ELEMENT datum;
	struct NODE *next;
} node, *ptNode;

typedef struct LISTCAEN {
	ptNode first,last,current;
	int nEl;
} listcaen, *ptListcaen;


#define listcaenLength(l)      ((l).nEl)

#define isEmptyListcaen(l)     (listcaenLength(l) == 0)

#define resetCurrent(pL)   ((pL)->current = (pL)->first)

#define advanceCurrent(pL) (if (!isEmptyListcaen(*(pL))) (pL)->current = (pL)->current->next)

#define lookLeft(l) 	   ((!isEmptyListcaen(l)) ? (l).first->datum : NULL)

#define lookRight(l)       ((!isEmptyListcaen(l)) ? (l).last->datum : NULL)

#define lookCurrent(l)     ((!isEmptyListcaen(l)) ? (l).current->datum : NULL)


STATUS createEmpty(ptListcaen pL);

STATUS createUnitary(ptListcaen pL, ELEMENT e);

STATUS createUnitaryC(ptListcaen pL, ELEMENT e, int DatumSize);

STATUS addLeft(ELEMENT e, ptListcaen pL);

STATUS addLeftC(ELEMENT e, ptListcaen pL, int DatumSize);

STATUS addRight(ELEMENT e, ptListcaen pL);

STATUS addRightC(ELEMENT e, ptListcaen pL, int DatumSize);

ELEMENT lookLeftC(listcaen l, int DatumSize);

ELEMENT lookRightC(listcaen l, int DatumSize);

ELEMENT lookCurrentC(listcaen l, int DatumSize);

/* Before using 'lookAdvanceCurrent' functions, use 'resetCurrent'.
   After using 'lookNth' functions, and before calling any of the
   'lookAdvanceCurrent' functions, a function call to 'advanceCurrent'
   should be carried out in order to avoid reading twice the same element. */
ELEMENT lookAdvanceCurrent(ptListcaen pL);

ELEMENT lookAdvanceCurrentC(ptListcaen pL, int DatumSize);

ELEMENT lookNth(ptListcaen pL, int p);

ELEMENT lookNthC(ptListcaen pL, int p, int DatumSize);

STATUS deleteLeft(ptListcaen pL);

STATUS deleteLeftC(ptListcaen pL, size_t size);

STATUS deleteRight(ptListcaen pL);

STATUS deleteRightC(ptListcaen pL, size_t size);

STATUS concatenate(ptListcaen pL1, listcaen l2);

STATUS concatenateOn1Copying2(ptListcaen ptL1, listcaen l2);

STATUS concatenateOn1Copying2C(ptListcaen ptL1, listcaen l2, int DatumSize);

STATUS listcaenConcatenation(listcaen l1, listcaen l2, ptListcaen result);

STATUS listcaenConcatenationC(listcaen l1, listcaen l2, ptListcaen result, int DatumSize);

STATUS listcaenIntersection(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*));

STATUS listcaenIntersectionC(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*), int DatumSize);

STATUS listcaenUnion(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*));

STATUS listcaenUnionC(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*), int DatumSize);

STATUS listcaenUnionOnFirst(ptListcaen ptL1, listcaen l2, int (*IG)(void *, void*));

STATUS listcaenUnionOnFirstC(ptListcaen ptL1, listcaen l2, int (*IG)(void *, void*), int datumSize);

STATUS listcaenSymmetricDifference(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*));

STATUS listcaenSymmetricDifferenceC(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*), int DatumSize);

STATUS listcaenExclusiveElements(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*));

STATUS listcaenExclusiveElementsC(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*), int DatumSize);

int isPresent(ELEMENT e, listcaen l, int (*IG)(void *, void *));

ELEMENT search(ELEMENT e, listcaen l, int (*IG)(void *, void *));

int position(ELEMENT e, listcaen l, int (*IG)(void *, void *));

STATUS insert(ptListcaen pL, int i, ELEMENT e);

STATUS insertC(ptListcaen pL, int i, ELEMENT e, int DatumSize);

STATUS ce_delete(ptListcaen pL, int i);
/* STATUS delete(ptListcaen pL, int i); */

STATUS deleteC(ptListcaen pL, int i, size_t size);

STATUS update(ptListcaen pL, int i, ELEMENT e);

STATUS updateC(ptListcaen pL, int i, ELEMENT e, int DatumSize);

STATUS assign(ptListcaen, listcaen);
/* STATUS assign(ptListcaen new, listcaen old); */

STATUS assignC(ptListcaen, listcaen, int);
/* STATUS assignC(ptListcaen new, listcaen old, int DatumSize); */

STATUS liberate(ptListcaen pL);

STATUS liberateC(ptListcaen pL, size_t size);

/**********/
/* diag.h */
/**********/

#define MAXLINE 79

/* A 'conflict' is a triplet <name ,activation condition, time-instant>. */
typedef struct Conflict {
  /* Either the name of the component to which the influence belongs or the own influence's name.
     For a given influence A ---> B, its name will be the string "A->B". */
  char * name;
  /* By keeping track of the 'influence', we have access to its activation-condition and 
     we can make up a name for it in case it belongs to no component. */    
  influence infl;
  /* (t - instant) tells us about the moment when the influence was active. */
  int instant;
  /* For implementation purposes */
  int index;
} *ptConflict, conflict;

#define allocConflict()                          ((ptConflict) cealloc (sizeof (Conflict)))
#define getConflictName(ptC)                     ((ptC)->name)
#define getConflictInfluence(ptC)                ((ptC)->infl)
#define getConflictInstant(ptC)                  ((ptC)->instant)
#define getConflictIndex(ptC)                    ((ptC)->index)
#define setConflictName(ptC,s)                   (getConflictName(ptC) = (s))
#define setConflictInfluence(ptC,ptI)            (getConflictInfluence(ptC) = (ptI))
#define setConflictInstant(ptC,i)                (getConflictInstant(ptC) = (i))
#define setConflictIndex(ptC,i)                  (getConflictIndex(ptC) = (i))

inline ptConflict copyConflict (ptConflict a) {
  if (isNull (a))
    return NULL;
  else {
    ptConflict c = allocConflict();
    setConflictName (c, getConflictName (a));
    setConflictInstant (c, getConflictInstant (a));
    setConflictIndex (c, getConflictIndex (a));
    return c;
  }
}

typedef listcaen *ptConflictSet, conflictSet;

#define createConflictSet(ptCS)                  (createEmpty(ptCS))
#define resetConflictSet(ptCS)                   (resetCurrent(ptCS))
#define getConflictSetNumber(ptCS)               (listcaenLength(*(ptCS)))
#define getFirstConflictCS(ptCS)                 ((ptConflict) lookLeft(*(ptCS)))
#define getFirstConflictCSC(ptCS)                ((ptConflict) lookLeftC(*(ptCS), sizeof(conflict)))
#define getNextConflictCS(ptCS)                  ((ptConflict) lookAdvanceCurrent(ptCS))
#define getNextConflictCSC(ptCS)                 ((ptConflict) lookAdvanceCurrentC(ptCS, sizeof(conflict)))
#define setFirstConflictCS(ptC,ptCS)             (addLeft(ptC, ptCS))
#define setFirstConflictCSC(ptC,ptCS)            (addLeftC(ptC, ptCS, sizeof(conflict)))
#define setNextConflictCS(ptC,ptCS)              (addRight(ptC, ptCS))
#define setNextConflictCSC(ptC,ptCS)             (addRightC(ptC, ptCS, sizeof(conflict)))
#define assignConflictSet(ptCS,CS)               (assign(ptCS, CS))
#define assignConflictSetC(ptCS,CS)              (assignC(ptCS, CS, sizeof(conflict)))
#define deleteFirstConflictCS(ptCS)              (deleteLeft(ptCS))
#define deleteFirstConflictCSC(ptCS)             (deleteLeftC(ptCS, sizeof(conflict)))
#define liberateConflictSet(ptCS)                (liberate(ptCS))
#define liberateConflictSetC(ptCS)               (liberateC(ptCS, sizeof(conflict)))

/* In these four functions, the conflict's name is taken
   into account when copying and deleting a conflict */
STATUS setFirstConflictWithNameCS(ptConflict, ptConflictSet);
STATUS setNextConflictWithNameCS(ptConflict, ptConflictSet);
STATUS assignConflictSetWithName(ptConflictSet, conflictSet);
STATUS deleteFirstConflictWithNameCS(ptConflictSet);
STATUS liberateConflictSetWithName(ptConflictSet);


typedef listcaen *ptConflictSetListcaen, conflictSetListcaen;

#define createConflictSetListcaen(ptCSL)             (createEmpty(ptCSL))
#define resetConflictSetListcaen(ptCSL)              (resetCurrent(ptCSL))
#define getConflictSetListcaenNumber(ptCSL)          (listcaenLength(*(ptCSL)))
#define getFirstConflictSetCSL(ptCSL)            ((ptConflictSet) lookLeft(*(ptCSL)))
#define getFirstConflictSetCSLC(ptCSL)           ((ptConflictSet) lookLeftC(*(ptCSL), sizeof(conflictSet)))
#define getNextConflictSetCSL(ptCSL)             ((ptConflictSet) lookAdvanceCurrent(ptCSL))
#define getNextConflictSetCSLC(ptCSL)            ((ptConflictSet) lookAdvanceCurrentC(ptCSL, sizeof(conflictSet)))
#define setFirstConflictSetCSL(ptCS,ptCSL)       (addLeft(ptCS, ptCSL))
#define setFirstConflictSetCSLC(ptCS,ptCSL)      (addLeftC(ptCS, ptCSL, sizeof(conflictSet)))
#define setNextConflictSetCSL(ptCS,ptCSL)        (addRight(ptCS, ptCSL))
#define setNextConflictSetCSLC(ptCS,ptCSL)       (addRightC(ptCS, ptCSL, sizeof(conflictSet)))
#define assignConflictSetListcaen(ptCSL,CSL)         (assign(ptCSL, CSL))
#define assignConflictSetListcaenC(ptCSL,CSL)        (assignC(ptCSL, CSL, sizeof(conflictSet)))
#define deleteFirstConflictSetCSL(ptCSL)         (deleteLeft(ptCSL))
#define deleteFirstConflictSetCSLC(ptCSL)        (deleteLeftC(ptCSL, sizeof(conflictSet)))
#define liberateConflictSetListcaen(ptCSL)           (liberate(ptCSL))
#define liberateConflictSetListcaenC(ptCSL)          (liberateC(ptCSL, sizeof(conflictSet)))


typedef struct MisbVar {
  variable var;
  conflictSet hardConflSet;
  conflictSetListcaen conflSetListcaen;
} *ptMisbVar, misbVar;

#define allocMisbVar()                                 ((ptMisbVar) cealloc (sizeof (MisbVar)))
#define getVarMisbVar(ptMV)                            ((ptMV)->var)
#define setVarMisbVar(ptMV,MV)                         (getVarMisbVar(ptMV) = (MV))
#define getHardConflictSetMisbVar(ptMV)                ((ptMV)->hardConflSet)
#define getHardConflictSetNumberMisbVar(ptMV)          (getConflictSetNumber(&getHardConflictSetMisbVar(ptMV)))
#define createHardConflictSetMisbVar(ptMV)             (createConflictSet(&getHardConflictSetMisbVar(ptMV)))
#define resetHardConflictSetMisbVar(ptMV)              (resetConflictSet(&getHardConflictSetMisbVar(ptMV)))
#define getFirstHardConflictMisbVar(ptMV)              (getFirstConflictCS(&getHardConflictSetMisbVar(ptMV)))
#define getFirstHardConflictMisbVarC(ptMV)             (getFirstConflictCSC(&getHardConflictSetMisbVar(ptMV)))
#define getNextHardConflictMisbVar(ptMV)               (getNextConflictCS(&getHardConflictSetMisbVar(ptMV)))
#define getNextHardConflictMisbVarC(ptMV)              (getNextConflictCSC(&getHardConflictSetMisbVar(ptMV)))
#define setFirstHardConflictMisbVar(ptHC,ptMV)         (setFirstConflictCS(ptHC, &getHardConflictSetMisbVar(ptMV)))
#define setFirstHardConflictMisbVarC(ptHC,ptMV)        (setFirstConflictCSC(ptHC, &getHardConflictSetMisbVar(ptMV)))
#define setFirstHardConflictWithNameMisbVar(ptHC,ptMV) (setFirstConflictWithNameCS(ptHC, &getHardConflictSetMisbVar(ptMV)))
#define setNextHardConflictMisbVar(ptHC,ptMV)          (setNextConflictCS(ptHC, &getHardConflictSetMisbVar(ptMV)))
#define setNextHardConflictMisbVarC(ptHC,ptMV)         (setNextConflictCSC(ptHC, &getHardConflictSetMisbVar(ptMV)))
#define setNextHardConflictWithNameMisbVar(ptHC,ptMV)  (setNextConflictWithNameCS(ptHC, &getHardConflictSetMisbVar(ptMV)))
#define assignHardConflictSetMisbVar(ptMV,HCS)         (assignConflictSet(&getHardConflictSetMisbVar(ptMV), HCS))
#define assignHardConflictSetMisbVarC(ptMV,HCS)        (assignConflictSetC(&getHardConflictSetMisbVar(ptMV), HCS))
#define assignHardConflictSetWithNameMisbVar(ptMV,HCS) (assignConflictSetWithName(&getHardConflictSetMisbVar(ptMV), HCS))
#define deleteFirstHardConflictMisbVar(ptMV)           (deleteFirstConflictCS(&getHardConflictSetMisbVar(ptMV)))
#define deleteFirstHardConflictMisbVarC(ptMV)          (deleteFirstConflictCSC(&getHardConflictSetMisbVar(ptMV)))
#define deleteFirstHardConflictWithNameMisbVar(ptMV)   (deleteFirstConflictWithNameCS(&getHardConflictSetMisbVar(ptMV)))
#define liberateHardConflictSetMisbVar(ptMV)           (liberateConflictSet(&getHardConflictSetMisbVar(ptMV)))
#define liberateHardConflictSetMisbVarC(ptMV)          (liberateConflictSetC(&getHardConflictSetMisbVar(ptMV)))
#define liberateHardConflictSetWithNameMisbVar(ptMV)   (liberateConflictSetWithName(&getHardConflictSetMisbVar(ptMV)))
#define getConflictSetListcaenMisbVar(ptMV)                ((ptMV)->conflSetListcaen)
#define getConflictSetListcaenNumberMisbVar(ptMV)          (getConflictSetListcaenNumber(&getConflictSetListcaenMisbVar(ptMV)))
#define createConflictSetListcaenMisbVar(ptMV)             (createConflictSetListcaen(&getConflictSetListcaenMisbVar(ptMV)))
#define resetConflictSetListcaenMisbVar(ptMV)              (resetConflictSetListcaen(&getConflictSetListcaenMisbVar(ptMV)))
#define getFirstConflictSetMisbVar(ptMV)               (getFirstConflictSetCSL(&getConflictSetListcaenMisbVar(ptMV)))
#define getFirstConflictSetMisbVarC(ptMV)              (getFirstConflictSetCSLC(&getConflictSetListcaenMisbVar(ptMV)))
#define getNextConflictSetMisbVar(ptMV)                (getNextConflictSetCSL(&getConflictSetListcaenMisbVar(ptMV)))
#define getNextConflictSetMisbVarC(ptMV)               (getNextConflictSetCSLC(&getConflictSetListcaenMisbVar(ptMV)))
#define setFirstConflictSetMisbVar(ptCS,ptMV)          (setFirstConflictSetCSL(ptCS, &getConflictSetListcaenMisbVar(ptMV)))
#define setFirstConflictSetMisbVarC(ptCS,ptMV)         (setFirstConflictSetCSLC(ptCS, &getConflictSetListcaenMisbVar(ptMV)))
#define setNextConflictSetMisbVar(ptCS,ptMV)           (setNextConflictSetCSL(ptCS, &getConflictSetListcaenMisbVar(ptMV)))
#define setNextConflictSetMisbVarC(ptCS,ptMV)          (setNextConflictSetCSLC(ptCS, &getConflictSetListcaenMisbVar(ptMV)))
#define assignConflictSetListcaenMisbVar(ptMV,CSL)         (assignConflictSetListcaen(&getConflictSetListcaenMisbVar(ptMV), CSL))
#define assignConflictSetListcaenMisbVarC(ptMV,CSL)        (assignConflictSetListcaenC(&getConflictSetListcaenMisbVar(ptMV), CSL))
#define deleteFirstConflictSetMisbVar(ptMV)            (deleteFirstConflictSetCSL(&getConflictSetListcaenMisbVar(ptMV)))
#define deleteFirstConflictSetMisbVarC(ptMV)           (deleteFirstConflictSetCSLC(&getConflictSetListcaenMisbVar(ptMV)))
#define liberateConflictSetListcaenMisbVar(ptMV)           (liberateConflictSetListcaen(&getConflictSetListcaenMisbVar(ptMV)))
#define liberateConflictSetListcaenMisbVarC(ptMV)          (liberateConflictSetListcaenC(&getConflictSetListcaenMisbVar(ptMV)))

typedef ptMisbVar *ptMisbVarArray, misbVarArray;

#define getMisbVarPt(rt,n)                       ((ptMisbVar)(getRTMisbVarArray(rt) + (n)))
#define getMisbVar(rt,n)                         (*(getMisbVarPt(rt, n)))
#define setMisbVar(MV,rt,n)                      (getMisbVar(rt, n) = (MV))

typedef listcaen *ptMisbVarListcaen, misbVarListcaen;

#define createMisbVarListcaen(ptMVL)                 (createEmpty(ptMVL))
#define resetMisbVarListcaen(ptMVL)                  (resetCurrent(ptMVL))
#define getMisbVarListcaenNumber(ptMVL)              (listcaenLength(*(ptMVL)))
#define getFirstMisbVarMVL(ptMVL)                ((ptMisbVar) lookLeft(*(ptMVL)))
#define getFirstMisbVarMVLC(ptMVL)               ((ptMisbVar) lookLeftC(*(ptMVL), sizeof(misbVar)))
#define getNextMisbVarMVL(ptMVL)                 ((ptMisbVar) lookAdvanceCurrent(ptMVL))
#define getNextMisbVarMVLC(ptMVL)                ((ptMisbVar) lookAdvanceCurrentC(ptMVL, sizeof(misbVar)))
#define setFirstMisbVarMVL(ptMV,ptMVL)           (addLeft(ptMV, ptMVL))
#define setFirstMisbVarMVLC(ptMV,ptMVL)          (addLeftC(ptMV, ptMVL, sizeof(misbVar)))
#define setNextMisbVarMVL(ptMV,ptMVL)            (addRight(ptMV, ptMVL))
#define setNextMisbVarMVLC(ptMV,ptMVL)           (addRightC(ptMV, ptMVL, sizeof(misbVar)))
#define assignMisbVarListcaen(ptMVL,MVL)             (assign(ptMVL, MVL))
#define assignMisbVarListcaenC(ptMVL,MVL)            (assignC(ptMVL, MVL, sizeof(misbVar)))
#define deleteFirstMisbVarMVL(ptMVL)             (deleteLeft(ptMVL))
#define deleteFirstMisbVarMVLC(ptMVL)            (deleteLeftC(ptMVL, sizeof(misbVar)))
#define liberateMisbVarListcaen(ptMVL)               (liberate(ptMVL))
#define liberateMisbVarListcaenC(ptMVL)              (liberateC(ptMVL, sizeof(misbVar)))


typedef conflict *ptAscendantInfluence, ascendantInfluence;

#define getAscInflName(ptAI)                     ((ptAI)->name)
#define getAscInflInfluence(ptAI)                ((ptAI)->infl)
#define getAscInflInstant(ptAI)                  ((ptAI)->instant)
#define getAscInflIndex(ptAI)                    ((ptAI)->index)
#define setAscInflName(ptAI,s)                   (getAscInflName(ptAI) = (s))
#define setAscInflInfluence(ptAI,ptI)            (getAscInflInfluence(ptAI) = (ptI))
#define setAscInflInstant(ptAI,i)                (getAscInflInstant(ptAI) = (i))
#define setAscInflIndex(ptAI,i)                  (getAscInflIndex(ptAI) = (i))


typedef listcaen *ptAscendantInfluenceSet, ascendantInfluenceSet;

#define createAscInflSet(ptAIS)                  (createEmpty(ptAIS))
#define resetAscInflSet(ptAIS)                   (resetCurrent(ptAIS))
#define getAscInflSetNumber(ptAIS)               (listcaenLength(*(ptAIS)))
#define getFirstAscInflAIS(ptAIS)                ((ptAscendantInfluence) lookLeft(*(ptAIS)))
#define getFirstAscInflAISC(ptAIS)               ((ptAscendantInfluence) lookLeftC(*(ptAIS), sizeof(ascendantInfluence)))
#define getNextAscInflAIS(ptAIS)                 ((ptAscendantInfluence) lookAdvanceCurrent(ptAIS))
#define getNextAscInflAISC(ptAIS)                ((ptAscendantInfluence) lookAdvanceCurrentC(ptAIS, sizeof(ascendantInfluence)))
#define setFirstAscInflAIS(ptAI,ptAIS)           (addLeft(ptAI, ptAIS))
#define setFirstAscInflAISC(ptAI,ptAIS)          (addLeftC(ptAI, ptAIS, sizeof(ascendantInfluence)))
#define setFirstAscInflWithNameAIS(ptAI,ptAIS)   (setFirstConflictWithNameCS(ptAI,ptAIS))
#define setNextAscInflAIS(ptAI,ptAIS)            (addRight(ptAI, ptAIS))
#define setNextAscInflAISC(ptAI,ptAIS)           (addRightC(ptAI, ptAIS, sizeof(ascendantInfluence)))
#define setNextAscInflWithNameAIS(ptAI,ptAIS)    (setNextConflictWithNameCS(ptAI,ptAIS))
#define assignAscInflSet(ptAIS,AIS)              (assign(ptAIS, AIS))
#define assignAscInflSetC(ptAIS,AIS)             (assignC(ptAIS, AIS, sizeof(ascendantInfluence)))
#define assignAscInflSetWithName(ptAIS,AIS)      (assignConflictSetWithName(ptAIS, AIS))
#define deleteFirstAscInflAIS(ptAIS)             (deleteLeft(ptAIS))
#define deleteFirstAscInflAISC(ptAIS)            (deleteLeftC(ptAIS, sizeof(ascendantInfluence)))
#define deleteFirstAscInflWithNameAIS(ptAIS)     (deleteFirstConflictWithNameCS(ptAIS))
#define liberateAscInflSet(ptAIS)                (liberate(ptAIS))
#define liberateAscInflSetC(ptAIS)               (liberateC(ptAIS, sizeof(ascendantInfluence)))
#define liberateAscInflSetWithName(ptAIS)        (liberateConflictSetWithName(ptAIS))


typedef struct OkVar {
  variable var;
  ascendantInfluenceSet ascInflSet;
} *ptOkVar, okVar;

#define getVarOkVar(ptOV)                        ((ptOV)->var)
#define setVarOkVar(ptOV,OV)                     (getVarOkVar(ptOV) = (OV))
#define getAscInflSetOkVar(ptOV)                 ((ptOV)->ascInflSet)
#define getAscInflSetNumberOkVar(ptOV)           (getAscInflSetNumber(&getAscInflSetOkVar(ptOV)))
#define createAscInflSetOkVar(ptOV)              (createAscInflSet(&getAscInflSetOkVar(ptOV)))
#define resetAscInflSetOkVar(ptOV)               (resetAscInflSet(&getAscInflSetOkVar(ptOV)))
#define getFirstAscInflOkVar(ptOV)               (getFirstAscInflAIS(&getAscInflSetOkVar(ptOV)))
#define getFirstAscInflOkVarC(ptOV)              (getFirstAscInflAISC(&getAscInflSetOkVar(ptOV)))
#define getNextAscInflOkVar(ptOV)                (getNextAscInflAIS(&getAscInflSetOkVar(ptOV)))
#define getNextAscInflOkVarC(ptOV)               (getNextAscInflAISC(&getAscInflSetOkVar(ptOV)))
#define setFirstAscInflOkVar(ptAI,ptOV)          (setFirstAscInflAIS(ptAI, &getAscInflSetOkVar(ptOV)))
#define setFirstAscInflOkVarC(ptAI,ptOV)         (setFirstAscInflAISC(ptAI, &getAscInflSetOkVar(ptOV)))
#define setFirstAscInflWithNameOkVar(ptAI,ptOV)  (setFirstAscInflWithNameAIS(ptAI, &getAscInflSetOkVar(ptOV)))
#define setNextAscInflOkVar(ptAI,ptOV)           (setNextAscInflAIS(ptAI, &getAscInflSetOkVar(ptOV)))
#define setNextAscInflOkVarC(ptAI,ptOV)          (setNextAscInflAISC(ptAI, &getAscInflSetOkVar(ptOV)))
#define setNextAscInflWithNameOkVar(ptAI,ptOV)   (setNextAscInflWithNameAIS(ptAI, &getAscInflSetOkVar(ptOV)))
#define assignAscInflSetOkVar(ptOV,AIS)          (assignAscInflSet(&getAscInflSetOkVar(ptOV), AIS))
#define assignAscInflSetOkVarC(ptOV,AIS)         (assignAscInflSetC(&getAscInflSetOkVar(ptOV), AIS))
#define assignAscInflSetWithNameOkVar(ptOV,AIS)  (assignAscInflSetWithName(&getAscInflSetOkVar(ptOV), AIS))
#define deleteFirstAscInflOkVar(ptOV)            (deleteFirstAscInflAIS(&getAscInflSetOkVar(ptOV)))
#define deleteFirstAscInflOkVarC(ptOV)           (deleteFirstAscInflAISC(&getAscInflSetOkVar(ptOV)))
#define deleteFirstAscInflWithNameOkVar(ptOV)    (deleteFirstAscInflWithNameAIS(&getAscInflSetOkVar(ptOV)))
#define liberateAscInflSetOkVar(ptOV)            (liberateAscInflSet(&getAscInflSetOkVar(ptOV)))
#define liberateAscInflSetOkVarC(ptOV)           (liberateAscInflSetC(&getAscInflSetOkVar(ptOV)))
#define liberateAscInflSetWithNameOkVar(ptOV)    (liberateAscInflSetWithName(&getAscInflSetOkVar(ptOV)))


typedef ptOkVar *ptOkVarArray, okVarArray;

#define getOkVarPt(rt,n)                         ((ptOkVar)(getRTOkVarArray(rt) + (n)))
#define getOkVar(rt,n)                           (*(getOkVarPt(rt, n)))
#define setOkVar(OV,rt,n)                        (getOkVar(rt, n) = (OV))


typedef listcaen *ptOkVarListcaen, okVarListcaen;

#define createOkVarListcaen(ptOVL)                   (createEmpty(ptOVL))
#define resetOkVarListcaen(ptOVL)                    (resetCurrent(ptOVL))
#define getOkVarListcaenNumber(ptOVL)                (listcaenLength(*(ptOVL)))
#define getFirstOkVarOVL(ptOVL)                  ((ptOkVar) lookLeft(*(ptOVL)))
#define getFirstOkVarOVLC(ptOVL)                 ((ptOkVar) lookLeftC(*(ptOVL), sizeof(okVar)))
#define getNextOkVarOVL(ptOVL)                   ((ptOkVar) lookAdvanceCurrent(ptOVL))
#define getNextOkVarOVLC(ptOVL)                  ((ptOkVar) lookAdvanceCurrentC(ptOVL, sizeof(okVar)))
#define setFirstOkVarOVL(ptOV,ptOVL)             (addLeft(ptOV, ptOVL))
#define setFirstOkVarOVLC(ptOV,ptOVL)            (addLeftC(ptOV, ptOVL, sizeof(okVar)))
#define setNextOkVarOVL(ptOV,ptOVL)              (addRight(ptOV, ptOVL))
#define setNextOkVarOVLC(ptOV,ptOVL)             (addRightC(ptOV, ptOVL, sizeof(okVar)))
#define assignOkVarListcaen(ptOVL,OVL)               (assign(ptOVL, OVL))
#define assignOkVarListcaenC(ptOVL,OVL)              (assignC(ptOVL, OVL, sizeof(okVar)))
#define deleteFirstOkVarOVL(ptOVL)               (deleteLeft(ptOVL))
#define deleteFirstOkVarOVLC(ptOVL)              (deleteLeftC(ptOVL, sizeof(okVar)))
#define liberateOkVarListcaen(ptOVL)                 (liberate(ptOVL))
#define liberateOkVarListcaenC(ptOVL)                (liberateC(ptOVL, sizeof(okVar)))


typedef ascendantInfluenceSet *ptSynonym, synonym;

#define createSynonym(ptS)                       (createAscInflSet(ptS))
#define resetSynonym(ptS)                        (resetAscInflSet(ptS))
#define getSynonymNumber(ptS)                    (getAscInflSetNumber(ptS))
#define getFirstAscInflSynonym(ptS)              (getFirstAscInflAIS(ptS))
#define getFirstAscInflSynonymC(ptS)             (getFirstAscInflAISC(ptS))
#define getNextAscInflSynonym(ptS)               (getNextAscInflAIS(ptS))
#define getNextAscInflSynonymC(ptS)              (getNextAscInflAISC(ptS))
#define setFirstAscInflSynonym(ptAI,ptS)         (setFirstAscInflAIS(ptAI, ptS))
#define setFirstAscInflSynonymC(ptAI,ptS)        (setFirstAscInflAISC(ptAI, ptS))
#define setFirstAscInflWithNameSynonym(ptAI,ptS) (setFirstAscInflWithNameAIS(ptAI, ptS))
#define setNextAscInflSynonym(ptAI,ptS)          (setNextAscInflAIS(ptAI, ptS))
#define setNextAscInflSynonymC(ptAI,ptS)         (setNextAscInflAISC(ptAI, ptS))
#define setNextAscInflWithNameSynonym(ptAI,ptS)  (setNextAscInflWithNameAIS(ptAI, ptS))
#define assignSynonym(ptS,S)                     (assignAscInflSet(ptS, S))
#define assignSynonymC(ptS,S)                    (assignAscInflSetC(ptS, S))
#define assignSynonymWithName(ptS,S)             (assignAscInflSetWithName(ptS, S))
#define deleteFirstAscInflSynonym(ptS)           (deleteFirstAscInflAIS(ptS))
#define deleteFirstAscInflSynonymC(ptS)          (deleteFirstAscInflAISC(ptS))
#define deleteFirstAscInflWithNameSynonym(ptS)   (deleteFirstAscInflWithNameAIS(ptS))
#define liberateSynonym(ptS)                     (liberateAscInflSet(ptS))
#define liberateSynonymC(ptS)                    (liberateAscInflSetC(ptS))
#define liberateSynonymWithName(ptS)             (liberateConflictSetWithName(ptS))


typedef listcaen *ptSynonymListcaen, synonymListcaen;

#define createSynonymListcaen(ptSL)                  (createEmpty(ptSL))
#define resetSynonymListcaen(ptSL)                   (resetCurrent(ptSL))
#define getSynonymListcaenNumber(ptSL)               (listcaenLength(*(ptSL)))
#define getFirstSynonymSL(ptSL)                  ((ptSynonym) lookLeft(*(ptSL)))
#define getFirstSynonymSLC(ptSL)                 ((ptSynonym) lookLeftC(*(ptSL), sizeof(synonym)))
#define getNextSynonymSL(ptSL)                   ((ptSynonym) lookAdvanceCurrent(ptSL))
#define getNextSynonymSLC(ptSL)                  ((ptSynonym) lookAdvanceCurrentC(ptSL, sizeof(synonym)))
#define setFirstSynonymSL(ptS,ptSL)              (addLeft(ptS, ptSL))
#define setFirstSynonymSLC(ptS,ptSL)             (addLeftC(ptS, ptSL, sizeof(synonym)))
#define setNextSynonymSL(ptS,ptSL)               (addRight(ptS, ptSL))
#define setNextSynonymSLC(ptS,ptSL)              (addRightC(ptS, ptSL, sizeof(synonym)))
#define assignSynonymListcaen(ptSL,SL)               (assign(ptSL, SL))
#define assignSynonymListcaenC(ptSL,SL)              (assignC(ptSL, SL, sizeof(synonym)))
#define deleteFirstSynonymSL(ptSL)               (deleteLeft(ptSL))
#define deleteFirstSynonymSLC(ptSL)              (deleteLeftC(ptSL, sizeof(synonym)))
#define liberateSynonymListcaen(ptSL)                (liberate(ptSL))
#define liberateSynonymListcaenC(ptSL)               (liberateC(ptSL, sizeof(synonym)))


typedef struct SoftVar {
  boolean misbehaving;
  union {
    ptMisbVar mbVar;
    ptOkVar okVar;
    } var;
  int *numOkVarClassArray;
  struct SoftVar ***okVarClassMatrix;
  synonymListcaen *synonymListcaenArray;
  int *numElemUnionSetArray;
  struct SoftVar ***unionSetMatrix;
} *ptSoftVar, softVar;

#define getMisbehavingSoftVar(ptSV)              ((ptSV)->misbehaving)
#define setMisbehavingSoftVar(ptSV,b)            (getMisbehavingSoftVar(ptSV) = (b))
#define getMisbVarSoftVar(ptSV)                  ((ptSV)->var.mbVar)
#define setMisbVarSoftVar(ptSV,ptMV)             (getMisbVarSoftVar(ptSV) = (ptMV))
#define getOkVarSoftVar(ptSV)                    ((ptSV)->var.okVar)
#define setOkVarSoftVar(ptSV,ptOV)               (getOkVarSoftVar(ptSV) = (ptOV))
#define getNumOkVarClassArraySoftVar(ptSV)       ((ptSV)->numOkVarClassArray)
#define setNumOkVarClassArraySoftVar(ptSV,a)     (getNumOkVarClassArraySoftVar(ptSV) = (a))
#define getNumOkVarClassSoftVar(ptSV,m)          (*(getNumOkVarClassArraySoftVar(ptSV) + (m)))
#define setNumOkVarClassSoftVar(ptSV,m,n)        (getNumOkVarClassSoftVar(ptSV, m) = (n))
#define getOkVarClassMatrixSoftVar(ptSV)         ((ptSV)->okVarClassMatrix)
#define setOkVarClassMatrixSoftVar(ptSV,m)       (getOkVarClassMatrixSoftVar(ptSV) = (m))
#define getOkVarClassSoftVar(ptSV,m,ncol,n)      (*(((ptSoftVar *)(getOkVarClassMatrixSoftVar(ptSV) + ((m) * (ncol)))) + (n)))
#define setOkVarClassSoftVar(ptSV,m,ncol,n,v)    (getOkVarClassSoftVar(ptSV, m, ncol, n) = (v))
#define getSynonymListcaenArraySoftVar(ptSV)         ((ptSV)->synonymListcaenArray)
#define setSynonymListcaenArraySoftVar(ptSV,a)       (getSynonymListcaenArraySoftVar(ptSV) = (a))
#define getSynonymListcaenPtSoftVar(ptSV,m)          (getSynonymListcaenArraySoftVar(ptSV) + (m))
#define getSynonymListcaenSoftVar(ptSV,m)            (*(getSynonymListcaenPtSoftVar(ptSV, m)))
#define setSynonymListcaenSoftVar(ptSV,m,s)          (getSynonymListcaenSoftVar(ptSV,n) = (s))
#define getNumSynonymSoftVar(ptSV,m)             (getSynonymListcaenNumber(getSynonymListcaenPtSoftVar(ptSV,m)))
#define getNumElemUnionSetArraySoftVar(ptSV)     ((ptSV)->numElemUnionSetArray)
#define setNumElemUnionSetArraySoftVar(ptSV,a)   (getNumElemUnionSetArraySoftVar(ptSV) = (a))
#define getNumElemUnionSetSoftVar(ptSV,m)        (*(getNumElemUnionSetArraySoftVar(ptSV) + (m)))
#define setNumElemUnionSetSoftVar(ptSV,m,n)      (getNumElemUnionSetSoftVar(ptSV, m) = (n))
#define getUnionSetMatrixSoftVar(ptSV)           ((ptSV)->unionSetMatrix)
#define setUnionSetMatrixSoftVar(ptSV,m)         (getUnionSetMatrixSoftVar(ptSV) = (m))
#define getUnionSetSoftVar(ptSV,m,ncol)          ((ptSoftVar *)(getUnionSetMatrixSoftVar(ptSV) + ((m) * (ncol))))
#define setUnionSetSoftVar(ptSV,m,ncol,u)        (getUnionSetSoftVar(ptSV, m, ncol) = (u))
#define getUnionSetElemSoftVar(ptSV,m,ncol,n)    (*(((ptSoftVar *)(getUnionSetMatrixSoftVar(ptSV) + ((m) * (ncol)))) + (n)))
#define setUnionSetElemSoftVar(ptSV,m,ncol,n,v)  (getUnionSetElemSoftVar(ptSV, m, ncol, n) = (v))


typedef ptSoftVar *ptSoftVarArray, softVarArray;

#define getSoftVarPt(rt,n)                       ((ptSoftVar)(getRTSoftVarArray(rt) + (n)))
#define getSoftVar(rt,n)                         (*(getSoftVarPt(rt, n)))
#define setSoftVar(SV,rt,n)                      (getSoftVar(rt, n) = (SV))


typedef listcaen *ptSoftVarListcaen, softVarListcaen;

#define createSoftVarListcaen(ptSVL)                 (createEmpty(ptSVL))
#define resetSoftVarListcaen(ptSVL)                  (resetCurrent(ptSVL))
#define getSoftVarListcaenNumber(ptSVL)              (listcaenLength(*(ptSVL)))
#define getFirstSoftVarSVL(ptSVL)                ((ptSoftVar) lookLeft(*(ptSVL)))
#define getFirstSoftVarSVLC(ptSVL)               ((ptSoftVar) lookLeftC(*(ptSVL), sizeof(softVar)))
#define getNextSoftVarSVL(ptSVL)                 ((ptSoftVar) lookAdvanceCurrent(ptSVL))
#define getNextSoftVarSVLC(ptSVL)                ((ptSoftVar) lookAdvanceCurrentC(ptSVL, sizeof(softVar)))
#define setFirstSoftVarSVL(ptSV,ptSVL)           (addLeft(ptSV, ptSVL))
#define setFirstSoftVarSVLC(ptSV,ptSVL)          (addLeftC(ptSV, ptSVL, sizeof(softVar)))
#define setNextSoftVarSVL(ptSV,ptSVL)            (addRight(ptSV, ptSVL))
#define setNextSoftVarSVLC(ptSV,ptSVL)           (addRightC(ptSV, ptSVL, sizeof(softVar)))
#define assignSoftVarListcaen(ptSVL,SVL)             (assign(ptSVL, SVL))
#define assignSoftVarListcaenC(ptSVL,SVL)            (assignC(ptSVL, SVL, sizeof(softVar)))
#define deleteFirstSoftVarSVL(ptSVL)             (deleteLeft(ptSVL))
#define deleteFirstSoftVarSVLC(ptSVL)            (deleteLeftC(ptSVL, sizeof(softVar)))
#define liberateSoftVarListcaen(ptSVL)               (liberate(ptSVL))
#define liberateSoftVarListcaenC(ptSVL)              (liberateC(ptSVL, sizeof(softVar)))


typedef ascendantInfluenceSet *ptEdgeSet, edgeSet;

#define createEdgeSet(ptES)                      (createAscInflSet(ptES))
#define resetEdgeSet(ptES)                       (resetAscInflSet(ptES))
#define getEdgeSetNumber(ptES)                   (getAscInflSetNumber(ptES))
#define getFirstEdgeES(ptES)                     (getFirstAscInflAIS(ptES))
#define getFirstEdgeESC(ptES)                    (getFirstAscInflAISC(ptES))
#define getNextEdgeES(ptES)                      (getNextAscInflAIS(ptES))
#define getNextEdgeESC(ptES)                     (getNextAscInflAISC(ptES))
#define setFirstEdgeES(ptE,ptES)                 (setFirstAscInflAIS(ptE, ptES))
#define setFirstEdgeESC(ptE,ptES)                (setFirstAscInflAISC(ptE, ptES))
#define setFirstEdgeWithNameES(ptE,ptES)         (setFirstAscInflWithNameAIS(ptE, ptES))
#define setNextEdgeES(ptE,ptES)                  (setNextAscInflAIS(ptE, ptES))
#define setNextEdgeESC(ptE,ptES)                 (setNextAscInflAISC(ptE, ptES))
#define setNextEdgeWithNameES(ptE,ptES)          (setNextAscInflWithNameAIS(ptE, ptES))
#define assignEdgeSet(ptES,ES)                   (assignAscInflSet(ptES, ES))
#define assignEdgeSetC(ptES,ES)                  (assignAscInflSetC(ptES, ES))
#define assignEdgeSetWithName(ptES,ES)           (assignAscInflSetWithName(ptES, ES))
#define deleteFirstEdgeES(ptES)                  (deleteFirstAscInflAIS(ptES))
#define deleteFirstEdgeSetESC(ptES)              (deleteFirstAscInflAISC(ptES))
#define deleteFirstEdgeWithNameES(ptES)          (deleteFirstAscInflWithNameAIS(ptES))
#define liberateEdgeSet(ptES)                    (liberateAscInflSet(ptES))
#define liberateEdgeSetC(ptES)                   (liberateAscInflSetC(ptES))
#define liberateEdgeSetWithName(ptES)            (liberateAscInflSetWithName(ptES))


typedef struct Leaf {
  int nodeNumber;
  edgeSet edges;
} *ptLeaf, leaf;

#define getNodeNumberLeaf(ptL)                   ((ptL)->nodeNumber)
#define setNodeNumberLeaf(ptL,n)                 (getNodeNumberLeaf(ptL) = (n))
#define getEdgeSetLeaf(ptL)                      ((ptL)->edges)
#define getEdgeSetNumberLeaf(ptL)                (getEdgeSetNumber(&getEdgeSetLeaf(ptL)))
#define createEdgeSetLeaf(ptL)                   (createEdgeSet(&getEdgeSetLeaf(ptL)))
#define resetEdgeSetLeaf(ptL)                    (resetEdgeSet(&getEdgeSetLeaf(ptL)))
#define getFirstEdgeLeaf(ptL)                    (getFirstEdgeES(&getEdgeSetLeaf(ptL)))
#define getFirstEdgeLeafC(ptL)                   (getFirstEdgeESC(&getEdgeSetLeaf(ptL)))
#define getNextEdgeLeaf(ptL)                     (getNextEdgeES(&getEdgeSetLeaf(ptL)))
#define getNextEdgeLeafC(ptL)                    (getNextEdgeESC(&getEdgeSetLeaf(ptL)))
#define setFirstEdgeLeaf(ptE,ptL)                (setFirstEdgeES(ptE, &getEdgeSetLeaf(ptL)))
#define setFirstEdgeLeafC(ptE,ptL)               (setFirstEdgeESC(ptE, &getEdgeSetLeaf(ptL)))
#define setFirstEdgeWithNameLeaf(ptE,ptL)        (setFirstEdgeWithNameES(ptE, &getEdgeSetLeaf(ptL)))
#define setNextEdgeLeaf(ptE,ptL)                 (setNextEdgeES(ptE, &getEdgeSetLeaf(ptL)))
#define setNextEdgeLeafC(ptE,ptL)                (setNextEdgeESC(ptE, &getEdgeSetLeaf(ptL)))
#define setNextEdgeWithNameLeaf(ptE,ptL)         (setNextEdgeWithNameES(ptE, &getEdgeSetLeaf(ptL)))
#define assignEdgeSetLeaf(ptL,L)                 (assignEdgeSet(&getEdgeSetLeaf(ptL), getEdgeSetLeaf(&L)))
#define assignEdgeSetLeafC(ptL,L)                (assignEdgeSetC(&getEdgeSetLeaf(ptL), getEdgeSetLeaf(&L)))
#define assignEdgeSetWithNameLeaf(ptL,L)         (assignEdgeSetWithName(&getEdgeSetLeaf(ptL), getEdgeSetLeaf(&L)))
#define deleteFirstEdgeLeaf(ptL)                 (deleteFirstEdgeES(&getEdgeSetLeaf(ptL)))
#define deleteFirstEdgeLeafC(ptL)                (deleteFirstEdgeESC(&getEdgeSetLeaf(ptL)))
#define deleteFirstEdgeWithNameLeaf(ptL)         (deleteFirstEdgeWithNameES(&getEdgeSetLeaf(ptL)))
#define liberateEdgeSetLeaf(ptL)                 (liberateEdgeSet(&getEdgeSetLeaf(ptL)))
#define liberateEdgeSetLeafC(ptL)                (liberateEdgeSetC(&getEdgeSetLeaf(ptL)))
#define liberateEdgeSetWithNameLeaf(ptL)         (liberateEdgeSetWithName(&getEdgeSetLeaf(ptL)))


typedef listcaen *ptLeafListcaen, leafListcaen;

#define createLeafListcaen(ptLL)                     (createEmpty(ptLL))
#define resetLeafListcaen(ptLL)                      (resetCurrent(ptLL))
#define getLeafListcaenNumber(ptLL)                  (listcaenLength(*(ptLL)))
#define getFirstLeafLL(ptLL)                     ((ptLeaf) lookLeft(*(ptLL)))
#define getFirstLeafLLC(ptLL)                    ((ptLeaf) lookLeftC(*(ptLL), sizeof(leaf)))
#define getNextLeafLL(ptLL)                      ((ptLeaf) lookAdvanceCurrent(ptLL))
#define getNextLeafLLC(ptLL)                     ((ptLeaf) lookAdvanceCurrentC(ptLL, sizeof(leaf)))
#define setFirstLeafLL(ptL,ptLL)                 (addLeft(ptL, ptLL))
#define setFirstLeafLLC(ptL,ptLL)                (addLeftC(ptL, ptLL, sizeof(leaf)))
#define setNextLeafLL(ptL,ptLL)                  (addRight(ptL, ptLL))
#define setNextLeafLLC(ptL,ptLL)                 (addRightC(ptL, ptLL, sizeof(leaf)))
#define assignLeafListcaen(ptLL,LL)                  (assign(ptLL, LL))
#define assignLeafListcaenC(ptLL,LL)                 (assignC(ptLL, LL, sizeof(leaf)))
#define deleteFirstLeafLL(ptLL)                  (deleteLeft(ptLL))
#define deleteFirstLeafLLC(ptLL)                 (deleteLeftC(ptLL, sizeof(leaf)))
#define liberateLeafListcaen(ptLL)                   (liberate(ptLL))
#define liberateLeafListcaenC(ptLL)                  (liberateC(ptLL, sizeof(leaf)))


typedef listcaen **ptLeafListcaenArray, *leafListcaenArray;

#define getLeafListcaenPtLLA(LLA,n)                  ((LLA) + (n))
#define createLeafListcaenLLA(LLA,n)                 (createLeafListcaen(getLeafListcaenPtLLA(LLA, n)))
#define resetLeafListcaenLLA(LLA,n)                  (resetLeafListcaen(getLeafListcaenPtLLA(LLA, n)))
#define getLeafListcaenNumberLLA(LLA,n)              (getLeafListcaenNumber(getLeafListcaenPtLLA(LLA, n)))
#define getLeafListcaenLLA(LLA,n)                    (*(getLeafListcaenPtLLA(LLA, n)))
#define getFirstLeafLLA(LLA,n)                   (getFirstLeafLL(getLeafListcaenPtLLA(LLA, n)))
#define getFirstLeafLLAC(LLA,n)                  (getFirstLeafLLC(getLeafListcaenPtLLA(LLA, n)))
#define getNextLeafLLA(LLA,n)                    (getNextLeafLL(getLeafListcaenPtLLA(LLA, n)))
#define getNextLeafLLAC(LLA,n)                   (getNextLeafLLC(getLeafListcaenPtLLA(LLA, n)))
#define setFirstLeafLLA(ptL,LLA,n)               (setFirstLeafLL(ptL, getLeafListcaenPtLLA(LLA, n)))
#define setFirstLeafLLAC(ptL,LLA,n)              (setFirstLeafLLC(ptL, getLeafListcaenPtLLA(LLA, n)))
#define setNextLeafLLA(ptL,LLA,n)                (setNextLeafLL(ptL, getLeafListcaenPtLLA(LLA, n)))
#define setNextLeafLLAC(ptL,LLA,n)               (setNextLeafLLC(ptL, getLeafListcaenPtLLA(LLA, n)))
#define assignLeafListcaenLLA(LLA,LL,n)              (assignLeafListcaen(getLeafListcaenPtLLA(LLA, n), LL))
#define assignLeafListcaenLLAC(LLA,LL,n)             (assignLeafListcaenC(getLeafListcaenPtLLA(LLA, n), LL))
#define deleteFirstLeafLLA(LLA,n)                (deleteFirstLeafLL(getLeafListcaenPtLLA(LLA, n)))
#define deleteFirstLeafLLAC(LLA,n)               (deleteFirstLeafLLC(getLeafListcaenPtLLA(LLA, n)))
#define liberateLeafListcaenLLA(LLA,n)               (liberateLeafListcaen(getLeafListcaenPtLLA(LLA, n)))
#define liberateLeafListcaenLLAC(LLA,n)              (liberateLeafListcaenC(getLeafListcaenPtLLA(LLA, n)))


typedef struct PreferenceClass {
  ptLeaf lf;
  int numSoftConflicts;
} *ptPreferenceClass, preferenceClass;

#define getPrefClassLeaf(ptPC)                   ((ptPC)->lf)
#define getNumSoftConflicts(ptPC)                ((ptPC)->numSoftConflicts)
#define setPrefClassLeaf(ptPC,ptL)               (getPrefClassLeaf(ptPC) = (ptL))
#define setNumSoftConflicts(ptPC,n)              (getNumSoftConflicts(ptPC) = (n))

/*******************/
/* Function macros */
/*******************/

#define findAscendantInfluences(rt,ptVarOk,path) \
     findConflictsAux(rt, getVarOkVar(ptVarOk), &getAscInflSetOkVar(ptVarOk),\
     path, 0, CE_TRUE)
#define findHardConflicts(rt,ptVarMisb,path) \
     findConflictsAux(rt, getVarMisbVar(ptVarMisb),\
     &getHardConflictSetMisbVar(ptVarMisb), path, 0, CE_TRUE)

#define compareAscendantInfluences               compareConflicts

#define sortPreferenceClasses(pfV,r)             quickSort(pfV, 0, r - 1, comparePreferenceClasses)


/**
 * \struct s_runtime
 * \brief the "runtime" type, along with its operators.
 */
typedef struct {
   const char *appli;            /**< application name */
   long num;                     /**< id number of the runtime trajectory (chiefly for display purposes) */
   double ts;                    /**< continuous time step */
   double dts;                   /**< discrete time step */
   int twc;                      /**< temporal continuous window */
   long nvar;                    /**< length of "dico" vector */
   variable *dico;               /**< vector of all variables */
   long nin;                     /**< length of "input" vector */
   variable *in;                 /**< vector of input variables */
   double *inval;
   long ndisp;                   /**< length of "display" vector  */
   variable *disp;               /**< vector of display variables  */
   double *dispinfval;
   double *dispsupval;
   long ndispall;
   double *dispall;              /**< vector of all displays (mostly for debug) */
   long nmes;                    /**< length of "measure" vector */
   variable *mes;                /**< vector of measure variables */
   double *mesval;
   long *lowalarm;
   long *highalarm;
   long *lowmisb;
   long *highmisb;
   long *misb;
   long ntemp;                   /**< length of "temporal" vector  */
   variable *temp;               /**< vector of "temporal" variables  */
   long nuntemp;                 /**< length of "untemporal" vector  */
   variable *untemp;             /**< vector of "untemporal" variables  */
   long nifl;
   influence *infl;
   long *condin;                 /**< set of true influences conditions */
   long *condout;                /**< set of false influences conditions */
   long ncg;
   cGraph *cg;
   long maxDepth;                /**< maximum depth in causal graph */
   long nit;
   initVar *iv;
   long ngv;
   gVar *vgv;
   long ngc;
   expr *glc;                    /**< array of global constraints */
   long *inag;                   /**< agenda for filtering procedures */
   long inited;                  /**< initializations done */
   long runGC;                   /**< global constraints evaluation required */
   long cumTime;
   double prec;                  /**< precision */
   long onlyFaultIso;            /**< Flag that indicates whether only fault isolation (and no simulation) is being carried out */
   int numMisb;                  /**< Number of misbehaving variables */
   int numOk;                    /**< Number of non-misbehaving variables */
   misbVarArray mbVA;            /**< Misbehaving variable array */
   okVarArray okVA;              /**< Ok variable array */
   softVarArray sfVA;            /**< Soft variable array */
   int oldNumMisb;               /**< oldMisbVarArray's cardinality */
   variable *oldMisbVarArray;    /**< Previous iteration's misbehaving variable set */
   int numNodesHST;              /**< Number of nodes created on the HSTree */
   leafListcaen HSTree;              /**< Listcaen of diagnosis (leaves from the HSTree) */
   conflictSet softConflictListcaen;    /**< Set of all soft-conflicts (used in diagnosis type 3) */
   char *diagOut;                       /**< Buffer where to print fault isolation results */
   long ncondi;
   condition* condi;                                 /**< Array of runtime conditions */
   multimap<bool, Atom*, eqbool>* bridgeEvents;      /**< changes on atoms */
   multimap<bool, Atom*, eqbool>* bridgeConds;       /**< changes on infs conditions */
  vector<Transition*>* bridgeTransitions;           /**< trigerred transitions */
   vector<vector<Atom*>* >* bridgeConflicts;             /**< current conflicts at continuous level */
  int ncdc;                                          /**< number of conditions constraints */
  expr *cdc;                                         /**< conditions constraints */
  int nmfc;
  expr *mfc;                                         /**< mapping functions constraints */
  long *valSpc;                                       /**< validity space conditions */
  long *inac;                                         /**< agenda for condC filtering */
 } s_runtime, *runtime;

#define getRTAppliName(r)             (r->appli)
#define getRTTimeStep(r)              (r->ts)
#define getRTDTimeStep(r)             (r->dts)
#define setRTDTimeStep(r,t)           (r->dts = t)
#define getRTNVar(r)                  (r->nvar)
#define getRTDicoNth(r,i)             (*(r->dico+(i)))
#define getRTNInput(r)                (r->nin)
#define getRTInputNth(r,i)            (*(r->in+(i)))
#define getRTInVal(r)                 (r->inval)
#define setRTInVal(r,d)               (r->inval = d)
#define getRTInValNth(r,i)            (*(r->inval+(i)))
#define getRTNDisplay(r)              (r->ndisp)
#define getRTDisplayNth(r,i)          (*(r->disp+(i)))
#define getRTDispInfVal(r)            (r->dispinfval)
#define getRTDispInfValNth(r,i)       (*(r->dispinfval+(i)))
#define setRTDispInfValNth(r,i,d)     (*(r->dispinfval+(i)) = d)
#define getRTDispSupVal(r)            (r->dispsupval)
#define getRTDispSupValNth(r,i)       (*(r->dispsupval+(i)))
#define setRTDispSupValNth(r,i,d)     (*(r->dispsupval+(i)) = d)
#define getRTNDispAll(r)              (r->ndispall)
#define getRTDispAllNth(r,i)          (*(r->dispall+(i)))
#define setRTDispAllNth(r,i,d)        (*(r->dispall+(i)) = d)
#define getRTNMeasure(r)              (r->nmes)
#define getRTMeasureNth(r,i)          (*(r->mes+(i)))
#define getRTMesVal(r)                (r->mesval)
#define setRTMesVal(r,d)              (r->mesval = d)
#define getRTMesValNth(r,i)           (*(r->mesval+(i)))
#define getRTMesLowAlarm(r)           (r->lowalarm)
#define getRTMesHighAlarm(r)          (r->highalarm)
#define getRTMesHighAlarmNth(r,i)     (*(r->highalarm+(i)))
#define getRTMesLowAlarmNth(r,i)      (*(r->lowalarm+(i)))
#define setRTMesLowAlarmNth(r,i,b)    (*(r->lowalarm+(i)) = b)
#define setRTMesHighAlarmNth(r,i,b)   (*(r->highalarm+(i)) = b)
#define getRTMesLowMisb(r)            (r->lowmisb)
#define getRTMesHighMisb(r)           (r->highmisb)
#define getRTMesMisb(r)               (r->misb)
#define getRTMesLowMisbNth(r,i)       (*(r->lowmisb+(i)))
#define getRTMesHighMisbNth(r,i)      (*(r->highmisb+(i)))
#define getRTMesMisbNth(r,i)          (*(r->misb+(i)))
#define setRTMesLowMisbNth(r,i,b)     (*(r->lowmisb+(i)) = b)
#define setRTMesHighMisbNth(r,i,b)    (*(r->highmisb+(i)) = b)
#define setRTMesMisbNth(r,i,b)        (*(r->misb+(i)) = b)
#define getRTNTempo(r)                (r->ntemp)
#define getRTTempoNth(r,i)            (*(r->temp+(i)))
#define getRTNUntemp(r)               (r->nuntemp)
#define getRTUntempNth(r,i)           (*(r->untemp+(i)))
#define getRTNInflu(r)                (r->nifl)
#define getRTInfluNth(r,i)            (*(r->infl+(i)))
#define getRTCondIn(r)                (r->condin)
#define getRTCondInNth(r,i)           (*(r->condin+(i)))
#define setRTCondInNth(r,i,b)         (*(r->condin+(i)) = b)
#define getRTCondOut(r)               (r->condout)
#define getRTCondOutNth(r,i)          (*(r->condout+(i)))
#define setRTCondOutNth(r,i,b)        (*(r->condout+(i)) = b)
#define getRTNCGraph(r)               (r->ncg)
#define getRTCGraphNth(r,i)           (*(r->cg+(i)))
#define getRTMaxDepth(r)              (r->maxDepth)
#define getRTNInit(r)                 (r->nit)
#define getRTInitNth(r,i)             (*(r->iv+(i)))
#define getRTNGVar(r)                 (r->ngv)
#define getRTGVarNth(r,i)             (*(r->vgv+(i)))
#define getRTNGlobC(r)                (r->ngc)
#define getRTGlobCNth(r,i)            (*(r->glc+(i)))
#define isGcInAgenda(r,i)             (*(r->inag+(i)))
#define setGcInAgenda(r,i)            (*(r->inag+(i)) = 1)
#define setGcOutOfAgenda(r,i)         (*(r->inag+(i)) = 0)
#define getRTNCondC(r)                (r->ncdc)
#define getRTCondCNth(r,i)            (*(r->cdc+(i)))
#define getRTNMFC(r)                  (r->nmfc)
#define getRTMFNth(r,i)               (*(r->mfc+(i)))
#define isCcInValSpc(r,i)             (*(r->valSpc+(i)))
#define isCcInAgenda(r,i)             (*(r->inac+(i)))
#define setCcInAgenda(r,i)            (*(r->inac+(i)) = 1)
#define setCcOutOfAgenda(r,i)         (*(r->inac+(i)) = 0)
#define getRTNValSpc(r)               (r->valspc)
#define getRTValSpcNth(r,i)           (*(r->valSpc+(i)))
#define setCcInValSpc(r,i)            (*(r->valSpc+(i)) = 1)
#define setCcOutOfValSpc(r, i)        (*(r->valSpc+(i)) = 0)
#define isInited(r)                   (r->inited)
#define setInited(r)                  (r->inited = 1)
#define areGCRequired(r)              (r->runGC)
#define setGCRequired(r)              (r->runGC = 1)
#define setGCUnRequired(r)            (r->runGC = 0)
#define getRTMisbCumTime(r)           (r->cumTime)
#define setRTMisbCumTime(r,c)         (r->cumTime = c)
#define getRTPrecision(r)             (r->prec)
#define setRTPrecision(r,p)           (r->prec = p)
#define isOnlyFaultIso(r)             (r->onlyFaultIso)
#define getRTNumMisb(r)               (r->numMisb)
#define getRTNumOk(r)                 (r->numOk)
#define getRTMisbVarArray(r)          (r->mbVA)
#define getRTOkVarArray(r)            (r->okVA)
#define getRTSoftVarArray(r)          (r->sfVA)
#define getRTOldNumMisb(r)            (r->oldNumMisb)
#define getRTOldMisbVarArray(r)       (r->oldMisbVarArray)
#define getVarOldMisbVarArrayPt(rt,n) \
        ((variable *)(getRTOldMisbVarArray(rt) + (n)))
#define getVarOldMisbVarArray(rt,n)   (*(getVarOldMisbVarArrayPt(rt, n)))
#define getRTNumNodesHST(r)           (r->numNodesHST)
#define getRTHSTree(r)                (r->HSTree)
#define getRTSoftConflictListcaen(r)      (r->softConflictListcaen)
#define setOnlyFaultIso(r,v)          (isOnlyFaultIso(r) = (v))
#define setRTNumMisb(r,n)             (getRTNumMisb(r) = (n))
#define setRTNumOk(r,n)               (getRTNumOk(r) = (n))
#define setRTMisbVarArray(r,a)        (getRTMisbVarArray(r) = (a))
#define setRTOkVarArray(r,a)          (getRTOkVarArray(r) = (a))
#define setRTSoftVarArray(r,a)        (getRTSoftVarArray(r) = (a))
#define setRTOldNumMisb(r,n)          (getRTOldNumMisb(r) = (n))
#define setRTOldMisbVarArray(r,a)     (getRTOldMisbVarArray(r) = (a))
#define setVarOldMisbVarArray(v,rt,n) (getVarOldMisbVarArray(rt, n) = (v))
#define setRTNumNodesHST(r,n)         (getRTNumNodesHST(r) = (n))
#define setRTHSTree(r,l)              (getRTHSTree(r) = (l))
#define setRTSoftConflictListcaen(r,s)    (getRTSoftConflictListcaen(r) = (s))
#define getRTDiagOut(r)               (r->diagOut)
#define setRTDiagOut(r,s)             (r->diagOut = s)
#define getRTNCondi(r)                (r->ncondi)
#define getRTCondiNth(r,i)            (*(r->condi+(i)))
#define freeRuntime(r)                (freeArrayVariables (r->dico, r->nvar), \
                                       cefree (r->in), \
                                       cefree (r->dispinfval), \
                                       cefree (r->dispsupval), \
                                       cefree (r->dispall), \
                                       cefree (r->disp), \
                                       cefree (r->lowalarm), \
                                       cefree (r->highalarm), \
                                       cefree (r->lowmisb), \
                                       cefree (r->highmisb), \
                                       cefree (r->misb), \
                                       cefree (r->temp), \
                                       cefree (r->untemp), \
                                       freeArrayInfluences (r->infl, r->nifl), \
                                       cefree (r->condin), \
                                       cefree (r->condout), \
                                       freeArrayCGraph (r->cg, r->ngc), \
                                       freeArrayIV (r->iv, r->nit), \
                                       freeArrayGVar (r->vgv, r->ngv), \
                                       freeArrayExpr (r->glc, r->ngc), \
                                       cefree (r->inag), \
                                       cefree (r->oldMisbVarArray), \
                                       freeArrayCond (r->condi, r->ncondi), \
                                       delete r->bridgeEvents, \
                                       delete r->bridgeConds, \
                                       delete r->bridgeTransitions, \
                                       freeArrayExpr (r->cdc, r->ncdc), \
                                       freeArrayExpr (r->mfc, r->nmfc), \
                                       cefree (r->inac), \
                                       cefree (r->valSpc), \
                                       cefree (r), \
                                       r = 0)

inline void freeBridgeConflicts (vector<vector<Atom*>* >* bcf) {
  if (isNull (bcf))
    return;
  for (unsigned int i=0; i<bcf->size (); i++) {
    for (unsigned int j=0; j<(*bcf)[i]->size (); j++) {
      delete (*(*bcf)[i])[j];
      (*(*bcf)[i])[j] = 0;
    }
    delete (*bcf)[i];
    (*bcf)[i] = 0;
  }
  delete bcf;
  bcf = 0;
}

extern variable ce_getVarDico (runtime, const char*);

extern void addBridgeEvent (condition cd, Atom* at, bool tr, runtime rt);
extern void removeBridgeEvent (condition cd, Atom* at, runtime rt);
extern void addBridgeConflict (runtime rt, ptConflict conf);

/**
 * eval condition.
 */
/*extern boolean ce_splitCond (condition, runtime);*/
extern boolean ce_evalCond (condition, runtime);

/*  */
extern boolean ce_pushCond (influence, runtime);

/**
 * print the variable value of v in runtime rt on the file descriptor.
 */
extern void ce_printVariable (FILE *, runtime, variable);
extern void ce_printVarHistory (FILE*, variable, runtime);

/**
 * print the vector of all variables in rt on the file descriptor.
 */
extern void ce_printDic (FILE *, runtime);
extern void ce_printConds (FILE *, runtime);
extern void ce_printSpace (FILE *, runtime);
extern void ce_printInputs (FILE *, runtime);
extern void ce_printDisplays (FILE *, runtime);
extern void ce_printMeasures (FILE *, runtime);
extern void ce_printInits (FILE *, runtime);

/**
 * initialize the vector of all variables.
 */
extern void ce_updateDico (runtime);

/**
 * fill history of the variable with its 'old' value.
 * init only.
 */
extern void ce_fillHisto (variable);

/**
 * fill history of measures of the variable with its measured value,
 * and reset alarms.
 * ?
 */
extern void ce_fillHistoMes (variable);
extern void ce_fillAllHistos (runtime);
extern void ce_fillAllHistosMes (runtime);

/**
 * initialize all variables of a runtime.
 */
extern void ce_initVars (runtime);

/**
 * eval variables conditions references.
 */
extern boolean ce_evalVarConds (variable, runtime);

/**
 * reset splits flag
 */
extern void resetSplits (runtime);

/**
 * eval coefficients of an influence.
 */
extern void ce_evalInfluCoefs (runtime, double);

/**
 * test if a variable is an input variable of the runtime rt.
 */
extern long ce_isInput (runtime, variable);
extern long ce_isDisplay (runtime, variable);
extern long ce_isMeasure (runtime, variable);

/**
 * move history values by one..
 */
extern void ce_pushHisto (variable);
extern void ce_pushHistoMes (variable);

/**
 * init a variable considering influences on it.
 */
extern void ce_compInitVar (runtime, variable);

/**
 * update temporal vector of a runtime.
 */
extern void ce_doStepTempo (runtime);

/**
 * update untemporal vector of a runtime.
 */
extern void ce_doStepUntemp (runtime);

/**
 * move all variables to the 'old' slot
 */
extern void ce_moveAllOld (runtime);

/**
 * update a variable value considering its influences.
 */
extern void ce_updateVariable (runtime, variable);

/**
 * put the value of the cause in TMP2.
 */
extern void ce_loadCauseValue (runtime, influence);

/**
 * eval a dynamic influence following :
 * \f$Y(t+1) = a*Y(t) + (1-a)*\Sigma(Ki*X_i(t-d_i))\f$.
 */
extern void ce_evalDynInfl (runtime, influence);

/**
 * TODO !
 */
extern void ce_evalDyn2Infl (runtime, influence);

/**
 * set a low alarm to the measured variables.
 * used to prevent evaluation of measured variables.
 */
extern void ce_setAlarms (runtime);

/**
 * set the new value of the variable as the old one.
 */
extern void ce_new2OldValue (runtime, variable);

/**
 * hybrid trajectories validity space filtering algorithms.
 *
 */
extern void viewValSpc (runtime);
extern void viewCcAgenda (runtime);
extern long ce_initValSpc (runtime);
extern void ce_fillCondAgenda (runtime);
extern void ce_setOutOfValSpc (runtime, condition);
extern void ce_setInValSpc (runtime, condition);
extern void ce_updateValSpc (runtime, condition, condition);
extern long ce_updateCondAgenda (runtime, gVar);
extern boolean ce_evalCondExpr (condition, runtime, bool);
extern vector<condition>* ce_findCondCover (gVar, runtime);
extern void ce_splitRt (runtime, objet*, long, queue<runtime>*);
extern void ce_emptyIsecValSpc (runtime, long);
extern void viewSplitState (gVar, objet, objet); /* debug */
extern void ce_splitValiditySpc (runtime);

/**
 * eval function in global constraints.
 * result is in TMP4.
 */
extern objet ce_evalExpr (runtime, expr);
extern void ce_evalSolFunc (runtime, expr);
extern void ce_fillAgenda (runtime);
extern long ce_updateAgenda (runtime, gVar);
extern void ce_globalConstraints (runtime);

/**
 * get inputs from the data in and affect it to variables in runtime.
 */
extern long ce_enterInputs (runtime);
extern void ce_moveInputs (runtime);

/**
 * get measures from the data in and affect them to the variables.
 */
extern long ce_enterMeasures (runtime);
extern long ce_enterMeasuresOld (runtime);
extern long ce_yieldMRTConds (runtime);
extern long ce_yieldConds (runtime);
extern long ce_yieldDisplays (runtime, long);
extern long ce_yieldAlarms (runtime);

/**
 * send diagnosis to the output. 
 */
extern long ce_yieldDiags (runtime);
extern long ce_enterMisbs (runtime);
extern long ce_enterConds (runtime);
extern void ce_resetMs (runtime);
extern long ce_checkMs (runtime, long);
extern void ce_printRules (FILE *, runtime);
extern void ce_printCGraph (FILE *, runtime);
extern void ce_printUpdateListcaen (FILE *, runtime);
extern void ce_printExpr (FILE *, expr);
extern void ce_printGContr (FILE *, expr);
extern void ce_printGVar (FILE *outs, gVar);
extern void ce_printGlobVs (FILE *, runtime);
extern void ce_printGlobCs (FILE *, runtime);
extern STATUS faultIsolationSetup(runtime, long);
extern STATUS hittingSetDiagnosis(runtime, int);
extern void hsTreeDeallocation(runtime);
extern void faultIsolationEnd(runtime);
extern void ce_razDiagBuf (runtime);
extern void ce_prepareDiag (runtime);
extern void ce_printDiagCond (runtime, condition);
extern void ce_printDiagExpr (runtime, expr);
extern void ce_printDiagVarNom (runtime, variable);
extern void ce_printDiagReal (runtime, double);
extern void ce_printDiagRange (runtime, range);

extern void caen_initialize (runtime, long, double, char *, long);
extern long caen_doCycle (runtime, long, long);
extern long caen_doCycle2 (runtime, long, long);
extern void caen_close (runtime);
extern void ce_detectAndIsolate (runtime, long, long);

extern multimap<long, runtime, less<long> >* nrt;
extern double* meas;
extern double* inps;
#define pushNrt(l, rt)         (nrt->insert(pair<long, runtime>(l, rt)))

/* copy of structures. */
inline variable copyVariable (variable v) {
 
  if (isNull (v))
    return NULL;
  else {
    variable Vac = allocVariable ();
    Vac->nom = mkString (getVarName (v));
    Vac->cst = v->cst;
    setVarDom (Vac, copyObjet (getVarDom (v)));
    Vac->deter = v->deter;
    Vac->oldVal = copyObjet (v->oldVal);
    Vac->newVal = copyObjet (v->newVal);
    Vac->mesVal = copyObjet (v->mesVal);
    Vac->lowalrm = v->lowalrm;
    Vac->highalrm = v->highalrm;
    Vac->lowmisbehaving = v->lowmisbehaving;
    Vac->highmisbehaving = v->highmisbehaving;
    Vac->misbehaving = v->misbehaving;
    Vac->histoMisb = copyLong (v->histoMisb, Vac->misbehaving);
    Vac->histoLn = v->histoLn;
    Vac->histo = copyArrayObj (v->histo, v->histoLn);
    Vac->histoMes = copyArrayObj (v->histoMes, v->histoLn);
    Vac->histoLowAlrm = copyLong (v->histoLowAlrm, v->histoLn);
    Vac->histoHighAlrm = copyLong (v->histoHighAlrm, v->histoLn);
    Vac->ncd = v->ncd;
    Vac->relCd = copyLong (v->relCd, v->ncd);
    return Vac;
  }
}

inline variable* copyArrayVariables (variable* a, long s) {
  
  if (isNull (a))
    return NULL;
  else {
    variable* Vanc = allocVVar (s);
    for (int i=0; i<s; i++) 
      Vanc[i] = copyVariable (a[i]);
    return Vanc;
  }
}

inline gVar copyGVar (gVar gv) {
  if (isNull (gv))
  return NULL;
  else {
    gVar gvc = allocGVar ();
    gvc->tgv = (gVarType) gv->tgv;
    gvc->uv.interm = gv->uv.interm;
    gvc->ob = copyObjet (gv->ob);
    gvc->nex = gv->nex;
    gvc->ex = copyLong (gv->ex, gv->nex);
    gvc->ncex = gv->ncex;
    gvc->cex = copyLong (gv->cex, gv->ncex);
    return gvc;
  }
}

inline gVar* copyArrayGVar (gVar* a, long s) {
  if (isNull (a))
    return NULL;
  else {
    gVar* gvnc = allocVGVar (s);
    for (int i=0; i<s; i++)
      gvnc[i] = copyGVar (a[i]);
    return gvnc;
  }
}

inline expr copyExpr (expr e) {
  if (isNull (e))
    return NULL;
  else {
    expr ec = allocExpr ();
    ec->tex = e->tex;
    return ec;
  }
}

inline condition copyCondition (condition f) {
    if (isNull (f))
    return NULL;
    else {
      condition fc = allocCondition();
      fc->type = (condType) f->type;
      fc->opr = (opRel) f->opr;
      fc->ob = copyObjet (f->ob);
      fc->val = f->val;
      fc->at = f->at;
      fc->num = f->num;
      fc->nle = f->nle;
      fc->le = copyLong (f->le, f->nle);
      return fc;
    }
  }

inline condition* copyArrayCond(condition* a, long s)   {
  
  if (isNull (a))
    return NULL;
  else {
    condition* Cdnc = allocVCondition (s);
    for (int i=0; i<s; i++)
      Cdnc[i] = copyCondition (a[i]);
    return Cdnc;
  }
  return NULL;
}

inline influence copyInfluence (influence i) {
  if (isNull (i))
  return NULL;
  else {
    influence Infc = allocInfluence ();
    Infc->name = mkString (i->name);
    Infc->it = (influType) i->it;
    Infc->length_act_cond_hist = i->length_act_cond_hist;
    Infc->act_cond_array = copyBoolean (i->act_cond_array, i->length_act_cond_hist);
    Infc->gain = copyObjet (i->gain);
    Infc->delinf = i->delinf;
    Infc->delsup = i->delsup;
    Infc->tRep = copyObjet (i->tRep);
    Infc->a = copyObjet (i->a);
    Infc->b = copyObjet (i->b);
    Infc->a2 = copyObjet (i->a2);
    Infc->b2 = copyObjet (i->b2);
    Infc->xi = copyObjet (i->xi);
    Infc->w = copyObjet (i->w);
    Infc->n1 = copyObjet (i->n1);
    Infc->n2 = copyObjet (i->n2);
    Infc->d0 = copyObjet (i->d0);
    Infc->d1 = copyObjet (i->d1);
    Infc->d2 = copyObjet (i->d2);
#ifdef ENV_OPTI
    Infc->inf_fct = i->inf_fct;
#endif
    return Infc;
  }
}


inline influence* copyArrayInf(influence* a, long s)   {
  
  if (isNull (a))
    return NULL;
  else {
    influence* Infnc = allocVInfluence (s);
    for (int i=0; i<s; i++)
      Infnc[i] = copyInfluence (a[i]);
    return Infnc;
  }
}

inline cGraph copyCGraph (cGraph cg) {
  if (isNull (cg))
    return NULL;
  else {
    cGraph Cgc = allocCGraph ();
    Cgc->ncause = cg->ncause;
    Cgc->ti = (influType) cg->ti;
    Cgc->neffet = cg->neffet;
    return Cgc;
  }
}

inline cGraph* copyArrayCGraph (cGraph* a, long s) {
  if (isNull (a))
    return NULL;
  else {
    cGraph* Cgnc = allocVCGraph (s);
    for (int i=0; i<s; i++)
      Cgnc[i] = copyCGraph (a[i]);
    return Cgnc;
  }
} 

inline initVar copyInitVar (initVar iv) {
  if (isNull (iv))
    return NULL;
  else {
    initVar ivc = allocInitVar ();
    ivc->val = copyObjet (iv->val);
    return ivc;
  }
}

inline initVar* copyArrayInitV (initVar* a, long s) {
  if (isNull (a))
    return NULL;
  else {
    initVar* ivnc = allocVInitV (s);
    for (int i=0; i<s; i++)
      ivnc[i] = copyInitVar (a[i]);
    return ivnc;
  }
} 

#ifdef ENV_OPTI
extern bool env_opti;                     /* envelopes optimization from options */
#endif
extern long disp_num;                     /* counter for display ordering */
extern long rt_num;                       /* current runtime number for creation */
extern vector<condition>* existing_conflicts;
extern runtime ce_copyRuntime (runtime run);
extern condition ce_copyCondition (condition cd, hash_map<const char*, variable, hash<const char*>, eqstr>* n_v, hash_map<int, condition, hash<int> >* n_c, hash_map<int, expr, hash<int> >* n_e);
extern expr ce_copyExpr (expr e, hash_map<int, gVar, hash<int> >* ngvars, hash_map<int, expr, hash<int> >* nexprs, hash_map<const char*, variable, hash<const char*>, eqstr>* nvars);

/*ELEMENT (*ptFuncConf) (ptListcaen, listcaen, ELEMENT, hash_map<int, ptConflict, hash<int> >*, hash_map<int, influence, hash<int> >*);
  extern misbVarArray ce_copyMisbVarArray (misbVarArray mva, long nm, hash_map<const char*, variable, hash<const char*>, eqstr>* nvars, hash_map<int, ptConflict, hash<int> >* nconfs, hash_map<int, influence, hash<int> >* ninfs);
  ELEMENT ce_copyConflict (ptListcaen pnew, listcaen old, ELEMENT e, hash_map<int, ptConflict, hash<int> >* nconfs, hash_map<int, influence, hash<int> >* ninfs);
  ELEMENT ce_copyConflictSet (ptListcaen pnew, listcaen old, ELEMENT e, hash_map<int, ptConflict, hash<int> >* nconfs, hash_map<int, influence, hash<int> >* ninfs);*/

extern runtime ce_copyPartialFaultyRt (runtime, int, double, double, int);
extern void moveHistory (runtime, variable, double, double);

extern void addNewRuntime (runtime rt);
void call_nelder_second (influence ifl, runtime rt, objet y, objet yp, objet non_opt_res);
void call_nelder_obs (variable v, runtime rt, objet y, objet yp, objet non_opt_res, enum influType it, bool upd);
void meanCoef (objet, double*, double*, double*, int);
void mean1stOrderCoefs (influence, double*, double*, double*);
void mean2ndOrderCoefs (influence, double*, double*, double*);
void setNelderResult (objet obj, double d1, double d2);

#endif  /* _CE_RUNT_H */
