/* Model checking on the model. */

#ifndef MDLCHK_H
#define MDLCHK_H

#include <map>
#include <math.h>
#include <stdlib.h>
#include "../system/DataManager.h"
#include "Clause.h"
#include "State.h"
#ifdef USE_BDD
#include "bdd.h"
#endif

#undef PROBAS

#ifndef EQSTR_S
#define EQSTR_S
struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};
#endif

class mdlChk : public DataManager {
 public:
#ifdef PROBAS
  bdd **var; /* BDD variable array */
  bdd pb; /* local problem expressed as a BDD. */
  bdd global; /* global bdd of the model. */
  hash_map<const char*, bdd*, hash<const char*>, eqstr>* corresp_var; /* correspondence table for state variables. */
  hash_map<const char*, bdd, hash<const char*>, eqstr>* corresp_atoms;
  hash_map<const char*, bdd, hash<const char*>, eqstr>* corresp_comps;
  //hash_map<int, Atom*, hash<int> >* table_atoms;
#endif

  mdlChk();

#ifdef PROBAS
  void encode(vector<Clause*>* cls, bdd &curr_bdd, unsigned int size);
  bdd* encodeClause(Clause* cla, int &var_i, int &count, bdd &curr_bdd);
  void encodeInternalClauses(StateVar* sta, int &var_i, int &count, bdd &curr_bdd);
  void runChk(vector<Clause*>* cls);
  int bdd_fnload(const char* fname, bdd* root);
  int bdd_fnsave(const char* fname, bdd r);
  void compileBddClauses(vector<Clause*>* cls, unsigned int size);
  double computeCondProb(TrajEl* elt, Transition* trans, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  double computeProb(Atom* at);
  void addNode(State* wkst, const char* id, string lab);
  void addNode(bdd* bdpb, const char* id, string lab);
  bdd* makeStateBdd(State* wkst);
  void addToStateBdd(State* wkst, const char* comp_id);
  double satcountset(bdd state_bdd, unsigned int unused);
#endif
};
#endif
