/* Prime implicants generation */
 
#ifndef PIGEN_H
#define PIGEN_H

#include "itms.h"
#include "pi_table.h"
#include "Trajectory.h"
#include <vector>
#include <time.h>

using namespace std;

/* s_component from caen-objet */
#ifndef S_COMPONENT
#define S_COMPONENT

typedef struct s_component {
  string *type;
  string *var_name;
  string *name;
  Comp* com;
  vector<string>* comp_args;
} S_component, *component;

#endif /* S_COMPONENT */

/**
 * \struct eqstr
 * Equality operator for const char* for hashtables (hash_map).
 */

#ifndef EQSTR_S
#define EQSTR_S
struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};
#endif

extern map<int, component>* oco;
extern char appliName[];

Trajectory* initFakeTrajectory (itms* tms);
Trajectory* setAtomsToUnknown (itms* tms);
PiTable* generateAllPI (itms* tms);
vector<vector<Atom*>*>* generatePI (Atom* at, Trajectory* tr, itms* tms);

void enumAllPI (itms* tms, int opt);
vector<vector<Atom*>*>* enumPI (Atom* at, Trajectory* tr, itms* tms, int opt);
vector<Atom*>* enumSingletons (vector<Comp*>* comps, multimap<int, vector<Atom*>*>* pi2test, Atom* at, Trajectory* tr, itms* tms, vector<vector<Atom*>*>* res, int opt);
bool enumNextSize (Atom* current_at, vector<Atom*>* single, multimap<int, vector<Atom*>*>* pi2test, unsigned int set_size, Trajectory* tr, itms* tms, vector<vector<Atom*>*>* res, unsigned int comp_size);
void enumCompleteSize (Atom* current_at, vector<Atom*>* single, Trajectory* tr, itms* tms, vector<vector<Atom*>*>* res);
bool cartVec (vector<vector<Atom*>*>* candis, vector<Atom*>* single, unsigned int num);
bool hasStVar (vector<Atom*>* candis, Atom* at, vector<Atom*>* store);
bool insertAtomInVec (Atom* current_at, Atom* at, vector<Atom*>* vec, Trajectory* tr, itms* tms, vector<vector<Atom*>*>* res);
bool inImplicants (vector<Atom*>* vec, itms* tms, vector<vector<Atom*>*>* res);
vector<Comp*>* getImportedComps (Comp* co, vector<Comp*>* res);
bool inVector (Comp* co, vector<Comp*>* comps);
bool inVector (vector<Atom*>* at, vector<Atom*>* vec);
bool inVector (vector<Atom*>* vec, Atom* at);
bool satisTms (vector<Atom*>* props, Atom* current_at, Trajectory* tr, itms* tms, vector<vector<Atom*>*>* res);

#endif
