#include "pi_table.h"
#include <fstream>
#include <sstream>

PiTable::PiTable () {
  implicants = new hash_map<int, vector<vector<Atom*>*>*, hash<int> > ();
}

PiTable::~PiTable () {
  for (hash_map<int, vector<vector<Atom*>*>*, hash<int> >::iterator it = implicants->begin ();
       it != implicants->end (); it++) {
    for (unsigned int i=0; i<(*it).second->size (); i++)
      delete (*(*it).second)[i];
    delete (*it).second;
  }
  delete implicants;
}

void PiTable::addPIAtom (Atom* at, vector<vector<Atom*>*>* pis) {
  (*implicants)[(int)at] = pis;
}

vector<vector<Atom*>*>* PiTable::getPIs (Atom* at) {
  if (implicants->count ((int) at) != 0)
    return (*implicants)[(int) at];
  else return NULL;
}

void PiTable::showPI (Atom* at) {
  cout << "\nPI(" << at->getId () << ") => ";
  vector<vector<Atom*>*>* pis = (*implicants)[(int)at];
  for (unsigned int i=0; i<pis->size (); i++) {
    if (i > 0)
      cout << "\n\t\t=>";
    for (unsigned int j=0; j<(*pis)[i]->size (); j++) {
      cout << (*(*pis)[i])[j]->getId () << " - ";
    }
  }
}

void PiTable::showAllPI () {
  for (hash_map<int, vector<vector<Atom*>*>*, hash<int> >::iterator it = implicants->begin ();
       it != implicants->end (); it++)
    showPI ((Atom*) (*it).first);
}



/************************************************************
  Write the prime implicants in the file '<model>.pi'
  The ASCII format is the following: 
  IndpteVar:at11 at12 ... at1n , ... , atm1 atm2 ... atmn;
  wich means:
  IndpteVar => at11/\at12/\.../\at1n
            => ...
            => atm1/\atm2/\.../\atmn
 ************************************************************/

void PiTable::writePI(char* name) {
  const char* PISUFFIX=".pi";
  char FileName[strlen(name)+strlen(PISUFFIX)+1];
  vector<vector<Atom*>*>* pis;
  
  strcpy(FileName,name);
  strcat(FileName,PISUFFIX);
  cout << "writing file " << FileName << endl;

  ofstream f(FileName); //create and open the file

  for (hash_map<int, vector<vector<Atom*>*>*, hash<int> >::iterator it = implicants->begin(); it != implicants->end (); it++) {
    f << ((Atom*)(*it).first)->getId() << ":";
    pis = (*implicants)[(int)(*it).first];

    for(unsigned int i=0; i<pis->size();i++) {
      for (unsigned int j=0; j<(*pis)[i]->size (); j++) {
	f << (*(*pis)[i])[j]->getId ();
	if (j<((*pis)[i]->size ())-1) {
	  f << " ";
	}
      }
      if (i<(pis->size())-1) {
	f << ",";
      }
    }
    f << "\n";
  }

  f.close();
  
}


