/*********************************************************************/
/*                                                                   */
/*                          --- nelder.h ---                         */
/*                                                                   */
/* fct: module des en-tetes des procedures d'optimisation            */
/*      non-lineaire avec le simplexe de Nelder & Mead.              */
/*                                                                   */
/* ver: 1 mars 1994                                                  */
/*                                                                   */
/*********************************************************************/

#include "optim.h"
#include <string.h>

/*** definitions globales du module ***/

#define INFINI			1.0e20

#define EN_COURS		1
#define MINIMUM_ATTEINT		0
#define ERREUR_MAX_EVAL_COUT	-1
#define ERREUR_CALCUL_COUT	-99

#define ERREUR_AUCUN_OPTIMUM	-1

#define FACTEUR_PRECISION_OPTIMUM		1.0e2

/*
 * boolean values
 */
/*typedef enum boolVal {FALSE, TRUE} boolean;*/

#define COPY_DOUBLE(ptr_cible, ptr_source, dim)   memcpy(ptr_cible, ptr_source, dim * sizeof(double))

/*** declaration des types globaux ***/

typedef struct
   {
   boolean TRACE;
   boolean TRACE_FICHIER;
   char    fichier_trace[80];
   boolean TRACE_OPT;
   boolean FICHIER_OPT;
   char    fichier_opt[80];
   boolean TRACE_STAT;
   boolean FICHIER_STAT;
   char    fichier_stat[80];
   double  alpha;
   double  beta;
   double  gamma;
   double  delta;
   int     type_initialisation;
   double  param_initialisation;
   int     type_pivot;
   int     type_critere_arret;
   double  precision_cout;
   double  precision_X;
   int     max_eval_cout;
   }  Nelder_options; 

typedef struct
   {
   int     n;      /* nombre de parametres d'optimisation */
   double *x;      /* ptr du vecteur des parametres       */
   double  fx;     /* valeur de la fonction cout          */
   }   Nelder_point;

typedef struct
   {
   int dim;             /* nb de points qui composent le simplexe */
   Nelder_point **pt;   /* vecteur de ptr des points du simplexe  */
   int min;             /* index du point a cout minimal          */
   int max;             /* index du point a cout maximal          */
   }   Nelder_simplexe;


/*** declaration des procedures et fonctions du module ***/

void Nelder_initial
(
int     type_init,   /* type d'initialisation du simplexe         */
double  param_init,  /* parametre d'initialisation                */
double *Xi,          /* ptr du point initial                      */
double *Xmin,        /* ptr des bornes minimales du domaine des X */
double *Xmax,        /* ptr des bornes maximales du domaine des X */
Nelder_simplexe *simplexe      /* ptr du simplexe                 */
);

void Nelder_pivot
(
int              type_pivot,   /* type de pivot */
Nelder_simplexe *simplexe,     /* ptr du simplexe */
Nelder_point    *pivot         /* ptr du point pivot */
);

void Nelder_tri_simplexe
(
Nelder_simplexe *simplexe   /* ptr du simplexe */
);

void Nelder_reflexion
(
double        alpha,      /* facteur de reflexion */
Nelder_point *pivot,      /* ptr du point pivot */
Nelder_point *fx_max,     /* ptr du point du simplexe avec fx_max */
Nelder_point *reflexion,  /* ptr du point resultant de la reflexion */
double* Xmin,
double* Xmax
);

void Nelder_expansion
(
double        gamma,       /* facteur d'expansion */
Nelder_point *pivot,       /* ptr du point pivot */
Nelder_point *reflexion,   /* ptr du point resultant de la reflexion */
Nelder_point *expansion,   /* ptr du point resultant de l'expansion */
double* Xmin,
double* Xmax
);

void Nelder_contraction
(
double        beta,         /* facteur de contraction */
Nelder_point *pivot,        /* ptr du point pivot */
Nelder_point *fx_max,       /* ptr du point du simplexe avec fx_max */
Nelder_point *contraction,  /* ptr du point resultant de la contraction */
double* Xmin,
double* Xmax
);

void Nelder_reduction
(
double           delta,     /* facteur de reduction */
Nelder_simplexe *simplexe   /* ptr du simplexe */
);

boolean Nelder_critere_arret
(
int     type_critere_arret,   /* type de critere d'arret */
double  precision_cout,       /* precision recherchee sur le cout */
double  precision_X,          /* precision recherchee sur les X */
double *delta_X,              /* ptr des largeurs des domaines */
Nelder_simplexe *simplexe     /* ptr du simplexe */
);

void Nelder_trace_initial
(
FILE *fich_trace,           /* flux de sortie */
Nelder_simplexe *simplexe   /* ptr du simplexe */
);

void Nelder_trace
(
FILE *fich_trace,           /* flux de sortie */
Nelder_simplexe *simplexe   /* ptr du simplexe */
);

void Nelder_trace_final
(
FILE *fich_trace,         /* flux de sortie */
int   etat_calcul,        /* etat des calculs */
int   nb_iter_simplexe,   /* nb iteration dans l'algorithme du simplexe */
int   nb_eval_cout        /* nb evaluation de la fonction cout */
);

extern void Nelder_init_options
(
Nelder_options *options   /* ptr du vecteur des options */
);

void Nelder_options_def
(
Nelder_options *options   /* ptr du vecteur des options */
);

void Nelder_optimisation
(
Nelder_options *options,   /* ptr des options du simplexe           */
double (*Nelder_cout)      /* ptr de la fonction de calcul du cout  */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)    */
int    *iuser,  /* ptr d'entiers (variables utilisateur)     */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      + = nb d'evaluations du cout   */
		/*	     -1 = max eval. cout atteint     */
		/*	    -99 = erreur dans Nelder_cout    */
);

extern void opt_nelder
(
double (*Nelder_cout)   /* ptr de la fonction de calcul du cout */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)    */
int    *iuser,  /* ptr d'entiers (variables utilisateur)     */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      + = nb d'evaluations du cout   */
		/*	     -1 = max iterations atteint     */
		/*	    -99 = erreur dans Nelder_cout    */
);

void mopt_nelder
(
int    *nb_opt,        /* ptr du nb de points de depart de l'optim.  */
double (*Nelder_cout)  /* ptr de la fonction de calcul du cout       */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)    */
int    *iuser,  /* ptr d'entiers (variables utilisateur)     */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      0 = ok			     */
		/*	     -1 = aucun d'optimum trouve     */
		/*	    -99 = erreur dans Nelder_cout    */
);

void opt_nelder_
(
double (*Nelder_cout)   /* ptr de la fonction de calcul du cout */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)    */
int    *iuser,  /* ptr d'entiers (variables utilisateur)     */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      + = nb d'evaluations du cout   */
		/*	     -1 = max iterations atteint     */
		/*	    -99 = erreur dans Nelder_cout    */
);

void mopt_nelder_
(
int     *nb_opt,       /* ptr du nb de points de depart de l'optim.  */
double (*Nelder_cout)  /* ptr de la fonction de calcul du cout       */
       (
       int    *n,      /* ptr du nombre de parametres de commande    */
       double *X,      /* ptr du vecteur des parametres de commande  */
       double *Xmin,   /* ptr des bornes minimales du domaine des X  */
       double *Xmax,   /* ptr des bornes maximales du domaine des X  */
       double *duser,  /* ptr de doubles (variables utilisateur)     */
       int    *iuser,  /* ptr d'entiers (variables utilisateur)      */
       int    *iret,   /* ptr du code de retour des calculs (0 = ok) */
       int    *ier     /* ptr du code d'erreur des calculs (0 = ok)  */
       ),
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax,   /* ptr des bornes maximales du domaine des X */
double *duser,  /* ptr de doubles (variables utilisateur)    */
int    *iuser,  /* ptr d'entiers (variables utilisateur)     */
int    *ier     /* ptr du code d'erreur de l'optimisation:   */
		/*	      0 = ok			     */
		/*	     -1 = aucun d'optimum trouve     */
		/*	    -99 = erreur dans Nelder_cout    */
);

