/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
- 18/12/00 : creation.

*/

#ifndef DIAGNOSER_H
#define DIAGNOSER_H

#include "DataManagerrt.h"
#include "../itms/itmsrt.h"

#ifndef CONFLICT_H
#define CONFLICT_H

/**
* \class disConflict
* \brief Represents conflicts between transitions and variables in states.
*/

class disConflict {
 public:
  int conf_time;
  TrajEl* conf;
  bool sign; /**< if it a negated proposition or not.*/
  
  disConflict(TrajEl* cf, int ct) {
    conf_time = ct;
    conf = cf;
    sign = true;
  }
  disConflict(TrajEl* cf, int ct, bool si) {
    conf_time = ct;
    conf = cf;
    sign = si;
  }
  ~disConflict() {
  }
};
#endif

#ifndef LIKELY_TRANS
#define LIKELY_TRANS

class likelyTransition {
 
 public:
  Atom* val; //start value.
  Atom* at; //end value.
  double cond_prob; //conditional probability = likeliness.
  double rank;
  Transition* trans; //val -> at transition.
  int start_step; /**< transition starting step. */
  Comp* del_comp; /**< component to be removed at del_step. */
  int del_step;

  likelyTransition(Transition* tr) {
    trans = tr;
    del_comp = NULL;
    del_step = 0;
  }
  ~likelyTransition() {
  }
  void showLikelyTransition() {
    std::cout << cond_prob << std::endl;
    std::cout << rank << std::endl;
    std::cout << val->getId() << std::endl;
    std::cout << at->getId() << std::endl;
    trans->showTransition();
    std::cout << std::endl;
  }
};
#endif

/**
 * \class Diagnoser
 * \brief Implements all needed routines for diagnosis over trajectories.
 */

class Diagnoser : public itms {

  static hash_map<int, Transition*, hash<int> >* redux_diag;

 public:
  Diagnoser();
  void setObservation(Atom* obs, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state);
  void setObservation(Atom* obs, State* sta);

  /****************************
   * Trajectories management. *
   ****************************/

  //lock elts from the state that has the same state variable as the argument.
  void lockTrajEls(TrajEl* elt, State* sta);
  void lockTrajElsPerm(TrajEl* elt, State* sta);
  void lockTrajElsPerm (TrajEl* elt, State* sta, bool neg);
  void lockTrajElsNegPerm (TrajEl*, State* sta);
  void unlockState(State* st);
  void unlockTrajectory(Trajectory* tr);
  bool checkTransitionsConsistency(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state);

  /******************************************** 
   *  CBFS. Methods for the diagnosis engine. * 
   ********************************************/
  multimap<double, Trajectory*, greater<double> >* constructSuccessors(likelyTransition* lktrans, Trajectory* tr, int time_step, multimap<double, Trajectory*, greater<double> >* res, double rank);
  multimap<double, Trajectory*, greater<double> >* immediateSuccessors(hash_map<const char*, disConflict*, hash<const char*>, eqstr>* soln, Trajectory* tr, int time_step, double rank, multimap<double, Trajectory*, greater<double> >* trajs, bool post);
  hash_map<const char*, disConflict*, hash<const char*>, eqstr>* getConflicts(Trajectory* tr, int time_step, bool post);
  multimap<double, Trajectory*, greater<double> >* coverTrack2(Trajectory* trajec, double rank_lim_low, double rank_lim_high, multimap<bool, Atom*, eqbool>* obs, int amp, double &current_rank);
  multimap<double, Trajectory*, greater<double> >* coverTrack3(Trajectory* trajec, double rank_lim_low, double rank_lim_high, multimap<bool, Atom*, eqbool>* obs, int amp, double &current_rank);
   
  /************************
   * Conflicts Management *
   ************************/
  void addConflictToDB(bool post, const char* id, Transition* trans);
  bool inConflictsDatabase(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* pre_state, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state, Transition* trans); //test for a conflict in the conflicts' database.
  
  /************************
   * Reducing the diag    *
   ************************/
  bool checkForDeadEnd (Transition* trans);
  bool checkForDeadEndExt (Transition* trans, Trajectory* tr);
  void addResult (Trajectory* res_tr, multimap<double, Trajectory*, greater<double> >* res, double &norm);

  /************************
   * Observation Management *
   ************************/
  multimap<double, Trajectory*, greater<double> >* observationResults(vector<Atom*>* obs, Trajectory* tr, int time_step, double rank, bool nom);
  multimap<double, Trajectory*, greater<double> >* observationResults(multimap<bool, Atom*, eqbool>* obs, Trajectory* tr, int time_step, double rank, bool nom);
  void viewHashMapConflicts(hash_map<const char*, disConflict*, hash<const char*>, eqstr>* confs);
};
#endif
  
