
/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* Koala options management */

#ifndef KOACONF_H
#define KOACONF_H

#include <string>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#if (__GNUC__ >= 3)
#include <ext/hash_map>
#else
#include <hash_map>
#endif

#if (__GNUC__ >= 3)
using __gnu_cxx::hash_map;
using __gnu_cxx::hash;
#else
using std::hash_map;
using std::hash;
#endif


#ifndef CESTRING
#define CESTRING
/* les macros pour les chai^nes de caracte`res */
#define allocString(s)      ((char*) malloc(strlen(s)+1))   /* room for the \0, too */
#define mkString(s)         (strcpy(allocString(s),s))
#define isEmpty(s)          (*(s) == '\0')
/* gets memory for s + t, then copies s, then appends t */
#define catString(s,t)      (isNull(t) ? s \
			     : (S1 = strcat (strcpy (cealloc (strlen(s) \
				  + strlen(t)  + 1), s), t), free (s), S1))
#endif

/**
 * \struct eqstr
 * Equality operator for const char* for hashtables (hash_map).
 */

#ifndef EQSTR_S
#define EQSTR_S
struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};
#endif

extern hash_map<const char*, int, hash<const char*>, eqstr>* config_pairs;
extern char* caenDir;

extern void config_init ();
extern int load_conf_file ();
extern int load_conf_file (FILE* fd);
extern void conf_cache_entry (const char* s);
extern void conf_set (const char* s, int v);
extern int conf_get (const char* s);
extern void print_conf ();

extern int koa_hybrid;
extern int koa_log;
extern int koa_evlps_opti;                 /* envelopes optimization for simulation */
extern int koa_code_opti;                  /* C++ optimization */
extern int koa_code_debug;                 /* C++ debug */
extern int twc; /* temporal window continuous */
extern int twd; /* temporal window discrete */
extern int i; /* prime implicants */

#endif
