/*    
 *  Copyright 1999, 2000, 2001, 2003 Emmanuel Benazera, Jean-Claude Gouyon, Jose Angel Jimenez Molina 
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* caen new runtime */

/* #ifdef VXWORKS
   #include <vxWorks.h>
   #endif*/ /* VXWORKS */

/*#define VERB_SPLITS*/
/*#define VERB_INFS*/
/*#define VERB_GLB*/
/*#define VERB_LOADCAUSE*/
/* #define VERB_OPTI */
/*#define VERB_MISB*/

#include "caen-runt.h"
#include "caen-face.h"
#include <iostream>

#ifndef DEBUG
static long L1;
#endif
static s_objet s_objZero={CE_REAL,{{0.,0.}}};
static objet objZero=&s_objZero;
static s_objet s_objUn={CE_REAL,{{1.,0.}}};
static objet objUn=&s_objUn;
static s_objet s_objDix={CE_REAL,{{10.,0.}}};
static objet objDix=&s_objDix;
static s_objet s_objMax={CE_RANGE,{{-DBL_MAX, DBL_MAX}}};
static objet objMax=&s_objMax;
static s_objet s_tmp1, s_tmp2, s_tmp3, s_tmp4, s_tmp5;
static objet TMP1=&s_tmp1, TMP2=&s_tmp2, TMP3=&s_tmp3, TMP4=&s_tmp4, TMP5=&s_tmp5;
/* TMP4 is bound to ce_evalSolFunc () */
long runt_count;
/* variables to store D2INF ranges for b1, b2. */
static s_objet s_tmpii, s_tmpis, s_tmpsi, s_tmpss;
static objet TMPII=&s_tmpii, TMPIS=&s_tmpis, TMPSI=&s_tmpsi, TMPSS=&s_tmpss;

/* copy a given runtime. */
runtime ce_copyRuntime (runtime run) {
  variable v;
  runtime res;
  hash_map<const char*, variable, hash<const char*>, eqstr>* new_vars = new hash_map<const char*, variable, hash<const char*>, eqstr>();
  hash_map<int, condition, hash<int> >* new_conditions = new hash_map<int, condition, hash<int> >();
  hash_map<int, influence, hash<int> >* new_influences = new hash_map<int, influence, hash<int> >();
  hash_map<int, cGraph, hash<int> >* new_nodes = new hash_map<int, cGraph, hash<int> >();
  hash_map<int, gVar, hash<int> >* new_gvars = new hash_map<int, gVar, hash<int> >();
  hash_map<int, expr, hash<int> >* new_exprs = new hash_map<int, expr, hash<int> >();
  hash_map<int, ptConflict, hash<int> >* new_confs = new hash_map<int, ptConflict, hash<int> >();

  res = (runtime) new_malloc (sizeof (s_runtime));
  res->appli = mkString (run->appli);
  res->num = rt_num++;
  res->ts = run->ts;
  res->dts = run->dts;
  res->twc = run->twc;
  res->nvar = run->nvar;
  res->dico = copyArrayVariables (run->dico, run->nvar);
  for (int i=0; i<getRTNVar(res); i++) {
    v = getRTDicoNth (res, i);
    (*new_vars)[getVarName (v)] = v;
  }
  res->nin = run->nin;
  if (!isNull (run->in)) {
    res->in = (variable*) new_malloc (getRTNInput (res) * sizeof (variable*));
    for (int i=0; i<getRTNInput (res); i++)
      res->in[i] = (*new_vars)[getVarName (getRTInputNth(run, i))];
  }
  else res->in = NULL;
  res->inval = run->inval;
  res->ndisp = run->ndisp;
  if (! isNull (run->disp)) {
    res->disp = (variable*) new_malloc (getRTNDisplay (res) * sizeof (variable*));
    for (int i=0; i<getRTNDisplay (res); i++)
      res->disp[i] = (*new_vars)[getVarName (getRTDisplayNth (run, i))];
  }
  else res->disp = NULL;
  res->dispinfval = copyDouble (run->dispinfval, run->ndisp);
  res->dispsupval = copyDouble (run->dispsupval, run->ndisp);
  res->ndispall = run->ndispall;
  res->dispall = copyDouble (run->dispall, run->ndispall);
  res->nmes = run->nmes;
  if (!isNull (run->mes)) {
    res->mes = (variable*) new_malloc (getRTNMeasure (res) * sizeof (variable*));
    for (int i=0; i<getRTNMeasure (res); i++)
      res->mes[i] = (*new_vars)[getVarName (getRTMeasureNth (run, i))]; 
  }
  else res->mes = NULL;
  res->mesval = run->mesval; /*copyDouble (run->mesval, run->nmes);*/
  res->lowalarm = copyLong (run->lowalarm, run->nmes);
  res->highalarm = copyLong (run->highalarm, run->nmes);
  res->lowmisb = copyLong (run->lowmisb, run->nmes);
  res->highmisb = copyLong (run->highmisb, run->nmes);
  res->misb = copyLong (run->misb, run->nmes);
  res->ntemp = run->ntemp;
  if (!isNull (run->temp)) {
    res->temp = (variable*) new_malloc (getRTNTempo (res) * sizeof (variable*));
    for (int i=0; i<getRTNTempo (res); i++)
      res->temp[i] = (*new_vars)[getVarName (getRTTempoNth (run, i))];
  }
  else res->temp = NULL;
  res->nuntemp = run->nuntemp;
  if (!isNull (run->untemp)) {
    res->untemp = (variable*) new_malloc (getRTNUntemp (res) * sizeof (variable*));
    for (int i=0; i<getRTNUntemp (res); i++)
      res->untemp[i] = (*new_vars)[getVarName (getRTUntempNth (run, i))];
  }
  else res->untemp = NULL;

  /* gVars */
  res->ngv = run->ngv;
  if (! isNull (run->vgv)) {
    res->vgv = copyArrayGVar (run->vgv, getRTNGVar (run));
    for (int i=0; i<getRTNGVar (res); i++) {
      if (! isNull (run->vgv[i]->uv.var))
	res->vgv[i]->uv.var = (*new_vars)[getVarName (run->vgv[i]->uv.var)];
      else res->vgv[i]->uv.var = NULL;
      (*new_gvars)[(int)run->vgv[i]] = res->vgv[i];
    }
  }
  else res->vgv = NULL;
  
  /* exprs */
  res->ncdc = run->ncdc;
  if (! isNull (run->cdc)) {
    res->cdc = allocVExpr (run->ncdc);
    for (int i=0; i<getRTNCondC (res); i++) {
      if (new_exprs->count ((int)run->cdc[i]) == 0)
	res->cdc[i] = ce_copyExpr (run->cdc[i], new_gvars, new_exprs, new_vars);
      else res->cdc[i] = (*new_exprs)[(int)run->cdc[i]];
    }
  }
  else res->cdc = NULL;
  
  /* conditions */
  res->ncondi = run->ncondi;
  res->condi = allocVCondition (run->ncondi);
  for (int i=0; i<run->ncondi; i++) {
    if (new_conditions->count ((int)run->condi[i]) == 0)
      res->condi[i] = ce_copyCondition (run->condi[i], new_vars, new_conditions, new_exprs);
    else res->condi[i] = (*new_conditions)[(int) run->condi[i]];;
  }

  /* mapping functions exprs */
  res->nmfc = run->nmfc;
  if (! isNull (run->mfc)) {
    res->mfc = allocVExpr (run->nmfc);
    for (int i=0; i<getRTNMFC (res); i++) { /* store it in hash_map ? */
      res->mfc[i] = ce_copyExpr (run->mfc[i], new_gvars, new_exprs, new_vars);
    }
  }
  else res->mfc = NULL;

  res->nifl = run->nifl;
  res->infl = copyArrayInf (run->infl, run->nifl);
  for (int i=0; i<res->nifl; i++) {
    res->infl[i]->cause = (*new_vars)[getVarName (getInfluCause (run->infl[i]))];
    res->infl[i]->effet = (*new_vars)[getVarName (getInfluEffet (run->infl[i]))];
    res->infl[i]->cd = (*new_conditions)[(int) getInfluCond (run->infl[i])];
    (*new_influences)[(int)run->infl[i]] = res->infl[i];
  }
  
  res->condin = copyLong (run->condin, run->nifl);
  res->condout = copyLong (run->condout, run->nifl);
  
  /* nodes */
  res->ncg = run->ncg;
  if (! isNull (run->cg)) {
    res->cg = copyArrayCGraph (run->cg, getRTNCGraph (run));
    for (int i=0; i<getRTNCGraph (run); i++) {
      if (! isNull (run->cg[i]->vcause))
	res->cg[i]->vcause = allocVInfluence (getCGNCause (res->cg[i]));
      else res->cg[i]->vcause = NULL;
      for (int j=0; j<getCGNCause (res->cg[i]); j++)
	res->cg[i]->vcause[j] = (*new_influences)[(int) run->cg[i]->vcause[j]];
      if (! isNull (run->cg[i]->veffet))
	res->cg[i]->veffet = allocVInfluence (getCGNEffet (res->cg[i]));
      else res->cg[i]->veffet = NULL;
      for (int k=0; k<getCGNEffet (res->cg[i]); k++)
	res->cg[i]->veffet[k] = (*new_influences)[(int) run->cg[i]->veffet[k]];
      (*new_nodes)[(int) run->cg[i]] = res->cg[i];
    }
  }
  else res->cg = NULL;
  
  /* add nodes to variables. */
  for (int i=0; i<getRTNVar (res); i++) {
    res->dico[i]->node = (*new_nodes)[(int) run->dico[i]->node];
  }
  
  res->maxDepth = run->maxDepth;
  res->nit = run->nit;
  if (! isNull (run->iv)) {
    res->iv = copyArrayInitV (run->iv, getRTNInit (run));
    for (int i=0; i<getRTNInit (res); i++) {
      res->iv[i]->var = (*new_vars)[getVarName (run->iv[i]->var)];
    }
  }
  else res->iv = NULL;

  /* TODO : global constraints */
  res->ngc = run->ngc;
  if (! isNull (run->glc)) {
    res->glc = allocVExpr (run->ngc);
    for (int i=0; i<getRTNGlobC (res); i++) {
      if (new_exprs->count ((int)run->glc[i]) == 0)
	res->glc[i] = ce_copyExpr (run->glc[i], new_gvars, new_exprs, new_vars);
      else res->glc[i] = (*new_exprs)[(int)run->glc[i]];
    }
  }
  else res->glc = NULL;
  
  res->inag = copyLong (run->inag, run->ngc); /* Warning : not sure about size */
  res->inited = run->inited;
  res->runGC = run->runGC;
  res->cumTime = run->cumTime;
  res->prec = run->prec;
  res->onlyFaultIso = run->onlyFaultIso;
  res->numMisb = run->numMisb;
  res->numOk = run->numOk;
  
  /* TODO : remaining variables and ... */
  /*if (isNull (run->mbVA))*/
    res->mbVA = NULL;
  /*else res->mbVA = ce_copyMisbVarArray (run->mbVA, run->numMisb, new_vars, new_confs, new_influences);*/
  res->okVA = NULL;
  res->sfVA = NULL;
  res->oldNumMisb = 0;
  res->oldMisbVarArray = NULL;
  res->numNodesHST = 0;
  createLeafListcaen (&res->HSTree);
  createConflictSet (&res->softConflictListcaen);
  if (isNull (run->diagOut))
    res->diagOut = NULL;
  else res->diagOut = mkString (run->diagOut);
 
  /* bridges sets */
  if (! isNull (run->bridgeEvents))
    res->bridgeEvents = new multimap<bool, Atom*, eqbool>(*run->bridgeEvents);
  else res->bridgeEvents = new multimap<bool, Atom*, eqbool>();
  if (! isNull (run->bridgeConds))
    res->bridgeConds = new multimap<bool, Atom*, eqbool>(*run->bridgeConds);
  else res->bridgeConds = new multimap<bool, Atom*, eqbool>();
  if (! isNull (run->bridgeTransitions))
    res->bridgeTransitions = new vector<Transition*>(*run->bridgeTransitions);
  else res->bridgeTransitions = new vector<Transition*>();
  if (! isNull (run->bridgeConflicts)) {
    res->bridgeConflicts = new vector<vector<Atom*>*>();
    for (unsigned int i=0; i<run->bridgeConflicts->size(); i++)
      res->bridgeConflicts->push_back (new vector<Atom*> (*(*run->bridgeConflicts)[i]));
  }
  else res->bridgeConflicts = NULL;

  /* validity space */
  res->valSpc = copyLong (run->valSpc, run->ncdc);
  res->inac = copyLong (run->inac, run->ncdc);

  delete new_vars;
  delete new_conditions;
  delete new_influences;
  delete new_nodes;
  delete new_gvars;
  delete new_exprs;
  delete new_confs;

  return res;
}

condition ce_copyCondition (condition cd, hash_map<const char*, variable, hash<const char*>, eqstr>* n_v, hash_map<int, condition, hash<int> >* n_c, hash_map<int, expr, hash<int> >* n_e) {
  condition op1, op2;
  condition cdc = copyCondition (cd);

  if (! isNull (getCondVar1 (cd)))
    cdc->var1 = (*n_v)[getVarName (getCondVar1 (cd))];
  else cdc->var1 = NULL;
  if (! isNull (getCondVar2 (cd)))
    cdc->var2 = (*n_v)[getVarName (getCondVar2 (cd))];
  else cdc->var2 = NULL;
  
  if (! isNull (getCondOp1 (cd))) {
    if (! isNull ((op1 = (*n_c)[(int)getCondOp1 (cd)])))
      cdc->op1 = op1;
    else {
      cdc->op1 = ce_copyCondition (getCondOp1 (cd), n_v, n_c, n_e);
      (*n_c)[(int)getCondOp1 (cd)] = getCondOp1 (cdc);
    }
  }
  else cdc->op1 = NULL;

  if (! isNull (getCondOp2 (cd))) {
    if (! isNull ((op2 = (*n_c)[(int)getCondOp2 (cd)])))
      cdc->op2 = op2;
    else {
      cdc->op2 = ce_copyCondition (getCondOp2 (cd), n_v, n_c, n_e);
      (*n_c)[(int)getCondOp2 (cd)] = getCondOp2 (cdc);
    }
  }
  else cdc->op2 = NULL;

  if (! isNull (cd->e))
    cdc->e = (*n_e)[(int) cd->e];
  else cdc->e = NULL;

  (*n_c)[(int) cd] = cdc;

  return cdc;
}

expr ce_copyExpr (expr e, hash_map<int, gVar, hash<int> >* ngvars, hash_map<int, expr, hash<int> >* nexprs, hash_map<const char*, variable, hash<const char*>, eqstr>* nvars) {
  expr op1, op2;
  expr ec = copyExpr (e);
  
  if (! isNull (e->gv)) {
    if (ngvars->count ((int)e->gv) != 0)
      ec->gv = (*ngvars)[(int)e->gv];
    else {
      ec->gv = copyGVar (e->gv);
      if (! isNull (e->gv->uv.var))
	ec->gv->uv.var = (*nvars)[getVarName (e->gv->uv.var)];
      else ec->gv->uv.var = NULL;
    }
  }
  else ec->gv = NULL;

  if (! isNull (getExprOp1 (e))) {
    if (!isNull ((op1 = (*nexprs)[(int)getExprOp1 (e)])))
      ec->op1 = op1;
    else {
      ec->op1 = ce_copyExpr (getExprOp1 (e), ngvars, nexprs, nvars);
      (*nexprs)[(int)getExprOp1 (e)] = getCondOp1 (ec);
    }
  }
  else ec->op1 = NULL;

  if (! isNull (getExprOp2 (e))) {
    if (!isNull ((op2 = (*nexprs)[(int)getExprOp2 (e)])))
      ec->op2 = op2;
    else {
      ec->op2 = ce_copyExpr (getExprOp2 (e), ngvars, nexprs, nvars);
      (*nexprs)[(int)getExprOp2 (e)] = getCondOp2 (ec);
    }
  }
  else ec->op2 = NULL;

  (*nexprs)[(int)e] = ec;

  return ec;
}

/*misbVarArray ce_copyMisbVarArray (misbVarArray mva, long nm, hash_map<const char*, variable, hash<const char*>, eqstr>* nvars, hash_map<int, ptConflict, hash<int> >* nconfs, hash_map<int, influence, hash<int> >* ninfs) {
  misbVarArray res = (misbVarArray) new_malloc (nm * sizeof (ptMisbVar));

  for (long i=0; i<nm; i++) {
    res[i] = allocMisbVar ();
    setVarMisbVar (res[i], (*nvars)[getVarName (getVarMisbVar (mva[i]))]);
    if (isNull (mva[i]->hardConflSet))
      res[i]->hardConflSet = NULL;
    else ce_copyConflictSet (res[i]->hardConflSet, mva[i]->hardConflSet, nconfs, ninfs);
    if (isNull (mva[i]->conflSetListcaen))
      res[i]->conflSetListcaen = NULL;
    else ce_copyConflictSetListcaen (res[i]->conflSetListcaen, mva[i]->conflSetListcaen);
  }
  return res;
}*/

/* pnew and old are unused (for code reduction) */
/*ELEMENT ce_copyConflict (ptListcaen pnew, listcaen old, ELEMENT e, hash_map<int, ptConflict, hash<int> >* nconfs, hash_map<int, influence, hash<int> >* ninfs) {
  ptConflict cn;
  ptConflict conf = (ptConflict) e;
  
  if (isNull ((cn = (*nconfs)[(int)conf]))) {
    cn = copyConflict (conf);
      setConflictInfluence (cn, (*ninfs)[(int)getConflictInfluence (conf)]);
      (*nconfs)[(int)conf] = cn;
  }
  return (ELEMENT) cn;
}*/

/* copy a listcaen of conflict. e unused */
/*ELEMENT ce_copyConflictSet (ptListcaen pnew, listcaen old, ELEMENT e, hash_map<int, ptConflict, hash<int> >* nconfs, hash_map<int, influence, hash<int> >* ninfs) {
  int i;

  if (createEmpty (pnew) == MEM_ERROR)
    return NULL;;
  
  for (i=1, i<=listcaenLength (old); i++) {
    if (addRightInsert (pnew, old, old.first->datum, nconfs, ninfs, &ce_copyConflict) == MEM_ERROR)
      return NULL;
    old.first = old.first->next;
  }
  return (ELEMENT) pnew;
}*/

/* copy a listcaen of conflictSet (listcaen of conflict) ! */
/*ELEMENT ce_copyConflicSetListcaen (ptListcaen pnew, listcaen old, hash_map<int, ptConflict, hash<int> >* nconfs, hash_map<int, influence, hash<int> >* ninfs) {
  int i;
  ptListcaen p2new;

  if (createEmpty (pnew) == MEMERROR)
    return MEMERROR;
  
  for (i=1; i<listcaenLength (old); i++) {
    if (addRightInsert (pnew, p2new, (listcaen) old.first->datum, NULL, nconfs, ninfs, &ce_copyConflictSet) == MEM_ERROR)
      return NULL;
    old.first = old.first->next;
  }
  return (ELEMENT) pnew;
}

STATUS addRightInsert (ptListcaen pL, ptListcaen pnew, listcaen old, ELEMENT e, hash_map<int, ptConflict, hash<int> >* nconfs, hash_map<int, influence, hash<int> >* ninfs, ELEMENT (*ptFuncConf) (ELEMENT, hash_map<int, ptConflict, hash<int> >*, hash_map<int, influence, hash<int> >*)) {
  void *allocAux;

  if ((allocAux = new_malloc(sizeof(node))) == NULL)
      return MEM_ERROR;

  if (isEmptyListcaen(*pL)) {

    pL->first = (ptNode) allocAux;
    pL->last = pL->first;
    
    pL->first->datum = ptFuncConf (pnew, old, e, nconfs, ninfs);
    pL->first->next = NULL;
  }
  else {

    pL->last->next = (ptNode) allocAux;
    pL->last = pL->last->next;

    pL->last->datum = ptFuncConf (pnew, old, e, nconfs, ninfs);;
    pL->last->next = NULL;
  }
  pL->nEl++;
  return OK;
}*/

runtime ce_copyPartialFaultyRt (runtime rt, int discrete_fault_time, double cont_fault_time, double stime, int twd) {
  variable v;
  runtime res_rt = ce_copyRuntime (rt);  /* full copy */
  
  if (discrete_fault_time < rt->ts - twd) { /* the fault isn't in the temporal window */
    fprintf (stdout, "The fault isn't in the discrete temporal window !\n");
    exit (1);
    /* what to do ? Beware when activating it again: used in applyMappingFunction (koala-runt) */
  }
  else if (cont_fault_time < stime - rt->twc) {
    std::cerr << "discrete fault time: " << discrete_fault_time << endl;
    std::cerr << "cont_fault_time: " << cont_fault_time << " -- stime: " << stime << "-- twc: " << rt->twc << endl;
    fprintf (stdout, "The fault isn't in the continuous temporal window !\n");
    exit (1);
  }
  else {
    for (long i=0; i<getRTNVar (res_rt); i++) {
      v = getRTDicoNth (res_rt, i);
      moveHistory (res_rt, v, cont_fault_time, stime);
    }
  }
    return res_rt;
}

void moveHistory (runtime rt, variable v, double fault_time, double stime) {
  long fault_pos = (long) stime - (long) fault_time;
  long j=0;

  /*cout << "stime: " << stime << endl;
    cout << "fault_time: " << fault_time << endl;
    cout << "moving history: " << getVarName (v) << " -- fault position: " << fault_pos << endl;
    ce_printVarHistory (stdout, v, rt);
    cout << endl;*/

  if (fault_pos == 0) return;

  if (! isParameter (v)) {
    ce_putObjet (varHistoNth (v, fault_pos-1), varNewValue (v));
    ce_putObjet (varHistoNth (v, fault_pos), varOldValue (v));
  }

  for (long i=fault_pos+1; i<getVarHistoLn (v); i++) {  /* moves history */
    ce_putObjet (varHistoNth (v, i), varHistoNth (v, j));
    ce_putObjet (objZero, varHistoNth (v, i)); /* TODO : get rid of the zero at the end of the history ? */
    j++;
  }
  
  if (ce_isMeasure (rt, v)) { /* move measures history */
    j = 0;
    ce_putObjet (varHistoMesNth (v, fault_pos-1), varMesValue (v));
    for (long i=fault_pos; i<getVarHistoLn (v); i++) {
      ce_putObjet (varHistoMesNth (v, i), varHistoMesNth (v, j));
      ce_putObjet (objZero, varHistoMesNth (v, i));
      if (! isNull (getVarHistoMisb (v))) {
	getVarHistoMisbNth (v, j) = getVarHistoMisbNth (v, i);
	getVarHistoMisbNth (v, i) = 0;
      }
      isVarHistoLowAlrm (v, j) = isVarHistoLowAlrm (v, i);
      isVarHistoLowAlrm (v, i) = 0;
      isVarHistoHighAlrm (v, j) = isVarHistoHighAlrm (v, i);
      isVarHistoHighAlrm (v, i) = 0;
      j++;
    }
  }

  /*ce_printVarHistory (stdout, v, rt);
    cout << endl;*/

}

/**
 * get a variable from the runtime directory.
 */
variable ce_getVarDico (runtime rt, const char* nv) {
  for (long i=0; i<getRTNVar (rt); i++)
    if (strcmp (nv, getVarName (getRTDicoNth (rt, i))) == 0)
      return getRTDicoNth (rt, i);
  return 0;
}

/**
 * eval and print real value to fd.
 */
void ce_printReal (FILE *fd, double vd) {

  if (stderr == fd) {
    if (-DBL_MAX == (vd))
      fprintf (fd, "-oo");
    else if (DBL_MAX == vd)
      fprintf (fd, "+oo");
    else
      fprintf (fd, "%g", vd);
  }
  else /* stderr */
    fprintf (fd, "%g", vd);
  return;
}

/**
 * eval and print range value to fd.
 */
void ce_printRange (FILE *fd, range r) {

  if (fd == stderr) {
    double vd;
    if (isIncludedInf (r))
      fprintf (fd, " [");
    else fprintf (fd, "]");
    if (-DBL_MAX == (vd = getRangeInf (r)))
      fprintf (fd, "-oo");
    else
      fprintf (fd, "%g", vd);
    fprintf (fd, ", ");
    if (DBL_MAX == (vd = getRangeSup (r)))
      fprintf (fd, "+oo");
    else
      fprintf (fd, "%g", vd);
    if (isIncludedSup (r))
      fprintf (fd, "]");
    else fprintf (fd, "[");
  }
  else
    fprintf (fd, "%g %g", getRangeInf (r), getRangeSup (r));
  return;
}

/**
 * eval and print object value to fd.
 */
void ce_printObjet (FILE *fd, objet ob) {

  switch (getObjType(ob)) {
  case CE_REAL :
    if (fd == stderr)
      fprintf (fd, " =");
    ce_printReal (fd, getObjRealVal (ob));
    break;
  case CE_RANGE :                                              /* in [o1, o2] */
    if (fd == stderr)
      fprintf (fd, " in ");
    ce_printRange (fd, getObjRangeVal (ob));
    break;
  default:
    fprintf (fd, "type ?\n");
    break;
  }
  return;
}

/**
 * copy p values into q.
 */
void ce_putObjet (objet p, objet q) {
  enum objType t=getObjType(p);

  setObjType (q, t);
  switch (t) {
    case CE_REAL :
      setObjRealVal (q, getObjRealVal (p));
      break;
    case CE_RANGE :
      setObjRangeVal (q, getObjRangeVal (p));
      break;
    case CE_BOOL :
      setObjBoolVal (q, getObjBoolVal (p));
      break;
  default:
    break;
  }
  return;
}

/**
 * add an object b to an object a and update valeur.
 */
void ce_addObjet (objet a, objet b) {
  enum objType ta=getObjType(a), tb=getObjType(b);

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    if (tb == CE_REAL)
      setObjRealVal (a, da + getObjRealVal (b));
    else { /* tb */
      double binf=getRangeInf(getObjRangeVal(b)),
	bsup=getRangeSup(getObjRangeVal(b));
      setObjType (a, CE_RANGE);
      setRangeInf (getObjRangeVal (a), da + binf);
      setRangeSup (getObjRangeVal (a), da + bsup);
      setBounds (getObjRangeVal (a), getObjRangeVal (b));
    } /* endif tb */
  }
  else { /* ta */
    double ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    if (tb == CE_REAL) {
      double db=getObjRealVal(b);
      setRangeInf (getObjRangeVal (a), ainf + db);
      setRangeSup (getObjRangeVal (a), asup + db);
    }
    else { /* tb */
      double binf=getRangeInf(getObjRangeVal(b)),
	bsup=getRangeSup(getObjRangeVal(b));
      setRangeInf (getObjRangeVal (a), ainf + binf);
      setRangeSup (getObjRangeVal (a), asup + bsup);
    } /* endif tb */
  } /* endif ta */
  return;
}

/**
 * substract an object b to an object a and update 'valeur' of a.
 */
void ce_subObjet (objet a, objet b) {
  enum objType ta=getObjType(a), tb=getObjType(b);

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    if (tb == CE_REAL)
      setObjRealVal (a, da - getObjRealVal (b));
    else { /* tb */
      double binf=getRangeInf(getObjRangeVal(b)),
	bsup=getRangeSup(getObjRangeVal(b));
      setObjType (a, CE_RANGE);
      setRangeInf (getObjRangeVal (a), da - binf);
      setRangeSup (getObjRangeVal (a), da - bsup);
      setBounds (getObjRangeVal (a), getObjRangeVal (b));
    } /* endif tb */
  }
  else { /* ta */
    double ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    if (tb == CE_REAL) {
      double db=getObjRealVal(b);
      setRangeInf (getObjRangeVal (a), ainf - db);
      setRangeSup (getObjRangeVal (a), asup - db);
    }
    else { /* tb */
      double binf=getRangeInf(getObjRangeVal(b)),
	bsup=getRangeSup(getObjRangeVal(b));
      setRangeInf (getObjRangeVal (a), ainf - bsup);
      setRangeSup (getObjRangeVal (a), asup - binf);
    } /* endif tb */
  } /* endif ta */
  return;
}

/**
 * reverse the sign of an object value.
 */
void ce_minusObjet (objet a) {
  enum objType ta=getObjType(a);

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    setObjRealVal (a, -da);
  }
  else { /* ta */
    double ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    setRangeInf (getObjRangeVal (a), -asup);
    setRangeSup (getObjRangeVal (a), -ainf);
    /* inclusion */
    if (isIncludedInf (getObjRangeVal (a)))
      setIncludedSup (getObjRangeVal (a));
    else setNotIncludedSup (getObjRangeVal (a));
    if (isIncludedSup (getObjRangeVal (a)))
      setIncludedInf (getObjRangeVal (a));
    else setNotIncludedInf (getObjRangeVal (a));
  } /* endif ta */
  return;
}

/**
 * set an object value to its exponential value.
 */
void ce_expObjet (objet a) {
  enum objType ta=getObjType(a);

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    setObjRealVal (a, exp (da));
  }
  else { /* ta */
    double ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    setRangeInf (getObjRangeVal (a), exp (ainf));
    setRangeSup (getObjRangeVal (a), exp (asup));
  } /* endif ta */
  return;
}

/**
 * set an object value to its log value.
 */
void ce_logObjet (objet a) {
  enum objType ta=getObjType(a);

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    setObjRealVal (a, log (da));
  }
  else { /* ta */
    double ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    setRangeInf (getObjRangeVal (a), log (ainf));
    setRangeSup (getObjRangeVal (a), log (asup));
  } /* endif ta */
  return;
}

/**
 * take the sqrt of an object value.
 */
void ce_sqrtObjet (objet a) {
  enum objType ta=getObjType(a);

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    setObjRealVal (a, sqrt (da));
  }
  else { /* ta */
    double ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    setRangeInf (getObjRangeVal (a), sqrt (ainf));
    setRangeSup (getObjRangeVal (a), sqrt (asup));
  }
  return;
}

/* sinus optimization fct pointer */
double (*sin_cost_ptr) (int*, double*, double*, double*, double*, int*, int*, int*);

/* sinus optimization function */
double sin_cost (int *n, double *X, double *Xmin, double *Xmax, double *duser, int *iuser, int *iret, int *ier) {
  ier[0] = 0;
  if (duser[0] == 0)
    return sin (X[0]);
  else return -sin (X[0]);
}

/**
 * take the sin of an objet value.
 */
void ce_sinObjet (objet a ) {
  int *n, *iuser=0, *iret, *ier;
  double *X, *Xmin, *Xmax, *duser;
  enum objType ta=getObjType(a);

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    setObjRealVal (a, sin(da));
  }
  else { /* ta */
    double sinf, sinup, binf=getRangeInf(getObjRangeVal(a)),
      bsup=getRangeSup(getObjRangeVal(a));
    
    sin_cost_ptr = &sin_cost;
    n = (int*) malloc (sizeof (int));
    n[0] = 1;
    X = (double*) malloc (sizeof (double));
    X[0] = 0.5 * (binf + bsup);
    Xmin = (double*) malloc (sizeof (double));
    Xmin[0] = binf;
    Xmax = (double*) malloc (sizeof (double));
    Xmax[0] = bsup;
    iret = (int*) malloc (sizeof (int));
    ier = (int*) malloc (sizeof (int));
    duser = (double*) malloc (sizeof (double));

    duser[0] = 0;

    opt_nelder (sin_cost_ptr, n, X, Xmin, Xmax, duser, iuser, ier);
    sinf = sin_cost_ptr (n, X, Xmin, Xmax, duser, iuser, iret, ier);

    duser[0] = 1;
    X[0] = 0.5 * (binf + bsup);
    opt_nelder (sin_cost_ptr, n, X, Xmin, Xmax, duser, iuser, ier);
    sinup = -sin_cost_ptr (n, X, Xmin, Xmax, duser, iuser, iret, ier);

    /*sinf = sin(binf);
      sinup = sin(bsup);*/
    
    setRangeInf (getObjRangeVal (a), min (sinf, sinup));
    setRangeSup (getObjRangeVal (a), max (sinf, sinup));

    free (n);
    free (X);
    free (Xmax);
    free (Xmin);
    free (iret);
    free (ier);
    free (duser);
  }
  return;
}

/* cosinus optimization fct pointer */
double (*cos_cost_ptr) (int*, double*, double*, double*, double*, int*, int*, int*);

/* cosinus optimization function */
double cos_cost (int *n, double *X, double *Xmin, double *Xmax, double *duser, int *iuser, int *iret, int *ier) {
  ier[0] = 0;
  if (duser[0] == 0)
    return cos (X[0]);
  else return -cos (X[0]);
}

/**
 * take the cos of an objet value.
 */
void ce_cosObjet (objet a ) {
  int *n, *iuser=0, *iret, *ier;
  double *X, *Xmin, *Xmax, *duser;
  enum objType ta=getObjType(a);

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    setObjRealVal (a, cos(da));
  }
  else { /* ta */
    double cinf, cosup, binf=getRangeInf(getObjRangeVal(a)),
      bsup=getRangeSup(getObjRangeVal(a));

    cos_cost_ptr = &cos_cost;
    n = (int*) malloc (sizeof (int));
    n[0] = 1;
    X = (double*) malloc (sizeof (double));
    X[0] = 0.5 * (binf + bsup);
    Xmin = (double*) malloc (sizeof (double));
    Xmin[0] = binf;
    Xmax = (double*) malloc (sizeof (double));
    Xmax[0] = bsup;
    iret = (int*) malloc (sizeof (int));
    ier = (int*) malloc (sizeof (int));
    duser = (double*) malloc (sizeof (double));

    duser[0] = 0;
    opt_nelder (cos_cost_ptr, n, X, Xmin, Xmax, duser, iuser, ier);
    cinf = cos_cost_ptr (n, X, Xmin, Xmax, duser, iuser, iret, ier);

    duser[0] = 1;
    X[0] = 0.5 * (binf + bsup);
    opt_nelder (cos_cost_ptr, n, X, Xmin, Xmax, duser, iuser, ier);
    cosup = -cos_cost_ptr (n, X, Xmin, Xmax, duser, iuser, iret, ier);

    /*cinf = cos(binf);
      cosup = cos(bsup);*/
    setRangeInf (getObjRangeVal (a), min (cinf, cosup));
    setRangeSup (getObjRangeVal (a), max (cinf, cosup));
  
    free (n);
    free (X);
    free (Xmax);
    free (Xmin);
    free (iret);
    free (ier);
    free (duser);
  }
  return;
}

/**
 * set a as a*b.
 */
void ce_mulObjet (objet a, objet b) {
  enum objType ta=getObjType(a), tb=getObjType(b);

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    if (da != 0.0) {
      if (tb == CE_REAL) {
	setObjRealVal (a, da * getObjRealVal (b));
      }
      else { /* tb */
	double dc, dd, de, binf=getRangeInf(getObjRangeVal(b)),
	  bsup=getRangeSup(getObjRangeVal(b));
	setObjType (a, CE_RANGE);
	dc = da*binf, dd = da*bsup;
	if (dc > dd)
	  de = dc, dc = dd, dd = de;
	setRangeInf (getObjRangeVal (a), dc);
	setRangeSup (getObjRangeVal (a), dd);
	setBounds (getObjRangeVal (a), getObjRangeVal (b));
      } /* endif tb */
    }
  }
  else { /* ta */
    double dc, dd, de, ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    if (tb == CE_REAL) {
      double db=getObjRealVal(b);
      if (db == 0.0) {
	setObjType (a, CE_REAL);               /* ??? range ??? was bug ?? */
	setObjRealVal (a, 0.0);
      }
      else { /* db */
	setObjType (a, CE_RANGE);
	dc = db*ainf, dd = db*asup;
	if (dc > dd)
	  de = dc, dc = dd, dd = de;
	setRangeInf (getObjRangeVal (a), dc);
	setRangeSup (getObjRangeVal (a), dd);
      }
    }
    else { /* tb */
      double df, dmin, dmax, binf=getRangeInf(getObjRangeVal(b)),
	bsup=getRangeSup(getObjRangeVal(b));
      setObjType (a, CE_RANGE);
      dc = ainf*binf, dd = ainf*bsup, de = asup*binf, df = asup*bsup;
      dmin = min (dc, min (dd, min (de, df)));
      dmax = max (dc, max (dd, max (de, df)));
      setRangeInf (getObjRangeVal (a), dmin);
      setRangeSup (getObjRangeVal (a), dmax);
    } /* endif tb */
  } /* endif ta */
  return;
}

/**
 * set a as a/b and return 0 if the division is impossible (1 otherwise).
 */
long ce_divObjet (objet a, objet b) {
  enum objType ta=getObjType(a), tb=getObjType(b);
  long r=1;

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    if (tb == CE_REAL) {
      double db=getObjRealVal(b);
      if (db == 0.0)
	r = 0;                                       /* division impossible */
      else
	setObjRealVal (a, da/db);
    }
    else { /* tb */
      double dc, dd, de, binf=getRangeInf(getObjRangeVal(b)),
	bsup=getRangeSup(getObjRangeVal(b));
      if (binf*bsup <= 0.0)
	r = 0;                                       /* division impossible */
      else {
	setObjType (a, CE_RANGE);
	dc = da/bsup, dd = da/binf;
	if (dc > dd)
	  de = dc, dc = dd, dd = de;
	setRangeInf (getObjRangeVal (a), dc);
	setRangeSup (getObjRangeVal (a), dd);
	setBounds (getObjRangeVal (a), getObjRangeVal (b));
      }
    } /* endif tb */
  }
  else { /* ta */
    double dc, dd, de, ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    if (tb == CE_REAL) {
      double db=getObjRealVal(b);
      if (db == 0.0)
	r = 0;                                       /* division impossible */
      else {
	setObjType (a, CE_RANGE);
	dc = ainf/db, dd = asup/db;
	if (dc > dd)
	  de = dc, dc = dd, dd = de;
	setRangeInf (getObjRangeVal (a), dc);
	setRangeSup (getObjRangeVal (a), dd);
      }
    }
    else { /* tb */
      double df, dmin, dmax, binf=getRangeInf(getObjRangeVal(b)),
	bsup=getRangeSup(getObjRangeVal(b));
      if (binf*bsup <= 0.0)
	r = 0;                                       /* division impossible */
      else {
	setObjType (a, CE_RANGE);
	dc = ainf/binf, dd = ainf/bsup, de = asup/binf, df = asup/bsup;
	dmin = min (dc, min (dd, min (de, df)));
	dmax = max (dc, max (dd, max (de, df)));
	setRangeInf (getObjRangeVal (a), dmin);
	setRangeSup (getObjRangeVal (a), dmax);
      }
    } /* endif tb */
  } /* endif ta */
  return r;
}

/**
 * intersect two objects values and returns CE_FALSE.
 * otherwise result is in first argument.
 */
boolean ce_interObjet (objet a, objet b) {       /* returns CE_FALSE */
  enum objType ta=getObjType(a), tb=getObjType(b);    /* if intersection empty */
  boolean r=CE_TRUE;

  /*cerr << "interObjet:\n";
    ce_printObjet (stderr, a);
    cerr << " -- ";
    ce_printObjet (stderr, b);
    cerr << endl;*/

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    if (tb == CE_REAL) {
      double db=getObjRealVal(b);
      if (da != db)
	r = CE_FALSE;
    }
    else { /* tb */
      if (! isInRange (da, getObjRangeVal (b)))
	r = CE_FALSE;
    } /* endif tb */
  }
  else { /* ta */
    double ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    if (tb == CE_REAL) {
      double db=getObjRealVal(b);
      if (isInRange (db, getObjRangeVal (a))) {
	setObjType (a, CE_REAL);
	setObjRealVal (a, db);
      }
      else
	r = CE_FALSE;
    }
    else { /* tb */
      double dmin, dmax, binf=getRangeInf(getObjRangeVal(b)),
	bsup=getRangeSup(getObjRangeVal(b));
      dmin = max (ainf, binf), dmax = min (asup, bsup);
      if (dmin < dmax) {
	setRangeInf (getObjRangeVal (a), dmin);
	setRangeSup (getObjRangeVal (a), dmax);
      }
      else if (dmin == dmax) {
	if ((((dmin == ainf) && isIncludedInf (getObjRangeVal (a)))
	     && ((dmax == asup) && isIncludedSup (getObjRangeVal (a))))
	    || (((dmin == ainf) && isIncludedInf (getObjRangeVal (a)))
		&& ((dmax == bsup) && isIncludedSup (getObjRangeVal (b))))
	    || (((dmin == binf) && isIncludedInf (getObjRangeVal (b)))
		&& ((dmax == asup) && isIncludedSup (getObjRangeVal (a))))
	    || (((dmin == binf) && isIncludedInf (getObjRangeVal (b)))
		&& ((dmax == bsup) && isIncludedSup (getObjRangeVal (b))))) {

	  setObjType (a, CE_REAL);
	  setObjRealVal (a, dmin);
	}
	else r = CE_FALSE;
      }
      else
	r = CE_FALSE;
    } /* endif tb */
  } /* endif ta */
  /*cerr << "interObjet:\n";
    ce_printObjet (stderr, a);
    cerr << endl;*/
  return r;
}

/**
 * set first object argument as the union of both objects.
 */
void ce_uniObjet (objet a, objet b) {
  enum objType ta=getObjType(a), tb=getObjType(b);

  /*cerr << "uniObjet:\n";
    ce_printObjet (stderr, a);
    cerr << " -- ";
    ce_printObjet (stderr, b);
    cerr << endl;*/

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    if (tb == CE_REAL) {
      double db=getObjRealVal(b);
      if (da != db) {
	setObjType (a, CE_RANGE);
	setRangeInf (getObjRangeVal (a), min (da, db));
	setRangeSup (getObjRangeVal (a), max (da, db));
      } /* endif da != db */
    }
    else { /* tb */                       /* inclusion on bounds are unchanged */
      ce_putObjet (b, a);
      if (! isInRange (da, getObjRangeVal (b))) {
	double bsup=getRangeSup(getObjRangeVal(b));
	if (da > bsup)
	  setRangeSup (getObjRangeVal (a), da);
	else
	  setRangeInf (getObjRangeVal (a), da);
      } /* endif isInRange da */
    } /* endif tb */
  }
  else { /* ta */                        
    double ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    if (tb == CE_REAL) {
      double db=getObjRealVal(b);
      if (! isInRange (db, getObjRangeVal (a))) {
	if (db > asup)
	  setRangeSup (getObjRangeVal (a), db);
	else if (db == asup)   /* inclusion pb */
	  setIncludedSup (getObjRangeVal (a));
	else if (db == ainf)  /* inclusion pb */
	  setIncludedInf (getObjRangeVal (a));
	else setRangeInf (getObjRangeVal (a), db);
      } /* endif isInRange */
    }
    else { /* tb */                     /* inclusion on bounds are unchanged */
      double dmin, dmax, binf=getRangeInf(getObjRangeVal(b)),
	bsup=getRangeSup(getObjRangeVal(b));
      dmin = min (ainf, binf), dmax = max (asup, bsup);
      setRangeInf (getObjRangeVal (a), dmin);
      setRangeSup (getObjRangeVal (a), dmax);
    } /* endif tb */
  } /* endif ta */
  /*cerr << "uniObjet:\n";
    ce_printObjet (stderr, a);
    cerr << endl;*/
  return;
}

/**
 * return the sign of the object value (-1 neg, 1 pos, 0 unknown).
 */
long ce_sgnObjet (objet a) {      /* retourne 1 si positif */
  enum objType ta=getObjType(a);  /*               -1 si negatif */
  long s=0;                       /*               0 sinon (nul, ou indetermine' */

  if (ta == CE_REAL) {
    double da=getObjRealVal(a);
    if (da > 0.0)
      s = 1;
    else if (da < 0.0)
      s = -1;
  }
  else { /* ta */
    double ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    if ((ainf >= 0.0) && (asup > 0.0))
      s = 1;
    else if ((ainf < 0.0) && (asup <= 0.0))
      s = -1;
  }
  return s;
}

/**
 * test if two objects have the same values.
 */
boolean ce_eqObjet (objet a, objet b) {
  boolean r=CE_FALSE;

  if ((getObjType (a) == CE_REAL) && (getObjType (b) == CE_REAL)
      && (getObjRealVal (a) == getObjRealVal (b)))
    r = CE_TRUE;
  else if ((getObjType (a) == CE_RANGE) && (getObjType (b) == CE_RANGE)
	   && (getRangeSup (getObjRangeVal (a)) == getRangeSup (getObjRangeVal (b)))
	   && (getRangeInf (getObjRangeVal (a)) == getRangeInf (getObjRangeVal (b)))
	   && (getBoundInf (getObjRangeVal (a)) == getBoundInf (getObjRangeVal (b)))
	   && (getBoundSup (getObjRangeVal (a)) == getBoundSup (getObjRangeVal (b))))
    r = CE_TRUE;
  return r;
}

/**
 * test if two objects have different types or values.
 */
boolean ce_diffObjet (objet a, objet b) {
  boolean r=CE_FALSE;

  if (CE_REAL == getObjType (a)) {
    double da=getObjRealVal(a);
    if (CE_REAL == getObjType (b)) {
      if (da != getObjRealVal (b))
	r = CE_TRUE;
    }
    else { /* CE_REAL b */
      range rb=getObjRangeVal(b);
      if ((da < getRangeInf (rb)) || (da > getRangeSup (rb)))
	r = CE_TRUE;
    } /* endif CE_REAL b */
  }
  else { /* CE_REAL a */
    range ra=getObjRangeVal(a);
    if (CE_REAL == getObjType (b)) {
      double db=getObjRealVal(b);
      if ((db < getRangeInf (ra)) || (db > getRangeSup (ra)))
	r = CE_TRUE;
    }
    else { /* CE_REAL b */
      range rb=getObjRangeVal(b);
      if ((getRangeInf (ra) > getRangeSup (rb))
	  || (getRangeInf (rb) > getRangeSup (ra))
	  || (getBoundInf (rb) != getBoundInf (ra))
	  || (getBoundSup (rb) != getBoundSup (ra)))
	r = CE_TRUE;
    } /* endif CE_REAL b */
  } /* endif CE_REAL a */
  return r;
}

/**
 * test if object a value is lower than object b value.
 */
boolean ce_ltObjet (objet a, objet b) {
  boolean r=CE_FALSE;

  if (CE_REAL == getObjType (a)) {
    double da=getObjRealVal(a);
    if (CE_REAL == getObjType (b)) {
      if (da < getObjRealVal (b))
	r = CE_TRUE;
    }
    else { /* CE_REAL b */
      range rb=getObjRangeVal(b);
      if ((da < getRangeInf (rb)))
	r = CE_TRUE;
    } /* endif CE_REAL b */
  }
  else { /* CE_REAL a */
    range ra=getObjRangeVal(a);
    if (CE_REAL == getObjType (b)) {
      double db=getObjRealVal(b);
      if (getRangeSup (ra) < db)
	r = CE_TRUE;
    }
    else { /* CE_REAL b */
      range rb=getObjRangeVal(b);
      if (getRangeSup (ra) < getRangeInf (rb))
	r = CE_TRUE;
    } /* endif CE_REAL b */
  } /* endif CE_REAL a */
  return r;
}

/**
 * test if object a value is greater than object b value.
 */
boolean ce_geObjet (objet a, objet b) {
  boolean r=CE_FALSE;

  if (CE_REAL == getObjType (a)) {
    double da=getObjRealVal(a);
    if (CE_REAL == getObjType (b)) {
      if (da >= getObjRealVal (b))
	r = CE_TRUE;
    }
    else { /* CE_REAL b */
      range rb=getObjRangeVal(b);
      if ((da > getRangeSup (rb))
	  || ((da == getRangeSup (rb))
	      && isIncludedSup (rb)))
	r = CE_TRUE;
    } /* endif CE_REAL b */
  }
  else { /* CE_REAL a */
    range ra=getObjRangeVal(a);
    if (CE_REAL == getObjType (b)) {
      double db=getObjRealVal(b);
      if ((getRangeInf (ra) > db)
	  || ((db == getRangeInf (ra))
	      && isIncludedInf (ra)))
	r = CE_TRUE;
    }
    else { /* CE_REAL b */
      range rb=getObjRangeVal(b);
      if ((getRangeInf (ra) > getRangeSup (rb))
	  || ((getRangeInf (ra) == getRangeSup (rb))
	      && isIncludedSup (rb)))
	r = CE_TRUE;
    } /* endif CE_REAL b */
  } /* endif CE_REAL a */
  return r;
}

/**
 * test if object a value is in object b value.
 */
boolean ce_inObjet (objet a, objet b) {
  boolean r=CE_FALSE;

  /*cerr << "a ";
    ce_printObjet (stderr, a);
    cerr<< "\nb ";
    ce_printObjet (stderr, b);
    cerr << endl;*/
  
  if (CE_RANGE == getObjType (b)) {
    range rb=getObjRangeVal(b);
    if (CE_REAL == getObjType (a)) {
      double da=getObjRealVal(a);
      if (((da > getRangeInf (rb)) && (da < getRangeSup (rb)))
	  || (((da == getRangeInf (rb)) && isIncludedInf (rb))
	      && (da < getRangeSup (rb)))
	  || (((da == getRangeSup (rb)) && isIncludedSup (rb))
	      && (da > getRangeInf (rb))))
	r = CE_TRUE;
    }
    else { /* CE_REAL a */
      range ra=getObjRangeVal(a);
      if ((((getRangeInf (ra) > getRangeInf (rb))
	    || ((getRangeInf (ra) == getRangeInf (rb))
		&& (isIncludedInf (rb)
		    || (!isIncludedInf (ra) && !isIncludedInf (rb)))))
	   && ((getRangeSup (ra) < getRangeSup (rb))))
	  || ((getRangeInf (ra) > getRangeInf (rb))
	      && ((getRangeSup (ra) == getRangeSup (rb)))
	      && (isIncludedSup (rb)
		  || (!isIncludedSup (ra) && !isIncludedSup (ra)))))
	r = CE_TRUE;
    } /* endif CE_REAL a */
  } /* endif CE_RANGE b */
  return r;
}

objet ce_mkSpObjet (double inf, double sup) {
  range rr = {inf, sup};
  objet res = (objet) new_malloc (sizeof (s_Objet));
  setObjType (res, (objType)2);
  setObjRangeVal (res, rr);
  return res;
}

/**
 * test if object a value is strictly included in object b value 
 * and compute the resulting intervals.
 */
objet* ce_inSpObjet (objet a, objet b) {
  objet* res = (objet*) calloc (2, sizeof(objet));

  if (CE_RANGE == getObjType (b)) {
    range rb=getObjRangeVal(b);
    if (CE_REAL == getObjType (a)) {
      double da=getObjRealVal(a);
      if (da > getRangeInf (rb)) {
	if (da < getRangeSup (rb)) {
	  res[0] = ce_mkSpObjet (getRangeInf (rb), da);
	  res[1] = ce_mkSpObjet (da, getRangeSup (rb));
	}
	else if (da == getRangeSup (rb) && isIncludedSup (rb)) {   /* equal to b */
	  res[0] = ce_mkSpObjet (getRangeInf (rb), da);
	  res[1] = 0;
	  return res;
	}
      }
      else if (da == getRangeInf (rb) && isIncludedInf (rb)) {
	if (da < getRangeSup (rb)) {
	  res[0] = ce_mkSpObjet (da, getRangeSup (rb));
	  res[1] = 0;
	  return res;
	}
      }
    }
    else { /* CE_REAL a */
      range ra=getObjRangeVal(a);
      if (getRangeInf (ra) > getRangeInf (rb)) {
	if (getRangeSup (ra) < getRangeSup (rb)) {
	  res[0] = ce_mkSpObjet (getRangeInf (rb), getRangeInf (ra));
	  res[1] = ce_mkSpObjet (getRangeSup (ra), getRangeSup (rb));
	  return res;
	}
	else if (getRangeSup (ra) == getRangeSup (rb)
		 && ((isIncludedSup (ra) && isIncludedSup (rb))
		     || (!isIncludedSup (ra) && !isIncludedSup (rb)))) {
	  res[0] = ce_mkSpObjet (getRangeInf (rb), getRangeInf (ra));
	  res[1] = 0;
	  return res;
	}
      }
      else if (getRangeInf (ra) == getRangeInf (rb)
	       && ((isIncludedInf (ra) && isIncludedInf (rb))
		   || (!isIncludedInf (ra) && !isIncludedInf (rb)))) {
	if (getRangeSup (ra) < getRangeSup (rb)) {
	  res[0] = ce_mkSpObjet (getRangeSup (ra), getRangeSup (rb));
	  res[1] = 0;
	  return res;
	}
      }
    } /* endif CE_REAL a */
  } /* endif CE_RANGE b */
  return 0;
}

/**
 * print the var name to the file descriptor.
 */
void ce_printVarNom (FILE *fd, variable v) {

  fprintf (fd, "%s", getVarName (v));
  return;
}


/**
 * print the var new value to the file descriptor.
 */
void ce_printVarValueNV (FILE *fd, variable v) {
  if (fd == stderr) {
    fprintf (fd, "  ");
    ce_printVarNom (fd, v);
  }
  if (isDetermined (v)) {
    objet ob=varNewValue(v);
    ce_printObjet (fd, ob);
    if ((fd != stderr) && (CE_REAL == getObjType (ob))) {
      fprintf (fd, " ");
    ce_printObjet (fd, ob);
    }
  } /* endif isDetermined v */
  else
    fprintf (fd, " ?");
  if (fd == stderr)
    fprintf (fd, "\n");
  else
    fprintf (fd, " ");
  return;
}

/**
 * print the var value to the file descriptor.
 */
void ce_printVarValue (FILE *fd, variable v) {

  if (fd == stderr) {
    fprintf (fd, "  ");
    ce_printVarNom (fd, v);
  }
  if (isDetermined (v)) {
    objet ob=varOldValue(v);
    ce_printObjet (fd, ob);
    if ((fd != stderr) && (CE_REAL == getObjType (ob))) {
      fprintf (fd, " ");
    ce_printObjet (fd, ob);
    }
  } /* endif isDetermined v */
  else
    fprintf (fd, " ?");
  if (fd == stderr)
    fprintf (fd, "\n");
  else
    fprintf (fd, " ");
  return;
}

boolean ce_evalVarConds (variable v, runtime rt) {
  /*int i;*/
  boolean sp = CE_TRUE;
  /*boolean sp = CE_FALSE;*/
  
  /*for (i=0; i<getVarNcd (v); i++) {
    if (sp == CE_FALSE)
    sp = ce_splitCond (getRTCondiNth (rt, getVarRelCdNth (v,i)), rt);
    else ce_splitCond (getRTCondiNth (rt, getVarRelCdNth (v,i)), rt);
    ce_printCondValues (stdout, getRTCondiNth (rt, getVarRelCdNth (v,i)));
    }*/
  
  return sp;
}

/**
 * eval condition.
 */
boolean ce_evalCond (condition cd, runtime rt) {
  boolean r1, r2, r=CE_FALSE;
  objet ob1, ob2;
  variable v;

  switch (getCondType (cd)) {
  case CE_CTE :
    r = getCondValue (cd);
    break;
  case CE_EXPR :
    v = getCondVar1 (cd);
    
    if (isNull (ob1 = varOldValue (v)))
      ob1 = varNewValue (v);
    
    /* evaluate the condition w.r.t. the current continuous vector. */
    ce_evalSolFunc (rt, getCondExpr (cd));
  
    ob2 = TMP4;
    if (getCondOpRel (cd) != CE_EQ)
      /* in or equal */
      r = (boolean) ((bool)ce_inObjet (ob1, ob2) || (bool)ce_eqObjet (ob1, ob2));
    else r= ce_eqObjet (ob1, ob2);
    break;
  case CE_COMP :  /* DEPRECATED */
    ob1 = varOldValue (getCondVar1 (cd));
    v = getCondVar2 (cd);
    if (isNull (v))
      ob2 = getCondObj (cd);
    else
      ob2 = varOldValue (v);
    switch (getCondOpRel (cd)) {
    case CE_EQ :
      r = ce_eqObjet (ob1, ob2);
      break;
    case CE_NE :
      r = ce_diffObjet (ob1, ob2);
      break;
    case CE_LE :
      r = (boolean)(ce_ltObjet (ob1, ob2) | ce_eqObjet (ob1, ob2));
      break;
    case CE_LT :
      r = ce_ltObjet (ob1, ob2);
      break;
    case CE_GE :
      r = ce_geObjet (ob1, ob2);
      break;
    case CE_GT :
      r = (boolean)(ce_geObjet (ob1, ob2) & !ce_eqObjet (ob1, ob2));
	break;
    case CE_IN :
      r = ce_inObjet (ob1, ob2);                     /* strictly included */
      break;
    default:
      break;
    }
    break;
  case CE_AT :  /* The condition is bound to a discrete proposition (mode, dependent) */
    if (getCondValue (cd))
      r = CE_TRUE;
    else r = CE_FALSE;
    break;
  case CE_OU :
      r1 = ce_evalCond (getCondOp1 (cd), rt);
      r2 = ce_evalCond (getCondOp2 (cd), rt);
      r = (r1 == CE_TRUE) ? CE_TRUE : ((r2 == CE_TRUE) ? CE_TRUE : CE_FALSE);
      break;
    case CE_ET :
      r1 = ce_evalCond (getCondOp1 (cd), rt);
      r2 = ce_evalCond (getCondOp2 (cd), rt);
      r = (r1 == CE_FALSE) ? CE_FALSE :
	((r2 == CE_FALSE) ? CE_FALSE : CE_TRUE);
      break;
    case CE_NON :
      r1 = ce_evalCond (getCondOp1 (cd), rt);
      r = (r1 == CE_TRUE) ? CE_FALSE : CE_TRUE;
      break;
  default:
    break;
  }
  return r;
}

/**
 * print the InitVar value on the file descriptor.
 */
void ce_printInitVar (FILE *outs, initVar iv) {

  fprintf (outs, "  ");
  ce_printVarNom (outs, getIVVar (iv));
  ce_printObjet (outs, getIVVal (iv));
  fprintf (outs, ";\n");
  return;
}

/**
 * print the variable value of v in runtime rt on the file descriptor.
 */
void ce_printVariable (FILE *fd, runtime rt, variable v) {
  long hLn;
 
  fprintf (fd, "  ");
  ce_printVarNom (fd, v);
  fprintf (fd, ": ");
  if (ce_isInput (rt, v)) {
    fprintf (fd, "input");                       /* fonction         */
    if (ce_isDisplay (rt, v))
      fprintf (fd, ", display");
    fprintf (fd, "; ");
  }
  else { /* input v */
    if (ce_isMeasure (rt, v)) {
      fprintf (fd, "measured");
      /*      fprintf (fd, " (%s) ", isVarMisbehaving(v) ? "MISB" : "OK"); */
      if (ce_isDisplay (rt, v))
        fprintf (fd, ", display");
      fprintf (fd, "; ");
    }
    else { /* measured v */
      if (ce_isDisplay (rt, v))
        fprintf (fd, "display; ");
      else
	fprintf (fd, " ");
    } /* endif measured v */
  } /* endif input v */
  if (isParameter (v)) {
    fprintf (fd, "parameter ");
    ce_printObjet (fd, varOldValue (v));
  }
  else {
    fprintf (fd, "variable");
    ce_printObjet (fd, getVarDom (v));
  }
  if ((hLn = getVarHistoLn (v))) {
    fprintf (fd, "; history length = %ld*Ts\n", hLn);
    /*ce_printVarHistory (fd, v);*/
  }
  else
    fprintf (fd, "; no history\n");
  return;
}
 
void ce_printVarHistory (FILE* fd, variable v, runtime rt) {
  if (ce_isMeasure (rt, v)) {
    for (int i=0; i<getVarHistoLn (v); i++) {
      ce_printObjet (fd, varHistoMesNth (v, i));
      cout << " * ";
    }
    cout << "\nmes: ";
    ce_printObjet (fd, varMesValue (v));
    cout << endl;
  }
  for (int i=0; i<getVarHistoLn (v); i++) {
    ce_printObjet (fd, varHistoNth (v, i));
    cout << " * ";
  }
  
  cout << "\nold: ";
  ce_printObjet (fd, varOldValue (v));
  if (! isParameter (v)) {
    cout << "\nnew: ";
    ce_printObjet (fd, varNewValue (v));
  }
}

/**
 * print the vector of all variables in rt on the file descriptor.
 */
void ce_printDic (FILE *fd, runtime rt) {
  long i;
  variable v;

  for (i=0; i<getRTNVar(rt); ++i) {
    v = getRTDicoNth (rt, i);
    ce_printVariable (fd, rt, v);
  } /* end for */
  return;
}

/**
 * print the conditions. (mostly for debug purposes).
 */
void ce_printConds (FILE* fd, runtime rt) {
  long i;
  condition cd;
  
  fprintf (fd, "size : %ld\n", rt->ncondi);
  for (i=0; i<getRTNCondi (rt); i++) {
    cd = getRTCondiNth (rt, i);
    ce_printCondition (fd, cd);
  }
  return;
}

void ce_printCondValues (FILE* fd, condition cd) {
  enum condType ct;
  variable v;
  ct = getCondType (cd);
  
  fprintf (fd, "(");
  switch (ct) {
  case CE_CTE :
    if (CE_TRUE == getCondValue (cd))
      fprintf (fd, "true");
    else
      fprintf (fd, "false");
    break;
  case CE_COMP :
    ce_printVarNom (fd, getCondVar1 (cd));
    cout << " : ";
    ce_printVarValueNV (fd, getCondVar1 (cd));
    switch (getCondOpRel (cd)) {
    case CE_EQ :
      fprintf (fd, " = ");
      break;
    case CE_NE :
      fprintf (fd, " != ");
      break;
    case CE_LE :
      fprintf (fd, " <= ");
      break;
    case CE_LT :
      fprintf (fd, " < ");
      break;
    case CE_GE :
      fprintf (fd, " >= ");
      break;
    case CE_GT :
      fprintf (fd, " > ");
      break;
    case CE_IN :
      break;
    default:
      break;
    }
    if (!isNull (v = getCondVar2 (cd))) {
      ce_printVarNom (fd, getCondVar2 (cd));
      cout << " : ";
      if (isParameter (v))
	ce_printObjet (stdout, varOldValue (v));
      else ce_printVarValueNV (fd, getCondVar2 (cd));
    }
    else { /* isNull */ 
       objet ob=getCondObj (cd);
       if (CE_RANGE == getObjType (ob))
	 ce_printRange (stdout, getObjRangeVal (ob));
       else /* RANGE */
	 ce_printReal (stdout, getObjRealVal (ob));
    }
    break;
  case CE_AT :
    getCondAtom (cd)->showAtom (0);
  case CE_OU :
    if (!isNull (getCondOp1 (cd))) {
      ce_printCondValues (fd, getCondOp1 (cd));
      fprintf (fd, " | ");
    }
    if (!isNull (getCondOp2 (cd)))
      ce_printCondValues (fd, getCondOp2 (cd));
    break;
  case CE_ET :
    if (!isNull (getCondOp1 (cd))) {
      ce_printCondValues (fd, getCondOp1 (cd));
      fprintf (fd, " & ");
    }
    if (!isNull (getCondOp2 (cd)))
      ce_printCondValues (fd, getCondOp2 (cd));
    break;
  case CE_NON :
    fprintf (fd, "! ");
    ce_printCondValues (fd, getCondOp1 (cd));
    break;
  default:
    break;
  } /* end switch ct */
  
  fprintf (fd, ")");
  cout << endl;
  return;
  
}

/* print all variables (dictionary) */
void ce_printSpace (FILE *fd, runtime rt) {
  long i;
  variable v;

  for (i=0; i<getRTNVar(rt); ++i) {
    v = getRTDicoNth (rt, i);
    if (! isParameter (v))
      ce_printVarValue (fd, v);
  }
  return;
}

/* print inputs variables */
void ce_printInputs (FILE *fd, runtime rt) {
  long i;
  variable v;

  for (i=0; i<getRTNInput(rt); ++i) {
    v = getRTInputNth (rt, i);
    ce_printVarValue (fd, v);
  } /* end for */
  return;
}

/* print display variables */
void ce_printDisplays (FILE *fd, runtime rt) {
  long i;
  variable v;

  for (i=0; i<getRTNDisplay(rt); ++i) {
    v = getRTDisplayNth (rt, i);
    ce_printVarValue (fd, v);
  } /* end for */
  return;
}

/* print measured variables */
void ce_printMeasures (FILE *fd, runtime rt) {
  long i;
  variable v;

  for (i=0; i<getRTNMeasure(rt); ++i) {
    v = getRTMeasureNth (rt, i);
    ce_printVarValue (fd, v);
  } /* end for */
  return;
}

/* print initialization values */
void ce_printInits (FILE *fd, runtime rt) {
  long i;
  initVar iv;

  for (i=0; i<getRTNInit(rt); ++i) {
    iv = getRTInitNth (rt, i);
    ce_printInitVar (fd, iv);
  } /* end for */
  return;
}

/* reset each observed variable to its measure */
void ce_resetMs (runtime rt) {
  long i;
  variable v;
  objet old, mes;
 
  for (i=0; i<getRTNMeasure (rt); ++i) {
    v = getRTMeasureNth (rt, i);
    old=varOldValue(v), mes=varMesValue(v);
    ce_putObjet (mes, old);
  } /* end for */
  return;
}

/* check for measures that are outside of the computed bounds */
long ce_checkMs (runtime rt, long seuil) {
  long i, low, high, nc;
  double dmes, dold;
  boolean test = CE_TRUE;
  variable v;
  cGraph cg;
  objet mes, old;
  
  //debug
  //std::cerr << "check seuil: " << seuil << std::endl;
  //debug

  for (i=0; i<getRTNMeasure (rt); ++i) {
    low = 0, high = 0;
    v = getRTMeasureNth (rt, i);
    old=varOldValue(v), mes=varMesValue(v);
    cg = getVarNode (v);
    nc = getCGNCause (cg);
    test = ce_inObjet (mes, old);                      /* strictly included */
    if (CE_FALSE == test) {
      dmes=getObjRealVal(mes);
      if (CE_REAL == getObjType (old)) {               /* mes is CE_REAL */
	dold=getObjRealVal(old);
	if ((0 == dmes) && (0 == dold))
	  test = CE_TRUE;
	else if (fabs (dmes-dold) / (fabs (dmes) + fabs (dold))
		 < getRTPrecision (rt))
	  test = CE_TRUE;                          
	else if (dmes > dold)
	  high = 1;
	else
	  low = 1;
      }
      else { /*  CE_REAL */
	range rold=getObjRangeVal(old);
	if (getRangeSup (rold) < dmes)
	  high = 1;
	else if (getRangeInf (rold) > dmes)
	  low = 1;
      } /* endif CE_REAL */
    } /* endif CE_FALSE */
    if (test == CE_TRUE) {
      /* measured value is inside predicted range : reset alarms counters */
      razVarLowAlrm (v);
      razVarHighAlrm (v);
      setVarLowOk (v);
      setVarHighOk (v);
      setVarOk (v);
    }
    else {
      if (high) {
	/* measured value is outside (up) predicted range: */
	/* increase high alarms counter */
	bool hAlrm = false;
	if ((hAlrm = incVarHighAlrm (v, seuil)))
	  setVarHighMisbehaving (v);
	
	//debug
	//std::cerr << "chechMs: high alrm: " << hAlrm << std::endl;
	//debug
      }
      else if (low) {
	/* measured value is outside (low) predicted range: */
	/* increase low alarms counter */
	bool lAlrm = false;
	if ((lAlrm = incVarLowAlrm (v, seuil)))
	  setVarLowMisbehaving (v);
	
	//debug
	//std::cerr << "chechMs: low alrm: " << lAlrm << std::endl;
	//debug
      } /* endif high */
      if (isVarLowMisbehaving (v) || isVarHighMisbehaving (v)) {
#ifdef VERB_MISB
	cout << "var: " << getVarName (v) << " is misbehaving!\n";
#endif
	setVarMisbehaving (v);
      }
    } /* endif test */
  } /* end for */
  return 1;
}

/* print all continuous influences */
void ce_printRules (FILE *fd, runtime rt) {
  long i;
  influence ifl;
  enum influType tp;
  const char *nom;
  condition cd;
 
  for (i=0; i<getRTNInflu(rt); ++i) {
    ifl = getRTInfluNth (rt, i);
    fprintf (fd, "  ");
    if ('\0' != *(nom=getInfluNom(ifl)))
      fprintf (fd, "%s: ", nom);
    if (! isNull (cd=getInfluCond(ifl))) {
      fprintf (fd, "condition ");
      ce_printCondition (fd, cd);
      fprintf (fd, " ");
    }
    ce_printVarNom (fd, getInfluCause (ifl));
    fprintf (fd, " -");
    switch (tp=getInfluType(ifl)) {
    case CE_DINFL :
      fprintf (fd, "D");
      break;
    case CE_D2INFL :
      fprintf (fd, "D2");
      break;
    case CE_IINFL :
      fprintf (fd, "I");
      break;
    case CE_SINFL :
      fprintf (fd, "S");
      break;
    case CE_CINFL :
      fprintf (fd, "C");
      break;
    default:
      break;
    }
    fprintf (fd, "-> ");
    ce_printVarNom (fd, getInfluEffet (ifl));
    fprintf (fd, ": gain");
    ce_printObjet (fd, getInfluGain (ifl));
    if (! isNull (getInfluXi (ifl))) {
      fprintf (stdout, ": xi ");
      ce_printObjet (stdout, getInfluXi (ifl));
    }
    if (! isNull (getInfluW (ifl))) {
      fprintf (fd, ": w ");
      ce_printObjet (fd, getInfluW (ifl));
    }
    if (! isNull (getInfluN1 (ifl))) {
      fprintf (fd, ": n1 ");
      ce_printObjet (fd, getInfluN1 (ifl));
      fprintf (fd, ": n2 ");
      ce_printObjet (fd, getInfluN2 (ifl));
      fprintf (fd, ": d0 ");
      ce_printObjet (fd, getInfluD0 (ifl));
      fprintf (fd, ": d1 ");
      ce_printObjet (fd, getInfluD1 (ifl));
      fprintf (fd, ": d2 ");
      ce_printObjet (fd, getInfluD2 (ifl));
    }
    fprintf (fd, "; delay");
    if (getInfluDelInf (ifl) == getInfluDelSup (ifl))
      fprintf (fd, " =%ld", getInfluDelInf (ifl));
    else
      fprintf (fd, " in [%ld, %ld]",
	       getInfluDelInf (ifl), getInfluDelSup (ifl));
    fprintf (fd, "; resp. time");
    ce_printObjet (fd, getInfluTRep (ifl));
    fprintf (fd, ";\n");
  } /* end for */
  return;
}
 
/**
 * print condition to the file descriptor.
 */
void ce_printCondition (FILE *fd, condition cd) {

  enum condType ct;
  variable v;
  Atom* at;

  if (cd == NULL) {
    fprintf(stderr, "Unable to print condition (NULL Pointer)\n");
    return;
  }

  ct = getCondType (cd);
  fprintf (fd, "(");
  switch (ct) {
  case CE_CTE :
    if (CE_TRUE == getCondValue (cd))
      fprintf (fd, "true");
    else
      fprintf (fd, "false");
    break;
  case CE_EXPR :
    ce_printGVar (fd, getExprGVar (getCondExpr (cd)));
    cout << " : ";
    ce_printExpr (fd, getCondExpr (cd));
    break;
  case CE_COMP :
    ce_printVarNom (fd, getCondVar1 (cd));
    switch (getCondOpRel (cd)) {
    case CE_EQ :
      fprintf (fd, " = ");
      break;
    case CE_NE :
      fprintf (fd, " != ");
      break;
    case CE_LE :
      fprintf (fd, " <= ");
      break;
    case CE_LT :
      fprintf (fd, " < ");
      break;
    case CE_GE :
      fprintf (fd, " >= ");
      break;
    case CE_GT :
      fprintf (fd, " > ");
      break;
    case CE_IN :
      break;
    default:
      break;
    }
    if (isNull (v = getCondVar2 (cd))) {
      objet ob=getCondObj (cd);
      if (CE_RANGE == getObjType (ob))
	ce_printRange (fd, getObjRangeVal (ob));
      else /* RANGE */
	ce_printReal (fd, getObjRealVal (ob));
    }
    else /* isNull v */
      ce_printVarNom (fd, v);
    break;
  case CE_AT :
    getCondAtom (cd)->showAtom (0);
    break;
  case CE_OU :
    if (!isNull (getCondOp1 (cd))) {
      ce_printCondition (fd, getCondOp1 (cd));
      fprintf (fd, " | ");
    }
    if (!isNull (getCondOp2 (cd)))
      ce_printCondition (fd, getCondOp2 (cd));
    break;
  case CE_ET :
    if (!isNull (getCondOp1 (cd))) {
      ce_printCondition (fd, getCondOp1 (cd));
      fprintf (fd, " & ");
    }
    if (!isNull (getCondOp2 (cd)))
      ce_printCondition (fd, getCondOp2 (cd));
    break;
  case CE_NON :
    fprintf (fd, "! ");
    ce_printCondition (fd, getCondOp1 (cd));
    break;
  default:
    break;
  } /* end switch ct */
  fprintf (fd, ")");
  if (!isNull (at = getCondAtom (cd)) && ! getCondType (cd) == CE_AT) {
    cout << "\t\t";
    at->showAtom (0);
    cout << endl;
  }
  cout << endl;
  return;
}

void ce_printCGraph (FILE *fd, runtime rt) {
  long i;
  variable v;
  cGraph cg;

  for (i=0; i<getRTNVar(rt); ++i) {
    v = getRTDicoNth (rt, i);
    if (! isNull (cg = getVarNode (v))) {
      long j, nc=getCGNCause(cg), ne=getCGNEffet(cg);
      fprintf (fd, "  ");
      ce_printVarNom (fd, v);
      if (nc) {
	fprintf (fd, ":\n    causes (");
	switch (getCGCauseType (cg)) {
	case CE_DINFL :
	  fprintf (fd, "dynamic");
	  break;
	case CE_D2INFL :
	  fprintf (fd, "dynamic2");
	  break;
	case CE_IINFL :
	  fprintf (fd, "integral");
	  break;
	case CE_SINFL :
	  fprintf (fd, "static");
	  break;
	default:
	  break;
	}
	fprintf (fd, "): ");
	for (j=0; j<nc; ++j) {
	  ce_printVarNom (fd, getInfluCause (getCGCauseNth (cg, j)));
	  fprintf (fd, "; ");
	}
      }
      else /* nc */
	fprintf (fd, ":\n    no causes;");
      /* endif nc */
      if (ne) {
	fprintf (fd, "\n    effects: ");
	for (j=0; j<ne; ++j) {
	  ce_printVarNom (fd, getInfluEffet (getCGEffetNth (cg, j)));
	  fprintf (fd, "; ");
	}
      }
      else /* ne */
	fprintf (fd, "\n    no effects;");
      /* endif ne */
      fprintf (fd, "\n");
    } /* endif ! isNull cg */
  } /* end for i */
  return;
}


void ce_printUpdateListcaen (FILE *fd, runtime rt) {
  /* imprimer successivement temporal et untemporal */
  long i, n;
  variable v;

  if (0 != (n = getRTNTempo(rt))) {
    fprintf (fd, "\n Temporal listcaen: ");
    for (i=0; i<n; ++i) {
      v = getRTTempoNth (rt, i);
      ce_printVarNom (fd, v);
      if (i < n-1)
	fprintf (fd, ", ");
    } /* end for */
    fprintf (fd, ";");
  } /* end if */
  fprintf (fd, "\n Untemporal listcaen: ");
  if (0 != (n = getRTNUntemp(rt))) {
    for (i=0; i<n; ++i) {
      v = getRTUntempNth (rt, i);
      ce_printVarNom (fd, v);
      if (i < n-1)
	fprintf (fd, ", ");
    } /* end for */
    fprintf (fd, ";");
  } /* end if */
  fprintf (fd, "\n");
  return;
}
 
/* print global variable */
void ce_printGVar (FILE *outs, gVar gv) {
  objet ob;

  switch (getGVType (gv)) {
  case CE_CONST :
    ob = gVVal (gv);
    switch (getObjType(ob)) {
    case CE_REAL :
      ce_printReal (outs, getObjRealVal (ob));
      break;
    case CE_RANGE :
      ce_printRange (outs, getObjRangeVal (ob));
      break;
    default:
      break;
    } /* end switch getObjType ob */
    break;
  case CE_VAR :
    ce_printVarNom (outs, getGVVar (gv));
    break;
  case CE_INTERM :
    fprintf (outs, "InTeRm%ld", getGVInterm (gv));
    break;
  default:
    break;
  } /* end switch */
  return;
}

  
void ce_printGlobVs (FILE *fd, runtime rt) {
  long i, k, ngc;
  gVar gv;

  for (i=0; i<getRTNGVar(rt); ++i) {
    gv = getRTGVarNth (rt, i);
    fprintf (fd, "  ");
    ce_printGVar (fd, gv);
    if (0 != (ngc = getGVNGlobC (gv))) {
      fprintf (fd, ":");
    for (k=1; k<= ngc; ++k)
      fprintf (fd, " %ld", getGVGlobCNth (gv, k) + 1);
    }/* endif ngc */
    fprintf (fd, ";\n");
  } /* end for */
  return;
}


void ce_printExpr (FILE *outs, expr e) {
  enum exprType tp=getExprType(e);
  expr e1, e2;
  objet ob;
 
  switch (tp) {
  case CE_CSTE :
    ob = gVVal (getExprGVar (e));
    ce_printObjet (outs, ob);
    break;
  case CE_VBL :
    ce_printGVar (outs, getExprGVar (e));
    break;
  case CE_PLUS :
  case CE_MOINS :
  case CE_MULT :
  case CE_DIV :
  case CE_EXP :
    e1 = getExprOp1 (e);
    if (isExprCte (e1) || hasName (e1))
      ce_printGVar (outs, getExprGVar (e1));
    else {
      fprintf (outs, "(");
      ce_printExpr (outs, e1);
      fprintf (outs, ")");
    }
    switch (tp) {
    case CE_PLUS :
      fprintf (outs, "+");
      break;
    case CE_MOINS :
      fprintf (outs, "-");
      break;
    case CE_MULT :
      fprintf (outs, "*");
      break;
    case CE_DIV :
      fprintf (outs, "/");
      break;
    case CE_EXP :
      fprintf (outs, "**");
      break;
    default:
      break;
    }
    e2 = getExprOp2 (e);
    if (isExprCte (e2) || hasName (e2))
      ce_printGVar (outs, getExprGVar (e2));
    else {
      fprintf (outs, "(");
      ce_printExpr (outs, e2);
      fprintf (outs, ")");
    }
    break;
  case CE_SGN :
    fprintf (outs, "-");
    e1 = getExprOp1 (e);
    if (isExprCte (e1) || hasName (e1))
      ce_printGVar (outs, getExprGVar (e1));
    else {
      fprintf (outs, "(");
      ce_printExpr (outs, e1);
      fprintf (outs, ")");
    }
    break;
  default:
    break;
  }
  return;
}


void ce_printGContr (FILE *outs, expr e) {

  ce_printGVar (outs, getExprGVar (e));
  fprintf (outs, " = ");
  ce_printExpr (outs, e);
  fprintf (outs, ";\n");
  return;
}

  
void ce_printGlobCs (FILE *fd, runtime rt) {
  long i;
  expr e;

  for (i=0; i<getRTNGlobC(rt); ++i) {
    fprintf (fd, "  (%ld) ", i+1);
    e = getRTGlobCNth (rt, i);
    ce_printGContr (fd, e);
  } /* end for */
  return;
}

/**
 * initialize the vector of all variables.
 */
void ce_updateDico (runtime rt) {
  long i, j;
  variable v;
  
  for (i=0; i<getRTNVar(rt); ++i) {
    v = getRTDicoNth (rt, i);
    if (! isParameter (v)) {
      setUndetermined (v);
      for (j=0; j<getVarHistoLn (v); ++j)
	razObjet (varHistoNth (v, j));
      if (ce_isMeasure (rt, v)) {
	for (j=0; j<getVarHistoLn (v); ++j) {
	  razObjet (varHistoNth (v, j));
	  razVarHistoLowAlrm (v, j);
	  razVarHistoHighAlrm (v, j);
	}
	razVarLowAlrm (v);
	razVarHighAlrm (v);
	setVarLowOk(v);
	setVarHighOk(v);
	setVarOk(v);
      } /* endif isMeasure v */
      /*ce_printVarHistory (stdout, v, rt);*/
    } /* endif ! isParameter */
  } /* end for i */
  return;
}

/**
 * initialize all variables of a runtime.
 */
void ce_initVars (runtime rt) {
  long i;
  initVar iv;
  variable v;
  objet ob;

  for (i=0; i<getRTNInit(rt); ++i) {   /* user initialization prevails on measures */
    iv = getRTInitNth (rt, i);
    v = getIVVar (iv);
    ob = getIVVal (iv);
    ce_putObjet (ob, varOldValue (v));
    ce_fillHisto (v);
    setDetermined (v);
  }
  for (i=0; i<getRTNMeasure(rt); ++i) {
    v = getRTMeasureNth (rt, i);
    if (! isDetermined (v)) {
      ce_putObjet (varMesValue (v), varOldValue (v));
      ce_fillHisto (v);
      setDetermined (v);
    } /* endif ! isDetermined v */
  } /* end for */
  for (i=0; i<getRTNTempo(rt); ++i) {
    v = getRTTempoNth (rt, i);
    if (! isDetermined (v)) {
      ce_compInitVar (rt, v);        /* result into old value */
    } /* endif ! isDetermined v */
  } /* end for */
  for (i=0; i<getRTNUntemp(rt); ++i) {
    v = getRTUntempNth (rt, i);
    if (! isDetermined (v))
      ce_compInitVar (rt, v);         /* result into old value */
  } /* end for */
  if (ce_initValSpc (rt) != 1)  /* splitting the state space */
    fprintf (stdout, "Error while splitting init state\n");
  setInited (rt);
  return;
}

/**
 * fill history of the variable with its 'old' value.
 */
void ce_fillHisto (variable v) {
  long j, hLn;
  objet ob;

  if ((hLn = getVarHistoLn (v)) > 0) {
    ob = varOldValue (v);
    for (j=0; j<hLn; ++j)
      ce_putObjet (ob, varHistoNth (v, j));
  } /* endif hLn */
  return;
}

/**
 * fill history of measures of the variable with its 'old' value,
 * and reset alarms.
 */
void ce_fillHistoMes (variable v) {
  long j, hLn;
  objet ob;

  if ((hLn = getVarHistoLn (v)) > 0) {
    ob = varMesValue (v);
    for (j=0; j<hLn; ++j) {
      ce_putObjet (ob, varHistoMesNth (v, j));
      razVarHistoLowAlrm (v, j);
      razVarHistoHighAlrm (v, j);
    }
  } /* endif hLn */
  return;
}

/* fill history of all variables */
void ce_fillAllHistos (runtime rt) {
  long i;
  variable v;

  for (i=0; i<getRTNVar (rt); ++i) {
    v = getRTDicoNth (rt, i);
    if (! isParameter (v))
      ce_fillHisto (v);
  } /* end for */
  return;
}

/* fill history of measured variables */
void ce_fillAllHistosMes (runtime rt) {
  long i;
  variable v;

  for (i=0; i<getRTNMeasure (rt); ++i) {
    v = getRTMeasureNth (rt, i);
      ce_fillHistoMes (v);
  } /* end for */
  return;
}

void printInfluence (influence inf) {
  enum influType tp=getInfluType(inf);
  const char *nom=getInfluNom(inf);
  condition cd=getInfluCond(inf);

  fprintf (stdout, "  ");
  if (! isNull (nom))
    fprintf (stdout, "%s: ", nom);
  if (! isNull (cd)) {
    fprintf (stdout, "condition ");
    ce_printCondition (stdout, cd);
    fprintf (stdout, " ");
  }
  ce_printVarNom (stdout, getInfluCause (inf));
  fprintf (stdout, " -");
  switch (tp) {
  case CE_DINFL:
    fprintf (stdout, "D");
    break;
  case CE_D2INFL:
    fprintf (stdout, "D2");
    break;
  case CE_IINFL:
    fprintf (stdout, "I");
    break;
  case CE_SINFL:
    fprintf (stdout, "S");
    break;
  case CE_CINFL:
    fprintf (stdout, "C");
    break;
  default:
    break;
  }
  fprintf (stdout, "-> ");
  ce_printVarNom (stdout, getInfluEffet (inf));
  fprintf (stdout, ": gain ");
  ce_printObjet (stdout, getInfluGain (inf));
  if (! isNull (getInfluXi (inf))) {
    fprintf (stdout, ": xi ");
    ce_printObjet (stdout, getInfluXi (inf));
  }
  if (! isNull (getInfluW (inf))) {
    fprintf (stdout, ": w ");
    ce_printObjet (stdout, getInfluW (inf));
  }
  if (! isNull (getInfluN1 (inf))) {
    fprintf (stdout, ": n1 ");
    ce_printObjet (stdout, getInfluN1 (inf));
    fprintf (stdout, ": n2 ");
    ce_printObjet (stdout, getInfluN2 (inf));
    fprintf (stdout, ": d0 ");
    ce_printObjet (stdout, getInfluD0 (inf));
    fprintf (stdout, ": d1 ");
    ce_printObjet (stdout, getInfluD1 (inf));
    fprintf (stdout, ": d2 ");
    ce_printObjet (stdout, getInfluD2 (inf));
  }
  fprintf (stdout, "; delay ");
  fprintf (stdout, "%ld, %ld\n", getInfluDelInf (inf), getInfluDelSup (inf));
  /*ce_printObjet (stdout, getInfluDelay (inf));*/
  fprintf (stdout, "; resp. time ");
  ce_printObjet (stdout, getInfluTRep (inf));
  fprintf (stdout, "; condHistoLn = %ld", getInfluNActCond(inf));
  fprintf (stdout, ";\n");
  return;
}

/**
 * eval coefficients of all influences in runtime.
 */
void ce_evalInfluCoefs (runtime rt, double sample) {
  long i;
  influence ifl;
  s_objet s_pmt1, s_pmt2, s_pmt3, s_pmt4, s_pmt5, s_pmt6, s_pmt7, s_pmt8, s_pmt9, s_pmt10, s_pmt11;
  objet PMT1=&s_pmt1, PMT2=&s_pmt2, PMT3=&s_pmt3, PMT4=&s_pmt4, PMT5=&s_pmt5, PMT6=&s_pmt6, 
    PMT7=&s_pmt7, PMT8=&s_pmt8, PMT9=&s_pmt9, PMT10=&s_pmt10, PMT11=&s_pmt11;

  for (i=0; i<getRTNInflu(rt); ++i) {
   
    ifl=getRTInfluNth (rt, i);
 
    switch (getInfluType (ifl)) {
    case CE_DINFL :
      razObjet (TMP1);
      setObjRealVal (TMP1, -3.0*sample);          /* -3*T */
      ce_divObjet (TMP1, getInfluTRep (ifl));     /* -3*T/Tr */
      ce_expObjet (TMP1);                         /* exp (-3*T/Tr) */
      ce_putObjet (TMP1, getInfluA (ifl));        /* a */
     
      break;
    case CE_D2INFL :  /* second order influences */
      if (! isNull (getInfluN1 (ifl))) { /* general case */

	/* PMT1 = a1/(2*a2) */
	/* PMT2 = sqrt(a0/a2 - PMT1^2) */
	/* PMT3 = 1-exp(-a*T)*cos(b*T)-(a/b)*exp(-a*T)*sin(b*T) */
	/* PMT4 = exp(-2*a*T)+(a/b)*exp(-a*T)*sin(b*T)-exp(-a*T)*cos(b*T) */
	/* PMT5 = exp(-a*T)*sin(b*T) */
	/* PMT6 = T */
	/* PMT7 = 2.0 */
	
	razObjet (PMT6);
	setObjType (PMT6, CE_REAL),
	  setObjRealVal (PMT6, sample);
	razObjet (PMT7);
	setObjType (PMT7, CE_REAL),
	  setObjRealVal (PMT7, 2.0);

	/* compute PMT1 */
	razObjet (PMT1);
	ce_putObjet (getInfluD1 (ifl), PMT1);
	razObjet (PMT2);
	ce_putObjet (getInfluD2 (ifl), PMT2);
	ce_mulObjet (PMT2, PMT7);
	ce_divObjet (PMT1, PMT2);
	
	/* compute PMT2 */
	razObjet (PMT2);
	razObjet (PMT3);
	ce_putObjet (getInfluD0 (ifl), PMT3);
	ce_divObjet (PMT3, getInfluD2 (ifl));
	ce_putObjet (PMT1, PMT2);
	ce_mulObjet (PMT2, PMT1);
	ce_minusObjet (PMT2);
	ce_addObjet (PMT2, PMT3);
	ce_sqrtObjet (PMT2);

	/* compute PMT5 */
	razObjet (PMT8);
	ce_putObjet (PMT1, PMT8);
	ce_mulObjet (PMT8, PMT6);
	ce_minusObjet (PMT8);
	ce_expObjet (PMT8);        /* PMT8 = exp (-a*T) */
	razObjet (PMT5);
	ce_putObjet (PMT2, PMT5);
	ce_mulObjet (PMT5, PMT6);
	razObjet (PMT9);
	ce_putObjet (PMT5, PMT9);
	ce_sinObjet (PMT5);
	ce_mulObjet (PMT5, PMT8);  /* PMT5 = exp(-a*T)*sin(b*T) */
	ce_cosObjet (PMT9);
	ce_mulObjet (PMT9, PMT8);  /* PMT9 = exp(-a*T)*cos(b*T) */

	/* compute PMT3 */
	razObjet (PMT10);
	ce_putObjet (PMT1, PMT10);
	ce_divObjet (PMT10, PMT2);  /* PMT1/PMT2 */
	razObjet (PMT3);
	ce_putObjet (PMT5, PMT3);
	ce_mulObjet (PMT3, PMT10);
	ce_putObjet (PMT3, PMT10); /* PMT10 = PMT1/PMT2 * PMT5 */
	ce_addObjet (PMT3, PMT9);
	ce_minusObjet (PMT3);
	ce_addObjet (PMT3, objUn);

	/* compute PMT4 */
	razObjet (PMT4);
	ce_putObjet (PMT9, PMT4);
	ce_minusObjet (PMT4);
	ce_addObjet (PMT4, PMT10);
	ce_mulObjet (PMT8, PMT8);   /* PMT8 = exp (-2*a*T) */
	ce_addObjet (PMT4, PMT8);

	/* compute a1 */
	razObjet (TMP1);
	ce_putObjet (PMT9, TMP1);
	ce_mulObjet (TMP1, PMT7);
	ce_putObjet (TMP1, getInfluA (ifl));
	
	/* compute a2 */
	razObjet (TMP1);
	ce_putObjet (PMT8, TMP1);
	ce_minusObjet (TMP1);
	ce_putObjet (TMP1, getInfluA2 (ifl));

	/* compute b1 */
	razObjet (TMP1);
	razObjet (PMT10);
	ce_putObjet (PMT5, PMT10);
	ce_mulObjet (PMT10, getInfluN2 (ifl));
	ce_divObjet (PMT10, PMT2);             /* PMT10 = PMT5*(n2/PMT2) */
	ce_divObjet (PMT10, getInfluD2 (ifl));
	razObjet (PMT11);
	ce_putObjet (getInfluN1 (ifl), PMT11);
	ce_divObjet (PMT11, getInfluD0 (ifl));  /* PMT11 = n1/d0 */
	ce_putObjet (PMT11, TMP1);
	ce_mulObjet (TMP1, PMT3);
	ce_addObjet (TMP1, PMT10);
	if (! isNull (getInfluGain (ifl)))
	  ce_mulObjet (TMP1, getInfluGain (ifl)); /* added 1/08/02 */
	ce_putObjet (TMP1, getInfluB (ifl));

	/* compute b2 */
	razObjet (TMP1);
	ce_minusObjet (PMT10);
	ce_putObjet (PMT11, TMP1);
	ce_mulObjet (TMP1, PMT4);
	ce_addObjet (TMP1, PMT10);
	if (! isNull (getInfluGain (ifl)))
	  ce_mulObjet (TMP1, getInfluGain (ifl)); /* added 1/08/02 */
	ce_putObjet (TMP1, getInfluB2 (ifl));
      }
	   
      else if (ce_ltObjet (getInfluXi (ifl), objUn)  /* computing coefs for underdumped 2nd order */
	       || ce_eqObjet (getInfluXi (ifl), objUn)) { 

	ce_putObjet (getInfluXi (ifl), PMT1);          /* xi */
	ce_putObjet (getInfluXi (ifl), PMT2);
	ce_minusObjet (PMT1);                           /* -xi */
	ce_mulObjet (PMT1, getInfluW (ifl));           /* -xi*w */
	setObjType (PMT6, CE_REAL),
	  setObjRealVal (PMT6, sample);
	ce_mulObjet (PMT1, PMT6);                    /* -xi*w*T */
	ce_mulObjet (PMT2, PMT2);                       /* xi^2 */
	ce_minusObjet (PMT2);
	ce_addObjet (PMT2, objUn);                     /* 1-xi^2 */
	ce_sqrtObjet (PMT2);                           /* sqrt(1-xi^2) */
	ce_putObjet (PMT2, PMT3);
	ce_mulObjet (PMT3, getInfluW (ifl));
	ce_mulObjet (PMT3, PMT6);                    /* w*sqrt(1-xi^2)*T */
	ce_putObjet (PMT3, PMT4);
	ce_putObjet (PMT3, PMT5);
	ce_cosObjet (PMT4);                            /* cos(w*sqrt(1-xi^2)*T) */
	ce_sinObjet (PMT5);                            /* sin(w*sqrt(1-xi^2)*T) */

	/* compute a. */
	razObjet (TMP1);
	razObjet (PMT6);
	setObjType (PMT6, CE_REAL),
	  setObjRealVal (PMT6, 2.0);
	ce_putObjet (PMT1, TMP1);                      /* -xi*w*T */
	ce_expObjet (TMP1);
	ce_mulObjet (TMP1, PMT6);                      /* 2*e^{-xi*w*T} */
	ce_mulObjet (TMP1, PMT4);                      /* 2*e^{-xi*w*T}*cos(w*sqrt(1-xi^2)*T) */
	ce_putObjet (TMP1, getInfluA (ifl));           /* a */

	/* compute a2. */
	razObjet (TMP1);
	ce_putObjet (PMT1, TMP1);                       /* -xi*w*T */
	ce_mulObjet (TMP1, PMT6);
	ce_expObjet (TMP1);                            
	ce_minusObjet (TMP1);                           /* -e^{-2*xi*w*T} */
	ce_putObjet (TMP1, getInfluA2 (ifl));
	
	/* compute b. */
	razObjet (TMP1);
	ce_putObjet (PMT1, TMP1);                      /* -xi*w*T */
	ce_expObjet (TMP1);
	ce_putObjet (TMP1, PMT7);
	ce_minusObjet (TMP1);                          /* -e^{-xi*w*T} */
	razObjet (PMT6);
	ce_putObjet (getInfluXi (ifl), PMT6);
	ce_divObjet (PMT6, PMT2);                     /* xi/sqrt(1-xi^2) */
	ce_mulObjet (PMT6, PMT5);                     /* xi/sqrt(1-xi^2)*sin(w*sqrt(1-xi^2)*T) */
	ce_addObjet (PMT6, PMT4);                     /* xi/sqrt(1-xi^2)*sin(w*sqrt(1-xi^2)*T) + cos(w*sqrt(1-xi^2)*T) */
	ce_mulObjet (TMP1, PMT6);
	ce_addObjet (TMP1, objUn);
	ce_mulObjet (TMP1, getInfluGain (ifl));
	ce_putObjet (TMP1, getInfluB (ifl));

	/* compute b2. */
	razObjet (TMP1);
	ce_putObjet (PMT1, TMP1);                      /* -xi*w*T */
	razObjet (PMT8);
	setObjType (PMT8, CE_REAL),
	  setObjRealVal (PMT8, 2.0);
	ce_mulObjet (TMP1, PMT8);                      /* -2*xi*w*T */
	ce_expObjet (TMP1);                            /* e^{-2*xi*w*T} */
      	razObjet (PMT6);
	ce_putObjet (getInfluXi (ifl), PMT6);
	ce_divObjet (PMT6, PMT2);                     /* xi/sqrt(1-xi^2) */
	ce_mulObjet (PMT6, PMT5);                     /* xi/sqrt(1-xi^2)*sin(w*sqrt(1-xi^2)*T) */
	ce_subObjet (PMT6, PMT4);                     /* xi/sqrt(1-xi^2)*sin(w*sqrt(1-xi^2)*T) - cos(w*sqrt(1-xi^2)*T) */
	ce_mulObjet (PMT6, PMT7);
	ce_addObjet (TMP1, PMT6);
	ce_mulObjet (TMP1, getInfluGain (ifl));
	ce_putObjet (TMP1, getInfluB2 (ifl));

      }
      else if (getObjType (getInfluXi (ifl)) == CE_REAL 
	       && getObjRealVal (getInfluXi (ifl)) == DBL_MAX) { /* F(s) = 1/s2 */
	
	razObjet (PMT5);
	setObjType(PMT5, CE_REAL),
	  setObjRealVal (PMT5, sample);
	razObjet (PMT6);
	setObjType(PMT6, CE_REAL),
	  setObjRealVal (PMT6, 2.0);

	/* compute a=2. */
	ce_putObjet (PMT6, getInfluA (ifl));

	/* compute a2=-1. */
	razObjet (TMP1);
	setObjType (TMP1, CE_REAL),
	  setObjRealVal (TMP1, -1.0);
	ce_putObjet (TMP1, getInfluA2 (ifl));

	/* set b. */
	razObjet (TMP1);
	ce_putObjet (getInfluGain (ifl), TMP1);
	ce_mulObjet (TMP1, PMT5);
	ce_mulObjet (TMP1, PMT5);
	ce_divObjet (TMP1, PMT6);               /* T^2/2 */
	ce_putObjet (TMP1, getInfluB (ifl));
	
	/* set b2. */
	razObjet (TMP1);
	ce_putObjet (getInfluGain (ifl), TMP1);
	ce_mulObjet (TMP1, PMT5);
	ce_mulObjet (TMP1, PMT5);
	ce_divObjet (TMP1, PMT6);               /* T^2/2 */
	ce_putObjet (TMP1, getInfluB2 (ifl));
      }
      else { /* xi */
	razObjet (PMT1);
	razObjet (PMT5);
	ce_putObjet (getInfluXi (ifl), PMT1);
	ce_mulObjet (PMT1, getInfluXi (ifl));      /* xi^2 */
	ce_putObjet (objUn, PMT5);
	ce_minusObjet (PMT5);
	ce_addObjet (PMT1, PMT5);                   /* xi^2-1 */
	ce_sqrtObjet (PMT1);
	ce_mulObjet (PMT1, getInfluW (ifl));        /* w*sqrt(xi^2-1) */
	ce_putObjet (PMT1, PMT2);
	ce_minusObjet (PMT1);
	razObjet (PMT5);
	ce_putObjet (getInfluXi (ifl), PMT5);
	ce_mulObjet (PMT5, getInfluW (ifl));
	ce_addObjet (PMT1, PMT5);                   /* PMT1 = xi*wi-w*sqrt(xi^2-1) */
	ce_addObjet (PMT2, PMT5);                   /* PMT2 = xi*wi+w*sqrt(xi^2-1) */

	razObjet (PMT5);
	setObjType(PMT5, CE_REAL),
	  setObjRealVal (PMT5, sample);
	setObjType(PMT6, CE_REAL),
	  setObjRealVal (PMT6, sample);
	ce_mulObjet (PMT5, PMT1);
	ce_minusObjet (PMT5);
	ce_expObjet (PMT5);                         /* PMT5 = exp(-PMT1*T) */
	ce_mulObjet (PMT6, PMT2);
	ce_minusObjet (PMT6);
	ce_expObjet (PMT6);                         /* PMT6 = exp(-PMT2*T) */

	/* compute a. */
	razObjet (TMP1);
	ce_putObjet (PMT5, TMP1);
	ce_addObjet (TMP1, PMT6);
	ce_putObjet (TMP1, getInfluA (ifl));        /* a = exp(-PMT1*T) + exp(-PMT2*T) */
	
	/* compute a2. */
	razObjet (TMP1);
	ce_putObjet (PMT1, TMP1);
	ce_addObjet (TMP1, PMT2);
	setObjType (PMT7, CE_REAL),
	  setObjRealVal (PMT7, sample);
	ce_mulObjet (TMP1, PMT7);
	ce_minusObjet (TMP1);
	ce_expObjet (TMP1);
	ce_minusObjet (TMP1);
	ce_putObjet (TMP1, getInfluA2 (ifl));      /* a2 = -exp(-(PMT1+PMT2)*T) */

	razObjet (PMT7);
	ce_putObjet (PMT5, PMT7);
	ce_minusObjet (PMT7);
	ce_addObjet (PMT7, objUn);                 
	ce_mulObjet (PMT7, PMT2);                  /* PMT7=PMT2*(1-exp(-PMT1*T)) */

	ce_putObjet (PMT6, PMT8);
	ce_minusObjet (PMT8);
	ce_addObjet (PMT8, objUn);                 
	ce_mulObjet (PMT8, PMT1);                  /* PMT8=PMT1*(1-exp(-PMT2*T)) */

	ce_putObjet (PMT1, PMT9);
	ce_minusObjet (PMT9);
	ce_addObjet (PMT9, PMT2);
	ce_mulObjet (PMT9, PMT2);
	ce_mulObjet (PMT9, PMT1);                  /* PMT9=PMT1*PMT2*(PMT2-PMT1) */

	/* compute b. */
	razObjet (TMP1);
	ce_putObjet (PMT8, TMP1);
	ce_minusObjet (TMP1);
	ce_addObjet (TMP1, PMT7);
	ce_divObjet (TMP1, PMT9);
	ce_mulObjet (TMP1, getInfluW (ifl));
	ce_mulObjet (TMP1, getInfluW (ifl));
	ce_mulObjet (TMP1, getInfluGain (ifl));
	ce_putObjet (TMP1, getInfluB (ifl));  /* b = K*w*(PMT7-PMT8)/PMT9 */

	/* compute b2. */
	razObjet (TMP1);
	ce_putObjet (PMT7, TMP1);
	ce_mulObjet (TMP1, PMT6);
	ce_minusObjet (TMP1);
	ce_mulObjet (PMT8, PMT5);
	ce_addObjet (TMP1, PMT8);
	ce_divObjet (TMP1, PMT9);
	ce_mulObjet (TMP1, getInfluW (ifl));
	ce_mulObjet (TMP1, getInfluW (ifl));
	ce_mulObjet (TMP1, getInfluGain (ifl));
	ce_putObjet (TMP1, getInfluB2 (ifl));

      }
      fprintf (stdout, "a1:\n");
      ce_printObjet (stdout, getInfluA (ifl));
      fprintf (stdout, "\n");
      fprintf (stdout, "a2 :\n");
      ce_printObjet (stdout, getInfluA2 (ifl));
      fprintf (stdout, "\n");
      fprintf (stdout, "b1 :\n");
      ce_printObjet (stdout, getInfluB (ifl));
      fprintf (stdout, "\n");
      fprintf (stdout, "b2 :\n");
      ce_printObjet (stdout, getInfluB2 (ifl));
      fprintf (stdout, "\n");
      break;
    case CE_IINFL :
      ce_putObjet (objUn, getInfluA (ifl));        /* a=1 */
      razObjet (TMP1);
      setObjRealVal (TMP1, sample);   /* T */
      ce_mulObjet (TMP1, getInfluGain (ifl));     /* K*T */
      ce_putObjet (TMP1, getInfluB (ifl));        /* b */
      break;
    case CE_SINFL :
      razObjet (getInfluA (ifl));                               /* a=0 */
      ce_putObjet (getInfluGain (ifl), getInfluB (ifl));        /* b=K */
      break;
    case CE_CINFL :
      razObjet (getInfluA (ifl));                               /* a=0 */
      ce_putObjet (objUn, getInfluB (ifl));                     /* b=1 */
      break;
    default:
      break;
    } /* end switch */
  } /* end for */
  return;
}

/**
 * test if a variable is an input variable of the runtime rt.
 */
long ce_isInput (runtime rt, variable v) {
  long i, r=0;

  for (i=0; i<getRTNInput(rt); ++i)
    if (getRTInputNth (rt, i) == v) {
      r = 1;
      break;
    }
  return r;
}


long ce_isDisplay (runtime rt, variable v) {
  long i, r=0;

  for (i=0; i<getRTNDisplay(rt); ++i)
    if (getRTDisplayNth (rt, i) == v) {
      r = 1;
      break;
    }
  return r;
}


long ce_isMeasure (runtime rt, variable v) {
  long i, r=0;

  for (i=0; i<getRTNMeasure(rt); ++i)
    if (getRTMeasureNth (rt, i) == v) {
      r = 1;
      break;
    }
  return r;
}

/**
 * move history values by one.
 */
void ce_pushHisto (variable v) {
  long j, hLn;

  if (0 != (hLn = getVarHistoLn (v))) {
    for (j=hLn-1; j>0; --j)
      ce_putObjet (varHistoNth (v, j-1), varHistoNth (v, j));
    ce_putObjet (varOldValue (v), varHistoNth (v, 0));
  } /* endif hLn */
  return;
}


void ce_pushHistoMes (variable v) { /* history of measures and alarms */
  long j, hLn;

  if (0 != (hLn = getVarHistoLn (v))) {
    for (j=hLn-1; j>0; --j) {
      ce_putObjet (varHistoMesNth (v, j-1), varHistoMesNth (v, j));
      razVarHistoLowAlrm (v, j);
      razVarHistoHighAlrm (v, j);
      if (isVarHistoLowAlrm (v, j-1))
	setVarHistoLowAlrm (v, j);
      else if (isVarHistoHighAlrm (v, j-1))
	setVarHistoHighAlrm (v, j);
    }
    ce_putObjet (varMesValue (v), varHistoMesNth (v, 0));
    razVarHistoLowAlrm (v, 0);
    razVarHistoHighAlrm (v, 0);
    if (isVarLowAlrm (v))
      setVarHistoLowAlrm (v, 0);
    else if (isVarHighAlrm (v))
      setVarHistoHighAlrm (v, 0);
  } /* endif hLn */
  return;
}

/**
 * init a variable considering influences on it.
 */
void ce_compInitVar (runtime rt, variable v) {

  s_objet s_pmt1, s_pmt2, s_pmt5;
  objet PMT1=&s_pmt1, PMT2=&s_pmt2, PMT5=&s_pmt5;
  cGraph cg = getVarNode (v);
  long i, nc=getCGNCause(cg);
  if (nc) {
    influence ifl = NULL;
    condition cd;
    variable cs;
    razObjet (PMT1);
    switch (getCGCauseType (cg)) {
    case CE_DINFL :
      for (i=0; i<nc; ++i) {
	ifl=getCGCauseNth(cg,i);
	cd=getInfluCond(ifl);
	if (! isNull (cd))
	  if (CE_FALSE == ce_pushCond (ifl, rt))
	    continue;
	ce_putObjet (varOldValue (getInfluCause(ifl)), PMT2);
	ce_mulObjet (PMT2, getInfluGain (ifl));   /* marginal influence */
	ce_addObjet (PMT1, PMT2);                 /* accumulation */
      } /* end for i */
      ce_putObjet (PMT1, TMP1); /* test it */
      ce_evalInitDynInfl (ifl);
      ce_putObjet (TMP1, PMT1);
      break;
    case CE_D2INFL :
      for (i=0; i<nc; i++) {
	ifl = getCGCauseNth(cg, i);
	cd = getInfluCond(ifl);
	if (! isNull (cd))
	  if (CE_FALSE == ce_pushCond (ifl, rt))
	    continue;
	ce_putObjet (varOldValue (getInfluCause (ifl)), PMT2);
	razObjet (PMT5);
	ce_putObjet (varHistoNth (getInfluCause (ifl), 0), PMT5); /* Xi(k-1) */
	ce_mulObjet (PMT2, getInfluB (ifl));
	ce_mulObjet (PMT5, getInfluB2 (ifl));
	ce_addObjet (PMT1, PMT2);
	ce_addObjet (PMT1, PMT5);
      } /* end for i */
      ce_putObjet (PMT1, TMP1); /* test it */
      ce_evalInitDynInfl (ifl);
      ce_putObjet (TMP1, PMT1);
      break;
    case CE_IINFL :
      for (i=0; i<nc; ++i) {
	ifl=getCGCauseNth(cg,i);
	cd=getInfluCond(ifl);
	if (! isNull (cd))
	  if (CE_FALSE == ce_pushCond (ifl, rt))
	    continue;
	ce_putObjet (varOldValue (getInfluCause(ifl)), PMT2);
	ce_mulObjet (PMT2, getInfluB (ifl));      /* marginal influence */
	ce_addObjet (PMT1, PMT2);                 /* accumulation */
      } /* end for i */
      ce_addObjet (PMT1, varOldValue (getInfluEffet (ifl)));
      break;
    case CE_SINFL :
      for (i=0; i<nc; ++i) {
	ifl=getCGCauseNth (cg,i);
	cs = getInfluCause (ifl);
	cd=getInfluCond(ifl);
	if (! isNull (cd))
	  if (CE_FALSE == ce_pushCond (ifl, rt))
	    continue;
	if (! isDetermined (cs)) {
	  if (ce_isMeasure (rt, cs)) {
	    ce_putObjet (varMesValue (cs), varOldValue (cs));
	    setDetermined (cs);
	  }
	  else
	    ce_compInitVar (rt, cs);
	}
	ce_putObjet (varOldValue (cs), PMT2);
	ce_mulObjet (PMT2, getInfluB (ifl));      /* marginal influence */
	ce_addObjet (PMT1, PMT2);                 /* accumulation */
      } /* end for i */
      break;
    default:
      break;
    } /* end switch */
    ce_putObjet (PMT1, varOldValue (v));
    ce_fillHisto (v);
    setDetermined (v);
  } /* endif nc */
  return;
}

/**
 * eval influence parameters at initialization.
 */
void ce_evalInitDynInfl (influence ifl) {
  /* Sigma (Ki * Xi) est dans TMP1, le resultat aussi */
  s_objet s_objA={CE_REAL,{{1.,0.}}};
  s_objet s_objA2={CE_REAL,{{1.,0.}}};
  s_objet s_objB={CE_REAL,{{1.,0.}}};
  s_objet s_objB2={CE_REAL,{{1.,0.}}};
  s_objet s_pmt1;
  objet a=&s_objA,
    a2=&s_objA2,
    b=&s_objB,
    b2=&s_objB2,
    PMT1=&s_pmt1,
    y=varOldValue(getInfluEffet(ifl)),
    yp=varHistoNth (getInfluEffet(ifl), 0);
  enum objType ta=getObjType(a);
  enum objType ta2=getObjType(a2);
  
  if (getInfluType (ifl) != CE_D2INFL) {  /* expand Tr close to infinity for static cycle */
    /* \f$a = e^{-3*\frac{T_s}{T_r}}. T_r = -3*\frac{T_s}{ln(a)}$\f. */
    ce_putObjet (getInfluA (ifl), TMP2); /* a is initialized in evalInfluCoefs. */
    ce_logObjet (TMP2); /* ln (a) */
    ce_mulObjet (TMP2, objDix); /* *10 */
    ce_expObjet (TMP2); /* exp(a) */
    ce_putObjet (TMP2, a); /* a -> a^10 */
  }
  else { /* TODO : it is wrong !!! -> must find a way to update coeffs for a static cycle ! */
    ce_putObjet (getInfluA (ifl), a);
    ce_putObjet (getInfluA2 (ifl), a2);
    ce_putObjet (getInfluB (ifl), b);
    ce_putObjet (getInfluB2 (ifl), b2);
  }
  
  if (getInfluType(ifl) != CE_D2INFL && ta == CE_REAL) {
    ce_putObjet (objUn, TMP2);              /* 1 */
    ce_subObjet (TMP2, a);                  /* 1-a */
    ce_mulObjet (TMP1, TMP2);               /* (1-a) * Sigma(Ki*Xi) */
    ce_putObjet (a, TMP2);                  /* a */
    ce_mulObjet (TMP2, y);                  /* a*Y */
    ce_addObjet (TMP1, TMP2);               /* a*Y + (1-a) * Sigma(Ki*Xi) */
  }
  else if (getInfluType(ifl) != CE_D2INFL) { /* ta, D2INFL */
    double f1inf, f1sup, f2inf, f2sup, rinf, rsup,
      ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    /* inf. */
    setObjType (TMP3, CE_REAL), setObjRealVal (TMP3, ainf);
    ce_putObjet (objUn, TMP2);              /* 1 */
    ce_subObjet (TMP2, TMP3);               /* 1-a */
    ce_mulObjet (TMP2, TMP1);               /* (1-a) * Sigma(Ki*Xi) */
    ce_mulObjet (TMP3, y);                  /* a*Y */
    ce_addObjet (TMP2, TMP3);               /* a*Y + (1-a) * Sigma(Ki*Xi) */
    if (getObjType (TMP2) == CE_RANGE)
      f1inf = getRangeInf (getObjRangeVal (TMP2)),
	f1sup = getRangeSup (getObjRangeVal (TMP2));
    else
      f1inf = f1sup = getObjRealVal (TMP2);
    /* sup. */
    setObjType (TMP3, CE_REAL), setObjRealVal (TMP3, asup);
    ce_putObjet (objUn, TMP2);              /* 1 */
    ce_subObjet (TMP2, TMP3);               /* 1-a */
    ce_mulObjet (TMP2, TMP1);               /* (1-a) * Sigma(Ki*Xi) */
    ce_mulObjet (TMP3, y);                  /* a*Y */
    ce_addObjet (TMP2, TMP3);               /* a*Y + (1-a) * Sigma(Ki*Xi) */
    if (getObjType (TMP2) == CE_RANGE)
      f2inf = getRangeInf (getObjRangeVal (TMP2)),
	f2sup = getRangeSup (getObjRangeVal (TMP2));
    else
      f2inf = f2sup = getObjRealVal (TMP2);
    rinf = min (f1inf, f2inf),
      rsup = max (f1sup, f2sup);
    if (rinf == rsup)
      setObjType (TMP1, CE_REAL),
	setObjRealVal (TMP1, rinf);
    else
      setObjType (TMP1, CE_RANGE),
	setRangeInf (getObjRangeVal (TMP1), rinf),
	setRangeSup (getObjRangeVal (TMP1), rsup);
  }
  else { /* D2INFL */
    if (ta == CE_REAL && ta2 == CE_REAL) {
      ce_putObjet (a, TMP2);                   /* a */
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      ce_putObjet (a2, PMT1);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      razObjet (PMT1);
      ce_addObjet (TMP1, TMP2);                /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
    }
    else { /* USELESS ? */
      double f1inf, f1sup, f2inf, f2sup, f3inf, f3sup, f4inf, f4sup, rinf, rsup,
	ainf=getRangeInf(getObjRangeVal(a)),
	asup=getRangeSup(getObjRangeVal(a)),
	a2inf=getRangeInf(getObjRangeVal(a2)),
	a2sup=getRangeInf(getObjRangeVal(a2));
      
      /* inf/inf. */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, ainf);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2inf);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      razObjet (PMT1);
      ce_addObjet (TMP1, TMP2);                /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP1) == CE_RANGE)
	f1inf = getRangeInf (getObjRangeVal (TMP1)),
	  f1sup = getRangeSup (getObjRangeVal (TMP1));
      else
	f1inf = f1sup = getObjRealVal (TMP1);

      /* inf/sup. */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, ainf);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2sup);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      razObjet (PMT1);
      ce_addObjet (TMP1, TMP2);                /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP1) == CE_RANGE)
	f2inf = getRangeInf (getObjRangeVal (TMP1)),
	  f2sup = getRangeSup (getObjRangeVal (TMP1));
      else
	f2inf = f2sup = getObjRealVal (TMP1);

      /* sup/inf. */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, asup);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2inf);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      razObjet (PMT1);
      ce_addObjet (TMP1, TMP2);                /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP1) == CE_RANGE)
	f3inf = getRangeInf (getObjRangeVal (TMP1)),
	  f3sup = getRangeSup (getObjRangeVal (TMP1));
      else
	f3inf = f3sup = getObjRealVal (TMP1);
      
      /* sup/sup. */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, asup);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2sup);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      razObjet (PMT1);
      ce_addObjet (TMP1, TMP2);                /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP1) == CE_RANGE)
	f4inf = getRangeInf (getObjRangeVal (TMP1)),
	  f4sup = getRangeSup (getObjRangeVal (TMP1));
      else
	f4inf = f4sup = getObjRealVal (TMP1);
     
      /* update. */
      rinf = min (min (f1inf, f2inf), min (f3inf, f4inf));
      rsup = max (max (f1sup, f2sup), max (f3sup, f4sup));
      if (rinf == rsup)
	setObjType (TMP1, CE_REAL),
	  setObjRealVal (TMP1, rinf);
      else
	setObjType (TMP1, CE_RANGE),
	  setRangeInf (getObjRangeVal (TMP1), rinf),
	  setRangeSup (getObjRangeVal (TMP1), rsup);
    }
  } /* endif ta, D2INFL */
  return;
}

boolean ce_pushCond (influence ifl, runtime rt) {
  boolean r;
  long j, cLn;
  condition cd=getInfluCond(ifl);

  if ((cLn = getInfluNActCond (ifl)) > 0) {
    for (j=cLn-1; j>0; --j)
      setInfluActCondNth (ifl, j, getInfluActCondNth (ifl, j-1));
  } /* endif cLn */
  r = ce_evalCond (cd, rt);
  setCondValue (cd, r);
  setInfluActCondNth (ifl, 0, r);
  return r;
}

void ce_pushExternCond (influence ifl, boolean r) {
  long j, cLn;
  condition cd=getInfluCond(ifl);

  if ((cLn = getInfluNActCond (ifl)) > 0) {
    for (j=cLn-1; j>0; --j)
      setInfluActCondNth (ifl, j, getInfluActCondNth (ifl, j-1));
  } /* endif cLn */
  setCondValue (cd, r);
  setInfluActCondNth (ifl, 0, r);
  return;
}

/**
 * update temporal vector of a runtime.
 */
void ce_doStepTempo (runtime rt) {
  long i;
  variable v;

  for (i=0; i<getRTNTempo(rt); ++i) { /* compute new value */
    v = getRTTempoNth (rt, i);
    ce_updateVariable (rt, v);        /* result into new value */
  } /* end for */
  /*for (i=0; i<getRTNTempo(rt); ++i) {*/ /* transfer new value to old value */
    /*v = getRTTempoNth (rt, i);
    ce_new2OldValue (rt, v);
    setDetermined (v);*/
  /*}*/ /* end for */
  return;
}

/**
 * update untemporal vector of a runtime.
 */
void ce_doStepUntemp (runtime rt) {
  long i;
  variable v;

  for (i=0; i<getRTNUntemp(rt); ++i) { /* compute new value */
    v = getRTUntempNth (rt, i);
    ce_updateVariable (rt, v);         /* result into new value */
  } /* end for */
  /*for (i=0; i<getRTNUntemp(rt); ++i) {*/ /* transfer new value to old value */
    /*v = getRTUntempNth (rt, i);*/
  return;
}

void ce_moveAllOld (runtime rt) {
  variable v;
  
  for (int i=0; i<getRTNVar (rt); ++i) {
    v = getRTDicoNth (rt, i);
    if (! isParameter (v) && ! ce_isInput (rt, v)) {
      ce_new2OldValue (rt, v);
      setDetermined (v);
    }
  } /* end for */
}

/**
 * update a variable value considering its influences.
 */
void ce_updateVariable (runtime rt, variable v) {

  cGraph cg=getVarNode (v);
  long i, nc=getCGNCause(cg);
  s_objet s_pmt1; /*, s_pmt2, s_pmt3;*/
  objet b, b2, PMT1=&s_pmt1;
  enum objType tb;
  enum objType tb2;
  
  if (nc) {
    influence ifl = NULL;
    condition cd;
    razObjet (TMP1);
    switch (getCGCauseType (cg)) {
    case CE_DINFL :
      for (i=0; i<nc; ++i) {
	ifl = getCGCauseNth(cg,i);
	cd = getInfluCond(ifl);
	if (! isNull (cd)) {
	  if (CE_FALSE == ce_pushCond (ifl, rt)) {
#ifdef VERB_INFS
	    cout << getInfluNom (ifl) << " non-active\n";
#endif
	    continue;
	  }
#ifdef VERB_INFS
	  else cout << getInfluNom (ifl) << " active\n";
#endif
	}
	ce_loadCauseValue (rt, ifl);                       /* -> TMP2 */
	ce_mulObjet (TMP2, getInfluGain (ifl));   /* marginal influence */
	ce_addObjet (TMP1, TMP2);                 /* accumulation */
      } /* end for i */
      ce_evalDynInfl (rt, ifl);
      break;
    case CE_D2INFL :
      for (i=0; i<nc; i++) {
	ifl = getCGCauseNth(cg, i);
	b = getInfluB (ifl);
	b2 = getInfluB2 (ifl);
	tb = getObjType(b);
	tb2 = getObjType(b2);
	cd = getInfluCond(ifl);
	if (! isNull (cd)) {
	  if (CE_FALSE == ce_pushCond (ifl, rt)) {
#ifdef VERB_INFS
	    cout << getInfluNom (ifl) << "non-active\n";
#endif
	    continue;
	  }
#ifdef VERB_INFS
	  else cout << getInfluNom (ifl) << " active\n";
#endif
	}
	ce_loadCauseValue (rt, ifl);                      /* -> TMP2 */
	
	if (tb == CE_REAL && tb2 == CE_REAL) {
	  ce_mulObjet (TMP2, getInfluB (ifl));
	  ce_mulObjet (TMP5, getInfluB2 (ifl));
	  ce_addObjet (TMP1, TMP2);
	  ce_addObjet (TMP1, TMP5);
	}
	else {
	  double  binf=getRangeInf(getObjRangeVal(b)),
	    bsup=getRangeSup(getObjRangeVal(b)),
	    b2inf=getRangeInf(getObjRangeVal(b2)),
	    b2sup=getRangeSup(getObjRangeVal(b2));

	  razObjet (TMPII);
	  razObjet (TMPIS);
	  razObjet (TMPSI);
	  razObjet (TMPIS);

	  /* inf/inf. */
	  setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, binf);
	  ce_mulObjet (PMT1, TMP2);
	  setObjType (TMPII, CE_REAL), setObjRealVal (TMPII, b2inf);
	  ce_mulObjet (TMPII, TMP5);
	  ce_addObjet (TMPII, PMT1);

	  /* inf/sup. */
	  setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, binf);
	  ce_mulObjet (PMT1, TMP2);
	  setObjType (TMPIS, CE_REAL), setObjRealVal (TMPIS, b2sup);
	  ce_mulObjet (TMPIS, TMP5);
	  ce_addObjet (TMPIS, PMT1);
	  
	  /* sup/inf. */
	  setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, bsup);
	  ce_mulObjet (PMT1, TMP2);
	  setObjType (TMPSI, CE_REAL), setObjRealVal (TMPSI, b2inf);
	  ce_mulObjet (TMPSI, TMP5);
	  ce_addObjet (TMPSI, PMT1);
	  
	  /* sup/sup. */
	  setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, bsup);
	  ce_mulObjet (PMT1, TMP2);
	  setObjType (TMPSS, CE_REAL), setObjRealVal (TMPSS, b2sup);
	  ce_mulObjet (TMPSS, TMP5);
	  ce_addObjet (TMPSS, PMT1);
	}
      } /* end for i */
      ce_evalDynInfl (rt, ifl);
      break;
    case CE_IINFL :
      for (i=0; i<nc; ++i) {
	ifl = getCGCauseNth(cg,i);
	cd = getInfluCond(ifl);
	if (! isNull (cd)) {
	  if (CE_FALSE == ce_pushCond (ifl, rt)) {
#ifdef VERB_INFS
	    cout << getInfluNom (ifl) << " non-active\n";
#endif
	    continue;
	  }
#ifdef VERB_INFS
	  else cout << getInfluNom (ifl) << " active\n";
#endif
	}
	ce_loadCauseValue (rt, ifl);                       /* -> TMP2 */
	ce_mulObjet (TMP2, getInfluB (ifl));      /* marginal influence */
	ce_addObjet (TMP1, TMP2);                 /* accumulation */
      } /* end for i */
      if (ce_isMeasure (rt, v) && ! (isVarLowAlrm (v) || isVarHighAlrm (v))) {
	ce_addObjet (TMP1, varMesValue (v));

/*#ifdef ENV_OPTI
	if (env_opti && getInfOptFct (ifl))  */ /* envelopes optimization */
	  /*call_nelder_obs (v, rt, varMesValue (v), NULL, TMP1, CE_IINFL, 1);
#endif      */
      }
      else
	ce_addObjet (TMP1, varOldValue (v));
      break;
    case CE_SINFL :
      for (i=0; i<nc; ++i) {
	ifl=getCGCauseNth(cg,i);
	cd=getInfluCond(ifl);
	if (! isNull (cd)) {
	  if (CE_FALSE == ce_pushCond (ifl, rt)) {
#ifdef VERB_INFS
	    cout << getInfluNom (ifl) << " non active\n";
#endif
	    continue;
	  }
#ifdef VERB_INFS
	  else cout << getInfluNom (ifl) << " active\n";
#endif
	}
	ce_loadCauseValue (rt, ifl);                       /* -> TMP2 */
	ce_mulObjet (TMP2, getInfluB (ifl));      /* marginal influence B */
	ce_addObjet (TMP1, TMP2);                 /* accumulation */
      } /* end for i */

#ifdef ENV_OPTI
      if (env_opti && getInfOptFct (ifl) && ce_isMeasure (rt, v) && ! (isVarLowAlrm (v) || isVarHighAlrm (v))) {
	call_nelder_obs (v, rt, varMesValue (v), NULL, TMP1, CE_SINFL, 1);
      }
#endif
      break;
    default:
      break;
    } /* end switch */
    ce_putObjet (TMP1, varNewValue (v));
#ifdef VERB_INFS
    cout << getVarName (v) << " updated: ";
    ce_printObjet (stdout, TMP1);
    cout << endl;
    if (ce_isMeasure (rt, v)) {
      ce_printObjet (stdout, varMesValue (v));
      cout << endl;
    }
#endif
  } /* endif nc */
  return;
}

/**
 * put the value of the cause in TMP2.
 * for second order influences, additionnal value is in TMP5.
 */
void ce_loadCauseValue (runtime rt, influence ifl) { /* value of cause -> TMP2 */
  variable cs=getInfluCause(ifl);
  long j, delinf=getInfluDelInf(ifl), delsup=getInfluDelSup(ifl);

  if (CE_SINFL == getInfluType (ifl))
    --delsup, --delinf;
  if (delsup > 0) {
    if (ce_isMeasure (rt, cs)
	/*&& ((cs != getInfluEffet(ifl))*/
	    && ! (isVarHistoLowAlrm (cs, delsup-1)
		  || isVarHistoHighAlrm (cs, delsup-1))) {
      ce_putObjet (varHistoMesNth (cs, delsup-1), TMP2);
      if (CE_D2INFL == getInfluType (ifl))
	ce_putObjet (varHistoMesNth (cs, delsup), TMP5);
#ifdef VERB_LOADCAUSE
      cerr << "loadCauseValue (measure): " << getVarName (cs) << " ";
      ce_printObjet (stderr, TMP2);
      cerr << endl;
      cerr << "influ effet: " << (cs != getInfluEffet(ifl)) << endl;
      cerr << "alrm: " << isVarLowAlrm (cs) << " -- " << isVarHighAlrm (cs) << endl;
#endif
    }
    else {
      ce_putObjet (varHistoNth (cs, delsup-1), TMP2);
#ifdef VERB_LOADCAUSE      
      cerr << "loadCauseValue: " << getVarName (cs) << " ";
      ce_printObjet (stderr, TMP2);
      cerr << endl;
#endif
      if (CE_D2INFL == getInfluType (ifl))
	ce_putObjet (varHistoNth (cs, delsup), TMP5); /* delsup - 2 ???. */
    }
    if (delinf < delsup) {
      for (j=delsup-2; j>=delinf-1; --j) {
	ce_uniObjet (TMP2, varHistoNth (cs, j));
      }
    } /* endif delinf < delsup */
  }
  else if (delsup == 0) { /* delsup */ /* delay is 1 */
    if (ce_isMeasure (rt, cs)
	&& ! (isVarLowAlrm (cs) || isVarHighAlrm (cs))) {
      ce_putObjet (varMesValue (cs), TMP2);
#ifdef VERB_LOADCAUSE      
      cerr << "loadCauseValue (measure): " << getVarName (cs) << " ";
      ce_printObjet (stderr, varMesValue (cs));
      cerr << endl;
      cerr << "influ effet: " << (cs != getInfluEffet(ifl)) << endl;
      cerr << "alrm: " << isVarLowAlrm (cs) << " -- " << isVarHighAlrm (cs) << endl;
#endif
      if (CE_D2INFL == getInfluType (ifl))
	ce_putObjet (varHistoMesNth (cs, 0), TMP5);
    }
    else {
      ce_putObjet (varOldValue (cs), TMP2);
#ifdef VERB_LOADCAUSE
      cerr << "loadCauseValue: " << getVarName (cs) << " ";
      ce_printObjet (stdout, TMP2);
      cerr << " type: " << getObjType (TMP2) << endl;
      /*cerr << endl;*/
#endif
      if (CE_D2INFL == getInfluType (ifl))
	ce_putObjet (varHistoNth (cs, 0), TMP5);
    }
  }
  else if (delsup == -1 && CE_SINFL == getInfluType(ifl)) {
    if (ce_isMeasure (rt, cs)
	/*&& ((cs != getInfluEffet(ifl))*/
	    && ! (isVarLowAlrm (cs) || isVarHighAlrm (cs))) {
      ce_putObjet (varMesValue (cs), TMP2);
#ifdef VERB_LOADCAUSE
      cerr << "loadCauseValue: " << getVarName (cs) << " (measure) (delsup==-1) : ";
      ce_printObjet (stderr, varMesValue (cs));
      cerr << endl;
      cerr << "alrm: " << isVarLowAlrm (cs) << " -- " << isVarHighAlrm (cs) << endl;
#endif
    }
    else {
      if (! isParameter (cs))
	ce_putObjet (varNewValue (cs), TMP2);
      else ce_putObjet (varOldValue (cs), TMP2);
#ifdef VERB_LOADCAUSE
      cerr << "loadCauseValue: " << getVarName (cs) << "(delsup==-1): ";
      if (! isParameter (cs))
	ce_printObjet (stderr, varNewValue (cs));
      else ce_printObjet (stderr, varOldValue (cs));
      cerr << endl;
      cerr << "alrm: " << isVarLowAlrm (cs) << " -- " << isVarHighAlrm (cs) << endl;
#endif
    }
  }
  return;
}

/**
 * eval a dynamic influence following :
 * first order: \f$Y(t+1) = a*Y(t) + (1-a)*\Sigma(Ki*X_i(t-d_i))\f$.
 */
void ce_evalDynInfl (runtime rt, influence ifl) {
  /* Sigma (Ki * Xi) est dans TMP1, le resultat aussi */
  variable v=getInfluEffet(ifl);
  s_objet s_pmt1;
  objet y=0, yp=0, a=getInfluA(ifl),
    a2=getInfluA2(ifl),
    b=getInfluB(ifl),
    PMT1=&s_pmt1;
  enum objType ta=getObjType(a);
  enum objType ta2=getObjType(a2);
  enum objType tb=getObjType(b);

  /*fprintf (stdout, "ce_evalDynInfl :\n");
    cout << "ifl : " << getInfluNom (ifl) << endl;
    cout << "a : ";
    ce_printObjet (stdout, a);
    cout << endl;
    cout << "b : ";
    ce_printObjet (stdout, b);
    cout << endl;
    cout << "a2 : ";
    ce_printObjet (stdout, a2);
    cout << endl;
    cout << "b2 : ";
    ce_printObjet (stdout, getInfluB2 (ifl));
    cout << endl;*/

  if (ce_isMeasure (rt, v) && ! (isVarLowAlrm (v) || isVarHighAlrm (v))) {
    y = varMesValue (v);
    yp = varHistoMesNth (v, 0);
  }
  else {
    y = varOldValue (v);
    yp = varHistoNth (v, 0);
  }
  if (getInfluType(ifl) != CE_D2INFL && ta == CE_REAL) {
    ce_putObjet (objUn, TMP2);              /* 1 */
    ce_subObjet (TMP2, a);                  /* 1-a */
    ce_mulObjet (TMP1, TMP2);               /* (1-a) * Sigma(Ki*Xi) */
    ce_putObjet (a, TMP2);                  /* a */
    ce_mulObjet (TMP2, y);                  /* a*Y */
    ce_addObjet (TMP1, TMP2);               /* a*Y + (1-a) * Sigma(Ki*Xi) */
    
    /*if (getInfOptFct (ifl) && ce_isMeasure (rt, v)) */      /* envelopes optimization */
	/*call_nelder_obs (v, rt, y, yp, TMP1, CE_DINFL);*/
  }
  else if (getInfluType(ifl) != CE_D2INFL) { /* ta */
    double f1inf, f1sup, f2inf, f2sup, rinf, rsup,
      ainf=getRangeInf(getObjRangeVal(a)),
      asup=getRangeSup(getObjRangeVal(a));
    /* inf. */
    setObjType (TMP3, CE_REAL), setObjRealVal (TMP3, ainf);
    ce_putObjet (objUn, TMP2);              /* 1 */
    ce_subObjet (TMP2, TMP3);               /* 1-a */
    ce_mulObjet (TMP2, TMP1);               /* (1-a) * Sigma(Ki*Xi) */
    ce_mulObjet (TMP3, y);                  /* a*Y */
    ce_addObjet (TMP2, TMP3);               /* a*Y + (1-a) * Sigma(Ki*Xi) */
    if (getObjType (TMP2) == CE_RANGE)
      f1inf = getRangeInf (getObjRangeVal (TMP2)),
	f1sup = getRangeSup (getObjRangeVal (TMP2));
    else
      f1inf = f1sup = getObjRealVal (TMP2);
    /* sup. */
    setObjType (TMP3, CE_REAL), setObjRealVal (TMP3, asup);
    ce_putObjet (objUn, TMP2);              /* 1 */
    ce_subObjet (TMP2, TMP3);               /* 1-a */
    ce_mulObjet (TMP2, TMP1);               /* (1-a) * Sigma(Ki*Xi) */
    ce_mulObjet (TMP3, y);                  /* a*Y */
    ce_addObjet (TMP2, TMP3);               /* a*Y + (1-a) * Sigma(Ki*Xi) */
    if (getObjType (TMP2) == CE_RANGE)
      f2inf = getRangeInf (getObjRangeVal (TMP2)),
	f2sup = getRangeSup (getObjRangeVal (TMP2));
    else
      f2inf = f2sup = getObjRealVal (TMP2);
    /* update. */
    rinf = min (f1inf, f2inf),
      rsup = max (f1sup, f2sup);
    if (rinf == rsup)
      setObjType (TMP1, CE_REAL),
	setObjRealVal (TMP1, rinf);
    else
      setObjType (TMP1, CE_RANGE),
	setRangeInf (getObjRangeVal (TMP1), rinf),
	setRangeSup (getObjRangeVal (TMP1), rsup);

#ifdef ENV_OPTI
    if (getInfOptFct (ifl)) {       /* envelopes optimization */
      if (env_opti && ce_isMeasure (rt, v))
	call_nelder_obs (v, rt, y, yp, TMP1, CE_D2INFL, 1);
      else if (env_opti) call_nelder_second (ifl, rt, y, yp, TMP1);
    }
#endif

  }
  else { /* D2INFL */
    if (ta == CE_REAL && ta2 == CE_REAL && tb == CE_REAL && tb == CE_REAL) {
      ce_putObjet (a, TMP2);                   /* a */
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      ce_putObjet (a2, PMT1);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      razObjet (PMT1);
      ce_addObjet (TMP1, TMP2);                /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
    }
    /* when K only is an interval... */
    else if (ta == CE_REAL && ta2 == CE_REAL && tb == CE_RANGE && tb == CE_RANGE) {
      double f1inf, f1sup, f2inf, f2sup, f3inf, f3sup, f4inf, f4sup, rinf, rsup;
      
      ce_putObjet (a, TMP2);                   /* a */
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      ce_putObjet (a2, PMT1);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */

      /* b inf/inf. */
      ce_addObjet (TMPII, TMP2);
      if (getObjType (TMPII) == CE_RANGE)
	f1inf = getRangeInf (getObjRangeVal (TMPII)),
	  f1sup = getRangeSup (getObjRangeVal (TMPII));
      else
	f1inf = f1sup = getObjRealVal (TMPII);
      
      /* b inf/sup. */
      ce_addObjet (TMPIS, TMP2);
      if (getObjType (TMPIS) == CE_RANGE)
	f2inf = getRangeInf (getObjRangeVal (TMPIS)),
	  f2sup = getRangeSup (getObjRangeVal (TMPIS));
      else
	f2inf = f2sup = getObjRealVal (TMPIS);
      
      /* b sup/inf. */
      ce_addObjet (TMPSI, TMP2);
      if (getObjType (TMPSI) == CE_RANGE)
	f3inf = getRangeInf (getObjRangeVal (TMPSI)),
	  f3sup = getRangeSup (getObjRangeVal (TMPSI));
      else
	f3inf = f3sup = getObjRealVal (TMPSI);
      
      /* b sup/sup. */
      ce_addObjet (TMPSS, TMP2);
      if (getObjType (TMPSS) == CE_RANGE)
	f4inf = getRangeInf (getObjRangeVal (TMPSS)),
	  f4sup = getRangeSup (getObjRangeVal (TMPSS));
      else
	f4inf = f4sup = getObjRealVal (TMPSS);

      /* update. */
      rinf = min (min (f1inf, f2inf), min (f3inf, f4inf));
      rsup = max (max (f1sup, f2sup), max (f3sup, f4sup));
      if (rinf == rsup)
	setObjType (TMP1, CE_REAL),
	  setObjRealVal (TMP1, rinf);
      else
	setObjType (TMP1, CE_RANGE),
	  setRangeInf (getObjRangeVal (TMP1), rinf),
	  setRangeSup (getObjRangeVal (TMP1), rsup);
      
      /*cout << "rinf: " << rinf << " -- rsup: " << rsup << endl;
	cout << "rinf: " << getRangeInf (getObjRangeVal (TMP1)) << " -- rsup: " << getRangeSup (getObjRangeVal (TMP1)) << endl;
	cout << "fct: " << getInfOptFct (ifl) << endl;*/
      
#ifdef ENV_OPTI
      if (getInfOptFct (ifl)) {       /* envelopes optimization */
	if (env_opti && ce_isMeasure (rt, v))
	  call_nelder_obs (v, rt, y, yp, TMP1, CE_D2INFL, 1);
	else if (env_opti) call_nelder_second (ifl, rt, y, yp, TMP1);
      }
#endif
    }
    else if (ta == CE_RANGE && ta2 == CE_RANGE) {
      double f1inf, f1sup, f2inf, f2sup, f3inf, f3sup, f4inf, f4sup, rinf, rsup, 
	f5inf, f5sup, f6inf, f6sup, f7inf, f7sup, f8inf, f8sup, f9inf, f9sup, f10inf,
	f10sup, f11inf, f11sup, f12inf, f12sup, f13inf, f13sup, f14inf, f14sup, f15inf, 
	f15sup, f16inf, f16sup,
	ainf=getRangeInf(getObjRangeVal(a)),
	asup=getRangeSup(getObjRangeVal(a)),
	a2inf=getRangeInf(getObjRangeVal(a2)),
	a2sup=getRangeSup(getObjRangeVal(a2));
      
      /*fprintf (stdout, "RANGE computation\n");
      fprintf (stdout, "y :\n");
      ce_printObjet (stdout, y);
	fprintf (stdout, "\n");
	fprintf (stdout, "yp :\n");
	ce_printObjet (stdout, yp);
	fprintf (stdout, "\n");
	fprintf (stdout, "ainf : %g\n", ainf);
	fprintf (stdout, "asup : %g\n", asup);
	fprintf (stdout, "a2inf : %g\n", a2inf);
	fprintf (stdout, "a2sup : %g\n", a2sup);*/
      
      /* inf/inf inf/inf */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, ainf);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2inf);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPII);            /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f1inf = getRangeInf (getObjRangeVal (TMP2)),
	  f1sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f1inf = f1sup = getObjRealVal (TMP2);

      /* inf/inf inf/sup. */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, ainf);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2inf);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPIS);            /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f2inf = getRangeInf (getObjRangeVal (TMP2)),
	  f2sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f2inf = f2sup = getObjRealVal (TMP2);
      
      /* inf/inf sup/inf. */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, ainf);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2inf);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPSI);            /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f3inf = getRangeInf (getObjRangeVal (TMP2)),
	  f3sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f3inf = f3sup = getObjRealVal (TMP2);

      /* inf/inf sup/sup. */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, ainf);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2inf);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPSS);            /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f4inf = getRangeInf (getObjRangeVal (TMP2)),
	  f4sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f4inf = f4sup = getObjRealVal (TMP2);

      /* inf/sup inf/inf  */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, ainf);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2sup);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPII);            /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f5inf = getRangeInf (getObjRangeVal (TMP2)),
	  f5sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f5inf = f5sup = getObjRealVal (TMP2);
      
      /* inf/sup sup/inf */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, ainf);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2sup);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPSI);            /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f6inf = getRangeInf (getObjRangeVal (TMP2)),
	  f6sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f6inf = f6sup = getObjRealVal (TMP2);
      
      /* inf/sup inf/sup */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, ainf);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2sup);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPIS);            /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f7inf = getRangeInf (getObjRangeVal (TMP2)),
	  f7sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f7inf = f7sup = getObjRealVal (TMP2);
      
      /* inf/sup sup/sup */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, ainf);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2sup);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPSS);            /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f8inf = getRangeInf (getObjRangeVal (TMP2)),
	  f8sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f8inf = f8sup = getObjRealVal (TMP2);
      
      /* sup/inf inf/inf */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, asup);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2inf);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPII);         /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f9inf = getRangeInf (getObjRangeVal (TMP2)),
	  f9sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f9inf = f9sup = getObjRealVal (TMP2);
      
      /* sup/inf sup/inf */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, asup);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2inf);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPSI);         /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f10inf = getRangeInf (getObjRangeVal (TMP2)),
	  f10sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f10inf = f10sup = getObjRealVal (TMP2);
      
      /* sup/inf inf/sup */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, asup);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2inf);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPIS);         /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f11inf = getRangeInf (getObjRangeVal (TMP2)),
	  f11sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f11inf = f11sup = getObjRealVal (TMP2);
      
      /* sup/inf sup/sup */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, asup);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2inf);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPSS);         /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f12inf = getRangeInf (getObjRangeVal (TMP2)),
	  f12sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f12inf = f12sup = getObjRealVal (TMP2);
      
      /* sup/sup inf/inf */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, asup);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2sup);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPII);              /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f13inf = getRangeInf (getObjRangeVal (TMP2)),
	  f13sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f13inf = f13sup = getObjRealVal (TMP2);
     
      /* sup/sup sup/inf */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, asup);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2sup);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPSI);              /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f14inf = getRangeInf (getObjRangeVal (TMP2)),
	  f14sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f14inf = f14sup = getObjRealVal (TMP2);
      
      /* sup/sup inf/sup */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, asup);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2sup);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPIS);              /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f15inf = getRangeInf (getObjRangeVal (TMP2)),
	  f15sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f15inf = f15sup = getObjRealVal (TMP2);
      
      /* sup/sup sup/sup */
      setObjType (TMP2, CE_REAL), setObjRealVal (TMP2, asup);
      ce_mulObjet (TMP2, y);                   /* a*y(k) */
      setObjType (PMT1, CE_REAL), setObjRealVal (PMT1, a2sup);
      ce_mulObjet (PMT1, yp);                  /* a2*y(k-1) */
      ce_addObjet (TMP2, PMT1);                /* a*y(k) + a2*y(k-1) */
      ce_addObjet (TMP2, TMPSS);              /* a*y(k) + a2*y(k-1) + Sigma (bi*Xi(k) + b2i*Xi(k-1)) */
      if (getObjType (TMP2) == CE_RANGE)
	f16inf = getRangeInf (getObjRangeVal (TMP2)),
	  f16sup = getRangeSup (getObjRangeVal (TMP2));
      else
	f16inf = f16sup = getObjRealVal (TMP2);

      /* update. */
      rinf = min (min(min(min(f1inf, f2inf), min(f3inf, f4inf)), min(min(f5inf, f6inf), 
		  min(f7inf, f8inf))), min(min(min(f9inf, f10inf), min(f11inf, f12inf)),
		  min(min(f13inf, f14inf), min(f15inf, f16inf))));
      rsup = max (max(max(max(f1sup, f2sup), max(f3sup, f4sup)), max(max(f5sup,f6sup),
		  max(f7sup,f8sup))), max(max(max(f9sup, f10sup), max(f11sup, f12sup)),
		  max(max(f13sup, f14sup), max(f15sup, f16sup))));
      if (rinf == rsup)
	setObjType (TMP1, CE_REAL),
	  setObjRealVal (TMP1, rinf);
      else
	setObjType (TMP1, CE_RANGE),
	  setRangeInf (getObjRangeVal (TMP1), rinf),
	  setRangeSup (getObjRangeVal (TMP1), rsup);

#ifdef ENV_OPTi
      if (getInfOptFct (ifl)) {       /* envelopes optimization */
	if (env_opti && ce_isMeasure (rt, v))
	  call_nelder_obs (v, rt, y, yp, TMP1, CE_D2INFL, 1);
	else if (env_opti) call_nelder_second (ifl, rt, y, yp, TMP1);
      }
#endif

    }
  } /* endif ta, D2INFL */
  return;
}

void ce_new2OldValue (runtime rt, variable v) {

  ce_pushHisto (v);
  ce_putObjet (varNewValue (v), varOldValue (v));
  (void) ce_interObjet (varOldValue (v), getVarDom (v));
  return;
}



/**
 * Envelopes optimization.
 **/

/* pointers on optimization functions */
double (*opti_static_obs_ptr) (int*, double*, double*, double*, double*, int*, int*, int*);
double (*opti_integral_obs_ptr) (int*, double*, double*, double*, double*, int*, int*, int*);
double (*opti_first_order_obs_ptr) (int*, double*, double*, double*, double*, int*, int*, int*);
double (*opti_second_order_ptr) (int*, double*, double*, double*, double*, int*, int*, int*);
double (*opti_second_order_obs_ptr) (int*, double*, double*, double*, double*, int*, int*, int*);

inline double opti_static_obs (int *n, double* X, double* Xmin, double* Xmax, double* duser, int *iuser, int *iret, int *ier) {
  ier[0] = 0;
  double res = 0, acc=0;

  for (int i=0; i<n[0]; i=i+2) {
    if (duser[1] == 0)
      /*res += X[i+1] * (X[i] - Xmin[i]) * (X[i] - Xmin[i]);*/
      res += (X[i] - Xmin[i]) * (X[i] - Xmin[i]);
    else
      /*res += X[i+1] * (X[i] - Xmax[i]) * (X[i] - Xmax[i]);*/
      res += (X[i] - Xmax[i]) * (X[i] - Xmax[i]);
    acc -= X[i+1] * X[i];
  }
  res += (duser[2] + acc) * (duser[2] + acc);
  return res;
}

inline objet opti_static_eval (int n, cGraph cg, runtime rt) {
  influence ifl;
  condition cd;
  objet res = mkObjReal (0), otmp;
  
  for (int i=0; i<n; i++) {
    ifl = getCGCauseNth (cg, i);
    cd = getInfluCond (ifl);
    if (! isNull (cd)) {
      if (CE_FALSE == ce_pushCond (ifl, rt))
	continue;
    }
    ce_loadCauseValue (rt, ifl);
    otmp = copyObjet (TMP2);
    ce_mulObjet (otmp, getInfluB (ifl));      /* marginal influence B */
    ce_addObjet (res, otmp);                 /* accumulation */
    freeObjet (otmp);
  }
  return res;
}

inline double opti_integral_obs (int *n, double* X, double* Xmin, double* Xmax, double* duser, int *iuser, int *iret, int *ier) {
  double res = 0, acc = 0;
  
  ier[0] = 0;
  if (duser[1] == 0) /* minimization cycle */
    res = (X[0] - Xmin[0]) * (X[0] - Xmin[0]) + (X[1] - Xmin[1]) * (X[1] - Xmin[1]);
  else /* maximization cycle */
    res = (X[0] - Xmax[0]) * (X[0] - Xmax[0]) + (X[1] - Xmax[1]) * (X[1] - Xmax[1]);
  acc = X[0] + X[2]*X[1];
  res += (duser[2] - acc) * (duser[2] - acc);
  return res;
}

inline objet opti_integral_eval (int n, cGraph cg, objet y, runtime rt) {
  objet res = mkObjReal (0), otmp;
  influence ifl = getCGCauseNth (cg, 0);

  ce_loadCauseValue (rt, ifl);
  otmp = copyObjet (TMP2);
  cout << "y: ";
  ce_printObjet (stdout, y);
  cout << endl;

  ce_mulObjet (otmp, getInfluB (ifl));
  ce_addObjet (res, otmp);
  ce_addObjet (res, y);
  
  freeObjet (otmp);

  return res;
}

inline double opti_first_order_obs (int *n, double* X, double* Xmin, double* Xmax, double* duser, int *iuser, int *iret, int *ier) {
  return 0;
}

inline double opti_second_order (int *n, double* X, double* Xmin, double* Xmax, double* duser, int *iuser, int *iret, int *ier) {
  ier[0] = 0;
  double res = 2*exp(-(X[3]/(2*X[4]))*duser[0])*cos(sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4]))))*X[5] 
    - exp(-(X[3]/X[4])*duser[0])*X[6] 
    + X[9]*((X[1]/(X[4]*sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4])))))*exp(-(X[3]/(2*X[4]))*duser[0])*sin(sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4])))) + (X[0]/X[2])*(1-exp(-(X[3]/(2*X[4]))*duser[0])*cos(sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4])))) - ((X[3]/(2*X[4]))/(sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4])))))*exp(-(X[3]/(2*X[4]))*duser[0])*sin(sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4]))))))*X[7] 
    + X[9]*(-(X[1]/(X[4]*sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4])))))*exp(-(X[3]/(2*X[4]))*duser[0])*sin(sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4])))) + (X[0]/X[2])*(exp(-(X[3]/X[4])*duser[0]) + ((X[3]/(2*X[4]))/(sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4])))))*exp(-(X[3]/(2*X[4]))*duser[0])*sin(sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4])))) - exp(-(X[3]/(2*X[4]))*duser[0])*cos(sqrt(X[2]/X[4] - (X[3]/(2*X[4]))*(X[3]/(2*X[4]))))))*X[8];
  if (duser[1] == 0) return res;
  else return -res;
}

inline double opti_second_order_obs (int *n, double* X, double* Xmin, double* Xmax, double* duser, int *iuser, int *iret, int *ier) {
  double res = 0, acc=0;
  
  ier[0] = 0;
  for (int i=5; i<n[0]-1; i++) {
    if (duser[1] == 0)
      res += (X[i] - Xmin[i]) * (X[i] - Xmin[i]);
    else
      res += (X[i] - Xmax[i]) * (X[i] - Xmax[i]);
  }
  duser[1] = 0;
  acc = opti_second_order (n, X, Xmin, Xmax, duser, iuser, iret, ier);
  res += (duser[2] - acc) * (duser[2] - acc);
  return res;
}

inline objet opti_second_order_eval (int n, cGraph cg, objet y, objet yp, runtime rt) {
  objet res = mkObjReal (0), otmp1, otmp2;
  influence ifl = getCGCauseNth (cg, 0);

  ce_loadCauseValue (rt, ifl);
  otmp1 = copyObjet (TMP2);
  otmp2 = copyObjet (TMP5);
  
  ce_mulObjet (otmp1, getInfluB (ifl));
  ce_mulObjet (otmp2, getInfluB2 (ifl));
  ce_addObjet (res, otmp1);
  ce_addObjet (res, otmp2);

  ce_putObjet (y, otmp1);
  ce_putObjet (yp, otmp2);
  ce_mulObjet (otmp1, getInfluA (ifl));
  ce_mulObjet (otmp2, getInfluA2 (ifl));
  ce_addObjet (res, otmp1);
  ce_addObjet (res, otmp2);
  
  freeObjet (otmp1);
  freeObjet (otmp2);

  return res;
}

void init_nelder () {
  opti_static_obs_ptr = &opti_static_obs;
  opti_integral_obs_ptr = &opti_integral_obs;
  opti_first_order_obs_ptr = &opti_first_order_obs;
  opti_second_order_ptr = &opti_second_order;
  opti_second_order_obs_ptr = &opti_second_order_obs;
}

/* second order optimization without observations */
void call_nelder_second (influence ifl, runtime rt, objet y, objet yp, objet non_opt_res) {
  int *n, *iuser=0, *iret, *ier;
  double *X, *X2, *Xmin, *Xmax, *duser;
  
  n = (int*) malloc (sizeof (int));
  duser = (double*) malloc (3 * sizeof (double));
  duser[0] = getRTTimeStep (rt); /* NO ! */
  duser[1] = 0;  /* minimization cycle */

  n[0]= 10;
  X = (double*) malloc (10 * sizeof (double));
  Xmin = (double*) malloc (10 * sizeof (double));
  Xmax = (double*) malloc (10 * sizeof (double));
  mean2ndOrderCoefs (ifl, X, Xmin, Xmax);  /* set the mean as init value */
  meanCoef (y, X, Xmin, Xmax, 5);
  meanCoef (yp, X, Xmin, Xmax, 6);
  ce_loadCauseValue (rt, ifl);
  meanCoef (TMP2, X, Xmin, Xmax, 7);
  meanCoef (TMP5, X, Xmin, Xmax, 8);
  
  iret = (int*) malloc (sizeof (int));
  ier = (int*) malloc (sizeof (int));   /* error code */
  
  opt_nelder (opti_second_order_ptr, n, X, Xmin, Xmax, duser, iuser, ier); /* optimize (minimize) bounds */

  double mine = opti_second_order (n, X, Xmin, Xmax, duser, iuser, iret, ier);

#ifdef VERB_OPTI
  for (int i=0; i<n[0]; i++)
    cout << "X[" << i << "] = " << X[i] << " -- min: " << Xmin[i] << " -- max: " << Xmax[i] << endl;
  if ((getObjType (non_opt_res) == CE_RANGE && (mine  > getRangeSup (getObjRangeVal (non_opt_res)) 
						|| mine  < getRangeInf (getObjRangeVal (non_opt_res))))
      || getObjType (non_opt_res) == CE_REAL && mine  != getObjRealVal (non_opt_res)) {
    cout << "ier: " << ier[0] << endl;
    cout << "Error in optimized bounds (min): " << mine << " not in [" << getRangeInf (getObjRangeVal (non_opt_res)) << ", " << getRangeSup (getObjRangeVal (non_opt_res)) << "]\n";
  }
  else cout << "mine: " << mine << " in [" << getRangeInf (getObjRangeVal (non_opt_res)) << ", " << getRangeSup (getObjRangeVal (non_opt_res)) << "]\n";
#endif
  
  X2 = (double*) malloc (10 * sizeof (double));
  mean2ndOrderCoefs (ifl, X2, Xmin, Xmax);  /* set the mean as init value */
  meanCoef (y, X2, Xmin, Xmax, 5);
  meanCoef (yp, X2, Xmin, Xmax, 6);
  ce_loadCauseValue (rt, ifl);
  meanCoef (TMP2, X2, Xmin, Xmax, 7);
  meanCoef (TMP5, X2, Xmin, Xmax, 8);
  
  duser[1] = 1;  /* maximization cycle */
  opt_nelder (opti_second_order_ptr, n, X2, Xmin, Xmax, duser, iuser, ier);   /* optimize (maximize) bounds */

  double maxe = - opti_second_order (n, X2, Xmin, Xmax, duser, iuser, iret, ier);
  
#ifdef VERB_OPTI
  for (int i=0; i<n[0]; i++)
    cout << "X2[" << i << "] = " << X2[i] << " -- min: " << Xmin[i] << " -- max: " << Xmax[i] << endl;

  if ((getObjType (non_opt_res) == CE_RANGE && (maxe > getRangeSup (getObjRangeVal (non_opt_res)) 
						|| maxe - duser[2] < getRangeInf (getObjRangeVal (non_opt_res))))
      || getObjType (non_opt_res) == CE_REAL && maxe != getObjRealVal (non_opt_res)) {
    cout << "ier: " << ier[0] << endl;
    cout << "Error in optimized bounds (max): " << maxe << " not in [" << getRangeInf (getObjRangeVal (non_opt_res)) << ", " << getRangeSup (getObjRangeVal (non_opt_res)) << "]\n";
  }
  else cout << "maxe: " << maxe << " in [" << getRangeInf (getObjRangeVal (non_opt_res)) << ", " << getRangeSup (getObjRangeVal (non_opt_res)) << "]\n";
#endif
  
  /* update other variables bounds */
  setRangeInf (getObjRangeVal (TMP1), mine);
  setRangeSup (getObjRangeVal (TMP1), maxe);
  setNelderResult (y, X[5], X2[5]);
  setNelderResult (yp, X[6], X2[6]);
  setNelderResult (TMP2, X[7], X2[7]);
  setNelderResult (TMP5, X[8], X2[8]); 

  free (n);
  free (X);
  free (X2);
  free (Xmin);
  free (Xmax);
  free (iret);
  free (ier);
  free (duser);
}

/*void call_nelder_obs (variable v, runtime rt, objet y, objet yp, objet non_opt_res) {
  int *n, *iuser, *iret, *ier;
  double *X, *Xmin, *Xmax, *duser, *Xresmax, *Xresmin, 
  *Xinit, *Xinitmin, *Xinitmax;
  cGraph cg=getVarNode (v);
  long ncg=getCGNCause (cg), nrcg=0;
  influence ifl;
  condition cd;
  
  n = (int*) malloc (sizeof (int));
  n[0] = 0;
  for (long i=0; i<ncg; i++) {  
  ifl=getCGCauseNth(cg,i);
  cd=getInfluCond(ifl);
  if (! isNull (cd))
  if (CE_FALSE == ce_pushCond (ifl, rt))
	continue;
    n[0] = n[0] + 2;
    nrcg++;
  }

  cout << "v: " << getVarName (v) << endl;
  cout << "n[0]: " << n[0] << endl;
  duser = (double*) malloc (3 * sizeof (double));
  duser[0] = getRTTimeStep (rt);
  duser[1] = 0; 
  for (long k=0; k<getRTNMeasure (rt); k++) 
    if (getRTMeasureNth (rt, k) == getInfluEffet (ifl))
      duser[2] = getRTMesValNth (rt, k);
  iuser = (int*) malloc (sizeof (int));

  Xinit = (double*) malloc (n[0] * sizeof (double));
  Xinitmin = (double*) malloc (n[0] * sizeof (double));
  Xinitmax = (double*) malloc (n[0] * sizeof (double));
  Xresmin = (double*) malloc (ncg * sizeof (double));
  Xresmax = (double*) malloc (ncg * sizeof (double));

  int count = -1;
  
  for (long i=0; i<ncg; i++) { 
    ifl=getCGCauseNth(cg,i);
    cd=getInfluCond(ifl);
    if (! isNull (cd))
      if (CE_FALSE == ce_pushCond (ifl, rt))
	continue;
    cout << getInfluNom (ifl) << " active\n";
    ce_loadCauseValue (rt, ifl);
    meanCoef (TMP2, Xinit, Xinitmin, Xinitmax, count + 1);
    meanCoef (getInfluB (ifl), Xinit, Xinitmin, Xinitmax, count + 2);
    count += 2;
  }

  iret = (int*) malloc (sizeof (int));
  ier = (int*) malloc (sizeof (int)); 
  iuser[0] = 0;

  for (int i=0; i<nrcg; i++) {
  
    X = copyDouble (Xinit, n[0]);
    Xmin = copyDouble (Xinitmin, n[0]);
    Xmax = copyDouble (Xinitmax, n[0]);
    for (int j=0; j<n[0]; j++)
      cout << "X[" << j << "] = " << X[j] << " -- min: " << Xmin[j] << " -- max: " << Xmax[j] << endl;
    cout << "iuser[0]: " << iuser[0] << endl;
    cout << "static minimization cycle\n";
    cout << "duser[2]: " << duser[2] << endl;
    opt_nelder (opti_static_obs_ptr, n, X, Xmin, Xmax, duser, iuser, ier);
    for (int j=0; j<n[0]; j++)
      cout << "X[" << j << "] = " << X[j] << " -- min: " << Xmin[j] << " -- max: " << Xmax[j] << endl;
    duser[1] = 2;
    Xresmin[i] = opti_static_obs (n, X, Xmin, Xmax, duser, iuser, iret, ier);
    cout << "min: " << Xresmin[i] << endl;
    free (X);
    free (Xmin);
    free (Xmax);

  
    cout << "static maximization cycle\n";
    cout << "duser[2]: " << duser[2] << endl;
    duser[1] = 1;
    X = copyDouble (Xinit, n[0]);
    Xmin = copyDouble (Xinitmin, n[0]);
    Xmax = copyDouble (Xinitmax, n[0]);
    opt_nelder (opti_static_obs_ptr, n, X, Xmin, Xmax, duser, iuser, ier); 
    duser[1] = 2;
    Xresmax[i] = opti_static_obs (n, X, Xmin, Xmax, duser, iuser, iret, ier);
    for (int j=0; j<n[0]; j++)
      cout << "X[" << j << "] = " << X[j] << " -- min: " << Xmin[j] << " -- max: " << Xmax[j] << endl;
    cout << "max: " << Xresmax[i] << endl;
    free (X);
    free (Xmin);
    free (Xmax);

    iuser[0] = iuser[0] + 2;
  }

 
  for (long i=0; i<nrcg; i++) {
    ifl=getCGCauseNth(cg,i);
    cd=getInfluCond(ifl);
    if (! isNull (cd))
      if (CE_FALSE == ce_pushCond (ifl, rt))
	continue;
    if (! isParameter (getInfluCause (ifl)))
      setNelderResult (varNewValue (getInfluCause (ifl)), Xresmin[i], Xresmax[i]);
  }
  
  cout << "non-opt result: ";
  ce_printObjet (stdout, non_opt_res);
  cout << endl;

  free (n);
  free (Xinit);
  free (Xresmin);
  free (Xresmax);
  free (iret);
  free (ier);
  free (duser);
}*/

void call_nelder_obs (variable v, runtime rt, objet y, objet yp, objet non_opt_res, enum influType it, bool upd) {
  /*return;*/
  int *n, *iuser=0, *iret, *ier;
  double *X=0, *Xmin=0, *Xmax=0, *duser, *Xresmax=0, *Xresmin=0, *X2, *Xinit=0;
  cGraph cg=getVarNode (v);
  long ncg=getCGNCause (cg);
  influence ifl=0;
  condition cd;
  
  n = (int*) malloc (sizeof (int));
  n[0] = 0;
  
  int count = -1;

  switch (it) {
  case CE_SINFL:
    
    for (long i=0; i<ncg; i++) { 
      ifl=getCGCauseNth(cg,i);
      cd=getInfluCond(ifl);
      if (! isNull (cd))
	if (CE_FALSE == ce_pushCond (ifl, rt))
	  continue;
      n[0] = n[0] + 2;
    }
    if (n[0] == 0) {
      return;
    }
    Xresmin = (double*) malloc (ncg * sizeof (double));
    Xresmax = (double*) malloc (ncg * sizeof (double));
    Xinit = (double*) malloc (n[0] * sizeof (double));
    Xmin = (double*) malloc (n[0] * sizeof (double));
    Xmax = (double*) malloc (n[0] * sizeof (double));
    for (long i=0; i<ncg; i++) { 
      ifl=getCGCauseNth(cg,i);
      cd=getInfluCond(ifl);
      if (! isNull (cd))
	if (CE_FALSE == ce_pushCond (ifl, rt))
	  continue;
      ce_loadCauseValue (rt, ifl);  /* result in TMP2 */
      meanCoef (TMP2, Xinit, Xmin, Xmax, count + 1);
      meanCoef (getInfluB (ifl), Xinit, Xmin, Xmax, count + 2);
      count += 2;
    }
    break;
  case CE_IINFL:
    ifl = getCGCauseNth (cg, 0);  /* only one first-order influence on a variable */
    cd = getInfluCond (ifl);
    if (! isNull (cd))
      if (CE_FALSE == ce_pushCond (ifl, rt))  /* the influence isn't activated */
	return;
    n[0] = 3;
    Xresmin = (double*) malloc (sizeof (double));
    Xresmax = (double*) malloc (sizeof (double));
    Xinit = (double*) malloc (3 * sizeof (double));
    Xmin = (double*) malloc (3 * sizeof (double));
    Xmax = (double*) malloc (3 * sizeof (double));
    mean1stOrderCoefs (ifl, X, Xmin, Xmax);  /* set the mean as init value */
    meanCoef (y, Xinit, Xmin, Xmax, 0);
    ce_loadCauseValue (rt, ifl);
    meanCoef (TMP2, Xinit, Xmin, Xmax, 1);
    meanCoef (getInfluGain (ifl), Xinit, Xmin, Xmax, 2);
    break;
  case CE_D2INFL:
    ifl = getCGCauseNth (cg, 0);  /* only one second-order influence on a variable */
    cd = getInfluCond (ifl);
    if (! isNull (cd))
      if (CE_FALSE == ce_pushCond (ifl, rt))  /* the influence isn't activated */
	return;
    n[0]= 10;
    Xresmin = (double*) malloc (4 * sizeof (double));
    Xresmax = (double*) malloc (4 * sizeof (double));
    Xinit = (double*) malloc (10 * sizeof (double));
    Xmin = (double*) malloc (10 * sizeof (double));
    Xmax = (double*) malloc (10 * sizeof (double));
    mean2ndOrderCoefs (ifl, Xinit, Xmin, Xmax);  /* set the mean as init value */
    meanCoef (y, Xinit, Xmin, Xmax, 5);
    meanCoef (yp, Xinit, Xmin, Xmax, 6);
    ce_loadCauseValue (rt, ifl);
    meanCoef (TMP2, Xinit, Xmin, Xmax, 7);
    meanCoef (TMP5, Xinit, Xmin, Xmax, 8);
    break;
  default:
    cout << "default: " << it << endl;
    break;
  } 

#ifdef VERB_OPTI
  cout << "Optimizing on v: " << getVarName (v) << endl;
#endif
  
  duser = (double*) malloc (3 * sizeof (double));
  duser[0] = getRTTimeStep (rt);
  duser[1] = 0;  /* minimization cycle */

  if (upd) { /* cycle with obs */
  for (long k=0; k<getRTNMeasure (rt); k++)   /* measure */
    if (getRTMeasureNth (rt, k) == getInfluEffet (ifl))
      duser[2] = getRTMesValNth (rt, k);
  }
  else duser[2] = getRangeInf (getObjRangeVal (non_opt_res));

  iret = (int*) malloc (sizeof (int));
  ier = (int*) malloc (sizeof (int));   /* error code */

    
  X = copyDouble (Xinit, n[0]); 
  X2 = copyDouble (Xinit, n[0]);
#ifdef VERB_OPTI
  cout << "n: " << *n << endl;
  cout << "it: " << it << endl;
  for (int j=0; j<n[0]; j++)
    cout << "X[" << j << "] = " << X[j] << " -- min: " << Xmin[j] << " -- max: " << Xmax[j] << endl;  
  cout << "static minimization cycle\n";
#endif
  switch (it) {
  case CE_SINFL:
    /* minimization cycle */  
    opt_nelder (opti_static_obs_ptr, n, X, Xmin, Xmax, duser, iuser, ier); /* optimize (minimize) bounds */
    count = -1;
    for (int j=0; j<n[0]; j=j+2) {
#ifdef VERB_OPTI
      cout << "X[" << j << "] = " << X[j] << " -- min: " << Xmin[j] << " -- max: " << Xmax[j] << endl;
#endif
      Xresmin[count+1] = X[j];
      count++;
    }

    /* maximization cycle */
#ifdef VERB_OPTI
  cout << "static maximization cycle\n";
#endif
  duser[1] = 1;
  if (! upd) duser[2] = getRangeSup (getObjRangeVal (non_opt_res));
  opt_nelder (opti_static_obs_ptr, n, X2, Xmin, Xmax, duser, iuser, ier); /* optimize (maximize) bounds */
  
  count = -1;
  for (int j=0; j<n[0]; j=j+2) {
#ifdef VERB_OPTI
    cout << "X2[" << j << "] = " << X2[j] << " -- min: " << Xmin[j] << " -- max: " << Xmax[j] << endl;
#endif
    Xresmax[count+1] = X2[j];
    count++;
  }
  break;
  case CE_IINFL:
    opt_nelder (opti_integral_obs_ptr, n, X, Xmin, Xmax, duser, iuser, ier); /* optimize (minimize) bounds */
#ifdef VERB_OPTI
    cout << "X[" << 1 << "] = " << X[1] << " -- min: " << Xmin[1] << " -- max: " << Xmax[1] << endl;
#endif
    Xresmin[0] = X[1];
    duser[1] = 1;
    if (! upd) duser[2] = getRangeSup (getObjRangeVal (non_opt_res));
    opt_nelder (opti_integral_obs_ptr, n, X2, Xmin, Xmax, duser, iuser, ier); /* optimize (maximize) bounds */
#ifdef VERB_OPTI
    cout << "X2[" << 1 << "] = " << X2[1] << " -- min: " << Xmin[1] << " -- max: " << Xmax[1] << endl;
#endif
    Xresmax[0] = X2[1];
    break;
  case CE_D2INFL:
    opt_nelder (opti_second_order_obs_ptr, n, X, Xmin, Xmax, duser, iuser, ier); /* optimize (minimize) bounds */
    for (int j=5; j<n[0]-1; j++) {
#ifdef VERB_OPTI
      cout << "X[" << j << "] = " << X[j] << " -- min: " << Xmin[j] << " -- max: " << Xmax[j] << endl;
#endif
      Xresmin[j-5] = X[j];
    }
    
    /* maximization cycle */
#ifdef VERB_OPTI  
    cout << "static maximization cycle\n";
#endif
    duser[1] = 1;
    opt_nelder (opti_second_order_obs_ptr, n, X2, Xmin, Xmax, duser, iuser, ier); /* optimize (maximize) bounds */
    
    for (int j=5; j<n[0]-1; j++) {
#ifdef VERB_OPTI
      cout << "X2[" << j << "] = " << X2[j] << " -- min: " << Xmin[j] << " -- max: " << Xmax[j] << endl;
#endif
      Xresmax[j-5] = X2[j];
    }
    break;
  default:
    break;
  } /* end switch */

  /* set results */
  /* TODO: must envision before making changes... */
  count = 0;
  for (long i=0; i<ncg; i++) {
    ifl=getCGCauseNth(cg,i);
    cd=getInfluCond(ifl);
    if (! isNull (cd))
      if (CE_FALSE == ce_pushCond (ifl, rt))
	continue;
    if (! isParameter (getInfluCause (ifl)) 
	&& ! ce_isMeasure (rt, getInfluCause (ifl)) && ! ce_isInput (rt, getInfluCause (ifl))
	&& ! (isVarLowAlrm (getInfluCause (ifl)) || isVarHighAlrm (getInfluCause (ifl)))) {
#ifdef VERB_OPTI
      cout << "Xi: " << getVarName (getInfluCause (ifl)) << endl;
      cout << "non-opt value: ";
      ce_printObjet (stdout, varNewValue (getInfluCause (ifl)));
      cout << endl;
      cout << "opt bounds = [" << Xresmin[count] << ", " << Xresmax[count] << "]\n"; 
#endif
      long delsup = getInfluDelSup (ifl);
      objet neo;
      if (CE_SINFL == getInfluType (ifl))
	delsup--;
      if (delsup > 0)
	neo = varHistoNth (getInfluCause (ifl), delsup-1);
      else if (delsup == 0)
	neo = varOldValue (getInfluCause (ifl));
      else neo = varNewValue (getInfluCause (ifl));
      setNelderResult (neo, Xresmin[count], Xresmax[count]);
      count++;
    }
  }
 
#ifdef VERB_OPTI 
  cout << "non-opt result: ";
  ce_printObjet (stdout, non_opt_res);
  cout << endl;
#endif
  objet ores=0;
  switch (it) {
  case CE_SINFL:
    ores = opti_static_eval (ncg, cg, rt);  /* Yk+1 */
    break;
  case CE_IINFL:
    ores = opti_integral_eval (ncg, cg, y, rt);
    break;
  case CE_D2INFL:
    /*ores = opti_second_order_eval (ncg, cg, y, yp, rt);*/
    ifl = getCGCauseNth (cg, 0);  /* only one second-order influence on a variable */
    call_nelder_second (ifl, rt, y, yp, non_opt_res);
    break;
  default:
    break;
  } /* end switch */

  if (it != CE_D2INFL && duser[2] <= getRangeSup (getObjRangeVal (ores))
      && duser[2] >= getRangeInf (getObjRangeVal (ores))) {
#ifdef VERB_OPTI 
    cout << "opt-result: ";
    ce_printObjet (stdout, ores);
    cout << endl;
#endif
    ce_putObjet (ores, non_opt_res);
  }
#ifdef VERB_OPTI 
  else if (! upd) {
    cout << "opt-result: ";
    ce_printObjet (stdout, ores);
    cout << endl;
  }
  else if (it != CE_D2INFL) cout << "opti alarming\n";
#endif
  
  if (it != CE_D2INFL)
    freeObjet (ores);

  free (X);
  free (Xmin);
  free (Xmax);
  free (X2);
  free (n);
  free (Xinit);
  free (Xresmin);
  free (Xresmax);
  free (iret);
  free (ier);
  free (duser);

  /* update previous influences */
  /*for (long i=0; i<ncg; i++) {
    ifl = getCGCauseNth (cg, i);
    cd=getInfluCond(ifl);
    if (! isNull (cd))
      if (CE_FALSE == ce_pushCond (ifl, rt))
	continue;
    if (! isParameter (getInfluCause (ifl)) && ! ce_isMeasure (rt, getInfluCause (ifl)) && getObjType (varNewValue (getInfluCause (ifl))) == CE_RANGE)
      call_nelder_obs (getInfluCause (ifl), rt, varNewValue (getInfluCause (ifl)), varOldValue (getInfluCause (ifl)), varNewValue ((getInfluCause (ifl))), getCGCauseType (getVarNode (getInfluCause (ifl))), 0);
  }*/

}

void meanCoef (objet ob, double* X, double* Xmin, double* Xmax, int i) {
  if (getObjType (ob) == CE_RANGE) {
    X[i] = 0.5*(getRangeSup (getObjRangeVal (ob)) + getRangeInf (getObjRangeVal (ob)));
    Xmin[i] = getRangeInf (getObjRangeVal (ob));
    Xmax[i] = getRangeSup (getObjRangeVal (ob));
  }
  else {
    X[i] = getObjRealVal (ob);
    Xmin[i] = getObjRealVal (ob);
    Xmax[i] = getObjRealVal (ob);
  }
}

void mean1stOrderCoefs (influence ifl, double* X, double* Xmin, double* Xmax) {
  meanCoef (getInfluTRep (ifl), X, Xmin, Xmax, 2);
  meanCoef (getInfluGain (ifl), X, Xmin, Xmax, 3);
}

/**
 * set the optimization vector values for 2nd order influencse:
 * X[0] = n1
 * X[1] = n2
 * X[2] = d0
 * X[3] = d1
 * X[4] = d2
 * X[5] = Y(k)
 * X[6] = Y(k-1)
 * X[7] = X(k)
 * X[8] = X(k-1)
 **/
void mean2ndOrderCoefs (influence ifl, double* X, double* Xmin, double* Xmax) {
  if (! isNull (getInfluN1 (ifl))) {       /* general case */
    meanCoef (getInfluN1 (ifl), X, Xmin, Xmax, 0);
    meanCoef (getInfluN2 (ifl), X, Xmin, Xmax, 1);
    meanCoef (getInfluD0 (ifl), X, Xmin, Xmax, 2);
    meanCoef (getInfluD1 (ifl), X, Xmin, Xmax, 3);
    meanCoef (getInfluD2 (ifl), X, Xmin, Xmax, 4);
    meanCoef (getInfluGain (ifl), X, Xmin, Xmax, 9);
  }
  else {
    
  }
}

/* set optimized bounds */
void setNelderResult (objet obj, double d1, double d2) {
  if (getObjType (obj) == CE_RANGE) {
    setRangeInf (getObjRangeVal (obj), min (d1, d2));
    setRangeSup (getObjRangeVal (obj), max (d1, d2));
  }
}

/**
 *
 * hybrid trajectories validity space filtering algorithm.
 *
 */

void viewValSpc (runtime rt) {
  cout << "rt " << rt->num << " valSpc => ";
  for (long i=0; i<getRTNCondC (rt); i++) {
    if (isCcInValSpc (rt,i)) {
      ce_printExpr (stdout, getRTCondCNth (rt, i));
      cout << " ";
    }
  }
}

long ce_initValSpc (runtime rt) {
  long r=0;
  condition cd;
  vector<condition>* cds;
  Atom* ev;
  runtime cur_rt;
  bool add;
  vector<condition>* nconds = new vector<condition>();

  if (getRTNGVar (rt) == 0)
    return 1;

  for (long i=0; i<getRTNGVar (rt); i++) {
    if (! isNull (getGVVar (getRTGVarNth (rt, i))) 
	&& getGVNCondC (getRTGVarNth (rt, i)) != 0) {
      if (! isNull (cds = ce_findCondCover (getRTGVarNth (rt, i), rt))) {
	
	for (unsigned int k=0; k<cds->size (); k++) {
	  cd  = (*cds)[k];
#ifdef VERB_SPLITS
	  cout << "init => condition:\n";
	  ce_printCondition (stdout, cd);
	  if (! isNull (getCondAtom (cd))) {
	    getCondAtom (cd)->showAtom (0);
	    cout << endl;
	  }
#endif
	  
	  nconds->push_back(cd);
	  for (long j=0; j<getCondNle (cd); j++)
	    setCcInValSpc (rt, getCondLeNth (cd, j));
	}
	delete cds;
	r = 1;
      }
    }
  }
  
  ce_splitValiditySpc (rt); /* splitting the state-space */
#ifdef VERB_SPLITS
  viewValSpc (rt);
#endif
  
  /* check for conflicts on initial atoms */
  for (multimap<long, runtime, less<long> >::iterator it = nrt->begin ();
       it != nrt->end (); it++) {
    cur_rt = (*it).second;
    for (unsigned int j=0; j<nconds->size (); j++) {
      add = true;
      for (multimap<bool, Atom*, eqbool>::iterator it2 = cur_rt->bridgeEvents->begin ();
	   it2 != cur_rt->bridgeEvents->end (); it2++) {
	ev = (*it2).second;
	if (! isNull (getCondAtom ((*nconds)[j]))
	    && ev->stateVar == getCondAtom ((*nconds)[j])->stateVar) {
	  add = false;
	  break;
	}
      }
      if (add)
	addBridgeEvent ((*nconds)[j], getCondAtom ((*nconds)[j]), 1, cur_rt);
    }
  }
  
  delete nconds;
  return r;
}

void ce_fillCondAgenda (runtime rt) {
  long i;

  for (i=0; i<getRTNCondC (rt); i++)
    if (isCcInValSpc (rt,i))
      setCcInAgenda (rt, i);

  return;
}

void ce_setOutOfValSpc (runtime rt, condition cd) {
  enum condType ct = getCondType (cd);
  long i;

  switch (ct) {
  case CE_NON:
  case CE_EXPR:
    for (i=0; i<getCondNle (cd); i++) {
      setCcOutOfValSpc (rt, getCondLeNth (cd, i));
      setCcOutOfAgenda (rt, getCondLeNth (cd, i));
    }
    break;
  case CE_ET:
    ce_setOutOfValSpc (rt, getCondOp1 (cd));
    ce_setOutOfValSpc (rt, getCondOp2 (cd));
    break;
  case CE_OU:
    ce_setOutOfValSpc (rt, getCondOp1 (cd));
    ce_setOutOfValSpc (rt, getCondOp2 (cd));
    break;
  case CE_RIEN:
  case CE_CTE:
  case CE_COMP:
  case CE_AT:
  default:
    break;
  } /* end switch */
}

void ce_setInValSpc (runtime rt, condition cdb) {
  enum condType ct = getCondType (cdb);
  long i;

  switch (ct) {
  case CE_NON:
  case CE_EXPR:
    for (i=0; i<getCondNle (cdb); i++)
      setCcInValSpc (rt, getCondLeNth (cdb, i));
    break;
  case CE_ET:
    ce_setInValSpc (rt, getCondOp1 (cdb));
    ce_setInValSpc (rt, getCondOp2 (cdb));
    break;
  case CE_OU:
    ce_setInValSpc (rt, getCondOp1 (cdb));
    ce_setInValSpc (rt, getCondOp2 (cdb));
    break;
  case CE_RIEN:
  case CE_CTE:
  case CE_COMP:
  case CE_AT:
  default:
    break;
  } /* end switch */
}

/* DEPRECATED ? */
void ce_updateValSpc (runtime rt, condition cd, condition cdb) {

  ce_setOutOfValSpc (rt, cd);
  ce_setInValSpc (rt, cdb);
  
  /*viewValSpc (rt);*/
  
  return;
}

long ce_updateCondAgenda (runtime rt, gVar gv) {
  long i, r=0;
  
  for (i=1; i<getGVNCondC (gv); ++i) {
    if (isCcInValSpc (rt, getGVCondCNth (gv, i))) {
      setCcInAgenda (rt, getGVCondCNth (gv, i));
      r = 1;
    }
  }
  return r;
}

boolean ce_evalCondExpr (condition cd, runtime rt, bool inter) {
  boolean r = CE_FALSE;
  boolean r1, r2;
  objet ob1, ob2;
  
  if (getCondType (cd) == CE_ET) {
    r1 = ce_evalCondExpr (getCondOp1 (cd), rt, 1);
    r2 = ce_evalCondExpr (getCondOp2 (cd), rt, 1);
    if (r1 == CE_TRUE && r2 == CE_TRUE)
      r = CE_TRUE;
  }
  else if (getCondType (cd) == CE_EXPR) {
    ob1 = gVVal (getExprGVar (getCondExpr (cd)));  /* get the expression from the condition */
    
    /*cerr << "[evalCond ->] ob1:";
      ce_printObjet (stderr, ob1);
      cerr << endl;*/
    
    ce_evalSolFunc (rt, getCondExpr (cd));

    /*cerr << "[evalCond ->] eval:";
      ce_printObjet (stderr, TMP4);
      cerr << endl;*/
    
    ob2 = copyObjet (TMP4);
    /*if (! inter && ((CE_TRUE == ce_inObjet (ob1, ob2)) 
		    || (CE_TRUE == ce_eqObjet (ob1, ob2)))) {
		    r = CE_TRUE;*/
      /*cefree (ob2);*/
    /*}
      else if (inter) {*/
      objet tmp = copyObjet (ob1);
      if ((CE_TRUE == ce_interObjet (tmp, ob2)) || (CE_TRUE == ce_eqObjet (ob1, ob2))) {
	/*cerr << "[evalCond ->] tmp:";
	  ce_printObjet (stderr, tmp);
	  cerr << endl;*/
	r = CE_TRUE;
      }
      freeObjet (tmp);
      /*}*/
    freeObjet (ob2);
  }
  /*cout << "[evalCondExpr->] returning: " << r << endl;*/
  return r;
}

/* find new condition as validity space cover */
vector<condition>* ce_findCondCover (gVar gvc, runtime rt) {
  condition cd;
  vector<condition>* res = new vector<condition> ();
  variable v = getGVVar (gvc);

  for (int i=0; i<getVarNcd (v); i++) {
    cd = getRTCondiNth (rt, getVarRelCdNth (v, i));
    /*cout << "findCover testing:\n";
      ce_printCondition (stdout, cd);
      cout << endl;*/
    if (CE_TRUE == ce_evalCondExpr (cd, rt, 0)) {
#ifdef VERB_SPLITS
      cout << "[findCover -->] returning :\n";
      ce_printCondition (stdout, cd);
      cout << endl;
#endif
      res->push_back (cd);
    }
  }
  if (! res->empty ())
    return res;
  else {
/*#ifdef VERB_SPLITS*/
    cout << "[findCover -->] returning NULL\n";
    exit (0);
/*#endif*/
    /*delete res;
      return 0;*/
  }
}

void ce_splitRt (runtime crt, objet* spcs, long i, queue<runtime>* trajs) {
  vector<condition>* cds;
  runtime srt;
  gVar gvc;
  expr e;

  for (int j=0; j<2; j++) {
    if (! isNull (spcs[j])) {
      srt = ce_copyRuntime (crt);
      e = getRTCondCNth (srt, i);
      gvc = getExprGVar (e);
#ifdef VERB_SPLITS
      cout << "splitting on " << getVarName (getGVVar (gvc)) << ": " << endl;
      viewSplitState (gvc, gVVal (gvc), spcs[j]);
#endif
      /*cout << "actual condition cover\n";*/
      cds = ce_findCondCover (gvc, srt);
      
      /*cerr << "splitted objet";
	ce_printObjet (stderr, spcs[j]);
	cout << endl;*/
      
      ce_putObjet (spcs[j], gVVal (gvc));
      for (unsigned int k=0; k<cds->size (); k++) {
	ce_setOutOfValSpc (srt, (*cds)[k]);
	removeBridgeEvent ((*cds)[k], getCondAtom ((*cds)[k]), srt);
      }
      delete cds;
      
      /*cout << "new condition cover\n";*/
      if (! isNull (cds = ce_findCondCover (gvc, srt))) {
	/*vector<runtime>* vecrt = new vector<runtime> ();
	  vecrt->push_back (srt);*/
	bool flag;
	for (unsigned int l=0; l<cds->size (); l++) {
	  flag = 0;
	  condition ctmp = (*cds)[l];
	  for (unsigned int m=0; m<l; m++) {
	    /*if (! isNull (getCondAtom ((*cds)[m])) && ! isNull (getCondAtom (ctmp))) {*/
	      if (getCondAtom((*cds)[m])->getStateVar () 
		  == getCondAtom (ctmp)->getStateVar ()) {  /* if conditions belong to the same CondVar */
		flag = 1;
		break;
	      }
	      /*}*/
	  } /* end for m */
	  if (! flag) {
	    ce_setInValSpc (srt, ctmp);
	    addBridgeEvent (ctmp, getCondAtom (ctmp), 1, srt);
	  }
	}
	ce_updateCondAgenda (srt, gvc);
	trajs->push (srt);

	/*for (unsigned int k=0; k<cds->size (); k++) {
	  ce_setInValSpc (srt, (*cds)[k]);
	  addBridgeEvent ((*cds)[k], getCondAtom ((*cds)[k]), 1, srt);
	  }*/
	delete cds;
	/*ce_updateCondAgenda (srt, gvc);
	  trajs->push (srt);*/
      } 
      /*cout << "after bridge gv value:\n";
	ce_printObjet (stdout, gVVal (gvc));
	cout << endl;*/
    }
  }
  cefree (spcs);
}

void ce_emptyIsecValSpc (runtime rt, long i) {
  /*condition cd, cdb;*/
  vector<condition>* cds;
  gVar gv;
  expr e;
  
  e = getRTCondCNth (rt, i);
  gv = getExprGVar (e);
#ifdef VERB_SPLITS
  cout << "empty intersec on " << getVarName (getGVVar (gv)) << ": " << endl;
  viewSplitState (gv, TMP1, TMP4);
#endif
  ce_putObjet (TMP4, gVVal (gv));
  cds = ce_findCondCover (gv, rt);
  for (unsigned int j=0; j<cds->size (); j++) {
    ce_setOutOfValSpc (rt, (*cds)[j]);
    removeBridgeEvent ((*cds)[j], getCondAtom ((*cds)[j]), rt);
  }
  delete cds;

  ce_putObjet (TMP1, gVVal (gv));
  cds = ce_findCondCover (gv, rt);
  bool flag;
  for (unsigned int j=0; j<cds->size (); j++) {
    flag = 0;
    for (unsigned int k=0; k<j; k++) {
      if (getCondAtom ((*cds)[k])->getStateVar ()
	  == getCondAtom ((*cds)[j])->getStateVar ()) {
	flag = 1;
	break;
      }
    }
    if (! flag) {
      ce_setInValSpc (rt, (*cds)[j]);
      addBridgeEvent ((*cds)[j], getCondAtom ((*cds)[j]), 1, rt);
    }
  }
  delete cds;
}

/* debug */
void viewSplitState (gVar gv, objet va, objet fil) {
  cout << "\t\t|variable: " << getVarName (getGVVar (gv)) << endl;
  cout << "\n\t\t|current value: ";
  ce_printObjet (stdout, va);
  cout << "\n\t\t|filtered value: ";
  ce_printObjet (stdout, fil);
  cout << endl;
}

void viewCcAgenda (runtime rt) {
  cout << "Cc agenda:\n ";
  for (int i=0; i<getRTNCondC (rt); i++) {
    if (! isNull (getExprGVar (getRTCondCNth (rt, i)))) {
      ce_printGVar (stdout, getExprGVar (getRTCondCNth (rt, i)));
      cout << endl;
    }
    ce_printExpr (stdout, getRTCondCNth (rt, i));
    cout << " -- " << rt->inac[i] << endl;
  }
  cout << endl;
    
}

void ce_splitValiditySpc (runtime rt) {
  long i, k, emptyAg, emptyFilter;
  gVar gv;
  expr sf, sg;
  queue<runtime>* trajs = new queue<runtime>();
  runtime  crt;             /* current runtime */
  objet* spcs;
  /* TMP1: filtered value */
  /* TMP4: computed value */

  trajs->push (rt);
  
  /*cout << "splitValiditySpc: \n";
    viewValSpc (rt);
    cout << endl;*/

   ce_fillCondAgenda (rt);
    for (i=0; i<getRTNGVar(rt); ++i) {     /* initialization of global variables */
      gv = getRTGVarNth (rt, i);
      if (getGVNCondC (gv) != 0) {
	switch (getGVType (gv)) {
	case CE_INTERM :                          /* intermediate variable value */
	  ce_putObjet (objMax, gVVal (gv));
	  break;
	case CE_VAR :                  /* value of the model's associated variable */
	  if (ce_isMeasure (rt, getGVVar (gv))
	      && ! (isVarLowAlrm (getGVVar (gv)) 
		    || isVarHighAlrm (getGVVar (gv))))
	    ce_putObjet (varMesValue (getGVVar (gv)), gVVal (gv));
	  ce_putObjet (varOldValue (getGVVar (gv)), gVVal (gv));
	  break;
	default:
	  break;
	} /* end switch */
      }
    } /* end for */
    while (! trajs->empty() ) {
    crt = trajs->front ();
    trajs->pop();
   
    do {
      emptyAg = 1;
      
      /*viewCcAgenda (crt);*/
      
      for (i=0; i<getRTNCondC(crt); ++i) {
	if (isCcInAgenda (crt, i)) {
	  sf = getRTCondCNth (crt, i);
	  gv = getExprGVar (sf);
	  /*cerr << "gv: \n";
	    ce_printGVar (stderr, gv);
	    ce_printObjet (stderr, gVVal (gv));
	    fprintf (stderr, "\n");*/
      /*for (i=0; i<getRTNGVar (crt); i++) {
	gv = getRTGVarNth (crt, i);*/
	  ce_putObjet (gVVal (gv), TMP1);             /* initialize filtered value */
	  emptyFilter = 0;
	  for (k=0; k<getRTNCondC(crt); ++k) {
	    if (isCcInAgenda (crt, k)) {
	      sg = getRTCondCNth (crt, k);
	      if (gv == getExprGVar (sg)) {
		if (! emptyFilter) {
		  ce_evalSolFunc (crt, sg);                   /* result in TMP4 */
		  /*cout << "filtered value:\n";
		  ce_printObjet (stdout, TMP4);
		  cout << endl;*/
		  if (CE_FALSE == ce_interObjet (TMP1, TMP4)) {  /* result is in TMP1 */
		    ce_emptyIsecValSpc (crt, i);
		    emptyFilter = 1;
		    emptyAg = 0;
		    /*cerr << "TMP1: ";
		    ce_printObjet (stderr, TMP1);
		    cerr << endl;*/
		  }
		} /* endif ! emptyFilter */
		setCcOutOfAgenda (crt, k);
	      } /* endif gv == ... */
	    } /* endif isCcInAgenda k */
	  } /* end for k */
	  emptyFilter = 0;  /* hack : TODO get rid of emptyFilter */
	  if (! emptyFilter &&
	      (! isNull (spcs = ce_inSpObjet (TMP1, gVVal (gv))))) { /* strictly included */ 
	    ce_splitRt (crt, spcs, i, trajs);             /* split trajectory */
	    ce_putObjet (TMP1, gVVal (gv));
#ifdef DEBUG
	      fprintf (stderr, "  (%ld) ", i+1);
	      ce_printGVar (stderr, gv);
	      ce_printObjet (stderr, gVVal (gv));
	      fprintf (stderr, "\n");
#endif
	      if (ce_updateCondAgenda (crt, gv))
		emptyAg = 0;
	  } /* endif !emptyFilter && ... */
	} /* endif isGcInAgenda i */
      } /* end for i */
      /*cout << "emptyAg: " << emptyAg << endl;*/
    } while (! emptyAg);
    /*cout << "tracked_trajs size: " << trajs->size() << endl;
      cout << "new traj added\n";*/
    /*if (crt != rt)*/
    pushNrt (crt->num, crt);
    /*cout << "view runtime:\n";
    for (long mm=0; mm<getRTNVar (crt); mm++) {
      if (! isParameter (getRTDicoNth (crt, mm))) {
	cout << getVarName (getRTDicoNth (crt, mm)) << ": ";
	ce_printObjet (stdout, varNewValue (getRTDicoNth (crt, mm)));
	cout << endl;
      }
    }*/
    /*nrt->push_back (crt);*/
    for (i=0; i<getRTNGVar(rt); ++i) {     /* updating model variables */
      gv = getRTGVarNth (crt, i);
      if (CE_VAR == getGVType (gv)) {
	ce_putObjet (gVVal (gv), varOldValue (getGVVar (gv)));
	ce_putObjet (gVVal (gv), varNewValue (getGVVar (gv)));
      }
    } /* end for */
    ce_yieldDisplays (crt, 0);
  } /* end while trajs */

  delete trajs;

  return;
}


/**
 * 
 * global constraints filtering algorithms.
 * 
 */

/* */
objet ce_evalExpr (runtime rt, expr e) {
  ce_evalSolFunc (rt, e);
  return TMP4;
}

/**
 * eval function in global constraints.
 * result is in TMP4.
 */
void ce_evalSolFunc (runtime rt, expr e) {                 /* result in TMP4 */
  s_objet s_pmt1, s_pmt2;
  objet PMT1=&s_pmt1, PMT2=&s_pmt2;
  expr e1, e2;

  switch (getExprType (e)) {
  case CE_CSTE :
  case CE_VBL :
    ce_putObjet (gVVal (getExprGVar (e)), PMT1);
    break;
  case CE_PLUS :
  case CE_MOINS :
  case CE_MULT :
  case CE_DIV :
  case CE_EXP :
    e2 = getExprOp2 (e);
    if (isExprCte (e2) || hasName (e2))
      ce_putObjet (gVVal (getExprGVar (e2)), PMT2);
    else {
      ce_evalSolFunc (rt, e2);                             /* result in TMP4 */
      ce_putObjet (TMP4, PMT2);
    }
 case CE_SGN :
    e1 = getExprOp1 (e);
    if (isExprCte (e1) || hasName (e1))
      ce_putObjet (gVVal (getExprGVar (e1)), PMT1);
    else {
      ce_evalSolFunc (rt, e1);                             /* result in TMP4 */
      ce_putObjet (TMP4, PMT1);
    }
    break;
  default:
    break;
  }
  switch (getExprType (e)) {
  case CE_PLUS :
    ce_addObjet (PMT1, PMT2);
    break;
  case CE_MOINS :
    ce_subObjet (PMT1, PMT2);
    break;
  case CE_MULT :
    ce_mulObjet (PMT1, PMT2);
    break;
  case CE_DIV :
    ce_divObjet (PMT1, PMT2);
    break;
  case CE_EXP :
    ce_logObjet (PMT2);
    ce_mulObjet (PMT1, PMT2);
    ce_expObjet (PMT1);
    break;
  case CE_SGN :
    ce_minusObjet (PMT1);
    break;
  default:
    break;
  }
  ce_putObjet (PMT1, TMP4);
  return;
}

void ce_fillAgenda (runtime rt) {
  long i;

  for (i=0; i<getRTNGlobC(rt); ++i)
    setGcInAgenda (rt, i);
  return;
}


long ce_updateAgenda (runtime rt, gVar gv) {
  long i, r=0;

  for (i=1; i<=getGVNGlobC(gv); ++i) {
    setGcInAgenda (rt, getGVGlobCNth (gv, i));
    r = 1;
  } /* end for */
  return r;
}


void ce_globalConstraints (runtime rt) {
  long i, k, emptyAg, emptyFilter;
  gVar gv;
  expr sf, sg;
  /* TMP1: filtered value */
  /* TMP2: computed value */

  ce_fillAgenda (rt);
  for (i=0; i<getRTNGVar(rt); ++i) {     /* initialization of global variables */
    gv = getRTGVarNth (rt, i);
    switch (getGVType (gv)) {
    case CE_INTERM :                          /* intermediate variable value */
      ce_putObjet (objMax, gVVal (gv));
      break;
    case CE_VAR :                  /* value of the model's associated variable */
      ce_putObjet (varOldValue (getGVVar (gv)), gVVal (gv));
      break;
    default:
      break;
    } /* end switch */
  } /* end for */
  do {
    emptyAg = 1;
    for (i=0; i<getRTNGlobC(rt); ++i) {
      if (isGcInAgenda (rt, i)) {
	sf = getRTGlobCNth (rt, i);
	gv = getExprGVar (sf);
	ce_putObjet (gVVal (gv), TMP1);             /* initialize filtered value */
	emptyFilter = 0;
	for (k=0; k<getRTNGlobC(rt); ++k) {
	  if (isGcInAgenda (rt, k)) {
	    sg = getRTGlobCNth (rt, k);
	    if (gv == getExprGVar (sg)) {
	      if (! emptyFilter) {
		ce_evalSolFunc (rt, sg);                   /* result in TMP4 */
		if (CE_FALSE == ce_interObjet (TMP1, TMP4)) {
#ifdef VERB_GLB
		  cerr << "setting " << getVarName (getGVVar (gv)) << " to ";
		  ce_printObjet (stderr, TMP4);
		  cerr << endl;
#endif
		  ce_putObjet (TMP4, gVVal (gv)); /* added 28/12/01. was bug ? */
		  emptyFilter = 1;
		}
	      } /* endif ! emptyFilter */
	      setGcOutOfAgenda (rt, k);
	    } /* endif gv == ... */
	  } /* endif isGcInAgenda k */
	} /* end for k */
	if (! emptyFilter &&
	    (CE_TRUE == ce_inObjet (TMP1, gVVal (gv)))) {/* strictly included */	  
	  ce_putObjet (TMP1, gVVal (gv));
#ifdef VERB_GLB
	  fprintf (stderr, "  (%ld) ", i+1);
	  ce_printGVar (stderr, gv);
	  ce_printObjet (stderr, gVVal (gv));
	  fprintf (stderr, "\n");
#endif
	  if (ce_updateAgenda (rt, gv))
	    emptyAg = 0;
	} /* endif !emptyFilter && ... */
      } /* endif isGcInAgenda i */
    } /* end for i */
  } while (! emptyAg);
  for (i=0; i<getRTNGVar(rt); ++i) {     /* updating model variables */
    gv = getRTGVarNth (rt, i);
    if (CE_VAR == getGVType (gv))
      ce_putObjet (gVVal (gv), varOldValue (getGVVar (gv)));
  } /* end for */
  return;
}


void caen_initialize (runtime rt, long global, double prec, char *diagBuf,
		      long cumul) {
  long i, k, oldCumul, nmes=getRTNMeasure(rt);
  void *allocAux;
  variable v;

  ce_updateDico (rt);
  ce_evalInfluCoefs (rt, getRTTimeStep (rt)); /* init a, etc... */
  ce_enterInputs (rt);
  ce_enterMeasures (rt);
  ce_initVars (rt);      /* splitting is performed here */
  ce_moveInputs (rt);

  std::cerr << "global: " << global << std::endl;
  if (global) {
    setGCRequired (rt);
    ce_globalConstraints (rt);
  }
  else
    setGCUnRequired (rt);
  ce_fillAllHistos (rt);
  ce_fillAllHistosMes (rt);
  ce_yieldDisplays (rt, 0);
  setRTPrecision (rt, prec);
  ce_checkMs (rt, 2);
  ce_yieldAlarms (rt);
  setRTDiagOut (rt, diagBuf);
  if ((nmes != 0) && (cumul != 0)) {
    oldCumul = getRTMisbCumTime (rt);
    for (i=0; i<nmes; ++i) {
      v=getRTMeasureNth (rt, i);
      if (! isNull (allocAux = getVarHistoMisb (v)))
	new_free (allocAux, oldCumul * sizeof (long));
      allocAux = new_malloc (cumul*sizeof (long));
      setVarHistoMisb (v, (long *) allocAux);
      for (k=0; k<cumul; ++k)
	setVarHistoMisbNth (v, k, 0);
    } /* end for */
    setRTMisbCumTime (rt, cumul);
  } /* endif */
  init_nelder (); /* optimization functions pointers initialization */
  return;
}

long caen_doCycle (runtime rt, long mode, long seuil) {
  long io=0, im=0;
  
  if (mode <= 1)                   /* 0: simul only; 1: simul and reset; */
    ce_setAlarms (rt);             /* to prevent evaluation from using measured */
  if (mode <= 5) {                 /* values, except at reset times */
    ce_doStepTempo (rt);          /* 2: detect only; 3, 4, 5: detect and isolate */
    if (! ce_enterInputs (rt))
      ++io;
    if (! ce_enterMeasures (rt))
      ++io;
    
    ce_doStepUntemp (rt);
    
    if (2 == ce_yieldMRTConds (rt))   { /* new runtimes are created here */
      im = 1;
    }
    for (multimap<long, runtime, less<long> >::iterator nrt_it = nrt->begin();
	 nrt_it != nrt->end(); nrt_it++) {
      if (areGCRequired ((*nrt_it).second))
	ce_globalConstraints ((*nrt_it).second);
      if (! ce_yieldDisplays ((*nrt_it).second, 0))
	++io;
    }
    switch (mode) {
    case 0 :                               /* 0: simul only */
      break;
    case 1 :                               /* 1: simul and reset */
      ce_resetMs (rt);
      break;
    default:                               /* 2: detect only */
      for (multimap<long, runtime, less<long> >::iterator nrt_it = nrt->begin();
	   nrt_it != nrt->end(); nrt_it++) {
	ce_checkMs ((*nrt_it).second, seuil);             
	if (! ce_yieldAlarms ((*nrt_it).second))
	  ++io;
      }
      /* mode 3 is not in the loop for now... */
      if (mode >= 3) {                     /* 3, 4, 5: detect and isolate */ 
	if (faultIsolationSetup (rt, 1) == MEM_ERROR)
	  ++io;
	ce_prepareDiag (rt);
	hittingSetDiagnosis (rt, mode-2);       /* procedure = mode-2 */
	ce_yieldDiags (rt);
      }
    } /* end switch */
  }
  else { /* mode > 5 */                      /* 6, 7, 8: isolate only */
    if (! ce_enterMisbs (rt))
      ++io;
    if (2 == ce_enterConds (rt))
      return -1;
    if (faultIsolationSetup (rt, 1) == MEM_ERROR)
      ++io;
    ce_razDiagBuf (rt);
    hittingSetDiagnosis (rt, mode-5);           /*  procedure = mode-5 */
    ce_yieldDiags (rt);
  }
  return (im == 0) ? io : -1;                 
}

long caen_doCycle2 (runtime rt, long mode, long seuil) {
  long io=0, im=0;
  
  if (mode <= 1)                   /* 0: simul only; 1: simul and reset; */
    ce_setAlarms (rt);             /* to prevent evaluation from using measures */
  if (mode <= 5) {                 /* values, except at reset times */
    ce_doStepTempo (rt);          /* 2: detect only; 3, 4, 5: detect and isolate */
    if (! ce_enterInputs (rt))
      ++io;
    if (! ce_enterMeasures (rt))
      ++io;
    ce_doStepUntemp (rt);
    ce_moveAllOld (rt);
    if (2 == ce_yieldConds (rt))
      im = 1;
  }
  if (areGCRequired (rt))
    ce_globalConstraints (rt);
  if (! ce_yieldDisplays (rt, 0))
    ++io;
  switch (mode) {
  case 0 :
    break;
  case 1 :
    ce_resetMs (rt);
    break;
  default:
    break;
    /*ce_checkMs (rt, seuil);
    if (! ce_yieldAlarms (rt))
      ++io;*/
  }
  ce_moveInputs (rt);
  return (im == 0) ? io : -1;
}

void ce_detectAndIsolate (runtime rt, long mode, long seuil) {
  long io = 0;
  if (mode >= 3) {             /* 3, 4, 5: detect and isolate */
    if (faultIsolationSetup (rt, 1) == MEM_ERROR)
      ++io;
    ce_prepareDiag (rt);
    hittingSetDiagnosis (rt, mode-2);
    ce_yieldDiags (rt);
  }
}

/**
 * set a low alarm to the measured variables.
 * used to prevent evaluation of measured variables.
 */
void ce_setAlarms (runtime rt) {
  long i;

  for (i=0; i<getRTNMeasure(rt); ++i)
    setVarLowAlrm (getRTMeasureNth (rt, i));
  return;
}

/**
 * get inputs from the data in and affect it to variables in runtime.
 */
long ce_enterInputs (runtime rt) {
  long ok=1, i, nin=getRTNInput(rt);

  if (nin) {
    /* ok = caen_getInputs (getRTAppliName (rt), nin, getRTInVal (rt)); */
    setRTInVal (rt, inps);
    for(i=0; i<nin; i++) {
      /*ce_pushHisto (getRTInputNth(rt, i));*/
      /*affectRealOld (getRTInputNth(rt, i), getRTInValNth (rt, i));*/
      affectRealNew (getRTInputNth(rt, i), getRTInValNth (rt, i));
      setDetermined(getRTInputNth(rt, i));
      /*cout << "input: " << getVarName (getRTInputNth (rt, i)) << endl;
	ce_printVarHistory (stdout, getRTInputNth (rt, i), rt);
	cout << endl;*/
    }
  }
  return ok;
}

/* move inputs to old slots */
void ce_moveInputs (runtime rt) {
  long nin=getRTNInput (rt);
  
  for (long i=0; i<nin; i++) {
    /*ce_putObjet (varNewValue (getRTInputNth (rt, i)), varOldValue(getRTInputNth (rt, i)));*/
    /* ce_printVarHistory (stdout, getRTInputNth (rt, i), rt);
       cout << endl;*/
    ce_new2OldValue (rt, getRTInputNth (rt, i));
  }
}

/**
 * get measures from the data in and affect them to the variables.
 */
long ce_enterMeasures (runtime rt) {
  long ok=1, i, nmes=getRTNMeasure(rt);

  if (nmes) {
    /*ok = caen_getMeasures (getRTAppliName (rt), nmes, getRTMesVal (rt));*/
    setRTMesVal (rt, meas);
    for(i=0; i<nmes; i++) {
      ce_pushHistoMes (getRTMeasureNth(rt, i));
      affectRealMes (getRTMeasureNth(rt, i), getRTMesValNth (rt, i));
    }
  }
  return ok;
}

/**
 * get measures from the data in and affect them to the variables at  k-1.
 */
long ce_enterMeasuresOld (runtime rt) {
  long ok=1, i, nmes=getRTNMeasure(rt);

  if (nmes) {
    setRTMesVal (rt, meas);
    for(i=0; i<nmes; i++) {
      ce_pushHistoMes (getRTMeasureNth(rt, i));
      affectRealOld (getRTMeasureNth(rt, i), getRTMesValNth (rt, i));
    }
  }
  return ok;
}

/*
 * Multi-runtime conditions evaluation and split
 */
long ce_yieldMRTConds (runtime rt) {
  long ok=1, allFalse=1, i, nifl=getRTNInflu (rt);
  condition cd;
  
  ce_splitValiditySpc (rt);     /* split the state space over variables */

  for (multimap<long, runtime, less<long> >::iterator nrt_j = nrt->begin();
       nrt_j != nrt->end(); nrt_j++) {
    if (nifl) {
      for (i=0; i<nifl; i++) {
	influence ifl=getRTInfluNth ((*nrt_j).second, i);
	if (isNull (cd = getInfluCond (ifl))) {
	  setRTCondOutNth ((*nrt_j).second, i, 1);
	  allFalse = 0;
	}
	else
	  if (setRTCondOutNth ((*nrt_j).second, i, getCondValue (cd)))
	    allFalse = 0;
      } /* end for */
      /*ok = caen_setConds (getRTAppliName ((*nrt_j).second), nifl, getRTCondOut ((*nrt_j).second), (*nrt_j).second);*/
    } /* endif nifl */
    if (allFalse)
      return 2;
    /*resetSplits ((*nrt_j).second);*/
  }

  return ok;
}

void resetSplits (runtime rt) {
  condition cd;

  for (int i=0; i<getRTNCondi (rt); i++) {
    cd = getRTCondiNth (rt, i);
    /*cd->splitted = 0;*/
  } 
}

/**
 * ? 
 */
long ce_yieldConds (runtime rt) {
  long ok=1, allFalse=1, i, nifl=getRTNInflu(rt);
  condition cd;

  /*if (nvars) {  */          /* changes in conditions to bridge with upper discrete level */
    /*for (i=0; i<nvars; i++) {
      v = getRTDicoNth (rt, i);
      sp = ce_evalVarConds (v, rt);
    }
  }*/

  if (nifl) {
    for (i=0; i<nifl; i++) {
      influence ifl=getRTInfluNth (rt, i);
      if (isNull (cd = getInfluCond (ifl))) {
	setRTCondOutNth (rt, i, 1);
	allFalse = 0;
      }
      else
	if (setRTCondOutNth (rt, i, getCondValue (cd)))
	  allFalse = 0;
    } /* end for */
    ok = 1;
    /*ok = caen_setConds (getRTAppliName (rt), nifl, getRTCondOut (rt), rt);*/
  } /* endif nifl */
  if (allFalse)
    ok = 2;
  return ok;
}

/**
 * 
 */
long ce_yieldDisplays (runtime rt, long j) {
  long ok=1, i, ndisp=getRTNDisplay(rt);
  objet ob;

  if (ndisp) {
    for(i=0; i<ndisp; ++i) {
      ob = varOldValue (getRTDisplayNth(rt, i));
      if (CE_REAL == getObjType (ob))
	setRTDispInfValNth (rt, i, getObjRealVal (ob)),
	  setRTDispSupValNth (rt, i, getObjRealVal (ob));
      else { /* CE_REAL */
	range r=getObjRangeVal(ob);
	setRTDispInfValNth (rt, i, getRangeInf (r)),
	  setRTDispSupValNth (rt, i, getRangeSup (r));
      } /* endif CE_REAL */
    } /* end for */
    /* now write displays to file or screen. */
    /*k = j;
      while (k+1 < rt->num) {
      caen_setBlankDisplays (ndisp, rt);
      k++;
    }*/
    /*ok = caen_setDisplays (getRTAppliName (rt), ndisp,
			   getRTDispInfVal (rt), getRTDispSupVal (rt), rt->num);*/
    ok = 1;
  }
  return ok;
}

/**
 *  
 */
long ce_yieldAlarms (runtime rt) {
  long ok=1, i, nalrm=getRTNMeasure(rt);

  if (nalrm) {
    for (i=0; i<nalrm; ++i) {
      variable v=getRTMeasureNth (rt, i);
      setRTMesLowAlarmNth (rt, i, isVarLowAlrm (v));
      setRTMesHighAlarmNth (rt, i, isVarHighAlrm (v));
      setRTMesLowMisbNth (rt, i, isVarLowMisbehaving (v));
      setRTMesHighMisbNth (rt, i, isVarHighMisbehaving (v));
      setRTMesMisbNth (rt, i, isVarMisbehaving (v));
    } /* end for */
    ok = caen_setAlarms (getRTAppliName (rt), nalrm,
			 getRTMesLowAlarm (rt), getRTMesHighAlarm (rt),
			 getRTMesLowMisb (rt), getRTMesHighMisb (rt));
  }
  return ok;
}

/**
 * send diagnosis to the output. 
 */
long ce_yieldDiags (runtime rt) {
  long ok=1;

  ok = caen_sendDiagnosis (getRTAppliName (rt), getRTDiagOut (rt));
  return ok;
}


long ce_enterMisbs (runtime rt) {
  long ok=1, i, nmes=getRTNMeasure(rt);

  if (nmes) {
    ok = caen_getMisbs (getRTAppliName (rt), nmes, getRTMesMisb (rt));
    for (i=0; i<nmes; i++) {
      variable v=getRTMeasureNth (rt, i);
      if (getRTMesMisbNth (rt, i))
	setVarMisbehaving (v);
      else
	setVarOk (v);
    }
  }
  return ok;
}


long ce_enterConds (runtime rt) {
  long ok=1, allFalse=1, val, i, nifl=getRTNInflu(rt);

  if (nifl) {
    ok = caen_getConds (getRTAppliName (rt), nifl, getRTCondIn (rt));
    if (ok == 0)
      return ok;
    for (i=0; i<nifl; i++) {
      influence ifl=getRTInfluNth (rt, i);
      if (isNull (getInfluCond (ifl)))
	allFalse = 0;
      else {
	if ((val = getRTCondInNth (rt, i)))
	  allFalse = 0;
	ce_pushExternCond (ifl, (boolean) val);
      }
    } /* end for */
  }
  if (allFalse)
    ok = 2;
  return ok;
}


void caen_close (runtime rt) {

  faultIsolationEnd (rt);
  return;
}


STATUS createEmpty(ptListcaen pL) {
  if (pL == NULL)
    return MEM_ERROR;

  pL->first = NULL;
  pL->last = NULL;
  pL->current = NULL;
  pL->nEl = 0;

  return OK;
}

STATUS createUnitary(ptListcaen pL, ELEMENT e) {
  void *allocAux;

  if (pL == NULL || e == NULL)
    return MEM_ERROR;

  if ((allocAux = new_malloc(sizeof(node))) == NULL)
    return MEM_ERROR;

  pL->first = (ptNode) allocAux;
  pL->last = pL->first;
  pL->current = NULL;
  pL->first->datum = e;
  pL->first->next = NULL;
  pL->nEl = 1;

  return OK;
}

STATUS createUnitaryC(ptListcaen pL, ELEMENT e, int datumSize) {
  void *allocAux;

  if (pL == NULL || e == NULL)
    return MEM_ERROR;
  
  if ((allocAux = new_malloc(sizeof(node))) == NULL)
    return MEM_ERROR;

  pL->first = (ptNode) allocAux;
  pL->last = pL->first;
  pL->current = NULL;

  if ((allocAux = new_malloc(datumSize)) == NULL)
    return MEM_ERROR;

  pL->first->datum = (ELEMENT) allocAux;
  memcpy(pL->first->datum, e, datumSize);
  pL->first->next = NULL;
  pL->nEl = 1;

  return OK;
}

STATUS addLeft(ELEMENT e, ptListcaen pL) {
  ptNode newNode;
  void *allocAux;

  if (e == NULL || pL == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*pL)) {

    if ((allocAux = new_malloc(sizeof(node))) == NULL)
      return MEM_ERROR;

    pL->first = (ptNode) allocAux;
    pL->last = pL->first;
    pL->first->datum = e;
    pL->first->next = NULL;
  }
  else {

    if ((allocAux = new_malloc(sizeof(node))) == NULL)
      return MEM_ERROR;

    newNode = (ptNode) allocAux;
    newNode->datum = e;
    newNode->next = pL->first;
    pL->first = newNode;
  }
  pL->nEl++;

  return OK;
}

STATUS addLeftC(ELEMENT e, ptListcaen pL, int datumSize) {
  ptNode newNode;
  void *allocAux;

  if (e == NULL || pL == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*pL)) {

    if ((allocAux = new_malloc(sizeof(node))) == NULL)
      return MEM_ERROR;

    pL->first = (ptNode) allocAux;
    pL->last = pL->first;

    if ((allocAux = new_malloc(datumSize)) == NULL)
      return MEM_ERROR;

    pL->first->datum = (ELEMENT) allocAux;
    memcpy(pL->first->datum, e, datumSize);
    pL->first->next = NULL;
  }
  else {

    if ((allocAux = new_malloc(sizeof(node))) == NULL)
      return MEM_ERROR;

    newNode = (ptNode) allocAux;

    if ((allocAux = new_malloc(datumSize)) == NULL)
      return MEM_ERROR;

    newNode->datum = (ELEMENT) allocAux;
    memcpy(newNode->datum, e, datumSize);
    newNode->next = pL->first;
    pL->first = newNode;
  }
  pL->nEl++;

  return OK;
}

STATUS addRight(ELEMENT e, ptListcaen pL) {
  void *allocAux;

  if (e == NULL || pL == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*pL)) {

    if ((allocAux = new_malloc(sizeof(node))) == NULL)
      return MEM_ERROR;

    pL->first = (ptNode) allocAux;
    pL->last = pL->first;
    pL->first->datum = e;
    pL->first->next = NULL;
  }
  else {

    if ((allocAux = new_malloc(sizeof(node))) == NULL)
      return MEM_ERROR;

    pL->last->next = (ptNode) allocAux;
    pL->last = pL->last->next;
    pL->last->datum = e;
    pL->last->next = NULL;
  }
  pL->nEl++;

  return OK;
}

STATUS addRightC(ELEMENT e, ptListcaen pL, int datumSize) {
  void *allocAux;

  if (e == NULL || pL == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*pL)) {

    if ((allocAux = new_malloc(sizeof(node))) == NULL)
      return MEM_ERROR;

    pL->first = (ptNode) allocAux;
    pL->last = pL->first;

    if ((allocAux = new_malloc(datumSize)) == NULL)
      return MEM_ERROR;

    pL->first->datum = (ELEMENT) allocAux;
    memcpy(pL->first->datum, e, datumSize);
    pL->first->next = NULL;
  }
  else {

    if ((allocAux = new_malloc(sizeof(node))) == NULL)
      return MEM_ERROR;

    pL->last->next = (ptNode) allocAux;
    pL->last = pL->last->next;

    if ((allocAux = new_malloc(datumSize)) == NULL)
      return MEM_ERROR;

    pL->last->datum = (ELEMENT) allocAux;
    memcpy(pL->last->datum, e, datumSize);
    pL->last->next = NULL;
  }
  pL->nEl++;

  return OK;
}

ELEMENT lookLeftC(listcaen l, int datumSize) {
  ELEMENT elm;

  if (!isEmptyListcaen(l)) {

    if ((elm = (ELEMENT)new_malloc(datumSize)) == NULL)
      return NULL;

    memcpy(elm, l.first->datum, datumSize);
    return elm;
  }

  return NULL;
}

ELEMENT lookRightC(listcaen l, int datumSize) {
  ELEMENT elm;

  if (!isEmptyListcaen(l)) {

    if ((elm = (ELEMENT)new_malloc(datumSize)) == NULL)
      return NULL;

    memcpy(elm, l.last->datum, datumSize);
    return elm;
  }

  return NULL;
}

ELEMENT lookCurrentC(listcaen l, int datumSize) {
  ELEMENT elm;

  if (l.current == NULL)
    return NULL;
  else {

    if ((elm = (ELEMENT)new_malloc(datumSize)) == NULL)
      return NULL;

    memcpy(elm, l.current->datum, datumSize);
    return elm;
  }
}

ELEMENT lookAdvanceCurrent(ptListcaen pL) {
  ptNode cur;

  if (pL == NULL)
    return NULL;

  if (pL->current == NULL)
    return NULL;
  else {
    cur = pL->current;
    pL->current = pL->current->next;
    return (ELEMENT)cur->datum;
  }
}

ELEMENT lookAdvanceCurrentC(ptListcaen pL, int datumSize) {
  ELEMENT elm;

  if (pL == NULL)
    return NULL;

  if (pL->current == NULL)
    return NULL;
  else {

    if ((elm = (ELEMENT)new_malloc(datumSize)) == NULL)
      return NULL;

    memcpy(elm, pL->current->datum, datumSize);
    pL->current = pL->current->next;
    return elm;
  }
}

ELEMENT lookNth(ptListcaen pL, int p) {
  int i;
  ptNode aux;

  if (pL == NULL)
    return NULL;
	
  if (isEmptyListcaen(*pL))
    return NULL;
  if (p == listcaenLength(*pL))
    return pL->last->datum;
  else {
    aux = pL->first;
    for (i = 1; i < p; i++) {
      aux = aux->next;
    }
    pL->current = aux;
    return aux->datum;
  }
}

ELEMENT lookNthC(ptListcaen pL, int p, int datumSize) {
  int i;
  ELEMENT elm;
  ptNode aux;

  if (pL == NULL)
    return NULL;

  if (isEmptyListcaen(*pL))
    return NULL;

  if ((elm = (ELEMENT)new_malloc(datumSize)) == NULL)
      return NULL;

  if (p == listcaenLength(*pL))
    memcpy(elm, pL->last->datum, datumSize);
  else {
    aux = pL->first;
    for (i = 1; i < p; i++) {
      aux = aux->next;
    }
    pL->current = aux;
    memcpy(elm, aux->datum, datumSize);
  }

  return elm;
}

STATUS deleteLeft(ptListcaen pL) {
  ptNode aux;

  if (pL == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*pL))
    return OK;

  aux = pL->first;
  pL->first = pL->first->next;
  if (pL->first == NULL)
    pL->last = NULL;
  new_free(aux, sizeof(node));
  aux = 0;
  pL->nEl--;

  return OK;
}

STATUS deleteLeftC(ptListcaen pL, size_t size) {
  ptNode aux;

  if (pL == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*pL))
    return OK;

  aux = pL->first;
  pL->first = pL->first->next;
  if (pL->first == NULL)
    pL->last = NULL;
  new_free(aux->datum, size);
  aux->datum = 0;
  new_free(aux, sizeof(node));
  aux = 0;
  pL->nEl--;

  return OK;
}

STATUS deleteRight(ptListcaen pL) {
  ptNode aux;

  if (pL == NULL)
    return MEM_ERROR;
  
  if (isEmptyListcaen(*pL))
    return OK;

  if (pL->first == pL->last) {
    new_free(pL->first, sizeof(node));
    pL->first = NULL;
    pL->last = NULL;				
  }
  else{
    aux = pL->first;
    while (aux->next != pL->last)
      aux = aux->next;
    aux->next = NULL;
    new_free(pL->last, sizeof(node));
    pL->last = aux;
  }
  pL->nEl--;

  return OK;
}

STATUS deleteRightC(ptListcaen pL, size_t size) {
  ptNode aux;

  if (pL == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*pL))
    return OK;

  if (pL->first == pL->last) {
    new_free(pL->first->datum, size);
    new_free(pL->first, sizeof(node));
    pL->first = NULL;
    pL->last = NULL;				
  }
  else{
    aux = pL->first;
    while (aux->next != pL->last)
      aux = aux->next;
    aux->next = NULL;
    new_free(pL->last->datum, size);
    new_free(pL->last, sizeof(node));
    pL->last = aux;
  }
  pL->nEl--;

  return OK;
}

STATUS concatenate(ptListcaen pL1, listcaen l2) {
  if (pL1 == NULL)
    return MEM_ERROR;

  if (!isEmptyListcaen(l2)) {
    if (isEmptyListcaen(*pL1)) {
      pL1->first = l2.first;
      pL1->last = l2.last;			
    }
    else {
      pL1->last->next = l2.first;
      pL1->last = l2.last;
      pL1->nEl = pL1->nEl+l2.nEl;			
    }
  }

  return OK;
}

STATUS concatenateOn1Copying2(ptListcaen ptL1, listcaen l2) {
  ptNode aux;

  if (ptL1 == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*ptL1)) {
    if (isEmptyListcaen(l2))
      return OK;
    else {
      if (assign(ptL1, l2) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }
  else
    if (isEmptyListcaen(l2))
      return OK;

  aux = l2.first;
  do {
    if (addRight(aux->datum, ptL1) == MEM_ERROR)
      return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

STATUS concatenateOn1Copying2C(ptListcaen ptL1, listcaen l2, int DatumSize) {
  ptNode aux;

  if (ptL1 == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*ptL1)) {
    if (isEmptyListcaen(l2))
      return OK;
    else {
      if (assignC(ptL1, l2, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }
  else
    if (isEmptyListcaen(l2))
      return OK;

  aux = l2.first;
  do {
    if (addRightC(aux->datum, ptL1, DatumSize) == MEM_ERROR)
      return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

STATUS listcaenConcatenation(listcaen l1, listcaen l2, ptListcaen result) {
  ptNode aux;

  if (isEmptyListcaen(l1)) {
    if (isEmptyListcaen(l2)) {
      if (createEmpty(result) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
    else {
      if (assign(result, l2) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }
  else {
    if (isEmptyListcaen(l2)) {
      if (assign(result, l1) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }

  if (assign(result, l1) == MEM_ERROR)
    return MEM_ERROR;
  aux = l2.first;
  do {
    if (addRight(aux->datum, result) == MEM_ERROR)
      return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

STATUS listcaenConcatenationC(listcaen l1, listcaen l2, ptListcaen result, int DatumSize) {
  ptNode aux;

  if (isEmptyListcaen(l1)) {
    if (isEmptyListcaen(l2)) {
      if (createEmpty(result) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
    else {
      if (assignC(result, l2, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }
  else {
    if (isEmptyListcaen(l2)) {
      if (assignC(result, l1, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }

  if (assignC(result, l1, DatumSize) == MEM_ERROR)
    return MEM_ERROR;
  aux = l2.first;
  do {
    if (addRightC(aux->datum, result, DatumSize) == MEM_ERROR)
      return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

/* O(N ^ 2) */
STATUS listcaenIntersection(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*)) {
  ptNode aux;

  if (createEmpty(result) == MEM_ERROR)
    return MEM_ERROR;

  if (isEmptyListcaen(l1) || isEmptyListcaen(l2))
    return OK;

  aux = l1.first;
  do {
    if (isPresent(aux->datum, l2, IG))
      if (addRight(aux->datum, result) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

/* O(N ^ 2) */
STATUS listcaenIntersectionC(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*), int DatumSize) {
   ptNode aux;

   if (createEmpty(result) == MEM_ERROR)
     return MEM_ERROR;

   if (isEmptyListcaen(l1) || isEmptyListcaen(l2))
     return OK;

   aux = l1.first;
   do {
     if (isPresent(aux->datum, l2, IG))
       if (addRightC(aux->datum, result, DatumSize) == MEM_ERROR)
	 return MEM_ERROR;
     aux = aux->next;
   } while (aux != NULL);

   return OK;
}

/* O(N ^ 2 + N) */
STATUS listcaenUnion(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*)) {
  ptNode aux;

  if (isEmptyListcaen(l1)) {
    if (isEmptyListcaen(l2)) {
      if (createEmpty(result) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
    else {
      if (assign(result, l2) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }
  else {
    if (isEmptyListcaen(l2)) {
      if (assign(result, l1) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }

  if (assign(result, l1) == MEM_ERROR)
    return MEM_ERROR;
  aux = l2.first;
  do {
    if (!isPresent(aux->datum, l1, IG))
      if (addRight(aux->datum, result) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

/* O(N ^ 2 + N) */
STATUS listcaenUnionC(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*), int DatumSize) {
  ptNode aux;

  if (isEmptyListcaen(l1)) {
    if (isEmptyListcaen(l2)) {
      if (createEmpty(result) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
    else {
      if (assignC(result, l2, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }
  else {
    if (isEmptyListcaen(l2)) {
      if (assignC(result, l1, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }

  if (assignC(result, l1, DatumSize) == MEM_ERROR)
    return MEM_ERROR;
  aux = l2.first;
  do {
    if (!isPresent(aux->datum, l1, IG))
      if (addRightC(aux->datum, result, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

STATUS listcaenUnionOnFirst(ptListcaen ptL1, listcaen l2, int (*IG)(void *, void*)) {
  ptNode aux;

  if (ptL1 == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*ptL1)) {
    if (isEmptyListcaen(l2)) {
      return OK;
    }
    else {
      if (assign(ptL1, l2) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }
  else {
    if (isEmptyListcaen(l2)) {
      return OK;
    }
  }

  aux = l2.first;
  do {
    if (!isPresent(aux->datum, *ptL1, IG))
      if (addRight(aux->datum, ptL1) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

STATUS listcaenUnionOnFirstC(ptListcaen ptL1, listcaen l2, int (*IG)(void *, void*), int datumSize) {
  ptNode aux;

  if (ptL1 == NULL)
    return MEM_ERROR;

  if (isEmptyListcaen(*ptL1)) {
    if (isEmptyListcaen(l2)) {
      return OK;
    }
    else {
      if (assignC(ptL1, l2, datumSize) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }
  else {
    if (isEmptyListcaen(l2)) {
      return OK;
    }
  }

  aux = l2.first;
  do {
    if (!isPresent(aux->datum, *ptL1, IG))
      if (addRightC(aux->datum, ptL1, datumSize) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

/* O(2N ^ 2) */
STATUS listcaenSymmetricDifference(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*)) {
  ptNode aux;

  if (createEmpty(result) == MEM_ERROR)
    return MEM_ERROR;

  if (isEmptyListcaen(l1)) {
    if (isEmptyListcaen(l2)) {
      return OK;
    }
    else {
      if (assign(result, l2) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }
  else {
    if (isEmptyListcaen(l2)) {
      if (assign(result, l1) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }

  aux = l1.first;
  do {
    if (!isPresent(aux->datum, l2, IG))
      if (addRight(aux->datum, result) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);
  aux = l2.first;
  do {
    if (!isPresent(aux->datum, l1, IG))
      if (addRight(aux->datum, result) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

/* O(2N ^ 2) */
STATUS listcaenSymmetricDifferenceC(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*), int DatumSize) {
  ptNode aux;

  if (createEmpty(result) == MEM_ERROR)
    return MEM_ERROR;

  if (isEmptyListcaen(l1)) {
    if (isEmptyListcaen(l2)) {
      return OK;
    }
    else {
      if (assignC(result, l2, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }
  else {
    if (isEmptyListcaen(l2)) {
      if (assignC(result, l1, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }

  aux = l1.first;
  do {
    if (!isPresent(aux->datum, l2, IG))
      if (addRightC(aux->datum, result, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);
  aux = l2.first;
  do {
    if (!isPresent(aux->datum, l1, IG))
      if (addRightC(aux->datum, result, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

STATUS listcaenExclusiveElements(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*)) {
  ptNode aux;

  if (createEmpty(result) == MEM_ERROR)
    return MEM_ERROR;

  if (isEmptyListcaen(l1)) {
    if (isEmptyListcaen(l2)) {
      return OK;
    }
    else {
      return OK;
    }
  }
  else {
    if (isEmptyListcaen(l2)) {
      if (assign(result, l1) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }

  aux = l1.first;
  do {
    if (!isPresent(aux->datum, l2, IG))
      if (addRight(aux->datum, result) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

STATUS listcaenExclusiveElementsC(listcaen l1, listcaen l2, ptListcaen result, int (*IG)(void *, void*), int DatumSize) {
  ptNode aux;

  if (createEmpty(result) == MEM_ERROR)
    return MEM_ERROR;

  if (isEmptyListcaen(l1)) {
    if (isEmptyListcaen(l2)) {
      return OK;
    }
    else {
      return OK;
    }
  }
  else {
    if (isEmptyListcaen(l2)) {
      if (assignC(result, l1, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
      else
	return OK;
    }
  }

  aux = l1.first;
  do {
    if (!isPresent(aux->datum, l2, IG))
      if (addRightC(aux->datum, result, DatumSize) == MEM_ERROR)
	return MEM_ERROR;
    aux = aux->next;
  } while (aux != NULL);

  return OK;
}

int isPresent(ELEMENT e, listcaen l, int (*IG)(void *, void *)) {
  ptNode aux;

  if (e == NULL)
    return NO;
  
  if (isEmptyListcaen(l))
    return NO;

  aux = l.first;
  do {
    if (IG == NULL) {
      if (aux->datum == e)
	return YES;
    }
    else {
      if ((*IG)(aux->datum, e))
	return YES;
    }
    aux = aux->next;
  } while (aux != NULL);
  
  return NO;
}

ELEMENT search(ELEMENT e, listcaen l, int (*IG)(void *, void *)) {
  ptNode aux;

  if (e == NULL)
    /*return NO;*/ /* bug ? */
    return NULL;

  if (isEmptyListcaen(l))
    return NULL;

  aux = l.first;
  do {
    if (IG == NULL) {
      if (aux->datum == e)
	return aux->datum;
    }
    else {
      if ((*IG)(aux->datum, e))
	return aux->datum;
    }
    aux = aux->next;
  } while (aux != NULL);
  
  return NULL;
}

int position(ELEMENT e, listcaen l, int (*IG)(void *, void *)) {
  ptNode aux;
  int pos = 1;

  if (e == NULL)
    return NO;

  if (isEmptyListcaen(l))
    return NO;

  aux = l.first;
  do {
    if (IG == NULL) {
      if (aux->datum == e)
	return pos;
    }
    else {
      if ((*IG)(aux->datum, e))
	return pos;
    }
    aux = aux->next;
    pos++;
  } while (aux != NULL);

  return NO;
}

STATUS insert(ptListcaen pL, int i, ELEMENT e) {
  ptNode aux, nextAux;
  int indNext;
  void *allocAux;

  if (pL == NULL || e == NULL)
    return MEM_ERROR;

  if (i == 1) {
    if (addLeft(e, pL) == MEM_ERROR)
      return MEM_ERROR;
  }
  else { 
    if (i == listcaenLength(*pL) + 1) {
      if (addRight(e, pL) == MEM_ERROR)
	return MEM_ERROR;
    }
    else {
      aux = pL->first;
      indNext = 2;
      while (indNext < i) {
	aux = aux->next;
	indNext++;
      }
      nextAux = aux->next;

      if ((allocAux = new_malloc(sizeof(node))) == NULL)
	return MEM_ERROR;

      aux->next = (ptNode) allocAux;
      aux = aux->next;
      aux->datum = e;
      aux->next = nextAux;
      pL->nEl++;
    }
  }

  return OK;
}

STATUS insertC(ptListcaen pL, int i, ELEMENT e, int datumSize) {
  ptNode aux, nextAux;
  int indNext;
  void *allocAux;

  if (pL == NULL || e == NULL)
    return MEM_ERROR;

  if (i == 1) {
    if (addLeftC(e, pL, datumSize) == MEM_ERROR)
      return MEM_ERROR;
  }
  else {
    if (i == listcaenLength(*pL) + 1) {
      if (addRightC(e, pL, datumSize) == MEM_ERROR)
	return MEM_ERROR;
    }
    else {
      aux = pL->first;
      indNext = 2;
      while (indNext < i) {
	aux = aux->next;
	indNext++;
      }
      nextAux = aux->next;

      if ((allocAux = new_malloc(sizeof(node))) == NULL)
	return MEM_ERROR;

      aux->next = (ptNode) allocAux;
      aux = aux->next;

      if ((allocAux = new_malloc(datumSize)) == NULL)
	return MEM_ERROR;

      aux->datum = (ELEMENT) allocAux;
      memcpy(aux->datum, e, datumSize);
      aux->next = nextAux;
      pL->nEl++;
    }
  }

  return OK;
}

STATUS ce_delete(ptListcaen pL, int i) {
  ptNode aux, nextAux;
  int indNext;

  if (pL == NULL)
    return MEM_ERROR;

  if (i == 1) {
    if (deleteLeft(pL) == MEM_ERROR)
      return MEM_ERROR;
  }
  else
    if (i == listcaenLength(*pL)) {
      if (deleteRight(pL) == MEM_ERROR)
	return MEM_ERROR;
    }
    else {
      aux = pL->first;
      indNext = 2;
      while (indNext < i) {
	aux = aux->next;
	indNext++;
      }
      nextAux = aux->next;
      aux->next = nextAux->next;
      new_free(nextAux, sizeof(node));
      pL->nEl--;
    }

  return OK;
}

STATUS deleteC(ptListcaen pL, int i, size_t size) {
  ptNode aux, nextAux;
  int indNext;

  if (pL == NULL)
    return MEM_ERROR;

  if (i == 1) {
    if (deleteLeftC(pL, size) == MEM_ERROR)
      return MEM_ERROR;
  }
  else
    if (i == listcaenLength(*pL)) {
      if (deleteRightC(pL, size) == MEM_ERROR)
	return MEM_ERROR;
    }
    else {
      aux = pL->first;
      indNext = 2;
      while (indNext < i) {
	aux = aux->next;
	indNext++;
      }
      nextAux = aux->next;
      aux->next = nextAux->next;
      new_free(nextAux->datum, size);
      new_free(nextAux, sizeof(node));
      pL->nEl--;
    }

  return OK;
}

STATUS update(ptListcaen pL, int i, ELEMENT e) {
  ptNode aux; 
  int ind;

  if (pL == NULL || e == NULL)
    return MEM_ERROR;

  if (i == listcaenLength(*pL))
    pL->last->datum = e;
  else {
    aux = pL->first;
    ind = 1;
    while (ind < i) {
      aux = aux->next;
      ind++;
    }
    aux->datum = e;
  }

  return OK;
}

STATUS updateC(ptListcaen pL, int i, ELEMENT e, int datumSize) {
  ptNode aux; 
  int ind;
  void *allocAux;

  if (pL == NULL || e == NULL)
    return MEM_ERROR;

  if ((allocAux = new_malloc(datumSize)) == NULL)
    return MEM_ERROR;

  if (i == listcaenLength(*pL)) {
    pL->last->datum = (ELEMENT) allocAux;
    memcpy(pL->last->datum, e, datumSize);
  }
  else {
    aux = pL->first;
    ind = 1;
    while (ind < i) {
      aux = aux->next;
      ind++;
    }
    aux->datum = (ELEMENT) allocAux;
    memcpy(aux->datum, e, datumSize);
  }

  return OK;
}

STATUS assign(ptListcaen pnew, listcaen old) {
  int i;

  if (createEmpty(pnew) == MEM_ERROR)
    return MEM_ERROR;

  for (i = 1; i <= listcaenLength(old); i++) {
    if (addRight(old.first->datum, pnew) == MEM_ERROR)
      return MEM_ERROR;
    old.first = old.first->next;
  }

  return OK;
}

STATUS assignC(ptListcaen pnew, listcaen old, int datumSize) {
  int i;

  if (createEmpty(pnew) == MEM_ERROR)
    return MEM_ERROR;

  for (i = 1; i <= listcaenLength(old); i++) {
    if (addRightC(old.first->datum, pnew, datumSize) == MEM_ERROR)
      return MEM_ERROR;
    old.first = old.first->next;
  }
  return OK;
}

STATUS liberate(ptListcaen pL) {
  int i, length;

  if (pL == NULL)
    return MEM_ERROR;

  length = listcaenLength(*pL);
  for (i = 1; i <= length; i++)
    deleteLeft(pL);

  return OK;
}

STATUS liberateC(ptListcaen pL, size_t size) {
  int i, length;

  if (pL == NULL)
    return MEM_ERROR;

  length = listcaenLength(*pL);
  for (i = 1; i <= length; i++)
    deleteLeftC(pL, size);

  return OK;
}

/*****************************************************************************/

void *new_malloc(size_t size) {
  void *allocp;

  if ((allocp = malloc(size)) == NULL) {
    fprintf(stderr, "\nImpossible to allocate any more dynamic memory!!!\n\n");
    return NULL;
  }
#ifdef DEBUG
  totalAllo += size;
#endif
  return allocp;
}


void *new_calloc(size_t nobj, size_t size) {
  void *allocp;

  if ((allocp = calloc(nobj, size)) == NULL) {
    fprintf(stderr, "\nImpossible to allocate any more dynamic memory!!!\n\n");
    return NULL;
  }
#ifdef DEBUG
  totalAllo += nobj * size;
#endif
  return allocp;
}

void new_free(void *p, size_t size) {

  if (! isNull (p)) {
    free(p);
#ifdef DEBUG
    totalDeallo += size;
#endif
  }
  return;
}


/*************************************************************/
/* Fault Isolation Module based on the Hitting-Set Algorithm */
/*                 Jose Angel Jimenez Molina                 */
/*                            1999                           */
/*************************************************************/

#ifdef DEBUG
long totalAllo, totalDeallo;
#endif

static int symb1 = '*';
static int symb2 = '-';
static char line[MAXLINE + 1];

/************************************************/
/* Hitting-Set fault isolation procedure set up */
/************************************************/
/* Called from: 'ceInit' and 'ceDiagnosis' in module 'caen-exec.c' */

STATUS faultIsolationSetup(runtime rt, long only) {
  void *allocAux;

  /* Set-up fault-isolation-only or simulation-and-fault-isolation mode*/
  setOnlyFaultIso(rt, only);
  /* If the simulation-and-fault-isolation is being carried out,
     then an historic array is created to keep the trace of
     each previous iteration's misbehaving variable set */
  if (!isOnlyFaultIso(rt)) {
      if ((allocAux = new_calloc(getRTNMeasure(rt), sizeof(variable))) == NULL)
	return MEM_ERROR;
      setRTOldMisbVarArray(rt, (variable *) allocAux);
  }
  /* Initialize Hittin-Set tree */
  createLeafListcaen(&getRTHSTree(rt));
  setRTNumNodesHST(rt, 0);
  
  return OK;
}


/*********************************************/
/* Hitting-Set tree heap memory deallocation */
/*********************************************/

void hsTreeDeallocation(runtime rt) {
  ptLeafListcaen hsTreePt = &getRTHSTree(rt);
  int numLeaves = getLeafListcaenNumber(hsTreePt);
  int i;

  setRTNumNodesHST(rt, 0);

  if (numLeaves == 0)
    return;

  /* Hitting-Set-Tree destruction */
  resetLeafListcaen(hsTreePt);
  for (i = 0; i < numLeaves; i++) {
    ptLeaf lfPt = getNextLeafLL(hsTreePt);

    /* Leaf (diagnostics) destruction */  
    liberateEdgeSetWithNameLeaf(lfPt);
  }
  liberateLeafListcaenC(hsTreePt);

  return;
}


/*********************************************/
/* Hitting-Set fault isolation procedure end */
/*********************************************/

/* At this point, just the memory used to store the 
   diagnostics (HS-Tree's leaves), and the misbehaving
   variable historic array, remain to be freed */

void faultIsolationEnd(runtime rt) {

  hsTreeDeallocation(rt);
  /* Old misbehaving variable array destruction (fault-isolation-only mode) */
  if (!isOnlyFaultIso(rt)) {
    new_free (getRTOldMisbVarArray(rt),
	      getRTNMeasure(rt) * sizeof(variable));
    setRTOldMisbVarArray (rt, NULL);
  }
  return;
}


/*****************************************/
/* Hitting-Set fault isolation procedure */
/*****************************************/
STATUS hittingSetDiagnosis(runtime rt, int diagnosisType) {

  boolean configurationHasChanged(runtime, boolean *, int);
  STATUS findConflictsAux(runtime, variable, ptConflictSet, ptListcaen, int, boolean);
  STATUS findClassTree(runtime, ptSoftVar, int, ptOkVarListcaen, ptSoftVarListcaen);
  STATUS generateSoftConflictSets(runtime, ptSoftVar, int);
  boolean closeBranch(ptLeaf, ptLeafListcaen);
  int compareConflicts(void *, void *);
  int compareAndUpdateConflicts(void *, void *);
  boolean compareConditions(condition, condition);
  int comparePreferenceClasses(const void *, const void *);
  void quickSort(void *[], int, int, int (*)(const void *, const void *));
  void printHSTree(runtime);
  void printClasses(runtime);
  void printConflictSets(runtime);

  boolean incremental;              /* Can the next HS Tree algorithm's iteration be done incrementally ? */
  int numMeas = getRTNMeasure(rt);  /* Number of measured variables */
  int numMisb = 0;                  /* Number of misbehaving variables */
  int numOk = 0;                    /* Number of Ok variables */
  int i, j, k, l;                   /* Indexing variables */
  void *allocAux;                   /* Heap memory allocation auxiliar variable */
#ifdef DEBUG
  totalAllo = totalDeallo = 0;      /* Reset the amount of allocated and deallocated dynamic memory */
#endif

  setRTNumMisb(rt, 0);
  setRTNumOk(rt, 0);
  /* Calculate the number of misbehaving variables and Ok variables */
  for (i = 0; i < numMeas; i++) {
    variable auxVar = getRTMeasureNth(rt, i);

    if (isVarMisbehaving(auxVar))
      getRTNumMisb(rt)++;
    else
      getRTNumOk(rt)++;
  }
  /* If there are no misbehaving variables, then the diagnosis procedure is not done */
  if (getRTNumMisb(rt) == 0)
    return OK;

  /* If the configuration of misbehaving variables has not changed between 
     two consecutive iterations, then the diagnosis procedure is not done */
  if (!configurationHasChanged(rt, &incremental, diagnosisType))
    return OK;

  /* HS-Tree destruction if the algorithm cannot evolve incrementally */
  if (!incremental)
    hsTreeDeallocation(rt);

  /* When all MEASURED variables in the system are MISBEHAVING, then all
     the diagnosis procedures are equivalent to DIAGNOSIS PROCEDURE 1 */
  if (numMeas == getRTNumMisb(rt))
    diagnosisType = 1;

  /* If the current context is considered to be incremental, then 
     only the new misbehaving variables are taken into account */
  if (incremental) {
    numMisb = getRTNumMisb(rt) - getRTOldNumMisb(rt);
    numOk = numMeas - numMisb;
    setRTNumMisb(rt, numMisb);
    setRTNumOk(rt, numOk);
    numMisb = numOk = 0;
  }
  else
    setRTOldNumMisb(rt, 0);

  /* Allocate memory for the misbehaving variable and OK variable arrays */
  if ((allocAux = new_calloc(getRTNumMisb(rt), sizeof(misbVar))) == NULL)
    return MEM_ERROR;
  setRTMisbVarArray(rt, (misbVarArray) allocAux);
  
  if (getRTNumOk (rt) > 0) {
    if ((allocAux = new_calloc(getRTNumOk(rt), sizeof(okVar))) == NULL)
      return MEM_ERROR;
    setRTOkVarArray(rt, (okVarArray) allocAux);
  }
  else setRTOkVarArray (rt, NULL);

  /* Seek misbehaving variables and OK variables */
  for (i = 0; i < numMeas; i++) {
    variable auxVar = getRTMeasureNth(rt, i);
    
    /* Initialize misbehaving variables */
    if (isVarMisbehaving(auxVar)) {
      ptMisbVar mbVarPt = getMisbVarPt(rt, numMisb);
      boolean newMisb = CE_TRUE;

      if (incremental) {
	for (j = 0; j < getRTOldNumMisb(rt); j++) {
	  variable oldVar = getVarOldMisbVarArray(rt, j);

	  if (auxVar == oldVar) {
	    newMisb = CE_FALSE;
	    break;
	  }
	}
      }
      if (newMisb) {
	setVarMisbVar(mbVarPt, auxVar);
	createHardConflictSetMisbVar(mbVarPt);
	createConflictSetListcaenMisbVar(mbVarPt);
	/* Update the old misbehaving variable array for the next iteration */
	if (!isOnlyFaultIso(rt))
	  setVarOldMisbVarArray(auxVar, rt , getRTOldNumMisb(rt)++);
	numMisb++;
      }
    }
    /* Initialize OK variables */
    else {
      ptOkVar okVarPt = getOkVarPt(rt, numOk);

      setVarOkVar(okVarPt, auxVar);
      createAscInflSetOkVar(okVarPt);
      numOk++;
    }
  }
  
  /************************/
  /* Hard conflict search */
  /************************/
  for (i = 0; i < numMisb; i++) {
    ptMisbVar mbVarPt = getMisbVarPt(rt, i);
    listcaen path;

    createEmpty(&path);
    if (findHardConflicts(rt, mbVarPt, &path) == MEM_ERROR)
      return MEM_ERROR;
    liberate(&path);
    /* The hard conflict set becomes the first conflict set for a given misbehaving variable */
    if (setNextConflictSetMisbVarC(&getHardConflictSetMisbVar(mbVarPt), mbVarPt) == MEM_ERROR)
      return MEM_ERROR;
  }

  /******************************/
  /* Ascendant influence search */
  /******************************/
  if (diagnosisType == 2 || diagnosisType == 3) {
    for (i = 0; i < numOk; i++) {
      ptOkVar okVarPt = getOkVarPt(rt, i);
      listcaen path;

      createEmpty(&path);
      if (findAscendantInfluences(rt, okVarPt, &path) == MEM_ERROR)
	return MEM_ERROR;
      liberate(&path);
    }
  }

  /**********************************/
  /* Soft conflict sets generation */
  /**********************************/
  if (diagnosisType == 2) {

#ifdef DEBUG
    char mes1[] = "* SOFT-CONFLICTS SETS Generation for variable ";
    int len1;

    len1 = strlen(mes1);
#endif

    /**************************************************/
    /* Find misbehaving variables' soft conflict sets */
    /**************************************************/

    /* Allocate memory for the soft variable array */
    if ((allocAux = new_calloc(numMeas, sizeof(softVar))) == NULL)
      return MEM_ERROR;
    setRTSoftVarArray(rt, (softVarArray) allocAux);

    /* Soft variables' memory allocation and initialization */
    for (i = 0, j = 0, k = 0; i < numMeas; i++) {
      ptSoftVar sfVarPt = getSoftVarPt(rt, i);
      int synIndex;

      if ((allocAux = new_calloc(numMisb, sizeof(int))) == NULL)
	return MEM_ERROR;
      setNumOkVarClassArraySoftVar(sfVarPt, (int *) allocAux);

      if ((allocAux = new_calloc(numMisb * numOk, sizeof(ptSoftVar))) == NULL)
	return MEM_ERROR;
      setOkVarClassMatrixSoftVar(sfVarPt, (ptSoftVar **) allocAux);

      if ((allocAux = new_calloc(numMisb, sizeof(synonymListcaen))) == NULL)
	return MEM_ERROR;
      setSynonymListcaenArraySoftVar(sfVarPt, (ptSynonymListcaen) allocAux);

      /* Initialize soft variable array first with misbehaving variables */
      if (j < numMisb) {
	ptMisbVar mbPt = getMisbVarPt(rt, j++);
	synonym hardCS = getHardConflictSetMisbVar(mbPt);

	setMisbehavingSoftVar(sfVarPt, CE_TRUE);
	setMisbVarSoftVar(sfVarPt, mbPt);
	for (synIndex = 0; synIndex < numMisb; synIndex++) {
	  ptSynonymListcaen synLPt = getSynonymListcaenPtSoftVar(sfVarPt, synIndex);
	  synonym firstSynonym;

	  if (assignSynonymC(&firstSynonym, hardCS) == MEM_ERROR)
	    return MEM_ERROR;
	  createSynonymListcaen(synLPt);
	  if (setNextSynonymSLC(&firstSynonym, synLPt) == MEM_ERROR)
	    return MEM_ERROR;
	}
      }
      /* Then with ok variables */
      else {
	ptOkVar okPt = getOkVarPt(rt, k++);
	synonym ascIS = getAscInflSetOkVar(okPt);

	setMisbehavingSoftVar(sfVarPt, CE_FALSE);
	setOkVarSoftVar(sfVarPt, okPt);
	for (synIndex = 0; synIndex < numMisb; synIndex++) {
	  ptSynonymListcaen synLPt = getSynonymListcaenPtSoftVar(sfVarPt, synIndex);
	  synonym firstSynonym;

	  if (assignSynonymC(&firstSynonym, ascIS) == MEM_ERROR)
	    return MEM_ERROR;
	  createSynonymListcaen(synLPt);
	  if (setNextSynonymSLC(&firstSynonym, synLPt) == MEM_ERROR)
	    return MEM_ERROR;
	}
      }
    }
    
    /***************************************************/
    /* Search of misbehaving and OK variables' classes */
    /***************************************************/

    /* For each misbehaving variable */
    for (i = 0; i < numMisb; i++) {
      ptSoftVar sfVarPt = getSoftVarPt(rt, i); /* A misbehaving variable in each iteration */
      softVarListcaen sfVL; /* Listcaen of soft variables to carry out a breadth-first class search */
      okVarListcaen okVL;   /* Listcaen of OK variables that are already present in some class for a given misbehaving variable */
      
      createSoftVarListcaen(&sfVL);
      if (setFirstSoftVarSVL(sfVarPt, &sfVL) == MEM_ERROR)
	return MEM_ERROR;
      createOkVarListcaen(&okVL);
      
      /* Breadth first class search */
      while (getSoftVarListcaenNumber(&sfVL) > 0) {
	ptSoftVar nextSoftVarPt = getFirstSoftVarSVL(&sfVL);

	if (findClassTree(rt, nextSoftVarPt, i, &okVL, &sfVL) == MEM_ERROR)
	  return MEM_ERROR;
      }

      /* Free 'okVarListcaen' before next misbehaving variable's class search */
      liberateOkVarListcaen(&okVL);
    }

#ifdef DEBUG
    printClasses(rt);
#endif

    /******************************************************************/
    /* Calculate the soft conflict sets for each misbehaving-variable */
    /******************************************************************/

    /* All the soft conflicts are also calculated at this stage */
    createConflictSet(&getRTSoftConflictListcaen(rt));
    /* For each misbehaving-variable */
    for (i = 0; i < numMisb; i++) {
      ptSoftVar sfVarPt = getSoftVarPt(rt, i); 

#ifdef DEBUG
      const char *name = getVarName(getVarMisbVar(getMisbVarSoftVar(sfVarPt)));
      int length;

      length = len1 + strlen(name) + 2;
      memset(line, symb1, length < MAXLINE ? length : MAXLINE - 1);
      line[length] = '\0';
      fprintf(stderr, "\n\n%s\n", line);
      fprintf(stderr, "%s%s *\n", mes1, name);
      fprintf(stderr, "%s\n", line);
#endif

      if (generateSoftConflictSets(rt, sfVarPt, i) == MEM_ERROR)
	return MEM_ERROR;
    }
    /* END For each misbehaving-variable */

#ifdef DEBUG
    printConflictSets(rt);
#endif

    /* Memory deallocation */
    for (i = 0; i < numMeas; i++) {
      ptSoftVar sfVarPt = getSoftVarPt(rt, i);

      new_free(getNumOkVarClassArraySoftVar(sfVarPt), numMisb * sizeof(int));
      new_free(getOkVarClassMatrixSoftVar(sfVarPt), numMisb * numOk * sizeof(ptSoftVar));
      
      for (j = 0; j < numMisb; j++) {
	ptSynonymListcaen synListcaenPt = getSynonymListcaenPtSoftVar(sfVarPt, j);
	int numSynonyms = getSynonymListcaenNumber(synListcaenPt);

	resetSynonymListcaen(synListcaenPt);
	for (k = 0; k < numSynonyms; k++) {
	  ptSynonym synPt = getNextSynonymSL(synListcaenPt);

	  liberateSynonymC(synPt);
	}
	liberateSynonymListcaenC(synListcaenPt);
      }
      new_free(getSynonymListcaenArraySoftVar(sfVarPt), numMisb * sizeof(synonymListcaen));
    }
    new_free(getRTSoftVarArray(rt), getRTNMeasure(rt) * sizeof(softVar));
  }
  /* END If (diagnosisType == 2) */

  /* ===================== */
  /* HITTING-SET ALGORITHM */
  /* ===================== */

  /* For each misbehaving-variable */
  for (i = 0; i < numMisb; i++) {
    ptMisbVar mbPt = getMisbVarPt(rt, i);
    ptConflictSetListcaen conflSetListcaenPt = &getConflictSetListcaenMisbVar(mbPt);
    int numConflictSets = getConflictSetListcaenNumber(conflSetListcaenPt);

    resetConflictSetListcaen(conflSetListcaenPt);
    /* For each conflict-set belonging to the misbehaving-variable */
    for (j = 0; j < numConflictSets; j++) {
      ptConflictSet conflSetPt = getNextConflictSetCSL(conflSetListcaenPt);
      int numConflicts = getConflictSetNumber(conflSetPt);
      /* added by jpg 08/02/2000 */
      if (numConflicts) {
	/* end addition jpg */

	leafListcaen newHSTree;
	leafListcaenArray newLeaves, oldLeaves;

	createLeafListcaen(&newHSTree);

	/* Create and initialize 'newLeaves' and 'oldLeaves' list arrays */
	if ((allocAux = new_calloc(numConflicts, sizeof(leafListcaen))) == NULL)
	  return MEM_ERROR;
	newLeaves = (ptListcaen) allocAux;

	if ((allocAux = new_calloc(numConflicts, sizeof(leafListcaen))) == NULL)
	  return MEM_ERROR;
	oldLeaves = (ptListcaen) allocAux;

	/* For each leaf 'l' of Hitting-Set tree 'T' */
	while (getLeafListcaenNumber(&getRTHSTree(rt)) || getRTNumNodesHST(rt) == 0) {
	  ptLeaf lf = NULL;
	  conflictSet intersection;

	  /* Get leaf 'l' from Hitting-Set tree 'T' */
	  if (getLeafListcaenNumber(&getRTHSTree(rt)) > 0) {
	    lf = getFirstLeafLL(&getRTHSTree(rt)); /* 'l' */

	    /* Intersection of 'H(l)' and Conflict Set 'S' */
	    if (listcaenIntersection(*conflSetPt, getEdgeSetLeaf(lf), &intersection, compareAndUpdateConflicts) == MEM_ERROR)
	      return MEM_ERROR;
	    /* Very important remark:
	       ----------------------
	       The elements that make up the intersection of two sets will be taken from 
	       the set that is in first position of the intersection function's parameter list.
	       In this case, these elements will be taken from the current conflict set 'conflSet'.
	       This is relevant when considering a conflict's index field. Look at the handling
	       of the 'oldLeaves' array and the 'singleton' conflict below for more information.
	    */
	  }
	  else {
	    /* If HS-tree is empty then intersection is the empty-set */
	    createConflictSet(&intersection);
	  }

	  /* If H(l) INTERSECTION 'S' = EMPTY-SET */
	  if (getConflictSetNumber(&intersection) == 0) {
	    resetConflictSet(conflSetPt);
	    /* For each conflict 's' in conflict-set 'S' */
	    for (k = 0; k < numConflicts; k++) {
	      /* Get conflict 's' from conflict-set 'S' */
	      ptConflict conflPt = getNextConflictCS(conflSetPt);
	      leaf newLeaf;

	      /* Add to 'l' an edge labeled with 's' and a successor node 'ns' */
	      if (lf != NULL) { /* Is HS-tree NOT empty? */
		/* Copy leaf 'l' into new leaf */
		if (assignEdgeSetWithNameLeaf(&newLeaf, *lf) == MEM_ERROR)
		  return MEM_ERROR;
	      }
	      else
		createEdgeSetLeaf(&newLeaf);
	      /* Add an edge to leaf 'l' */
	      if (setNextEdgeWithNameLeaf(conflPt, &newLeaf) == MEM_ERROR)
		return MEM_ERROR;
	      /* Add a successor node 'ns' to leaf 'l' */
	      setNodeNumberLeaf(&newLeaf, ++getRTNumNodesHST(rt));
	      /* Add the pair ('ns', H('l') UNION {'s'}) to new-leaves['s'] */
	      if (setNextLeafLLAC(&newLeaf, newLeaves, getConflictIndex(conflPt)) == MEM_ERROR)
		return MEM_ERROR;
	      /* END For each conflict 's' in conflict-set 'S' */
	    }
	  }
	  else {
	    leaf auxLeaf;

	    /* Checking for empty HS-tree is not necessary in this case,
	       since H(l) INTERSECTION 'S' is NOT empty */
	    if (assignEdgeSetWithNameLeaf(&auxLeaf, *lf) == MEM_ERROR)
	      return MEM_ERROR;
	    setNodeNumberLeaf(&auxLeaf, getNodeNumberLeaf(lf));
	    /* If H(l) INTERSECTION 'S' = {'s'} (a singleton) */
	    if (getConflictSetNumber(&intersection) == 1) {
	      /* Getting singleton 's' */
	      ptConflict singletonPt = getFirstConflictCS(&intersection);
	      /* The singleton must be the conflict in the current conflict set, 
		 not the one in the HSTree's leaf, to make sure that its index
		 is the correct one when handling 'oldLeaves' */

	      /* Add the pair ('l', H('l')) to old-leaves['s'] */
	      if (setNextLeafLLAC(&auxLeaf, oldLeaves, getConflictIndex(singletonPt)) == MEM_ERROR)
		return MEM_ERROR;
	    }
	    /* Add leaf to new HS-tree */
	    if (setNextLeafLLC(&auxLeaf, &newHSTree) == MEM_ERROR)
	      return MEM_ERROR;
	  }
	  /* END If H(l) INTERSECTION 'S' = EMPTY-SET */
	
	  /* Liberate conflict-set of leaf 'l', that is, H('l') */
	  if (lf != NULL)
	    liberateEdgeSetWithNameLeaf(lf);
	  /* Delete leaf 'l' from HS-tree */
	  if (getLeafListcaenNumber(&getRTHSTree(rt)) > 0)
	    deleteFirstLeafLLC(&getRTHSTree(rt));
	  /* Liberate (H('l') INTERSECTION 'S') list */
	  liberateConflictSet(&intersection);
	}
	/* END For each leaf 'l' of Hitting-Set tree 'T' */

	/* For each conflict 's' in conflict-set 'S' */
	resetConflictSet(conflSetPt);
	for (k = 0; k < numConflicts; k++) {
	  /* Get conflict 's' from conflict-set 'S' */
	  ptConflict conflPt = getNextConflictCS(conflSetPt);
	  /* Get leaf-list associated to conflict 's' from leaf-list-array 'new-leaves' */
	  ptLeafListcaen newLL = getLeafListcaenPtLLA(newLeaves, getConflictIndex(conflPt));
	  /* Get leaf-list associated to conflict 's' from leaf-list-array 'old-leaves' */
	  ptLeafListcaen oldLL = getLeafListcaenPtLLA(oldLeaves, getConflictIndex(conflPt));
	  int numNewLeaves = getLeafListcaenNumber(newLL);
	  int numOldLeaves = getLeafListcaenNumber(oldLL);

	  /* For each leaf 'n' of new-leaves['s'] */
	  /* resetLeafListcaen(newLL); */
	  for (l = 0; l < numNewLeaves; l++) {
	    /* Get leaf 'n' from new-leaves['s'] */
	    ptLeaf newLeaf = getFirstLeafLL(newLL);
	    /* If there are no leaves associated to conflict 's' in old-leaves,
	       then the branch in leaf 'n' will not have to be closed, that is,
	       it will be part of the new HS-tree */

	    /* If old-leaves['s'] is empty */
	    if (numOldLeaves == 0) {
	      /* Add new leaf 'n' to new HS-tree */
	      if (setNextLeafLLC(newLeaf, &newHSTree) == MEM_ERROR)
		return MEM_ERROR;
	    }
	    /* Not empty old-leaves['s'] */
	    else {
	      /* If H('n') contains H('l') for some leaf 'l' of old-leaves['s'],
		 then close the brach in 'n' */
	      if (closeBranch(newLeaf, oldLL)) {
		/* Liberate conflict-set of leaf 'n', that is, H('n') */
		liberateEdgeSetWithNameLeaf(newLeaf);
	      }
	      else {
		/* Add new leaf 'n' to new HS-tree */
		if (setNextLeafLLC(newLeaf, &newHSTree) == MEM_ERROR)
		  return MEM_ERROR;
	      }
	    }
	    /* Delete leaf from leaf-list in new-leaves['s'] */
	    deleteFirstLeafLLC(newLL);
	  }
	  /* END For each leaf 'n' of new-leaves['s'] */

	  /* Liberate leaf-list in old-leaves['s'] */
	  if (numOldLeaves > 0)
	    liberateLeafListcaenC(oldLL);
	} /* end for k */
	/* END For each conflict 's' in conflict-set 'S' */

	/* Update HSTree */
	setRTHSTree(rt, newHSTree);

	new_free(newLeaves, numConflicts * sizeof(leafListcaen));
	new_free(oldLeaves, numConflicts * sizeof(leafListcaen));

	/* added by jpg 08/02/2000 */
      } /* endif numConflicts */
      /* end addition jpg */
    } /* end for j */
    /* END For each conflict-set belonging to the misbehaving-variable */
  }
  /* END For each misbehaving-variable */

  /* ========================= */
  /* END HITTING-SET ALGORITHM */
  /* ========================= */

  /* Option 3 */
  if (diagnosisType == 3) {
    ptLeafListcaen hsTreePt = &getRTHSTree(rt);
    int numLeaves = getLeafListcaenNumber(hsTreePt);
    boolean *deleteLeaves;
    leafListcaen newHSTree;

    /* Initialize array of leaves to be deleted from HS-Tree */
    if ((allocAux = new_calloc(numLeaves, sizeof(boolean))) == NULL)
      return MEM_ERROR;
    deleteLeaves = (boolean *) allocAux;

    /* Which leaves must be deleted from the HS-Tree? */
    for (i = 0; i < numOk; i++) {
      ptOkVar okPt = getOkVarPt(rt, i);
      int numAscInfl = getAscInflSetNumberOkVar(okPt);

      resetAscInflSetOkVar(okPt);
      for (j = 0; j < numAscInfl; j++) {
	ptConflict ascInflPt = getNextAscInflOkVar(okPt);
	
	resetLeafListcaen(hsTreePt);
	for (k = 0; k < numLeaves; k++) {
	  ptLeaf lfPt = getNextLeafLL(hsTreePt);
	  int numEdges = getEdgeSetNumberLeaf(lfPt);

	  if (deleteLeaves[k] == CE_TRUE)
	    continue;

	  resetEdgeSetLeaf(lfPt);
	  for (l = 0; l < numEdges; l++) {
	    ptAscendantInfluence edgePt = getNextEdgeLeaf(lfPt);

	    if (compareConflicts(edgePt, ascInflPt)) {
	      if (getConflictInstant(edgePt) >= getAscInflInstant(ascInflPt)) {
		deleteLeaves[k] = CE_TRUE;
		break;
	      }
	    }
	  } /* End for each edge 'l' */
	} /* End for each leaf 'k' */
      } /* End for each ascendant influence 'j' */
    } /* End for each Ok variable 'i' */

    /* Generate new HS-Tree, without the leaves that must be eliminated */
    resetLeafListcaen(hsTreePt);
    createLeafListcaen(&newHSTree);
    for (i = 0; i < numLeaves; i++) {
      ptLeaf lfPt = getNextLeafLL(hsTreePt);
      leaf newLeaf = *lfPt;

      if (!deleteLeaves[i]) {
	if  (setNextLeafLLC(&newLeaf, &newHSTree) == MEM_ERROR)
	  return MEM_ERROR;
      }
      else {
	if (liberateEdgeSetWithNameLeaf(lfPt) == MEM_ERROR)
	  return MEM_ERROR;
      }
    }
    if (liberateLeafListcaenC(hsTreePt) == MEM_ERROR)
      return MEM_ERROR;

    /* Update Hitting-Set Tree with the new one */
    setRTHSTree(rt, newHSTree);

    new_free(deleteLeaves, numLeaves * sizeof(boolean));
  } /* endif diagnosisType == 3 */

  /* Option 2 */
  if (diagnosisType == 2) {
    ptLeafListcaen hsTreePt = &getRTHSTree(rt);
    int numLeaves = getLeafListcaenNumber(hsTreePt);
    ptPreferenceClass pcA;
    void **pcPtA;
    leafListcaen newHSTree;

    if ((allocAux = new_calloc(numLeaves, sizeof(preferenceClass))) == NULL)
      return MEM_ERROR;
    pcA = (ptPreferenceClass) allocAux;

    if ((allocAux = new_calloc(numLeaves, sizeof(ptPreferenceClass))) == NULL)
      return MEM_ERROR;
    pcPtA = (void **) allocAux;

    resetLeafListcaen(hsTreePt);
    for (i = 0; i < numLeaves; i++) {
      ptLeaf lfPt = getNextLeafLL(hsTreePt);
      int numEdges = getEdgeSetNumberLeaf(lfPt);

      setPrefClassLeaf(&pcA[i], lfPt);
      resetEdgeSetLeaf(lfPt);
      for (j = 0; j < numEdges; j++) {
	ptConflict conflPt = getNextEdgeLeaf(lfPt);

	if (isPresent(conflPt, getRTSoftConflictListcaen(rt), compareConflicts))
	  getNumSoftConflicts(&pcA[i])++;
      }
      pcPtA[i] = &pcA[i];
    }
    sortPreferenceClasses(pcPtA, numLeaves);
    
    createLeafListcaen(&newHSTree);
    for (i = 0; i < numLeaves; i++) {
      ptPreferenceClass pcPt = (ptPreferenceClass) pcPtA[i];
      ptLeaf lfPt = getPrefClassLeaf(pcPt);
      leaf newLeaf = *lfPt;

      if (setNextLeafLLC(&newLeaf, &newHSTree) == MEM_ERROR)
	return MEM_ERROR;
    }
    if (liberateLeafListcaenC(hsTreePt) == MEM_ERROR)
      return MEM_ERROR;

    /* Update Hitting-Set Tree with the new one */
    setRTHSTree(rt, newHSTree);

    new_free(pcA, numLeaves * sizeof(preferenceClass));
    new_free(pcPtA, numLeaves * sizeof(ptPreferenceClass));

    /* Liberate soft conflict list */
    liberateConflictSet(&getRTSoftConflictListcaen(rt));
  } /* endif diagnosisType == 2 */

  /*******************************************/
  /* Hitting Set Tree (DIAGNOSTICS) Printing */
  /*******************************************/
  printHSTree(rt);

  /***********************/
  /* Memory deallocation */
  /***********************/

  /* Misbehaving variables destruction */
  for (i = 0; i < numMisb; i++) {
    ptMisbVar mbPt = getMisbVarPt(rt, i);
    ptConflictSet csPt = &getHardConflictSetMisbVar(mbPt);
    ptConflictSetListcaen cslPt = &getConflictSetListcaenMisbVar(mbPt);

    /* Hard conflict set destruction */
    liberateConflictSetWithName(csPt);

    /* Delete hard conflict set from conflict set list */
    deleteFirstConflictSetCSLC(cslPt);
    resetConflictSetListcaen(cslPt);
    for (j = 0; j < getConflictSetListcaenNumber(cslPt); j++) {
      ptConflictSet csPt = getNextConflictSetCSL(cslPt);

      /* Soft conflict set destruction */
      liberateConflictSetC(csPt);
    }
    /* Soft conflict set list destruction */
    liberateConflictSetListcaenC(cslPt);
  }

  /* Ok variables destruction */
  if (diagnosisType == 2 || diagnosisType == 3) {
    for (i = 0; i < numOk; i++) {
      ptOkVar okPt = getOkVarPt(rt, i);
      ptAscendantInfluenceSet aisPt = &getAscInflSetOkVar(okPt);

      liberateConflictSetWithName(aisPt);
    }
  }

  /* Misbehaving variable array and Ok variable array destruction */  
  new_free(getRTOkVarArray(rt), numOk * sizeof(okVar));
  new_free(getRTMisbVarArray(rt), numMisb * sizeof(misbVar));

  /* added by jpg 03/02/2000 */
  /* Hitting-set tree and old misb var array destruction */
  faultIsolationEnd (rt);
  /* end addition jpg */

  return OK;
}


/* ******************* */
/* Auxiliary functions */
/* ******************* */

boolean configurationHasChanged(runtime rt, boolean *incrPt, int diagnosisType) {
  boolean oldChange = CE_FALSE;
  int i;

  *incrPt = CE_TRUE;
  /* HS-Tree algorithm cannot be incremental neither when using diagnosis type 2 nor 3 */
  if (isOnlyFaultIso(rt) || diagnosisType == 2 || diagnosisType == 3) {
    *incrPt = CE_FALSE;
    if (isOnlyFaultIso(rt))
      return CE_TRUE;
  }

  /* Are all of the old misbehaving variables in the precedent iteration already misbehaving
     in the current one?. If so, the HS-Tree algorithm can be applied incrementaly */
  for (i = 0; i < getRTOldNumMisb(rt); i++) {
    variable oldVar = getVarOldMisbVarArray(rt, i);

    if (!isVarMisbehaving(oldVar)) {
      oldChange = CE_TRUE;
      *incrPt = CE_FALSE;
    }
  }

  /* If any influences' activation condition has changed its value between two consecutive
     time instants, then the HS-Tree algorithm cannot be applied incrementaly */
  if (*incrPt != CE_FALSE) {
    for (i = 0; i < getRTNInflu(rt); i++) {
      influence inflPt = getRTInfluNth(rt, i);

      if ((getInfluNActCond(inflPt) > 1) && (getInfluActCondNth(inflPt, 0) != getInfluActCondNth(inflPt, 1))) {
	*incrPt = CE_FALSE;
	break;
      }
    }
  }

  /* If the number of misbehaving variables in the previous and in the current iteration is the same,
     and all of the old misbehaving variables remain so, then there is no need to apply the HST algo */
  if (!oldChange && getRTOldNumMisb(rt) == getRTNumMisb(rt))
    return CE_FALSE;
  else
    return CE_TRUE;
}


/* It finds the conflict-set for a misbehaving-variable or  the ascendant-influence-set for an ok-variable */

STATUS findConflictsAux(runtime rt, variable currentPt, ptConflictSet csPt, ptListcaen pathPt, int accDelay, boolean resetNumConflicts) {

  STATUS makeUpConflict(influence, int, int *, ptConflictSet);

  static int numConflicts;
  cGraph cg = getVarNode(currentPt);
  int i;

  /* If the current variable is isolated in the causal graph, then the search cannot proceed */
  if (cg == NULL)
    return OK;

  /* Update traversed path with our current position (variable 'from') */
  addRight(currentPt, pathPt);

  /* Reset 'numConflicts' to zero at the beginning of 
     each misbehaving variable's conflict search,
     since it is a static variable */
  if (resetNumConflicts)
    numConflicts = 0;

  /* For each ascendant influence ... */
  for (i = 0; i < getCGNCause(cg); i++) {
    /* Get the ascendant influence */
    influence inflPt = getCGCauseNth(cg, i);

    /* Was the influence's activation condition true at time instant (t - accDelay)? */
    if (getInfluActCondNth(inflPt, accDelay)) {
      /* Get the ascendant, influencing variable */
      variable ascendantPt = getInfluCause(inflPt);

      if (makeUpConflict(inflPt, accDelay, &numConflicts, csPt) == MEM_ERROR)
	return MEM_ERROR;

      /* If the ascendant influencing variable is neither measured, nor input, nor is 
	 already present in the traversed path, then the search continues recursively */
      if (!ce_isMeasure(rt, ascendantPt) && !ce_isInput(rt, ascendantPt) && !isPresent(ascendantPt, *pathPt, NULL)) {
	objet trep = getInfluTRep(inflPt);
	range respTime = getObjRangeVal(trep);
	long delay = getInfluDelSup(inflPt);
	int auxAccDelay;

	if (getObjType (trep) == CE_RANGE)
	  auxAccDelay = accDelay + (getRangeInf(respTime) == 0.0 && getRangeSup(respTime) == 0.0 ? delay : delay + 1);
	else auxAccDelay = accDelay + (getObjRealVal (trep) == 0.0 ? delay : delay + 1);
	/* Recursive search call for the ascendant, influencing variable */
	if (findConflictsAux(rt, ascendantPt, csPt, pathPt, auxAccDelay, CE_FALSE) == MEM_ERROR)
	  return MEM_ERROR;
      }
    }
  } /* end for i */

  return OK;
}


STATUS makeUpConflict(influence inflPt, int delay, int *numConflictsPt, ptConflictSet csPt) {

  STATUS createConflictName(ptConflict, influence, char *);
  int compareAndUpdateConflicts(void *, void *);
  
  /*static char *SEP = "_";*/
  /*const char *influName = getInfluNom(inflPt);*/
  /*char *auxString, *tokenName;*/

  /* The influence's name is copied in another portion of memory to avoid problems
     when using 'strtok', since influences' names are allocated in global variables
     as constant strings */
  /*if ((auxString = (char *)new_malloc(strlen(influName) + 1)) == NULL)
    return MEM_ERROR;
    strcpy(auxString, influName);*/

  /* An influence's label may point out the existence of several components
     gathered in the same influence due to causal graph design reasons. If so,
     the different component names are separated by '_' in the influence's 
     label. A conflict must be created for each of the components in the label */

  /*tokenName = strtok(auxString, SEP);
  do {*/
  char* tokenName = mkString (getInfluNom (inflPt));
    conflict confl;

    if (createConflictName(&confl, inflPt, tokenName) == MEM_ERROR)
    /*if (createConflictName (&confl, inflPt, getInfluNom (inflPt)) == MEM_ERROR)*/
      return MEM_ERROR;
    setConflictInfluence(&confl, inflPt);
    setConflictInstant(&confl, delay);

    /* Is the conflict already in the variable's conflict-list? */
    if (!isPresent(&confl, *csPt, compareAndUpdateConflicts)) {
      setConflictIndex(&confl, (*numConflictsPt)++);
      /* If the influence's activation-condition was true and it represents
	 a brand-new conflict, then the conflict (triplet) is inserted
	 into the conlict-list associated to the misbehaving variable.
	 Remark: the insertion is carried out by creating a new 'conflict'
	 variable in the heap memory (addRightC) */
      if (setNextConflictCSC(&confl, csPt) == MEM_ERROR)
	return MEM_ERROR;
    }
    else {
      char *conflName = getConflictName(&confl);

      new_free(conflName, strlen(conflName) + 1);
    }
  /*} while ((tokenName = strtok(NULL, SEP)) != NULL);*/

  /*new_free(auxString, strlen(influName) + 1);*/

  return OK;
}


/* It generates a name for a new conflict */
STATUS createConflictName(ptConflict conflPt, influence inflPt, char *conflName) {
  void *allocAux;

  /* Does the influence belong to a component? */
  /* If so, the argument 'conflName' should contain the component's name */
  if (conflName != NULL) {
    /* Copy of the component's name into the conflict's 'name' field */
    if ((allocAux = new_malloc(strlen(conflName) + 1)) == NULL)
      return MEM_ERROR;

    setConflictName(conflPt, (char *) allocAux);
    strcpy(getConflictName(conflPt), conflName);
  }
  else {
    /* If the influence belongs to no component, then we make up a name for it.
       The new name will be composed of the influencing variable's name,
       concatenated to the string "->" and the the influenced variable's name */
    int length = strlen(getVarName(getInfluCause(inflPt))) + strlen(getVarName(getInfluEffet(inflPt))) + 1;

    if ((allocAux = new_malloc(length)) == NULL)
      return MEM_ERROR;

    setConflictName(conflPt, (char *) allocAux);
    /* Copy of the influence's name into the conflict's 'name' field */      
    strcpy(getConflictName(conflPt), getVarName(getInfluCause(inflPt)));
    strcat(getConflictName(conflPt), "->");
    strcat(getConflictName(conflPt), getVarName(getInfluEffet(inflPt)));
  }
  return OK;
}


STATUS findClassTree(runtime rt, ptSoftVar originSoftVarPt, int mbVarIndex, ptOkVarListcaen okVLPt, ptSoftVarListcaen sfVLPt) {

  int compareAscendantInfluences(void *, void *);

  ptAscendantInfluenceSet inflSetPt;
  int okVarIndex;

  /* We delete from the 'fifo' queue the soft variable whose class is to be calculated */
  deleteFirstSoftVarSVL(sfVLPt);

  /* Is the soft variable a misbehaving or an Ok variable? */
  if (getMisbehavingSoftVar(originSoftVarPt)) {
    ptMisbVar mbPt = getMisbVarSoftVar(originSoftVarPt);

    inflSetPt = (ptAscendantInfluenceSet) &getHardConflictSetMisbVar(mbPt);
  }
  else {
    ptOkVar okPt = getOkVarSoftVar(originSoftVarPt);

    inflSetPt = &getAscInflSetOkVar(okPt);
  }

  /* For each OK variable */
  for (okVarIndex = 0; okVarIndex < getRTNumOk(rt); okVarIndex++) {
    boolean equal = CE_FALSE;
    ptOkVar currentOkPt = getOkVarPt(rt, okVarIndex);
    int numberAscInfl = getAscInflSetNumberOkVar(currentOkPt);

    /* If the current OK variable is present in the list of OK variables that belong to
       some variable's class, then it is not taken into account for the class search */
    if (isPresent(currentOkPt, *okVLPt, NULL))
      continue;

    resetAscInflSetOkVar(currentOkPt);
    /* Looks if the intersection of the origin (soft) variable's and the current OK variable's ascendant influence list is not NULL */
    while (numberAscInfl-- && !equal) {
      ptAscendantInfluence ascInflPt = getNextAscInflOkVar(currentOkPt);

      /* Is any ascendant influence of the Ok variable equal to any ascendant influence of the origin (soft) variable? */
      if (isPresent(ascInflPt, *inflSetPt, compareAscendantInfluences)) {

	/* Then, the current OK variable belongs to the origin variable's class */
	setOkVarClassSoftVar(originSoftVarPt, mbVarIndex, getRTNumOk(rt), getNumOkVarClassSoftVar(originSoftVarPt, mbVarIndex)++,
			     getSoftVarPt(rt, getRTNumMisb(rt) + okVarIndex));
	/* Update the list of OK variables that are already present in some class */
	if (setNextOkVarOVL(currentOkPt, okVLPt) == MEM_ERROR)
	  return MEM_ERROR;
	/* Update breadth first search fifo */
	if (setNextSoftVarSVL(getSoftVarPt(rt, getRTNumMisb(rt) + okVarIndex), sfVLPt) == MEM_ERROR)
	  return MEM_ERROR;
	equal = CE_TRUE;
      }
    }
  }
  /* END For each OK variable */

  return OK;
}


STATUS generateSoftConflictSets(runtime rt, ptSoftVar originSoftVarPt, int mbVarIndex) {

  void generateUnionSet(runtime, ptSoftVar, int, int *, int, int, int, ptSoftVar *);
  STATUS generateSynonyms(runtime, ptSoftVar, int, ptSoftVar *, int, int, ascendantInfluenceSet);
  int numberUnionSubsets(int);

  int nClass = getNumOkVarClassSoftVar(originSoftVarPt, mbVarIndex); /* Class cardinality */
  int nUnionSet; /* Number of existing union sets calculated from the elements of the class */
  ptSoftVar *unionSet; /* Auxiliar union set for union set generation procedure */
  int indexUS = 0; /* Index of the next union set to be calculated */
  int okVarIndex, unionIndex;
  void *allocAux;

#ifdef DEBUG
  if (!getMisbehavingSoftVar(originSoftVarPt)) {
    ptOkVar okPt = getOkVarSoftVar(originSoftVarPt);
    fprintf(stderr, "\n DESCENDING to OK variable %s\n", getVarName(getVarOkVar(okPt)));
  }
#endif

  if (nClass == 0) {

#ifdef DEBUG
    if (getMisbehavingSoftVar(originSoftVarPt)) {
      ptMisbVar mbPt = getMisbVarSoftVar(originSoftVarPt);
      fprintf(stdout, "\n The CLASS of misbehaving variable %s is EMPTY\n", getVarName(getVarMisbVar(mbPt)));      
    }
    else {
      ptOkVar okPt = getOkVarSoftVar(originSoftVarPt);
      fprintf(stdout, "\n The CLASS of OK variable %s is EMPTY --> GOING UP\n", getVarName(getVarOkVar(okPt)));      
    }
#endif
    
    return OK;
  }

  /*********************************************************************************************************/
  /* Post-order traversal of a misbehaving variable's class tree to calculate each variable's synonym list */
  /*********************************************************************************************************/

  /* For each OK variable in the current soft variable's class */
  for(okVarIndex = 0; okVarIndex < nClass; okVarIndex++) {
    ptSoftVar sfVarPt = getOkVarClassSoftVar(originSoftVarPt, mbVarIndex, getRTNumOk(rt), okVarIndex);

    if (generateSoftConflictSets(rt, sfVarPt, mbVarIndex) == MEM_ERROR)
      return MEM_ERROR;
  }

  /***************************************************************************/
  /* Calculate current soft variable's union sets from its OK variable class */
  /***************************************************************************/

  /* Initialization and memory allocation of data structures */
  nUnionSet = numberUnionSubsets(nClass);

  if ((allocAux = new_calloc(nUnionSet, sizeof(int))) == NULL)
    return MEM_ERROR;
  setNumElemUnionSetArraySoftVar(originSoftVarPt, (int *) allocAux);

  if ((allocAux = new_calloc(nUnionSet * nClass, sizeof(ptSoftVar))) == NULL)
    return MEM_ERROR;
  setUnionSetMatrixSoftVar(originSoftVarPt, (ptSoftVar **) allocAux);
    
  if ((allocAux = new_calloc(nClass, sizeof(ptSoftVar))) == NULL)
    return MEM_ERROR;
  unionSet = (ptSoftVar *) allocAux;

#ifdef DEBUG
  if (getMisbehavingSoftVar(originSoftVarPt)) {
    ptMisbVar mbPt = getMisbVarSoftVar(originSoftVarPt);
    fprintf(stdout, "\n CALCULATING UNION SETS for misbehaving variable %s\n", getVarName(getVarMisbVar(mbPt)));
  }
  else {
    ptOkVar okPt = getOkVarSoftVar(originSoftVarPt);
    fprintf(stdout, "\n CALCULATING UNION SETS for OK variable %s\n", getVarName(getVarOkVar(okPt)));
  }
#endif

  /* For each OK variable in the current soft variable's class */
  for (okVarIndex = 0; okVarIndex < nClass; okVarIndex++)
    generateUnionSet(rt, originSoftVarPt, mbVarIndex, &indexUS, okVarIndex, 0, 1, unionSet);

  new_free(unionSet, nClass * sizeof(ptSoftVar));
  
  /****************************************************/
  /* Generation of current soft variable's 'synonyms' */
  /****************************************************/

#ifdef DEBUG
  if (getMisbehavingSoftVar(originSoftVarPt)) {
    ptMisbVar mbPt = getMisbVarSoftVar(originSoftVarPt);
    fprintf(stdout, "\n GENERATING SOFT-CONFLICT SETS for misbehaving variable %s\n", getVarName(getVarMisbVar(mbPt)));
  }
  else {
    ptOkVar okPt = getOkVarSoftVar(originSoftVarPt);
    fprintf(stdout, "\n CALCULATING SYNONYMS for OK variable %s AND GOING UP\n", getVarName(getVarOkVar(okPt)));
  }
#endif

  for (unionIndex = 0; unionIndex < nUnionSet; unionIndex++) {
    int nElements = getNumElemUnionSetSoftVar(originSoftVarPt, unionIndex);
    ascendantInfluenceSet ascInflUnionSet;

    createAscInflSet(&ascInflUnionSet);
    unionSet = getUnionSetSoftVar(originSoftVarPt, unionIndex, nClass);
    if (generateSynonyms(rt, originSoftVarPt, mbVarIndex, unionSet, nElements, 0, ascInflUnionSet) == MEM_ERROR)
      return MEM_ERROR;
  }

  /* Memory deallocation */
  new_free(getNumElemUnionSetArraySoftVar(originSoftVarPt), nUnionSet * sizeof(int));
  new_free(getUnionSetMatrixSoftVar(originSoftVarPt), nUnionSet * nClass * sizeof(ptSoftVar));

  return OK;
}


void generateUnionSet(runtime rt, ptSoftVar originSoftVarPt, int indexMb, int *ptIndexUS,
		      int indexOk, int indexCurrent, int lengthUnionSet, ptSoftVar *unionSet) {

  int nClass = getNumOkVarClassSoftVar(originSoftVarPt, indexMb);
  int newIndex, i;

  unionSet[lengthUnionSet - 1] = getOkVarClassSoftVar(originSoftVarPt, indexMb, getRTNumOk(rt), indexOk + indexCurrent);
  setNumElemUnionSetSoftVar(originSoftVarPt, *ptIndexUS, lengthUnionSet);

#ifdef DEBUG
  fprintf(stderr, "\n");
#endif

  for (i = 0; i < lengthUnionSet; i++) {
    setUnionSetElemSoftVar(originSoftVarPt, *ptIndexUS, nClass , i, unionSet[i]);

#ifdef DEBUG
    fprintf(stderr, "\t%s", getVarName(getVarOkVar(getOkVarSoftVar(unionSet[i]))));
#endif

  }

#ifdef DEBUG
  fprintf(stderr, "\n");
#endif

  (*ptIndexUS)++;
  newIndex = indexCurrent + 1;
  while (indexOk + newIndex < nClass)
    generateUnionSet(rt, originSoftVarPt, indexMb, ptIndexUS, indexOk, newIndex++, lengthUnionSet + 1, unionSet);
}


STATUS generateSynonyms(runtime rt, ptSoftVar originSoftVarPt, int indexMb, ptSoftVar *unionSet, int nElements,
			int indexCurrentElement, ascendantInfluenceSet ascInflUnionSet) {

  STATUS updateSoftConflictListcaen(runtime, ptConflictSet);
  int compareConflicts(void *, void *);
  int compareAndUpdateConflicts(void *, void *);

  ptSoftVar ptCurrentSV = unionSet[indexCurrentElement];
  ptSynonymListcaen ptCurrentSynonymListcaen = getSynonymListcaenPtSoftVar(ptCurrentSV, indexMb);
  int nSynonyms = getSynonymListcaenNumber(ptCurrentSynonymListcaen);
  ascendantInfluenceSet resultUnionSet;
  int i, j;

  resetSynonymListcaen(ptCurrentSynonymListcaen);

  for (i = 0; i < nSynonyms; i++) {
    ptSynonym ptCurrentSynonym = getNextSynonymSL(ptCurrentSynonymListcaen);

    if (listcaenUnion(*ptCurrentSynonym, ascInflUnionSet, &resultUnionSet, compareAndUpdateConflicts) == MEM_ERROR)
      return MEM_ERROR;
      
    /* If the union set is completely generated */
    if (indexCurrentElement == nElements - 1) {
      /* If the origin soft variable was a misbehaving variable */
      if (getMisbehavingSoftVar(originSoftVarPt)) {
	ptMisbVar mbPt = getMisbVarSoftVar(originSoftVarPt);
	ptConflictSet hardCSPt = &getHardConflictSetMisbVar(mbPt);
	conflictSet softCS; /* New soft conflict set for the misbehaving variable */

	/* A new soft conflict set for the misbehaving variable is generated */
	if (listcaenSymmetricDifferenceC(*hardCSPt, resultUnionSet, &softCS, compareConflicts, sizeof(conflict)) == MEM_ERROR)
	  return MEM_ERROR;

	/* Set-up right conflict-index for each conflict in 'softCS' */
	resetConflictSet(&softCS);
	for (j = 0; j < getConflictSetNumber(&softCS); j++) {
	  ptConflict conflPt = getNextConflictCS(&softCS);

	  setConflictIndex(conflPt, j);
	}

	/* Add the new soft conflict set (symmetric difference) to the misbehaving variable's conflict set list */
	if (setNextConflictSetMisbVarC(&softCS, mbPt) == MEM_ERROR)
	  return MEM_ERROR;

	/*****************************/
	/* Update soft conflict list */
	/*****************************/
	if (updateSoftConflictListcaen(rt, &softCS) == MEM_ERROR)
	  return MEM_ERROR;
      }
      /* If the origin soft variable was an OK variable */
      else {
	ptSynonymListcaen ptOriginSynonymListcaen = getSynonymListcaenPtSoftVar(originSoftVarPt, indexMb);
	ptSynonym ptOriginSynonym; /* Original synonym for the current soft variable */
	synonym newSynonym; /* New synonym for the current soft variable */

	resetSynonym(ptOriginSynonymListcaen);
	/* The original synonym is always the first synonym on the synonym list */
	ptOriginSynonym = getNextSynonymSL(ptOriginSynonymListcaen);
	
	/* A new synonym for the current soft variable is generated */
	if (listcaenSymmetricDifferenceC(*ptOriginSynonym, resultUnionSet, &newSynonym, compareConflicts, sizeof(conflict)) == MEM_ERROR)
	  return MEM_ERROR;

	/* Add the new synonym (symmetric difference) to the soft variable's synonym list */
	if (setNextSynonymSLC(&newSynonym, ptOriginSynonymListcaen) == MEM_ERROR)
	  return MEM_ERROR;
      }
    }
    /* Generating a possible combination for the current union set */
    else {
      if (generateSynonyms(rt, originSoftVarPt, indexMb, unionSet, nElements, indexCurrentElement + 1, resultUnionSet) == MEM_ERROR)
	return MEM_ERROR;
    }
    liberateAscInflSet(&resultUnionSet);
  }

  return OK;
}


int numberUnionSubsets(int nElem) {
  int i, accum = 1;

  for(i = 0; i < nElem; i++)
    accum *= 2;

  return accum - 1;
}


STATUS updateSoftConflictListcaen(runtime rt, ptConflictSet ptSoftCS) {

  int compareConflicts(void *, void *);

  int numConflicts = getConflictSetNumber(ptSoftCS);
  int x, y;

  resetConflictSet(ptSoftCS);
  for (x = 0; x < numConflicts; x++) {
    ptConflict softConflPt = getNextConflictCS(ptSoftCS);
    boolean isSoftConflict = CE_TRUE;

    for (y = 0; y < getRTNumMisb(rt); y++) {
      ptMisbVar mbPt = getMisbVarPt(rt, y);
      conflictSet hardConflSet = getHardConflictSetMisbVar(mbPt);

      if (isPresent(softConflPt, hardConflSet, compareConflicts)) {
	isSoftConflict = CE_FALSE;
	break;
      }
    }
    if(isSoftConflict && !isPresent(softConflPt, getRTSoftConflictListcaen(rt), compareConflicts))
      if (setNextConflictCS(softConflPt, &getRTSoftConflictListcaen(rt)) == MEM_ERROR)
	return MEM_ERROR;
  }
  return OK;
}


/* Comparison of conflicts */
int compareConflicts(void *cf1, void* cf2) {

  boolean compareConditions(condition, condition);

  if (cf1 == NULL || cf2 == NULL)
    return CE_FALSE;

  /* Two conflicts are considered to be equal, only and only if, both have identical 
     names (they belong to the same component), and identical activation conditions */

  /********************************************************************************/

  /* Same accumulated-delay? */
  /* if (getConflictInstant((ptConflict)cf1) != getConflictInstant((ptConflict)cf2))
     return CE_FALSE; */
  
  /********************************************************************************/

  /* Same component? */
  if (strcmp(getConflictName((ptConflict)cf1), getConflictName((ptConflict)cf2)))
    return CE_FALSE; /* Different components */

  /* Same activation-condition? */
  if (!compareConditions(getInfluCond(getConflictInfluence((ptConflict)cf1)),
			 getInfluCond(getConflictInfluence((ptConflict)cf2))))
    return CE_FALSE; /* Different activation-conditions */

  return CE_TRUE; /* Identical conflicts */
}


/* Comparison and failure time instant update of conflicts */
int compareAndUpdateConflicts(void *cf1, void* cf2) {

  int compareConflicts(void *, void *);

  int result = compareConflicts(cf1, cf2);
  int inst1 = getConflictInstant((ptConflict)cf1);
  int inst2 = getConflictInstant((ptConflict)cf2);

  /* Update accumulated delay */
  if (result == CE_TRUE) {
    if (inst1 > inst2)
      setConflictInstant((ptConflict)cf2, inst1);
    else if (inst1 < inst2)
      setConflictInstant((ptConflict)cf1, inst2);
  }

  return result;
}


/* Comparison of activation-conditions */
boolean compareConditions(condition cd1, condition cd2) {

  boolean compareConditions(condition, condition);

  variable var1, var2;
  objet ob1, ob2;

  /*************************************************
     Remark: A NULL condition-pointer is considered
             to be a pointer on a TRUE condition
  **************************************************/

  /* Both condition pointers point to the same condition-objet? */
  if (cd1 == cd2)
    return CE_TRUE;

  /* If any condition pointer is NULL, but not both of them,
     then conditions are not equal */
  if (cd1 == NULL || cd2 == NULL)
    return CE_FALSE;

  if (getCondType(cd1) != getCondType(cd2))
    return CE_FALSE;
  
  switch (getCondType(cd1)) {
  case CE_CTE :
    if ( getCondValue(cd1) == getCondValue(cd2))
      return CE_TRUE;
    else
      return CE_FALSE;
    break; /* For clarity reasons */
  case CE_COMP :
    if (getCondOpRel(cd1) != getCondOpRel(cd2))
      return CE_FALSE;
    var1 = getCondVar1(cd1);
    var2 = getCondVar1(cd2);
    if (var1 != var2)
      return CE_FALSE;
    var1 = getCondVar2(cd1);
    var2 = getCondVar2(cd2);
    if (var1 != var2)
      return CE_FALSE;
    /* The second operand of a comparison operator can be 
       either a variable or an object (real number) */
    if (var1 == NULL) {
      ob1 = getCondObj(cd1);
      ob2 = getCondObj(cd2);
      /* Important Remark: Variables appearing on activation-conditions can only be of type 'measured'.
	 This means that their value, used for evaluation purposes, will be a real number (not a range)
	 and therefore, so will be the other objects appearing on the activation-condition. So the function
	 'ce_eqObjet(object1, object2)', which looks for the equality of two objects, since it is only used
	 when comparing objects belonging to activation-conditions, only takes into account comparison
	 for objets that are real numbers */
      return ce_eqObjet(ob1, ob2);
    }
    break;
  case CE_OU : case CE_ET :
    if (compareConditions(getCondOp1(cd1), getCondOp1(cd2)) && compareConditions(getCondOp2(cd1), getCondOp2(cd2)))
      return CE_TRUE;
    else
      return CE_FALSE;
    break;
  case CE_NON:
    return compareConditions(getCondOp1(cd1), getCondOp1(cd2));
    break;
  default :
    break;
  }

  return CE_TRUE;
}


/* For a given leaf 'new', it checks whether H(new) 
   contains H(l) for some leaf 'l' of leaf-list 'old', 
   where H(leaf) stands for the leaf's conflict-set */
boolean closeBranch(ptLeaf newLPt, ptLeafListcaen oldLLPt) {

  int compareConflicts(void *, void *);

  int numOldLeaves = getLeafListcaenNumber(oldLLPt);
  conflictSet newEdge = getEdgeSetLeaf(newLPt);
  int i, j;

  resetLeafListcaen(oldLLPt);
  for (i = 0; i < numOldLeaves; i++) {
    ptLeaf oldLeafPt = getNextLeafLL(oldLLPt);
    int numOldEdges = getEdgeSetNumberLeaf(oldLeafPt);
    boolean contained = CE_TRUE;

    resetEdgeSetLeaf(oldLeafPt);
    for (j = 0; j < numOldEdges; j++) {
      ptConflict oldConflPt = getNextEdgeLeaf(oldLeafPt);

      if (!isPresent(oldConflPt, newEdge, compareConflicts)) {
	contained = CE_FALSE;
	break;
      }
    } /* For each old-leaf's edge 'j' */

    if (contained)
      return CE_TRUE;

  } /* For each old-leaf 'i' */

  return CE_FALSE;
}


void quickSort(void *v[], int left, int right, int (*comp)(const void *, const void *)) {
  int i, last;

  void swap(void *v[], int, int);

  if (left >= right)
    return;

  swap(v, left, (left + right) / 2);
  last = left;
  for (i = left + 1; i <= right; i++)
    if ((*comp)(v[i], v[left]) < 0)
      swap(v, ++last, i);
  swap(v, left,last);
  quickSort(v, left, last - 1, comp);
  quickSort(v, last + 1, right, comp);
  return;
}


void swap(void *v[], int i, int j){
  void * temp;
  
  temp = v[i];
  v[i] = v[j];
  v[j] = temp;

  return;
}    


int comparePreferenceClasses(const void* pc1, const void* pc2) { 
  if (getNumSoftConflicts((ptPreferenceClass)pc1) < getNumSoftConflicts((ptPreferenceClass)pc2))
    return -1;
  else
    if (getNumSoftConflicts((ptPreferenceClass)pc1) > getNumSoftConflicts((ptPreferenceClass)pc2))
      return +1;
    else
      if (getEdgeSetNumberLeaf(getPrefClassLeaf((ptPreferenceClass)pc1)) <
	  getEdgeSetNumberLeaf(getPrefClassLeaf((ptPreferenceClass)pc2)))
	return -1;
      else
	if (getEdgeSetNumberLeaf(getPrefClassLeaf((ptPreferenceClass)pc1)) >
	    getEdgeSetNumberLeaf(getPrefClassLeaf((ptPreferenceClass)pc2)))
	  return +1;
	else
	  return 0;
}


void printHSTree(runtime rt) {

  int numLength(int);
  void printConflict(runtime, ptConflict);
  
  ptLeafListcaen hsTreePt = &getRTHSTree(rt);
  int numLeaves = getLeafListcaenNumber(hsTreePt);
  char mes1[] = "* DIAGNOSIS' RESULTS *";
  char mes2[] = "DIAGNOSTIC ";
  int i, j, length, len2;

  existing_conflicts = new vector<condition> ();
  freeBridgeConflicts (rt->bridgeConflicts);
  rt->bridgeConflicts = new vector<vector<Atom*>* > ();

  length = strlen(mes1);
  memset(line, symb1, length < MAXLINE ? length : MAXLINE - 1);
  line[length] = '\0';
  /*  fprintf(stderr, "\n\n%s\n", line);
  fprintf(stderr, "%s\n", mes1);
  fprintf(stderr, "%s\n", line); */

  len2 = strlen(mes2);
  resetLeafListcaen(hsTreePt);
  for (i = 0; i < numLeaves; i++) {
    ptLeaf lfPt = getNextLeafLL(hsTreePt);
    int numConflicts = getEdgeSetNumberLeaf(lfPt);

    length = len2 + numLength(i + 1);
    memset(line, symb2, length < MAXLINE ? length : MAXLINE - 1);
    line[length] = '\0';
    /*    fprintf(stderr, "\n%s%d\n", mes2, i + 1);
	  fprintf(stderr, "%s\n", line); */
    runt_count += sprintf (getRTDiagOut (rt) + runt_count, "[");

    resetEdgeSetLeaf(lfPt);
    for (j = 0; j < numConflicts; j++) {
      ptConflict conflPt = getNextEdgeLeaf(lfPt);

      addBridgeConflict (rt, conflPt);

      printConflict(rt, conflPt);
    } /* end for j */
    runt_count += sprintf (getRTDiagOut (rt) + runt_count, "]");
  } /* end for i */
  /*  fprintf(stderr, "\n"); */
  delete existing_conflicts;

  return;
}


void printConflict(runtime rt, ptConflict conflPt) {
  if (conflPt == NULL) {
    fprintf(stderr, "Unable to print conflict (NULL Pointer)\n");
    return;
  }

  /*  fprintf(stderr, "{ %-10s , %-5d , ", getConflictName(conflPt), getConflictInstant(conflPt)); */
  runt_count += sprintf (getRTDiagOut (rt) + runt_count, "{%s,%d,", getConflictName(conflPt), getConflictInstant(conflPt));
  if (getInfluCond(getConflictInfluence(conflPt)))
    ce_printDiagCond (rt, getInfluCond(getConflictInfluence(conflPt)));
  else
    /*    fprintf(stderr, "NO act-cond"); */
    runt_count += sprintf (getRTDiagOut (rt) + runt_count, "()");
  /*  fprintf(stderr, " }\n"); */
  runt_count += sprintf (getRTDiagOut (rt) + runt_count, "}");
  return;
}

int numLength(int n){
  int i;

  for (i = 1; n /= 10; i++);
  return i;
}


/***************************************************************************************************/


STATUS setFirstConflictWithNameCS(ptConflict conflPt, ptConflictSet csPt) {
  char *conflName;
  int nameLength;
  ptConflict newConflPt;
  void *allocAux;

  if (addLeftC(conflPt, csPt, sizeof(conflict)) == MEM_ERROR)
    return MEM_ERROR;

  conflName = getConflictName(conflPt);
  nameLength = strlen(conflName);
  if ((allocAux = new_malloc(nameLength + 1)) == NULL)
    return MEM_ERROR;

  newConflPt = (ptConflict) lookLeft(*csPt);
  setConflictName(newConflPt, (char *) allocAux);
  strcpy(getConflictName(newConflPt), conflName);

  return OK;
}


STATUS setNextConflictWithNameCS(ptConflict conflPt, ptConflictSet csPt) {
  char *conflName;
  int nameLength;
  ptConflict newConflPt;
  void *allocAux;

  if (addRightC(conflPt, csPt, sizeof(conflict)) == MEM_ERROR)
    return MEM_ERROR;

  conflName = getConflictName(conflPt);
  nameLength = strlen(conflName);
  if ((allocAux = new_malloc(nameLength + 1)) == NULL)
    return MEM_ERROR;

  newConflPt = (ptConflict) lookRight(*csPt);
  setConflictName(newConflPt, (char *) allocAux);
  strcpy(getConflictName(newConflPt), conflName);

  return OK;
}


STATUS assignConflictSetWithName(ptConflictSet newCSPt, conflictSet oldCS) {
  int numConflicts = getConflictSetNumber(&oldCS);
  int i;

  if (createConflictSet(newCSPt) == MEM_ERROR)
    return MEM_ERROR;
  resetConflictSet(&oldCS);
  for (i = 1; i <= numConflicts; i++)
    if (setNextConflictWithNameCS(getNextConflictCS(&oldCS), newCSPt) == MEM_ERROR)
      return MEM_ERROR;

  return OK;
}


STATUS deleteFirstConflictWithNameCS(ptConflictSet csPt) {
  ptConflict conflPt;
  char *conflName;
  int nameLength;

  if (csPt == NULL)
    return MEM_ERROR;

  conflPt = (ptConflict) lookLeft(*csPt);
  conflName = getConflictName(conflPt);
  nameLength = strlen(conflName);

  new_free(conflName, nameLength + 1);
  if (deleteLeftC(csPt, sizeof(conflict)) == MEM_ERROR)
    return MEM_ERROR;

  return OK;
}


STATUS liberateConflictSetWithName(ptConflictSet csPt) {
  int numConflicts, i;

  numConflicts = getConflictSetNumber(csPt);
  for (i = 1; i <= numConflicts; i++)
    if (deleteFirstConflictWithNameCS(csPt) == MEM_ERROR)
      return MEM_ERROR;

  return OK;
}


/***************************************************************************************************/

#ifdef DEBUG

void printClasses(runtime rt) {

  void printClass(runtime, ptSoftVar, int);

  char mes1[] = "* CLASSES OF VARIABLES *";
  int length, i;

  length = strlen(mes1);
  memset(line, symb1, length < MAXLINE ? length : MAXLINE - 1);
  line[length] = '\0';
  fprintf(stderr, "\n%s\n", line);
  fprintf(stderr, "%s\n", mes1);
  fprintf(stderr, "%s\n", line);
  for (i = 0; i < getRTNumMisb(rt); i++) {
    ptSoftVar sfVarPt = getSoftVarPt(rt, i);

    printClass(rt, sfVarPt, i);
  }
}


void printClass(runtime rt, ptSoftVar originSoftVarPt, int mbVarIndex) {
  int nClass = getNumOkVarClassSoftVar(originSoftVarPt, mbVarIndex);
  
  char mes2[] = "MISBEHAVING VARIABLE ";
  char mes3[] = "OK Variable ";
  int length, i;
  
  if (getMisbehavingSoftVar(originSoftVarPt)) {
    ptMisbVar mbPt = getMisbVarSoftVar(originSoftVarPt);
    variable varPt = getVarMisbVar(mbPt);
    const char *varName = getVarName(varPt);

    length = strlen(mes2) + strlen(varName) + 8;
    memset(line, symb2, length < MAXLINE ? length : MAXLINE - 1);
    line[length] = '\0';
    fprintf(stderr, "\n%s\n", line);
    fprintf(stderr, "%s%s's class\n", mes2, varName);
    fprintf(stderr, "%s\n", line);
  }
  else {
    ptOkVar okPt = getOkVarSoftVar(originSoftVarPt);
    variable varPt = getVarOkVar(okPt);
    const char *varName = getVarName(varPt);

    length = strlen(mes3) + strlen(varName) + 8;
    memset(line, symb2, length < MAXLINE ? length : MAXLINE - 1);
    line[length] = '\0';
    fprintf(stderr, "\n%s%s's class\n", mes3, varName);
    fprintf(stderr, "%s\n", line);
  }

  if (nClass == 0) {
    fprintf(stderr, "\n  --->\tEMPTY Class\n");
    return;
  }

  for (i = 0; i < nClass; i++) {
    ptSoftVar sfVarPt = getOkVarClassSoftVar(originSoftVarPt, mbVarIndex, getRTNumOk(rt), i);
    ptOkVar okPt = getOkVarSoftVar(sfVarPt);
    variable varPt = getVarOkVar(okPt);
    const char *varName = getVarName(varPt);

    fprintf(stderr, "\n  --->\tOK Variable %s", varName);
  }
  fprintf(stderr, "\n");

  for (i = 0; i < nClass; i++) {
    ptSoftVar sfVarPt = getOkVarClassSoftVar(originSoftVarPt, mbVarIndex, getRTNumOk(rt), i);

    printClass(rt, sfVarPt, mbVarIndex);
  }

  return;
}


void printConflictSets(runtime rt) {

  int numLength(int);
  void printConflict(runtime, ptConflict);

  char mes1[] = "* CONFLICTS SETS for MISBEHAVING VARIABLE ";
  char mes2[] = "HARD-CONFLICT SET";
  char mes3[] = "SOFT-CONFLICT SET";
  int length, len1, len2, len3;
  int i, j, k;
  
  len1 = strlen(mes1);
  len2 = strlen(mes2);
  len3 = strlen(mes3);
  for (i = 0; i < getRTNumMisb(rt); i++) {
    ptMisbVar mbPt = getMisbVarPt(rt, i);
    const char *name = getVarName(getVarMisbVar(mbPt));
    ptConflictSetListcaen conflSetListcaenPt = &getConflictSetListcaenMisbVar(mbPt);
    int numConflictSets = getConflictSetListcaenNumber(conflSetListcaenPt);

    length = len1 + strlen(name) + 2;
    memset(line, symb1, length < MAXLINE ? length : MAXLINE - 1);
    line[length] = '\0';
    fprintf(stderr, "\n\n%s\n", line);
    fprintf(stderr, "%s%s *\n", mes1, name);
    fprintf(stderr, "%s\n", line);

    resetConflictSetListcaen(conflSetListcaenPt);
    for (j = 0; j < numConflictSets; j++) {
      ptConflictSet conflSetPt = getNextConflictSetCSL(conflSetListcaenPt);
      int numConflicts = getConflictSetNumber(conflSetPt);

      length = (j == 0 ? len2 : len3) + numLength(j + 1);
      memset(line, symb2, length < MAXLINE ? length : MAXLINE - 1);
      line[length] = '\0';
      fprintf(stderr, "\n%s%d\n", (j == 0 ? mes2 : mes3), j + 1);
      fprintf(stderr, "%s\n", line);
      resetConflictSet(conflSetPt);
      for (k = 0; k < numConflicts; k++) {
	ptConflict confl = getNextConflictCS(conflSetPt);

	printConflict(rt, confl);
      }
      fflush(stderr);
    }
  }
}

#endif


void ce_razDiagBuf (runtime rt) {

  /* RAZ diagnostic buffer */
  runt_count = 0;
  /* *(getRTDiagOut (rt)) = '\0';*/
  if (! isNull (getRTDiagOut (rt)))
    cefree (getRTDiagOut (rt));
  getRTDiagOut (rt) = (char*) new_malloc (1024*sizeof (char*));
}


void ce_prepareDiag (runtime rt) {
  long i, k, nmes=getRTNMeasure(rt), cumul, misb, r;
  variable v;

  ce_razDiagBuf (rt);
  /* push history of behaviours */
  if ((cumul = getRTMisbCumTime (rt)) > 0) {
    for (i=0; i<nmes; i++) {
      v=getRTMeasureNth (rt, i);
      misb = 0;
      for (k=cumul-1; k>0; --k) {
	r = getVarHistoMisbNth (v, k-1);
	setVarHistoMisbNth (v, k, r);
	misb = misb || r;
      } /* end for k */
      r = isVarMisbehaving (v);
      setVarHistoMisbNth (v, 0, r);
      misb = misb || r;
      if (misb)
	setVarMisbehaving (v);
      else
	setVarOk (v);
    } /* end for i */
  } /* endif cumul */
  return;
}


void ce_printDiagCond (runtime rt, condition cd) {

  enum condType ct;
  /*variable v;*/

  ct = getCondType (cd);
  runt_count += sprintf (getRTDiagOut (rt) + runt_count, "(");
  switch (ct) {
  case CE_CTE :
    if (CE_TRUE == getCondValue (cd))
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, "true");
    else
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, "false");
    break;
  case CE_EXPR :
    ce_printDiagExpr (rt, getCondExpr (cd));
  case CE_COMP :
    ce_printDiagVarNom (rt, getCondVar1 (cd));
    switch (getCondOpRel (cd)) {
    case CE_EQ :
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, " = ");
      break;
    case CE_NE :
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, " != ");
      break;
    case CE_LT :
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, " < ");
      break;
    case CE_GE :
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, " >= ");
      break;
    case CE_IN :
      break;
    default:
      break;
    }
    /*if (isNull (v = getCondVar2 (cd))) {
      objet ob=getCondObj (cd);
      if (CE_RANGE == getObjType (ob))
	ce_printDiagRange (rt, getObjRangeVal (ob));
      else
	ce_printDiagReal (rt, getObjRealVal (ob));
    }
    else 
      ce_printDiagVarNom (rt, v);*/
    break;
  case CE_AT :
    runt_count += sprintf (getRTDiagOut (rt) + runt_count, "(%s)", getCondAtom (cd)->getId ());
  case CE_OU :
    if (! isNull (getCondOp1 (cd))) {
      ce_printDiagCond (rt, getCondOp1 (cd));
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, " | ");
    }
    if (! isNull (getCondOp2 (cd)))
      ce_printDiagCond (rt, getCondOp2 (cd));
    break;
  case CE_ET :
    if (! isNull (getCondOp1 (cd))) {
      ce_printDiagCond (rt, getCondOp1 (cd));
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, " & ");
    }
    if (! isNull (getCondOp2 (cd)))
      ce_printDiagCond (rt, getCondOp2 (cd));
    break;
  case CE_NON :
    runt_count += sprintf (getRTDiagOut (rt) + runt_count, "! ");
    ce_printDiagCond (rt, getCondOp1 (cd));
    break;
  default:
    break;
  } /* end switch ct */
  runt_count += sprintf (getRTDiagOut (rt) + runt_count, ")");
  return;
}

void ce_printDiagExpr (runtime rt, expr e) {
  enum exprType tp=getExprType (e);
  expr e1, e2;
  objet ob;
  
  switch (tp) {
  case CE_CSTE :
    ob = gVVal (getExprGVar (e));
    if (CE_RANGE == getObjType (ob))
      ce_printDiagRange (rt, getObjRangeVal (ob));
    else
      ce_printDiagReal (rt, getObjRealVal (ob));
    break;
  case CE_VBL :
    ce_printDiagVarNom (rt, getGVVar (getExprGVar (e)));
    break;
  case CE_PLUS :
  case CE_MOINS :
  case CE_MULT :
  case CE_DIV :
  case CE_EXP :
    e1 = getExprOp1 (e);
    /*ce_printExpr (stdout, e1);*/
    /*if (isExprCte (e1) || hasName (e1))
      ce_printDiagVarNom (rt, getGVVar (getExprGVar (e1)));
    else {*/
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, "(");
      ce_printDiagExpr (rt, e1);
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, ")");
    /*}*/
    switch (tp) {
    case CE_PLUS :
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, "+");
      break;
    case CE_MOINS :
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, "-");
      break;
    case CE_MULT :
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, "*");
      break;
    case CE_DIV :
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, "/");
      break;
    case CE_EXP :
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, "**");
      break;
    default:
      break;
    }
    e2 = getExprOp2 (e);
    /*if (isExprCte (e2) || hasName (e2))
      ce_printDiagVarNom (rt, getGVVar (getExprGVar (e2)));
    else {*/
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, "(");
      ce_printDiagExpr (rt, e2);
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, ")");
    /*}*/
    break;
  case CE_SGN :
    runt_count += sprintf (getRTDiagOut (rt) + runt_count, "-");
    e1 = getExprOp1 (e);
    /*if (isExprCte (e1) || hasName (e1))
      ce_printDiagVarNom (rt, getGVVar (getExprGVar (e1)));
    else {*/
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, "(");
      ce_printDiagExpr (rt, e1);
      runt_count += sprintf (getRTDiagOut (rt) + runt_count, ")");
    /*}*/
    break;
  default:
    break;
  }
  return ;
}

void ce_printDiagVarNom (runtime rt, variable v) {

  runt_count += sprintf (getRTDiagOut (rt) + runt_count, "%s", getVarName (v));
  return;
}

void ce_printDiagReal (runtime rt, double vd) {

  runt_count += sprintf (getRTDiagOut (rt) + runt_count, "%g", vd);
  return;
}


void ce_printDiagRange (runtime rt, range r) {

  runt_count += sprintf (getRTDiagOut (rt) + runt_count, "%g %g", getRangeInf (r), getRangeSup (r));
  return;
}


/*****************************************************************************************/
